import Qieyun from 'qieyun';

/**
 * 白一平轉寫
 * Baxter, W. H. (1992). A Handbook of Old Chinese Phonology. De Gruyter Mouton.
 * Baxter, W. H., & Sagart, L. (2014). Old Chinese: A New Reconstruction. Oxford University Press.
 *
 * 説明
 * 以下內容為生成白一平轉寫的函數體
 * 函數接受音韻地位，返回對應的白一平轉寫
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的白一平轉寫
 */
export function baxter(音韻地位, 字頭) {
  const 開關 = {
    // 版本可選 '1992' 或 '2014'，預設值為 '2014'
    版本: '2014',
  };

  const is = (x) => 音韻地位.屬於(x);

  let 聲母 = {
    幫: 'p',   滂: 'ph',   並: 'b',   明: 'm',
    端: 't',   透: 'th',   定: 'd',   泥: 'n',  來: 'l',
    知: 'tr',  徹: 'trh',  澄: 'dr',  孃: 'nr',
    精: 'ts',  清: 'tsh',  從: 'dz',                     心: 's',  邪: 'z',
    莊: 'tsr', 初: 'tsrh', 崇: 'dzr',                    生: 'sr', 俟: 'zr',
    章: 'tsy', 昌: 'tsyh', 常: 'dzy', 日: 'ny',          書: 'sy', 船: 'zy', 以: 'y',
    見: 'k',   溪: 'kh',   羣: 'g',   疑: 'ng',
    影: "'",   曉: 'x',    匣: 'h',                                          云: 'h',
  }[音韻地位.母];

  if (開關.版本 === '1992' && 聲母 === "'") {
    聲母 = 'ʔ';
  }

  let 韻母 = {
    // 一等韻
    東: 'uwng',
    冬: 'owng',
    模: 'u',
    泰: 'aj',
    灰: 'oj',
    咍: 'oj',
    魂: 'on',
    痕: 'on',
    寒: 'an',
    豪: 'aw',
    歌: 'a',
    唐: 'ang',
    登: 'ong',
    侯: 'uw',
    覃: 'om',
    談: 'am',

    // 二等韻
    江: 'aewng',
    佳: 'ea',
    皆: 'eaj',
    夬: 'aej',
    刪: 'aen',
    山: 'ean',
    肴: 'aew',
    麻: 'ae',
    庚: 'aeng',
    耕: 'eang',
    咸: 'eam',
    銜: 'aem',

    // 四等韻
    齊: 'ej',
    先: 'en',
    蕭: 'ew',
    青: 'eng',
    添: 'em',

    // 三等陰聲韻
    支: 'je',
    脂: 'ij',
    之: 'i',
    微: 'j+j',
    魚: 'jo',
    虞: 'ju',
    祭: 'jej',
    廢: 'joj',
    宵: 'jew',
    // 歌: 'ja',
    // 麻: 'jae',
    尤: 'juw',
    幽: 'jiw',

    // 三等陽聲韻
    // 東: 'juwng',
    鍾: 'jowng',
    眞: 'in',
    臻: 'in',
    文: 'jun',
    欣: 'j+n',
    元: 'jon',
    仙: 'jen',
    陽: 'jang',
    // 庚: 'jaeng',
    清: 'jeng',
    蒸: 'ing',
    侵: 'im',
    鹽: 'jem',
    嚴: 'jaem',
    凡: 'jom',
  }[音韻地位.韻];

  // 東歌麻庚韻同時含三等與非三等，上文僅處理非三等，此處處理三等
  if (is('東歌麻庚韻 三等')) {
    韻母 = 'j' + 韻母;
  }

  if (開關.版本 === '1992') {
    if (韻母 === 'ea') 韻母 = 'ɛɨ';
    韻母 = 韻母.replace('+', 'ɨ').replace('ae', 'æ').replace('ea', 'ɛ');
  }

  // 章組或日以母只與三等韻相拼，省去韻母起始的 j
  if (is('章組 或 日以母') && 韻母.startsWith('j')) {
    韻母 = 韻母.slice(1);
  }

  // 重紐 A 類添加 j 或 i
  if (is('重紐A類')) {
    if (韻母.startsWith('j')) 韻母 = 'ji' + 韻母.slice(1);
    else 韻母 = 'j' + 韻母;
  }

  // 合口字添加 w
  if (is('合口 或 灰魂韻') && !is('文凡韻')) {
    if (韻母.startsWith('j')) 韻母 = 'jw' + 韻母.slice(1);
    else 韻母 = 'w' + 韻母;
  }

  if (is('入聲')) {
    if (韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'p';
    else if (韻母.endsWith('n')) 韻母 = 韻母.slice(0, -1) + 't';
    else if (韻母.endsWith('ng')) 韻母 = 韻母.slice(0, -2) + 'k';
  }

  const 聲調 = {
    上: 'X',
    去: 'H',
  }[音韻地位.聲] || '';

  return 聲母 + 韻母 + 聲調;
}

/**
 * 有女羅馬字
 * https://github.com/BYVoid/ytenx/blob/master/ytenx/sync/kyonh/PrengQim.txt
 *
 * 以下內容為生成有女羅馬字的函數體
 * 函數接受音韻地位，返回對應的有女羅馬字
 *
 * 本版說明：
 * - 基於韻典網數據，並參照有女同車原版《〈廣韻〉全字表》（下稱「原表」）有所校正
 * - 目前版本推導器的音韻地位分析（基於 poem《廣韻字音表》）與原表有數十小韻在聲母、開合、等、重紐類等歸類有所不同，除個別外均不影響拼寫體系
 * - 旨韻（脂上聲）若干精、章組小韻，韻母寫為B類，與他處精章組均用A類不合，現修正
 * - 崇開三眞上（濜）：真諄上去聲莊組開口，原表及拼寫均析為臻韻上去聲，惟此小韻遺漏，現修正
 * - 莊合三真入（𠭴）：韻母寫為A類，與他處莊組均用B類不合，現修正（雖然不排除是因該小韻列於術韻而非質韻，但術韻下亦有「率」小韻用B類拼作 shwyt，故當以B類為正）
 * - 以開一寒入（䔾）：原表作明母，誤。原表亦無寒韻拼章組日以母例，但有談韻書母例與之平行，故可拼作 jat
 * - 幫三B清入（碧）：原表視為清A合口作 pvek，與「辟」小韻開合對立。原表清韻無B類，但可據體系推出處理為清B時的拼法 pyek。此外該小韻亦可按庚三處理為 pyak。本代碼同時支持三種處理法
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的有女羅馬字
 */
export function blankego(音韻地位, 字頭) {
  // 可調整參數
  const options = {
    清B脣音: null, // 清B脣音（「碧」小韻）處理法：默認（`null`）按清B拼寫
                   // 其他可選值：
                   // - `'合口'`：按清A合口拼寫，同原表處理法
                   // - `'庚三'`：按庚三拼寫，同韻典處理法
    莊組眞欣開口歸臻: true, // 莊組眞欣韻開口上、去聲均視作臻韻（原表、韻典均如此處理）
    脣音咍韻作開口: 1, // 脣音咍韻是否視作開口，可選值：
                       // - `0`：一律視作合口
                       // - `1`（默認）：視 `音韻地位.韻` 而定，
    凡韻非脣音作合口: 1, // 凡韻非脣音是否視作合口，可選值：
                         // - `0`：一律視作開口
                         // - `1`（默認）：視 `音韻地位.韻` 而定，
  };

  // 調整音韻地位

  if (options.清B脣音 === '庚三' && 音韻地位.屬於('幫組 清韻 重紐B類')) {
    const { 母, 聲 } = 音韻地位;
    音韻地位 = new 音韻地位.constructor(母, null, '三', null, '庚', 聲);
  }

  if (options.莊組眞欣開口歸臻 && 音韻地位.屬於('莊組 眞欣韻 開口')) {
    const { 母, 等, 聲 } = 音韻地位;
    音韻地位 = new 音韻地位.constructor(母, '開', 等, null, '臻', 聲);
  }

  if (!options.脣音咍韻作開口 && 音韻地位.屬於('幫組 咍韻')) {
    const { 母, 等, 聲 } = 音韻地位;
    音韻地位 = new 音韻地位.constructor(母, null, 等, null, '灰', 聲);
  }

  if (!options.凡韻非脣音作合口 && 音韻地位.屬於('凡韻') && !音韻地位.屬於('幫組')) {
    const { 母, 等, 聲 } = 音韻地位;
    音韻地位 = new 音韻地位.constructor(母, '開', 等, null, '嚴', 聲);
  }

  const is = (x) => 音韻地位.屬於(x);

  /**
   * 依次匹配列表 `rules` 中的規則，返回第一項測試為 true 的對應結果。
   * 無匹配規則時抛出異常，`name` 會包含於異常信息中。
   * 
   * `rules` 每項為 `[條件, 結果]`
   * 條件可為：
   * - 字串，會用 `音韻地位.屬於` 測試
   * - 函數，會直接調用並測試返回結果
   * - 其他，直接測試真值
   */
  function 規則(name, rules) {
    for (const [rule, res] of rules) {
      if (typeof rule === 'string') {
        if (is(rule)) {
          return res;
        }
      } else if (typeof rule === 'function') {
        if (rule()) {
          return res;
        }
      } else if (rule) {
        return res;
      }
    }
    throw new Error(`無${name}規則：${音韻地位.描述}`);
  }

  // 聲母

  let 母 = {
    幫: 'p',  滂: 'ph', 並: 'b',  明: 'm',
    端: 't',  透: 'th', 定: 'd',  泥: 'n',
    知: 't',  徹: 'th', 澄: 'd',  孃: 'n',
    精: 'z',  清: 'c',  從: 'dz',           心: 's',  邪: 'sz',
    莊: 'tr', 初: 'ch', 崇: 'dr',           生: 'sh', 俟: 'zh',
    章: 'tj', 昌: 'tc', 船: 'dj', 日: 'r',  書: 'sj', 常: 'zj',
    見: 'k',  溪: 'q',  羣: 'g',  疑: 'ng',
    影: '',   曉: 'x',  匣: 'h',  云: 'h',
                                  以: 'j',
                                  來: 'l',
  }[音韻地位.母];

  if (母 === undefined) {
    throw new Error(`無聲母規則: ${音韻地位.描述}`);
  }

  // 韻尾

  let 尾 = 規則('韻尾', [
    ['遇果假攝 或 支脂之佳韻', ''],
    ['通江宕梗曾攝', 'ng'],
    ['蟹攝 或 微韻', 'i'], // 已排除佳韻
    ['臻山攝', 'n'],
    ['效流攝', 'u'],
    ['深咸攝', 'm'],
  ]);

  // 主元音

  let 元 = 規則('主元音', [
    ['歌麻　唐庚陽　泰夬廢　寒刪元　豪肴　談銜嚴凡韻', 'a'],
    ['佳支　青耕清　齊皆祭　先山仙　蕭宵　添咸鹽　韻', 'e'],
    ['　脂　　　　　　　　　眞　　　　幽　　　侵　韻', 'i'],
    ['　　　　　　　　　　　臻　　　　　　　　　　韻', 'yi'],
    ['　　　登江蒸　咍灰微　痕魂欣　侯尤　覃　　　韻', 'o'],
    ['　之　　　　　　　　　　　　　　　　　　　　韻', 'io'], // 之韻 io 的 i 視為主元音
    ['模虞　東　　　　　　　　　文　　　　　　　　韻', 'u'],
    ['　魚　冬　鍾　　　　　　　　　　　　　　　　韻', 'v'],
  ]);

  // 介音

  // 拼寫上使用合口介音的場合
  const 合口 = (is('合口') && !is('文韻'))
    || (is('凡韻 或 脣音 歌陽韻'))
    || (is('脣音 一三等')
      && ['a', 'o'].includes(元)
      && !['', 'ng', 'u', 'm'].includes(尾)
      && !is('咍泰韻'))
    || (is('幫組 清韻 重紐B類') && options.清B脣音 === '合口');

  let 介 = {
    一: ['', 'u'],
    二: ['e', 'o'],
    三: ['i', 'v'],
    四: ['', 'u'],
  }[音韻地位.等][Number(合口)];

  if (介 === 'i' && (元.startsWith('i') || 元.startsWith('y'))) {
    介 = '';
  }

  // 二等開口 e 元音寫作 ae
  if (介 === 'e' && 元 === 'e') {
    介 = '';
    元 = 'ae';
  }
  // oe 上聲要雙寫 o，亦視作整體
  if (介 === 'o' && 元 === 'e') {
    介 = '';
    元 = 'oe';
  }

  // 三等重紐、重韻
  if ((is('重紐B類') && (options.清B脣音 !== '合口' || !is('幫組 清韻')))
    || (['i', 'e'].includes(元)
      && is('三等 知莊組 或 云母')
      && !is('知組 清韻 或 知組 眞韻 合口')) // 清、諄韻知組視作A類
    || is('麻庚韻 三等 或 幽韻') // 重韻
  ) {
    介 = {
      i: 'y',
      v: 'w',
      '': '', // 主元音為 i 時
    }[介];
    if (介 === undefined) {
      throw new Error(`無重紐重韻規則: ${音韻地位.描述}`);
    }
    if (元 === 'i') {
      元 = 'y';
    }
  }

  // 拼寫規則

  // 章組、日以母省略 i 介音，麻三亦省略 y 介音
  if (is('章組 或 日以母')) {
    if (介.startsWith('i')) {
      介 = 介.slice(1);
    } else if (is('麻韻 三等') && !合口) {
      介 = '';
    }
  }
  // 莊組拼 ea 省 e，拼 io, iu, iv 省 i（不含之韻），拼 ye 省 y（支韻除外）
  if (is('莊組')) {
    if ((介 === 'e' && 元 === 'a')
      || (介 === 'i' && ['o', 'u', 'v'].includes(元))
      || (介 === 'y' && 元 === 'e' && 尾 !== '')
    ) {
      介 = '';
    }
  }

  // 聲調
  if (音韻地位.聲 === '入') {
    尾 = {
      m: 'p',
      n: 't',
      ng: 'k',
    }[尾];
  } else if (音韻地位.聲 !== '平') {
    if (['i', 'u', 'ng'].includes(尾)) {
      尾 = {
        i: ['j', 'y'],
        u: ['v', 'w'],
        ng: ['nk', 'nq'],
      }[尾][Number(音韻地位.聲 === '去')];
    } else if (音韻地位.聲 === '上') {
      元 = (['io', 'ae', 'oe'].includes(元)) ? 元[0] + 元 : 元 + 元.slice(-1);
    } else if (音韻地位.聲 === '去' && ['m', 'n'].includes(尾)) {
      尾 = 尾 + 尾;
    } else if (音韻地位.聲 === '去' && 尾 === '') {
      元 = 元 + 'h';
    } else {
      throw new Error(`無聲調規則: ${音韻地位.描述}`);
    }
  }
  if (尾 === undefined || 元 === undefined) {
    throw new Error(`無聲調規則: ${音韻地位.描述}`);
  }

  return `${母}${介}${元}${尾}`;
}

/**
 * 古韻羅馬字
 * https://zh.wikipedia.org/wiki/User:Polyhedron/中古漢語拼音
 *
 * 説明
 * 以下內容為生成古韻羅馬字的函數體
 * 函數接受音韻地位，返回對應的古韻羅馬字
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的古韻羅馬字
 */
export function kyonh(音韻地位, 字頭) {
  const is = (x) => 音韻地位.屬於(x);

  function 聲母規則() {
    switch (音韻地位.母) {
      case '幫': return 'p';
      case '滂': return 'ph';
      case '並': return 'b';
      case '明': return 'm';
      case '端': return 't';
      case '透': return 'th';
      case '定': return 'd';
      case '泥': return 'n';
      case '知': return 'tr';
      case '徹': return 'thr';
      case '澄': return 'dr';
      case '孃': return 'nr';
      case '精': return 'c';
      case '清': return 'ch';
      case '從': return 'z';
      case '心': return 's';
      case '邪': return 'zs';
      case '莊': return 'cr';
      case '初': return 'chr';
      case '崇': return 'zr';
      case '生': return 'sr';
      case '俟': return 'zsr';
      case '章': return 'cj';
      case '昌': return 'chj';
      case '船': return 'zsj';
      case '書': return 'sj';
      case '常': return 'zj';
      case '見': return 'k';
      case '溪': return 'kh';
      case '羣': return 'g';
      case '疑': return 'ng';
      case '影': return 'q';
      case '曉': return 'h';
      case '匣': return 'gh';
      case '云': return '';
      case '以': return 'j';
      case '來': return 'l';
      case '日': return 'nj';
      default: throw new Error('無聲母規則');
    }
  }

  function 韻母規則() {
    // 通攝
    if (is('東韻 一等')) return 'ung';
    if (is('東韻 三等')) return 'iung';
    if (is('冬韻')) return 'uung';
    if (is('鍾韻')) return 'yung';
    // 江攝
    if (is('江韻')) return 'rung';
    // 止攝
    if (is('支韻 合口')) return is('重紐A類') ? 'jye' : 'ye';
    if (is('支韻')) return is('重紐A類') ? 'je' : 'ie';
    if (is('脂韻 合口')) return is('重紐A類') ? 'jyi' : 'yi';
    if (is('脂韻')) return is('重紐A類') ? 'jii' : 'ii';
    if (is('之韻')) return 'i';
    if (is('微韻 開口')) return 'ioi';
    if (is('微韻')) return 'yoi';
    // 遇攝
    if (is('魚韻')) return 'io';
    if (is('虞韻')) return 'yo';
    if (is('模韻')) return 'o';
    // 蟹攝
    if (is('齊韻 合口')) return 'ue';
    if (is('齊韻')) return 'e';
    if (is('祭韻 合口')) return 'yed';
    if (is('祭韻')) return is('重紐A類') ? 'jed' : 'ied';
    if (is('泰韻 合口')) return 'uad';
    if (is('泰韻')) return 'ad';
    if (is('佳韻 合口')) return 'rue';
    if (is('佳韻')) return 're';
    if (is('皆韻 合口')) return 'ruai';
    if (is('皆韻')) return 'rai';
    if (is('夬韻 合口')) return 'ruad';
    if (is('夬韻')) return 'rad';
    if (is('咍韻')) return 'ai';
    if (is('灰韻')) return 'uai';
    if (is('廢韻 開口')) return 'iad';
    if (is('廢韻')) return 'yad';
    // 臻攝
    if (is('眞韻 合口')) return is('重紐A類') ? 'jyn' : 'yn';
    if (is('眞韻')) return is('重紐A類') ? 'jin' : 'in';
    if (is('臻韻')) return 'in';
    if (is('欣韻')) return 'ion';
    if (is('文韻')) return 'yon';
    if (is('元韻 開口')) return 'ian';
    if (is('元韻')) return 'yan';
    if (is('痕韻')) return 'on';
    if (is('魂韻')) return 'uon';
    // 山攝
    if (is('寒韻 開口')) return 'an';
    if (is('寒韻')) return 'uan';
    if (is('刪韻 合口')) return 'ruan';
    if (is('刪韻')) return 'ran';
    if (is('山韻 合口')) return 'ruen';
    if (is('山韻')) return 'ren';
    if (is('仙韻 合口')) return is('重紐A類') ? 'jyen' : 'yen';
    if (is('仙韻')) return is('重紐A類') ? 'jen' : 'ien';
    if (is('先韻 合口')) return 'uen';
    if (is('先韻')) return 'en';
    // 效攝
    if (is('蕭韻')) return 'eu';
    if (is('宵韻')) return is('重紐A類') ? 'jeu' : 'ieu';
    if (is('肴韻')) return 'rau';
    if (is('豪韻')) return 'au';
    // 果攝
    if (is('歌韻 一等 開口')) return 'a';
    if (is('歌韻 一等')) return 'ua';
    if (is('歌韻 三等 開口')) return 'ia';
    if (is('歌韻 三等')) return 'ya';
    // 假攝
    if (is('麻韻 二等 合口')) return 'rua';
    if (is('麻韻 二等')) return 'ra';
    if (is('麻韻 三等')) return 'ia';
    // 宕攝
    if (is('陽韻 開口')) return 'iang';
    if (is('陽韻')) return 'yang';
    if (is('唐韻 合口')) return 'uang';
    if (is('唐韻')) return 'ang';
    // 梗攝
    if (is('庚韻 二等 合口')) return 'ruang';
    if (is('庚韻 二等')) return 'rang';
    if (is('庚韻 三等 合口')) return 'yeng';
    if (is('庚韻 三等')) return 'ieng';
    if (is('耕韻 合口')) return 'rueng';
    if (is('耕韻')) return 'reng';
    if (is('清韻 合口')) return is('重紐A類') ? 'jyeng' : 'yeng';
    if (is('清韻')) return is('重紐A類') ? 'jeng' : 'ieng';
    if (is('青韻 合口')) return 'ueng';
    if (is('青韻')) return 'eng';
    // 曾攝
    if (is('蒸韻 合口')) return 'yng';
    if (is('蒸韻')) return 'ing';
    if (is('登韻 合口')) return 'uong';
    if (is('登韻')) return 'ong';
    // 流攝
    if (is('尤韻')) return 'iu';
    if (is('侯韻')) return 'u';
    if (is('幽韻')) return 'y';
    // 深攝
    if (is('侵韻')) return is('重紐A類') ? 'jim' : 'im';
    // 咸攝
    if (is('覃韻')) return 'om';
    if (is('談韻')) return 'am';
    if (is('鹽韻')) return is('重紐A類') ? 'jem' : 'iem';
    if (is('添韻')) return 'em';
    if (is('咸韻')) return 'rem';
    if (is('銜韻')) return 'ram';
    if (is('嚴韻')) return 'iam';
    if (is('凡韻')) return 'yam';
    throw new Error('無韻母規則');
  }

  function 聲調規則() {
    if (is('平入聲')) return '';
    if (is('上聲')) return 'x';
    if (is('去聲')) return 'h';
    throw new Error('無聲調規則');
  }

  let 聲母 = 聲母規則();
  let 隔音符號 = "'";
  let 韻母 = 韻母規則();
  let 聲調 = 聲調規則();

  if (is('入聲')) {
    if (韻母.endsWith('m')) 韻母 = `${韻母.slice(0, -1)}p`;
    else if (韻母.endsWith('n')) 韻母 = `${韻母.slice(0, -1)}t`;
    else if (韻母.endsWith('ng')) 韻母 = `${韻母.slice(0, -2)}k`;
  }

  if (韻母.endsWith('d')) {
    聲調 = '';
  }

  if (聲母.endsWith('r') && 韻母.startsWith('r')) {
    韻母 = 韻母.substr(1);
  }

  if (聲母.endsWith('j') && 韻母.startsWith('i') && [...'aeou'].some((x) => 韻母.includes(x))) {
    韻母 = 韻母.substr(1);
  }

  if (is(`幫組 一二三四等 \
  或 端組 一四等 \
  或 知組 二三等 \
  或 精組 一三四等 \
  或 莊組 二三等 \
  或 章組 三等 \
  或 見溪疑母 一二三四等 \
  或 羣母 二三等 \
  或 影曉母 一二三四等 \
  或 匣母 一二四等 \
  或 云以母 三等 \
  或 來母 一二三四等 \
  或 日母 三等`)) {
    隔音符號 = '';
  }

  if (is('云母 一等')) 聲母 = 'i'; // 1444 倄小韻 i'uaix

  if (is('定母 三等')) 隔音符號 = ''; // 2237 地小韻 diih

  return 聲母 + 隔音符號 + 韻母 + 聲調;
}

/**
 * 隋拼
 * https://github.com/Magnezone462/Zyevio
 *
 * 説明
 * 以下內容為生成隋拼的函數體
 * 函數接受音韻地位，返回對應的隋拼
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的隋拼
 */
export function zyepheng(音韻地位, 字頭) {
  return {
    EAA: 'tuŋ',
    GAA: 'duŋ',
    JBA: 'tryuŋ',
    LBA: 'dryuŋ',
    XBA: 'tyuŋ',
    KBA: 'thryuŋ',
    UBA: 'dzryuŋ',
    QBA: 'syuŋ',
    cBA: 'nyuŋ',
    dBA: 'kyuŋ',
    lBA: 'jyuŋ',
    kBA: 'xyuŋ',
    DBA: 'myuŋ',
    eBA: 'khyuŋ',
    fBA: 'gyuŋ',
    CBA: 'byuŋ',
    ABA: 'pyuŋ',
    BBA: 'phyuŋ',
    YBA: 'thyuŋ',
    IBA: 'lyuŋ',
    eAA: 'khuŋ',
    dAA: 'kuŋ',
    DAA: 'muŋ',
    IAA: 'luŋ',
    jAA: 'xuŋ',
    PAA: 'dzuŋ',
    hAA: 'quŋ',
    OAA: 'chuŋ',
    FAA: 'thuŋ',
    NAA: 'cuŋ',
    CAA: 'buŋ',
    iAA: 'qhuŋ',
    gAA: 'ŋuŋ',
    QAA: 'suŋ',
    ECA: 'toŋ',
    GCA: 'doŋ',
    PCA: 'dzoŋ',
    HCA: 'noŋ',
    dCA: 'koŋ',
    jCA: 'xoŋ',
    ICA: 'loŋ',
    NCA: 'coŋ',
    QCA: 'soŋ',
    FCA: 'thoŋ',
    XDA: 'tyoŋ',
    IDA: 'lyoŋ',
    aDA: 'shyoŋ',
    RDA: 'zyoŋ',
    YDA: 'thyoŋ',
    lDA: 'jyoŋ',
    ADA: 'pyoŋ',
    iDA: 'qhyoŋ',
    gDA: 'ŋyoŋ',
    hDA: 'qyoŋ',
    MDA: 'nryoŋ',
    LDA: 'dryoŋ',
    PDA: 'dzyoŋ',
    KDA: 'thryoŋ',
    CDA: 'byoŋ',
    BDA: 'phyoŋ',
    NDA: 'cyoŋ',
    cDA: 'nyoŋ',
    fDA: 'gyoŋ',
    ZDA: 'dyoŋ',
    dDA: 'kyoŋ',
    QDA: 'syoŋ',
    ODA: 'chyoŋ',
    eDA: 'khyoŋ',
    dEA: 'kroŋ',
    DEA: 'mroŋ',
    MEA: 'nroŋ',
    TEA: 'chroŋ',
    AEA: 'proŋ',
    jEA: 'xroŋ',
    BEA: 'phroŋ',
    IEA: 'lroŋ',
    VEA: 'sroŋ',
    CEA: 'broŋ',
    iEA: 'qhroŋ',
    hEA: 'qroŋ',
    eEA: 'khroŋ',
    LEA: 'droŋ',
    KEA: 'throŋ',
    JEA: 'troŋ',
    gEA: 'ŋroŋ',
    UEA: 'dzroŋ',
    XFA: 'tie',
    lFA: 'jie',
    kFI: 'xrye',
    dFM: 'krye',
    iFM: 'qhrye',
    hFM: 'qrye',
    DFE: 'mrie',
    iFI: 'qhye',
    LFI: 'drye',
    ZFI: 'dye',
    IFI: 'lye',
    YFI: 'thye',
    BFE: 'phrie',
    AFE: 'prie',
    RFI: 'zye',
    eFM: 'khrye',
    eFI: 'khye',
    fFE: 'grie',
    fFA: 'gie',
    iFE: 'qhrie',
    eFE: 'khrie',
    gFE: 'ŋrie',
    CFE: 'brie',
    ZFA: 'die',
    cFA: 'nie',
    IFA: 'lie',
    PFA: 'dzie',
    NFA: 'cie',
    dFE: 'krie',
    AFA: 'pie',
    CFA: 'bie',
    aFA: 'shie',
    QFA: 'sie',
    TFA: 'chrie',
    KFA: 'thrie',
    DFA: 'mie',
    OFA: 'chie',
    JFA: 'trie',
    hFE: 'qrie',
    LFA: 'drie',
    QFI: 'sye',
    gFM: 'ŋrye',
    iFA: 'qhie',
    VFA: 'srie',
    VFI: 'srye',
    cFI: 'nye',
    dFI: 'kye',
    NFI: 'cye',
    TFI: 'chrye',
    JFI: 'trye',
    YFA: 'thie',
    SFA: 'crie',
    BFA: 'phie',
    lFI: 'jye',
    UFA: 'dzrie',
    XGA: 'ti',
    lGA: 'ji',
    VGA: 'sri',
    CGA: 'bi',
    NGA: 'ci',
    dGE: 'kri',
    YGA: 'thi',
    KGA: 'thri',
    OGA: 'chi',
    PGA: 'dzi',
    MGA: 'nri',
    LGA: 'dri',
    QGA: 'si',
    aGA: 'shi',
    hGA: 'qi',
    IGA: 'li',
    fGI: 'gyi',
    JGI: 'tryi',
    dGM: 'kryi',
    cGI: 'nyi',
    VGI: 'sryi',
    lGI: 'jyi',
    IGI: 'lyi',
    QGI: 'syi',
    fGM: 'gryi',
    DGE: 'mryi',
    AGE: 'pryi',
    XGI: 'tyi',
    ZGI: 'dyi',
    kGI: 'xryi',
    CGE: 'bryi',
    BGE: 'phryi',
    iGI: 'qhyi',
    LGI: 'dryi',
    YGI: 'thyi',
    JGA: 'tri',
    BGA: 'phi',
    NGI: 'cyi',
    eGM: 'khryi',
    gGE: 'ŋri',
    iGA: 'qhi',
    XHA: 'tiə',
    lHA: 'jiə',
    ZHA: 'diə',
    gHA: 'ŋiə',
    QHA: 'siə',
    THA: 'chriə',
    fHA: 'giə',
    aHA: 'shiə',
    cHA: 'niə',
    eHA: 'khiə',
    dHA: 'kiə',
    RHA: 'ziə',
    IHA: 'liə',
    SHA: 'criə',
    iHA: 'qhiə',
    hHA: 'qiə',
    KHA: 'thriə',
    LHA: 'driə',
    YHA: 'thiə',
    PHA: 'dziə',
    NHA: 'ciə',
    UHA: 'dzriə',
    WHA: 'zriə',
    DIA: 'myui',
    iII: 'qhyui',
    kII: 'xyui',
    BIA: 'phyui',
    AIA: 'pyui',
    CIA: 'byui',
    hII: 'qyui',
    fIA: 'giəi',
    dIA: 'kiəi',
    iIA: 'qhiəi',
    hIA: 'qiəi',
    gIA: 'ŋiəi',
    gII: 'ŋyui',
    dII: 'kyui',
    eII: 'khyui',
    gJA: 'ŋio',
    TJA: 'chrio',
    aJA: 'shio',
    dJA: 'kio',
    fJA: 'gio',
    lJA: 'jio',
    QJA: 'sio',
    OJA: 'chio',
    UJA: 'dzrio',
    KJA: 'thrio',
    VJA: 'srio',
    iJA: 'qhio',
    RJA: 'zio',
    hJA: 'qio',
    JJA: 'trio',
    IJA: 'lio',
    XJA: 'tio',
    LJA: 'drio',
    cJA: 'nio',
    NJA: 'cio',
    eJA: 'khio',
    SJA: 'crio',
    ZJA: 'dio',
    MJA: 'nrio',
    gKA: 'ŋyo',
    TKA: 'chryo',
    DKA: 'myo',
    kKA: 'xyo',
    iKA: 'qhyo',
    fKA: 'gyo',
    cKA: 'nyo',
    QKA: 'syo',
    JKA: 'tryo',
    KKA: 'thryo',
    ZKA: 'dyo',
    lKA: 'jyo',
    eKA: 'khyo',
    XKA: 'tyo',
    OKA: 'chyo',
    IKA: 'lyo',
    CKA: 'byo',
    UKA: 'dzryo',
    SKA: 'cryo',
    BKA: 'phyo',
    NKA: 'cyo',
    AKA: 'pyo',
    hKA: 'qyo',
    aKA: 'shyo',
    YKA: 'thyo',
    LKA: 'dryo',
    dKA: 'kyo',
    VKA: 'sryo',
    DLA: 'mo',
    CLA: 'bo',
    jLA: 'xo',
    dLA: 'ko',
    GLA: 'do',
    HLA: 'no',
    iLA: 'qho',
    gLA: 'ŋo',
    NLA: 'co',
    ILA: 'lo',
    QLA: 'so',
    PLA: 'dzo',
    hLA: 'qo',
    ALA: 'po',
    eLA: 'kho',
    OLA: 'cho',
    FLA: 'tho',
    ELA: 'to',
    BLA: 'pho',
    PMA: 'dzei',
    IMA: 'lei',
    OMA: 'chei',
    EMA: 'tei',
    GMA: 'dei',
    AMA: 'pei',
    dMA: 'kei',
    jMA: 'xei',
    hMA: 'qei',
    gMA: 'ŋei',
    iMA: 'qhei',
    QMA: 'sei',
    FMA: 'thei',
    CMA: 'bei',
    BMA: 'phei',
    NMA: 'cei',
    DMA: 'mei',
    HMA: 'nei',
    eMA: 'khei',
    dMI: 'kuei',
    eMI: 'khuei',
    jMI: 'xuei',
    cMA: 'nьʼei',
    ZMA: 'dьʼei',
    hMI: 'quei',
    iMI: 'qhuei',
    dPA: 'kre',
    jPA: 'xre',
    CPA: 'bre',
    dPI: 'krue',
    hPI: 'qrue',
    ePI: 'khrue',
    UPA: 'dzre',
    TPA: 'chre',
    iPI: 'qhrue',
    MPA: 'nre',
    gPA: 'ŋre',
    hPA: 'qre',
    VPA: 'sre',
    iPA: 'qhre',
    KPA: 'thre',
    DPA: 'mre',
    jPI: 'xrue',
    dQA: 'krei',
    hQA: 'qrei',
    jQA: 'xrei',
    CQA: 'brei',
    dQI: 'kruei',
    jQI: 'xruei',
    eQI: 'khruei',
    UQA: 'dzrei',
    TQA: 'chrei',
    iQI: 'qhruei',
    LQI: 'druei',
    DQA: 'mrei',
    SQA: 'crei',
    hQI: 'qruei',
    JQA: 'trei',
    IQA: 'lrei',
    eQA: 'khrei',
    MQA: 'nrei',
    VQA: 'srei',
    gQA: 'ŋrei',
    KQA: 'threi',
    iQA: 'qhrei',
    IQI: 'lruei',
    UQI: 'dzruei',
    iSI: 'qhui',
    eSI: 'khui',
    hSI: 'qui',
    jSI: 'xui',
    DSA: 'mui',
    dSI: 'kui',
    ISI: 'lui',
    GSI: 'dui',
    OSI: 'chui',
    ESI: 'tui',
    QSI: 'sui',
    PSI: 'dzui',
    CSA: 'bui',
    ASA: 'pui',
    BSA: 'phui',
    gSI: 'ŋui',
    FSI: 'thui',
    HSI: 'nui',
    NSI: 'cui',
    iTA: 'qhəi',
    eTA: 'khəi',
    hTA: 'qəi',
    GTA: 'dəi',
    dTA: 'kəi',
    PTA: 'dzəi',
    ITA: 'ləi',
    NTA: 'cəi',
    OTA: 'chəi',
    FTA: 'thəi',
    jTA: 'xəi',
    QTA: 'səi',
    gTA: 'ŋəi',
    HTA: 'nəi',
    ETA: 'təi',
    YTA: 'thьʼəi',
    XVA: 'tin',
    KVA: 'thrin',
    hVA: 'qin',
    QVA: 'sin',
    ZVA: 'din',
    cVA: 'nin',
    bVA: 'zhin',
    OVA: 'chin',
    kVA: 'xrin',
    aVA: 'shin',
    AVA: 'pin',
    IVA: 'lin',
    fVE: 'grin',
    JVA: 'trin',
    LVA: 'drin',
    NVA: 'cin',
    YVA: 'thin',
    PVA: 'dzin',
    lVA: 'jin',
    MVA: 'nrin',
    BVA: 'phin',
    CVA: 'bin',
    gVE: 'ŋrin',
    dVE: 'krin',
    dVM: 'kryn',
    kVI: 'xryn',
    eVM: 'khryn',
    DVE: 'mryn',
    CVE: 'bryn',
    hVE: 'qrin',
    hVM: 'qryn',
    AVE: 'pryn',
    DVA: 'min',
    XVI: 'tyn',
    KVI: 'thryn',
    LVI: 'dryn',
    QVI: 'syn',
    ZVI: 'dyn',
    cVI: 'nyn',
    bVI: 'zhyn',
    IVI: 'lyn',
    JVI: 'tryn',
    OVI: 'chyn',
    NVI: 'cyn',
    YVI: 'thyn',
    PVI: 'dzyn',
    lVI: 'jyn',
    RVI: 'zyn',
    dVI: 'kyn',
    fVA: 'gin',
    BVE: 'phryn',
    SWA: 'crin',
    VWA: 'srin',
    UWA: 'dzrin',
    DXA: 'myun',
    kXI: 'xyun',
    hXI: 'qyun',
    CXA: 'byun',
    AXA: 'pyun',
    fXI: 'gyun',
    iXI: 'qhyun',
    dXI: 'kyun',
    BXA: 'phyun',
    iYA: 'qhiən',
    hYA: 'qiən',
    fYA: 'giən',
    dYA: 'kiən',
    gYA: 'ŋiən',
    gZI: 'ŋyon',
    kZI: 'xyon',
    CZA: 'byon',
    BZA: 'phyon',
    iZI: 'qhyon',
    hZI: 'qyon',
    gZA: 'ŋian',
    eZA: 'khian',
    iZA: 'qhian',
    dZA: 'kian',
    hZA: 'qian',
    AZA: 'pyon',
    fZA: 'gian',
    DZA: 'myon',
    jaI: 'xun',
    daI: 'kun',
    haI: 'qun',
    DaA: 'mun',
    QaI: 'sun',
    NaI: 'cun',
    PaI: 'dzun',
    EaI: 'tun',
    FaI: 'thun',
    GaI: 'dun',
    OaI: 'chun',
    gaI: 'ŋun',
    CaA: 'bun',
    AaA: 'pun',
    IaI: 'lun',
    eaI: 'khun',
    iaI: 'qhun',
    BaA: 'phun',
    HaI: 'nun',
    jbA: 'xən',
    dbA: 'kən',
    hbA: 'qən',
    FbA: 'thən',
    gbA: 'ŋən',
    jcA: 'xan',
    gcA: 'ŋan',
    EcA: 'tan',
    hcA: 'qan',
    HcA: 'nan',
    OcA: 'chan',
    FcA: 'than',
    QcA: 'san',
    GcA: 'dan',
    PcA: 'dzan',
    dcA: 'kan',
    IcA: 'lan',
    ecA: 'khan',
    icA: 'qhan',
    HcI: 'nuan',
    jcI: 'xuan',
    gcI: 'ŋuan',
    EcI: 'tuan',
    hcI: 'quan',
    FcI: 'thuan',
    QcI: 'suan',
    GcI: 'duan',
    PcI: 'dzuan',
    dcI: 'kuan',
    IcI: 'luan',
    icI: 'qhuan',
    ecI: 'khuan',
    NcI: 'cuan',
    CcA: 'ban',
    DcA: 'man',
    BcA: 'phan',
    AcA: 'pan',
    VdA: 'sran',
    ddI: 'kruan',
    hdI: 'qruan',
    VdI: 'sruan',
    jdI: 'xruan',
    AdA: 'pran',
    DdA: 'mran',
    gdA: 'ŋran',
    ddA: 'kran',
    BdA: 'phran',
    MdI: 'nruan',
    gdI: 'ŋruan',
    edA: 'khran',
    SdI: 'cruan',
    VeA: 'sren',
    deI: 'kruen',
    deA: 'kren',
    jeA: 'xren',
    eeA: 'khren',
    UeA: 'dzren',
    ieA: 'qhren',
    geA: 'ŋren',
    heA: 'qren',
    AeA: 'pren',
    IeA: 'lren',
    MeA: 'nren',
    YeA: 'thьʼren',
    heI: 'qruen',
    LeI: 'druen',
    IeI: 'lruen',
    feI: 'gruen',
    JeA: 'tren',
    LeA: 'dren',
    jeI: 'xruen',
    QgA: 'sen',
    PgA: 'dzen',
    OgA: 'chen',
    NgA: 'cen',
    FgA: 'then',
    dgA: 'ken',
    jgA: 'xen',
    hgA: 'qen',
    IgA: 'len',
    GgA: 'den',
    HgA: 'nen',
    EgA: 'ten',
    egA: 'khen',
    ggA: 'ŋen',
    DgA: 'men',
    CgA: 'ben',
    hgI: 'quen',
    dgI: 'kuen',
    igI: 'qhuen',
    AgA: 'pen',
    jgI: 'xuen',
    igA: 'qhen',
    UgI: 'dzrʼuen',
    QfA: 'sien',
    PfA: 'dzien',
    OfA: 'chien',
    NfA: 'cien',
    cfA: 'nien',
    lfA: 'jien',
    XfA: 'tien',
    dfA: 'kien',
    JfA: 'trien',
    UfA: 'dzrien',
    afA: 'shien',
    KfA: 'thrien',
    ZfA: 'dien',
    LfA: 'drien',
    ifE: 'qhrien',
    IfA: 'lien',
    BfA: 'phien',
    CfA: 'bien',
    DfA: 'mien',
    PfI: 'dzyen',
    QfI: 'syen',
    NfI: 'cyen',
    ifI: 'qhyen',
    cfI: 'nyen',
    YfI: 'thyen',
    lfI: 'jyen',
    RfI: 'zyen',
    hfI: 'qyen',
    bfI: 'zhyen',
    AfA: 'pien',
    RfA: 'zien',
    OfI: 'chyen',
    XfI: 'tyen',
    ZfI: 'dyen',
    kfI: 'xryen',
    SfI: 'cryen',
    VfI: 'sryen',
    KfI: 'thryen',
    ffE: 'grien',
    efE: 'khrien',
    ffM: 'gryen',
    LfI: 'dryen',
    IfI: 'lyen',
    efM: 'khryen',
    hfE: 'qrien',
    kfA: 'xrien',
    hfM: 'qryen',
    YfA: 'thien',
    JfI: 'tryen',
    dfM: 'kryen',
    QhA: 'seu',
    FhA: 'theu',
    EhA: 'teu',
    GhA: 'deu',
    dhA: 'keu',
    IhA: 'leu',
    ghA: 'ŋeu',
    ihA: 'qheu',
    hhA: 'qeu',
    ehA: 'kheu',
    QiA: 'sieu',
    KiA: 'thrieu',
    JiA: 'trieu',
    LiA: 'drieu',
    iiE: 'qhrieu',
    PiA: 'dzieu',
    diE: 'krieu',
    NiA: 'cieu',
    ciA: 'nieu',
    aiA: 'shieu',
    liA: 'jieu',
    ZiA: 'dieu',
    XiA: 'tieu',
    AiA: 'pieu',
    AiE: 'prieu',
    CiA: 'bieu',
    DiA: 'mieu',
    DiE: 'mrieu',
    hiA: 'qieu',
    kiA: 'xrieu',
    fiE: 'grieu',
    OiA: 'chieu',
    hiE: 'qrieu',
    eiA: 'khieu',
    YiA: 'thieu',
    BiA: 'phieu',
    fiA: 'gieu',
    IiA: 'lieu',
    eiE: 'khrieu',
    jjA: 'xrau',
    djA: 'krau',
    UjA: 'dzrau',
    MjA: 'nrau',
    VjA: 'srau',
    DjA: 'mrau',
    ijA: 'qhrau',
    AjA: 'prau',
    BjA: 'phrau',
    ejA: 'khrau',
    gjA: 'ŋrau',
    SjA: 'crau',
    JjA: 'trau',
    TjA: 'chrau',
    CjA: 'brau',
    hjA: 'qrau',
    KjA: 'thrau',
    IjA: 'lrau',
    LjA: 'drau',
    jkA: 'xau',
    dkA: 'kau',
    IkA: 'lau',
    ikA: 'qhau',
    DkA: 'mau',
    FkA: 'thau',
    EkA: 'tau',
    QkA: 'sau',
    CkA: 'bau',
    AkA: 'pau',
    GkA: 'dau',
    NkA: 'cau',
    gkA: 'ŋau',
    PkA: 'dzau',
    hkA: 'qau',
    HkA: 'nau',
    ekA: 'khau',
    OkA: 'chau',
    BkA: 'phau',
    dlA: 'ka',
    OlA: 'cha',
    ElA: 'ta',
    QlA: 'sa',
    GlA: 'da',
    PlA: 'dza',
    glA: 'ŋa',
    FlA: 'tha',
    IlA: 'la',
    HlA: 'na',
    jlA: 'xa',
    ilA: 'qha',
    elA: 'kha',
    hlA: 'qa',
    dlI: 'kua',
    OlI: 'chua',
    ElI: 'tua',
    QlI: 'sua',
    ClA: 'ba',
    GlI: 'dua',
    DlA: 'ma',
    PlI: 'dzua',
    glI: 'ŋua',
    FlI: 'thua',
    IlI: 'lua',
    HlI: 'nua',
    AlA: 'pa',
    BlA: 'pha',
    jlI: 'xua',
    elI: 'khua',
    hlI: 'qua',
    imI: 'qhya',
    hmI: 'qya',
    emI: 'khya',
    fmA: 'gia',
    emA: 'khia',
    dmA: 'kia',
    OmI: 'chya',
    NmI: 'cya',
    fmI: 'gya',
    ImI: 'lya',
    DnA: 'mra',
    YoA: 'thia',
    aoA: 'shia',
    loA: 'jia',
    XoA: 'tia',
    NoA: 'cia',
    boA: 'zhia',
    jnI: 'xrua',
    dnI: 'krua',
    inI: 'qhrua',
    enI: 'khrua',
    MnA: 'nra',
    dnA: 'kra',
    jnA: 'xra',
    BnA: 'phra',
    hnA: 'qra',
    AnA: 'pra',
    TnA: 'chra',
    VnA: 'sra',
    gnA: 'ŋra',
    SnA: 'cra',
    LnA: 'dra',
    RoA: 'zia',
    ZoA: 'dia',
    hnI: 'qrua',
    SnI: 'crua',
    JnI: 'trua',
    CnA: 'bra',
    UnA: 'dzra',
    KnA: 'thra',
    JnA: 'tra',
    inA: 'qhra',
    enA: 'khra',
    PoA: 'dzia',
    coA: 'nia',
    QoA: 'sia',
    JoA: 'tria',
    gnI: 'ŋrua',
    lpA: 'jiaŋ',
    RpA: 'ziaŋ',
    IpA: 'liaŋ',
    ipA: 'qhiaŋ',
    apA: 'shiaŋ',
    CpA: 'byaŋ',
    XpA: 'tiaŋ',
    YpA: 'thiaŋ',
    epA: 'khiaŋ',
    dpA: 'kiaŋ',
    LpA: 'driaŋ',
    JpA: 'triaŋ',
    cpA: 'niaŋ',
    ApA: 'pyaŋ',
    QpA: 'siaŋ',
    NpA: 'ciaŋ',
    TpA: 'chriaŋ',
    DpA: 'myaŋ',
    MpA: 'nriaŋ',
    UpA: 'dzriaŋ',
    SpA: 'criaŋ',
    ZpA: 'diaŋ',
    VpA: 'sriaŋ',
    PpA: 'dziaŋ',
    OpA: 'chiaŋ',
    epI: 'khyaŋ',
    kpI: 'xyaŋ',
    hpA: 'qiaŋ',
    fpA: 'giaŋ',
    KpA: 'thriaŋ',
    BpA: 'phyaŋ',
    fpI: 'gyaŋ',
    GqA: 'daŋ',
    IqA: 'laŋ',
    EqA: 'taŋ',
    OqA: 'chaŋ',
    dqA: 'kaŋ',
    QqA: 'saŋ',
    eqA: 'khaŋ',
    iqI: 'qhuaŋ',
    jqI: 'xuaŋ',
    dqI: 'kuaŋ',
    FqA: 'thaŋ',
    BqA: 'phaŋ',
    hqI: 'quaŋ',
    hqA: 'qaŋ',
    iqA: 'qhaŋ',
    jqA: 'xaŋ',
    DqA: 'maŋ',
    NqA: 'caŋ',
    HqA: 'naŋ',
    CqA: 'baŋ',
    gqA: 'ŋaŋ',
    PqA: 'dzaŋ',
    eqI: 'khuaŋ',
    AqA: 'paŋ',
    drA: 'kraŋ',
    erA: 'khraŋ',
    DrA: 'mraŋ',
    jrI: 'xruaŋ',
    ArA: 'praŋ',
    irI: 'qhruaŋ',
    drI: 'kruaŋ',
    CrA: 'braŋ',
    irA: 'qhraŋ',
    KrA: 'thraŋ',
    TrA: 'chraŋ',
    UrA: 'dzraŋ',
    hsA: 'qrieŋ',
    BrA: 'phraŋ',
    CsA: 'brieŋ',
    dsA: 'krieŋ',
    DsA: 'mrieŋ',
    LrA: 'draŋ',
    JrA: 'traŋ',
    ksI: 'xryeŋ',
    AsA: 'prieŋ',
    isI: 'qhryeŋ',
    esA: 'khrieŋ',
    fsA: 'grieŋ',
    gsA: 'ŋrieŋ',
    jrA: 'xraŋ',
    MrA: 'nraŋ',
    dtA: 'kreŋ',
    etA: 'khreŋ',
    DtA: 'mreŋ',
    AtA: 'preŋ',
    jtI: 'xrueŋ',
    jtA: 'xreŋ',
    JtA: 'treŋ',
    htA: 'qreŋ',
    UtA: 'dzreŋ',
    TtA: 'chreŋ',
    MtA: 'nreŋ',
    BtA: 'phreŋ',
    itI: 'qhrueŋ',
    LtA: 'dreŋ',
    htI: 'qrueŋ',
    CtA: 'breŋ',
    gtA: 'ŋreŋ',
    StA: 'creŋ',
    OuA: 'chieŋ',
    PuA: 'dzieŋ',
    NuA: 'cieŋ',
    luA: 'jieŋ',
    luI: 'jyeŋ',
    huA: 'qieŋ',
    JuA: 'trieŋ',
    KuA: 'thrieŋ',
    ZuA: 'dieŋ',
    LuA: 'drieŋ',
    auA: 'shieŋ',
    XuA: 'tieŋ',
    euA: 'khieŋ',
    DuA: 'mieŋ',
    IuA: 'lieŋ',
    AuA: 'pieŋ',
    euI: 'khyeŋ',
    RuA: 'zieŋ',
    huI: 'qyeŋ',
    fuI: 'gyeŋ',
    QuI: 'syeŋ',
    fuA: 'gieŋ',
    iuI: 'qhyeŋ',
    OvA: 'cheŋ',
    dvA: 'keŋ',
    jvA: 'xeŋ',
    GvA: 'deŋ',
    EvA: 'teŋ',
    ivA: 'qheŋ',
    QvA: 'seŋ',
    BvA: 'pheŋ',
    IvA: 'leŋ',
    HvA: 'neŋ',
    FvA: 'theŋ',
    DvA: 'meŋ',
    CvA: 'beŋ',
    jvI: 'xueŋ',
    dvI: 'kueŋ',
    XwA: 'tiəŋ',
    ZwA: 'diəŋ',
    LwA: 'driəŋ',
    IwA: 'liəŋ',
    hwA: 'qiəŋ',
    CwA: 'bryəŋ',
    AwA: 'pryəŋ',
    lwA: 'jiəŋ',
    bwA: 'zhiəŋ',
    awA: 'shiəŋ',
    cwA: 'niəŋ',
    dwA: 'kiəŋ',
    JwA: 'triəŋ',
    PwA: 'dziəŋ',
    gwA: 'ŋiəŋ',
    iwA: 'qhiəŋ',
    YwA: 'thiəŋ',
    fwA: 'giəŋ',
    VwA: 'sriəŋ',
    ewA: 'khiəŋ',
    KwA: 'thriəŋ',
    UwA: 'dzriəŋ',
    BwA: 'phryəŋ',
    ExA: 'təŋ',
    IxA: 'ləŋ',
    QxA: 'səŋ',
    AxA: 'pəŋ',
    NxA: 'cəŋ',
    DxA: 'məŋ',
    PxA: 'dzəŋ',
    CxA: 'bəŋ',
    jxI: 'xuəŋ',
    dxI: 'kuəŋ',
    ixI: 'qhuəŋ',
    HxA: 'nəŋ',
    GxA: 'dəŋ',
    jxA: 'xəŋ',
    dxA: 'kəŋ',
    FxA: 'thəŋ',
    BxA: 'phəŋ',
    kyA: 'xyu',
    hyA: 'qyu',
    IyA: 'lyu',
    OyA: 'chyu',
    lyA: 'jyu',
    gyA: 'ŋyu',
    NyA: 'cyu',
    PyA: 'dzyu',
    QyA: 'syu',
    KyA: 'thryu',
    eyA: 'khyu',
    YyA: 'thyu',
    XyA: 'tyu',
    ZyA: 'dyu',
    cyA: 'nyu',
    ayA: 'shyu',
    ByA: 'phyu',
    dyA: 'kyu',
    AyA: 'pyu',
    VyA: 'sryu',
    TyA: 'chryu',
    SyA: 'cryu',
    UyA: 'dzryu',
    iyA: 'qhyu',
    RyA: 'zyu',
    LyA: 'dryu',
    JyA: 'tryu',
    fyA: 'gyu',
    CyA: 'byu',
    DyA: 'myu',
    jzA: 'xu',
    hzA: 'qu',
    HzA: 'nu',
    IzA: 'lu',
    QzA: 'su',
    ezA: 'khu',
    izA: 'qhu',
    NzA: 'cu',
    FzA: 'thu',
    GzA: 'du',
    gzA: 'ŋu',
    dzA: 'ku',
    EzA: 'tu',
    PzA: 'dzu',
    CzA: 'bu',
    OzA: 'chu',
    DzA: 'mu',
    h0A: 'qiu',
    f0A: 'griu',
    A0A: 'piu',
    I0A: 'liu',
    d0A: 'kriu',
    C0A: 'briu',
    N0A: 'ciu',
    V0A: 'sriu',
    g0A: 'ŋiu',
    i0A: 'qhriu',
    D0A: 'miu',
    O1A: 'chim',
    R1A: 'zim',
    I1A: 'lim',
    K1A: 'thrim',
    X1A: 'tim',
    L1A: 'drim',
    J1A: 'trim',
    Z1A: 'dim',
    c1A: 'nim',
    a1A: 'shim',
    l1A: 'jim',
    Q1A: 'sim',
    h1A: 'qim',
    N1A: 'cim',
    P1A: 'dzim',
    M1A: 'nrim',
    f1E: 'grim',
    e1E: 'khrim',
    g1E: 'ŋrim',
    i1E: 'qhrim',
    d1E: 'krim',
    h1E: 'qrim',
    V1A: 'srim',
    U1A: 'dzrim',
    S1A: 'crim',
    T1A: 'chrim',
    Y1A: 'thim',
    G2A: 'dəm',
    O2A: 'chəm',
    H2A: 'nəm',
    h2A: 'qəm',
    j2A: 'xəm',
    I2A: 'ləm',
    P2A: 'dzəm',
    N2A: 'cəm',
    F2A: 'thəm',
    E2A: 'təm',
    e2A: 'khəm',
    i2A: 'qhəm',
    Q2A: 'səm',
    d2A: 'kəm',
    g2A: 'ŋəm',
    G3A: 'dam',
    d3A: 'kam',
    E3A: 'tam',
    Q3A: 'sam',
    I3A: 'lam',
    e3A: 'kham',
    F3A: 'tham',
    P3A: 'dzam',
    j3A: 'xam',
    D3A: 'mam',
    i3A: 'qham',
    l4A: 'jiem',
    I4A: 'liem',
    A4E: 'priem',
    Q4A: 'siem',
    O4A: 'chiem',
    X4A: 'tiem',
    Z4A: 'diem',
    a4A: 'shiem',
    Y4A: 'thiem',
    c4A: 'niem',
    M4A: 'nriem',
    k4A: 'xriem',
    J4A: 'triem',
    K4A: 'thriem',
    h4E: 'qriem',
    e4E: 'khriem',
    g4E: 'ŋriem',
    N4A: 'ciem',
    P4A: 'dziem',
    f4E: 'griem',
    h4A: 'qiem',
    R4A: 'ziem',
    V4A: 'sriem',
    L4A: 'driem',
    f4A: 'giem',
    F5A: 'them',
    E5A: 'tem',
    G5A: 'dem',
    I5A: 'lem',
    e5A: 'khem',
    d5A: 'kem',
    j5A: 'xem',
    H5A: 'nem',
    i5A: 'qhem',
    j6A: 'xrem',
    d6A: 'krem',
    V6A: 'srem',
    h6A: 'qrem',
    g6A: 'ŋrem',
    i6A: 'qhrem',
    J6A: 'trem',
    M6A: 'nrem',
    U6A: 'dzrem',
    e6A: 'khrem',
    j7A: 'xram',
    U7A: 'dzram',
    g7A: 'ŋram',
    T7A: 'chram',
    V7A: 'sram',
    d7A: 'kram',
    C7A: 'bram',
    e7A: 'khram',
    g8A: 'ŋiam',
    i8A: 'qhiam',
    h8A: 'qiam',
    e8A: 'khiam',
    C9A: 'byom',
    B9A: 'phyom',
    EAB: 'túŋ',
    DAB: 'múŋ',
    eAB: 'khúŋ',
    QAB: 'súŋ',
    FAB: 'thúŋ',
    NAB: 'cúŋ',
    jAB: 'xúŋ',
    hAB: 'qúŋ',
    HAB: 'núŋ',
    AAB: 'púŋ',
    IAB: 'lúŋ',
    iAB: 'qhúŋ',
    GAB: 'dúŋ',
    CAB: 'búŋ',
    XDB: 'tyóŋ',
    KDB: 'thryóŋ',
    IDB: 'lyóŋ',
    hDB: 'qyóŋ',
    cDB: 'nyóŋ',
    LDB: 'dryóŋ',
    JDB: 'tryóŋ',
    CDB: 'byóŋ',
    BDB: 'phyóŋ',
    lDB: 'jyóŋ',
    eDB: 'khyóŋ',
    ZDB: 'dyóŋ',
    dDB: 'kyóŋ',
    QDB: 'syóŋ',
    iDB: 'qhyóŋ',
    ADB: 'pyóŋ',
    ECB: 'tóŋ',
    DCB: 'móŋ',
    fDB: 'gyóŋ',
    YDB: 'thyóŋ',
    NDB: 'cyóŋ',
    dEB: 'króŋ',
    CEB: 'bróŋ',
    hEB: 'qróŋ',
    DEB: 'mróŋ',
    jEB: 'xróŋ',
    AEB: 'próŋ',
    iEB: 'qhróŋ',
    XFB: 'tié',
    ZFB: 'dié',
    DFF: 'mrié',
    AFF: 'prié',
    CFF: 'brié',
    iFN: 'qhryé',
    hFN: 'qryé',
    eFN: 'khryé',
    dFN: 'kryé',
    QFJ: 'syé',
    IFJ: 'lyé',
    fFF: 'grié',
    hFF: 'qrié',
    dFF: 'krié',
    eFF: 'khrié',
    gFF: 'ŋrié',
    kFJ: 'xryé',
    NFJ: 'cyé',
    cFJ: 'nyé',
    OFB: 'chié',
    LFB: 'drié',
    QFB: 'sié',
    lFB: 'jié',
    IFB: 'lié',
    VFB: 'srié',
    AFB: 'pié',
    cFB: 'nié',
    DFB: 'mié',
    CFB: 'bié',
    YFB: 'thié',
    aFB: 'shié',
    NFB: 'cié',
    XFJ: 'tyé',
    TFJ: 'chryé',
    RFJ: 'zyé',
    bFB: 'zhié',
    SFB: 'crié',
    BFF: 'phrié',
    BFB: 'phié',
    lFJ: 'jyé',
    PFJ: 'dzyé',
    eFJ: 'khyé',
    MFB: 'nrié',
    gFN: 'ŋryé',
    fFN: 'gryé',
    KFB: 'thrié',
    JFB: 'trié',
    iFF: 'qhrié',
    eFB: 'khié',
    ZFJ: 'dyé',
    dFB: 'kié',
    XGB: 'tí',
    ZGB: 'dí',
    DGF: 'mrýi',
    AGF: 'prýi',
    RGB: 'zí',
    dGF: 'krí',
    NGB: 'cí',
    AGB: 'pí',
    dGN: 'krýi',
    kGJ: 'xrýi',
    aGB: 'shí',
    LGB: 'drí',
    QGB: 'sí',
    CGB: 'bí',
    IGB: 'lí',
    aGJ: 'shýi',
    IGJ: 'lýi',
    fGJ: 'gýi',
    OGJ: 'chýi',
    MGB: 'nrí',
    dGJ: 'kýi',
    CGF: 'brýi',
    PGJ: 'dzýi',
    BGF: 'phrýi',
    cGJ: 'nýi',
    lGJ: 'jýi',
    hGF: 'qrí',
    NGJ: 'cýi',
    JGB: 'trí',
    KGB: 'thrí',
    iGJ: 'qhýi',
    fGN: 'grýi',
    fGF: 'grí',
    eGN: 'khrýi',
    XHB: 'tiə́',
    ZHB: 'diə́',
    JHB: 'triə́',
    iHB: 'qhiə́',
    dHB: 'kiə́',
    lHB: 'jiə́',
    RHB: 'ziə́',
    VHB: 'sriə́',
    cHB: 'niə́',
    IHB: 'liə́',
    QHB: 'siə́',
    aHB: 'shiə́',
    LHB: 'driə́',
    eHB: 'khiə́',
    UHB: 'dzriə́',
    WHB: 'zriə́',
    NHB: 'ciə́',
    kHB: 'xiə́',
    gHB: 'ŋiə́',
    YHB: 'thiə́',
    KHB: 'thriə́',
    THB: 'chriə́',
    SHB: 'criə́',
    hHB: 'qiə́',
    MHB: 'nriə́',
    DIB: 'myúi',
    hIB: 'qiə́i',
    eIB: 'khiə́i',
    dIB: 'kiə́i',
    BIB: 'phyúi',
    AIB: 'pyúi',
    kIJ: 'xyúi',
    dIJ: 'kyúi',
    iIJ: 'qhyúi',
    gIB: 'ŋiə́i',
    iIB: 'qhiə́i',
    hIJ: 'qyúi',
    CIB: 'byúi',
    gJB: 'ŋió',
    IJB: 'lió',
    LJB: 'drió',
    lJB: 'jió',
    XJB: 'tió',
    cJB: 'nió',
    aJB: 'shió',
    YJB: 'thió',
    JJB: 'trió',
    QJB: 'sió',
    KJB: 'thrió',
    MJB: 'nrió',
    iJB: 'qhió',
    fJB: 'gió',
    VJB: 'srió',
    TJB: 'chrió',
    SJB: 'crió',
    UJB: 'dzrió',
    PJB: 'dzió',
    hJB: 'qió',
    dJB: 'kió',
    RJB: 'zió',
    eJB: 'khió',
    bJB: 'zhió',
    ZJB: 'dió',
    OJB: 'chió',
    NJB: 'ció',
    gKB: 'ŋyó',
    kKB: 'xyó',
    PKB: 'dzyó',
    AKB: 'pyó',
    DKB: 'myó',
    CKB: 'byó',
    BKB: 'phyó',
    LKB: 'dryó',
    iKB: 'qhyó',
    ZKB: 'dyó',
    lKB: 'jyó',
    XKB: 'tyó',
    hKB: 'qyó',
    eKB: 'khyó',
    JKB: 'tryó',
    cKB: 'nyó',
    fKB: 'gyó',
    VKB: 'sryó',
    dKB: 'kyó',
    OKB: 'chyó',
    IKB: 'lyó',
    QKB: 'syó',
    UKB: 'dzryó',
    DLB: 'mó',
    FLB: 'thó',
    GLB: 'dó',
    ILB: 'ló',
    OLB: 'chó',
    ELB: 'tó',
    dLB: 'kó',
    gLB: 'ŋó',
    CLB: 'bó',
    PLB: 'dzó',
    NLB: 'có',
    iLB: 'qhó',
    hLB: 'qó',
    eLB: 'khó',
    HLB: 'nó',
    jLB: 'xó',
    BLB: 'phó',
    ALB: 'pó',
    PMB: 'dzéi',
    IMB: 'léi',
    FMB: 'théi',
    BMB: 'phéi',
    NMB: 'céi',
    EMB: 'téi',
    GMB: 'déi',
    HMB: 'néi',
    QMB: 'séi',
    OMB: 'chéi',
    eMB: 'khéi',
    DMB: 'méi',
    CMB: 'béi',
    hMB: 'qéi',
    gMB: 'ŋéi',
    AMB: 'péi',
    jPB: 'xré',
    DPB: 'mré',
    ePB: 'khré',
    LPB: 'dré',
    MPB: 'nré',
    CPB: 'bré',
    hPB: 'qré',
    APB: 'pré',
    dPB: 'kré',
    VPB: 'sré',
    dPJ: 'krué',
    LPJ: 'drué',
    jPJ: 'xrué',
    iPJ: 'qhrué',
    fPB: 'gré',
    jQB: 'xréi',
    eQB: 'khréi',
    gQB: 'ŋréi',
    hQB: 'qréi',
    iSJ: 'qhúi',
    hSJ: 'qúi',
    ISJ: 'lúi',
    GSJ: 'dúi',
    PSJ: 'dzúi',
    DSB: 'múi',
    FSJ: 'thúi',
    jSJ: 'xúi',
    eSJ: 'khúi',
    ESJ: 'túi',
    HSJ: 'núi',
    gSJ: 'ŋúi',
    OSJ: 'chúi',
    CSB: 'búi',
    NSJ: 'cúi',
    iTB: 'qhə́i',
    eTB: 'khə́i',
    NTB: 'cə́i',
    GTB: 'də́i',
    HTB: 'nə́i',
    dTB: 'kə́i',
    jTB: 'xə́i',
    OTB: 'chə́i',
    YTB: 'thьʼə́i',
    ETB: 'tə́i',
    PTB: 'dzə́i',
    hTB: 'qə́i',
    lTB: 'jьʼə́i',
    FTB: 'thə́i',
    cTB: 'nьʼə́i',
    ITB: 'lə́i',
    XVB: 'tín',
    KVB: 'thrín',
    ZVB: 'dín',
    cVB: 'nín',
    aVB: 'shín',
    IVB: 'lín',
    LVB: 'drín',
    dVB: 'kín',
    PVB: 'dzín',
    NVB: 'cín',
    CVB: 'bín',
    gVF: 'ŋrín',
    OVB: 'chín',
    fVN: 'grýn',
    lVB: 'jín',
    DVF: 'mrýn',
    DVB: 'mín',
    kVJ: 'xrýn',
    XVJ: 'týn',
    lVJ: 'jýn',
    QVJ: 'sýn',
    cVJ: 'nýn',
    YVJ: 'thýn',
    bVJ: 'zhýn',
    KVJ: 'thrýn',
    IVJ: 'lýn',
    eVB: 'khín',
    aVJ: 'shýn',
    UVB: 'dzrín',
    JVB: 'trín',
    DXB: 'myún',
    AXB: 'pyún',
    CXB: 'byún',
    BXB: 'phyún',
    hXJ: 'qyún',
    gXJ: 'ŋyún',
    kXJ: 'xyún',
    eXJ: 'khyún',
    hYB: 'qiə́n',
    dYB: 'kiə́n',
    eYB: 'khiə́n',
    fYB: 'giə́n',
    gYB: 'ŋiə́n',
    iYB: 'qhiə́n',
    gZJ: 'ŋyón',
    kZJ: 'xyón',
    hZB: 'qián',
    dZB: 'kián',
    fZB: 'gián',
    eZB: 'khián',
    gZB: 'ŋián',
    iZB: 'qhián',
    DZB: 'myón',
    AZB: 'pyón',
    fZJ: 'gyón',
    hZJ: 'qyón',
    eZJ: 'khyón',
    iZJ: 'qhyón',
    CZB: 'byón',
    jaJ: 'xún',
    BaB: 'phún',
    OaJ: 'chún',
    AaB: 'pún',
    QaJ: 'sún',
    NaJ: 'cún',
    haJ: 'qún',
    GaJ: 'dún',
    PaJ: 'dzún',
    daJ: 'kún',
    FaJ: 'thún',
    eaJ: 'khún',
    IaJ: 'lún',
    CaB: 'bún',
    DaB: 'mún',
    HaJ: 'nún',
    iaJ: 'qhún',
    jbB: 'xə́n',
    ebB: 'khə́n',
    dbB: 'kə́n',
    jcB: 'xán',
    EcB: 'tán',
    FcB: 'thán',
    QcB: 'sán',
    GcB: 'dán',
    PcB: 'dzán',
    dcB: 'kán',
    IcB: 'lán',
    ecB: 'khán',
    icB: 'qhán',
    NcB: 'cán',
    jcJ: 'xuán',
    EcJ: 'tuán',
    hcJ: 'quán',
    FcJ: 'thuán',
    QcJ: 'suán',
    dcJ: 'kuán',
    IcJ: 'luán',
    ecJ: 'khuán',
    HcJ: 'nuán',
    NcJ: 'cuán',
    CcB: 'bán',
    DcB: 'mán',
    AcB: 'pán',
    RcJ: 'zuán',
    GcJ: 'duán',
    HcB: 'nán',
    BcB: 'phán',
    VdB: 'srán',
    hdJ: 'qruán',
    AdB: 'prán',
    SdB: 'crán',
    MdB: 'nrán',
    jdB: 'xrán',
    jdJ: 'xruán',
    CdB: 'brán',
    DdB: 'mrán',
    UdB: 'dzrán',
    gdB: 'ŋrán',
    TdB: 'chrán',
    BdB: 'phrán',
    UdJ: 'dzruán',
    VeB: 'srén',
    jeB: 'xrén',
    deB: 'krén',
    TeB: 'chrén',
    UeB: 'dzrén',
    geB: 'ŋrén',
    SeB: 'crén',
    TeJ: 'chruén',
    eeB: 'khrén',
    QgB: 'sén',
    FgB: 'thén',
    EgB: 'tén',
    hgB: 'qén',
    GgB: 'dén',
    dgB: 'kén',
    jgB: 'xén',
    igB: 'qhén',
    DgB: 'mén',
    HgB: 'nén',
    AgB: 'pén',
    jgJ: 'xuén',
    dgJ: 'kuén',
    CgB: 'bén',
    egJ: 'khuén',
    egB: 'khén',
    ggB: 'ŋén',
    QfB: 'sién',
    lfB: 'jién',
    PfB: 'dzién',
    JfB: 'trién',
    XfB: 'tién',
    MfB: 'nrién',
    OfB: 'chién',
    YfB: 'thién',
    efB: 'khién',
    dfF: 'krién',
    ZfB: 'dién',
    NfB: 'cién',
    cfB: 'nién',
    RfB: 'zién',
    IfB: 'lién',
    gfF: 'ŋrién',
    ffF: 'grién',
    CfF: 'bryén',
    DfB: 'mién',
    AfB: 'pién',
    NfJ: 'cyén',
    PfJ: 'dzyén',
    AfF: 'pryén',
    lfJ: 'jyén',
    IfJ: 'lyén',
    JfJ: 'tryén',
    dfN: 'kryén',
    ffN: 'gryén',
    cfJ: 'nyén',
    YfJ: 'thyén',
    ZfJ: 'dyén',
    LfJ: 'dryén',
    XfJ: 'tyén',
    QfJ: 'syén',
    UfJ: 'dzryén',
    ffJ: 'gyén',
    ifJ: 'qhyén',
    CfB: 'bién',
    DfF: 'mryén',
    KfB: 'thrién',
    BfF: 'phryén',
    afB: 'shién',
    hfF: 'qrién',
    LfB: 'drién',
    UfB: 'dzrién',
    QhB: 'séu',
    dhB: 'kéu',
    EhB: 'téu',
    IhB: 'léu',
    FhB: 'théu',
    ihB: 'qhéu',
    hhB: 'qéu',
    HhB: 'néu',
    jhB: 'xéu',
    GhB: 'déu',
    ehB: 'khéu',
    NhB: 'céu',
    QiB: 'siéu',
    LiB: 'driéu',
    XiB: 'tiéu',
    hiF: 'qriéu',
    KiB: 'thriéu',
    aiB: 'shiéu',
    ciB: 'niéu',
    CiB: 'biéu',
    YiB: 'thiéu',
    BiB: 'phiéu',
    DiB: 'miéu',
    ZiB: 'diéu',
    diF: 'kriéu',
    AiF: 'priéu',
    AiB: 'piéu',
    CiF: 'briéu',
    liB: 'jiéu',
    OiB: 'chiéu',
    NiB: 'ciéu',
    fiF: 'griéu',
    IiB: 'liéu',
    BiF: 'phriéu',
    hiB: 'qiéu',
    ejB: 'khráu',
    jjB: 'xráu',
    AjB: 'práu',
    MjB: 'nráu',
    DjB: 'mráu',
    djB: 'kráu',
    SjB: 'cráu',
    hjB: 'qráu',
    CjB: 'bráu',
    gjB: 'ŋráu',
    UjB: 'dzráu',
    TjB: 'chráu',
    JjB: 'tráu',
    VjB: 'sráu',
    jkB: 'xáu',
    CkB: 'báu',
    IkB: 'láu',
    FkB: 'tháu',
    GkB: 'dáu',
    HkB: 'náu',
    QkB: 'sáu',
    EkB: 'táu',
    OkB: 'cháu',
    NkB: 'cáu',
    PkB: 'dzáu',
    dkB: 'káu',
    ikB: 'qháu',
    DkB: 'máu',
    AkB: 'páu',
    hkB: 'qáu',
    ekB: 'kháu',
    gkB: 'ŋáu',
    dlB: 'ká',
    OlB: 'chá',
    ElB: 'tá',
    QlB: 'sá',
    GlB: 'dá',
    glB: 'ŋá',
    FlB: 'thá',
    IlB: 'lá',
    HlB: 'ná',
    jlB: 'xá',
    ilB: 'qhá',
    elB: 'khá',
    hlB: 'qá',
    NlB: 'cá',
    dlJ: 'kuá',
    ElJ: 'tuá',
    QlJ: 'suá',
    GlJ: 'duá',
    FlJ: 'thuá',
    DlB: 'má',
    PlJ: 'dzuá',
    glJ: 'ŋuá',
    IlJ: 'luá',
    hlJ: 'quá',
    HlJ: 'nuá',
    AlB: 'pá',
    BlB: 'phá',
    jlJ: 'xuá',
    ilJ: 'qhuá',
    elJ: 'khuá',
    ClB: 'bá',
    OlJ: 'chuá',
    DnB: 'mrá',
    XoB: 'tiá',
    loB: 'jiá',
    gnB: 'ŋrá',
    dnB: 'krá',
    VnB: 'srá',
    hnB: 'qrá',
    RoB: 'ziá',
    jnB: 'xrá',
    QoB: 'siá',
    DoB: 'miá',
    OoB: 'chiá',
    inB: 'qhrá',
    ZoB: 'diá',
    enB: 'khrá',
    CnB: 'brá',
    aoB: 'shiá',
    NoB: 'ciá',
    AnB: 'prá',
    jnJ: 'xruá',
    dnJ: 'kruá',
    gnJ: 'ŋruá',
    coB: 'niá',
    SnB: 'crá',
    JnB: 'trá',
    UnB: 'dzrá',
    YoB: 'thiá',
    MnB: 'nrá',
    enJ: 'khruá',
    TnJ: 'chruá',
    KnJ: 'thruá',
    VnJ: 'sruá',
    KnB: 'thrá',
    InB: 'lrá',
    lpB: 'jiáŋ',
    RpB: 'ziáŋ',
    NpB: 'ciáŋ',
    IpB: 'liáŋ',
    hpB: 'qiáŋ',
    fpB: 'giáŋ',
    gpB: 'ŋiáŋ',
    TpB: 'chriáŋ',
    QpB: 'siáŋ',
    XpB: 'tiáŋ',
    VpB: 'sriáŋ',
    ipB: 'qhiáŋ',
    YpB: 'thiáŋ',
    dpB: 'kiáŋ',
    LpB: 'driáŋ',
    KpB: 'thriáŋ',
    dpJ: 'kyáŋ',
    cpB: 'niáŋ',
    apB: 'shiáŋ',
    BpB: 'phyáŋ',
    DpB: 'myáŋ',
    ApB: 'pyáŋ',
    hpJ: 'qyáŋ',
    kpJ: 'xyáŋ',
    ipJ: 'qhyáŋ',
    OpB: 'chiáŋ',
    JpB: 'triáŋ',
    ZpB: 'diáŋ',
    fpJ: 'gyáŋ',
    CpB: 'byáŋ',
    GqB: 'dáŋ',
    QqB: 'sáŋ',
    dqJ: 'kuáŋ',
    AqB: 'páŋ',
    NqB: 'cáŋ',
    HqB: 'náŋ',
    jqB: 'xáŋ',
    FqB: 'tháŋ',
    DqB: 'máŋ',
    EqB: 'táŋ',
    IqB: 'láŋ',
    hqB: 'qáŋ',
    eqB: 'kháŋ',
    hqJ: 'quáŋ',
    jqJ: 'xuáŋ',
    BqB: 'pháŋ',
    iqJ: 'qhuáŋ',
    dqB: 'káŋ',
    gqB: 'ŋáŋ',
    PqB: 'dzáŋ',
    OqB: 'cháŋ',
    iqB: 'qháŋ',
    eqJ: 'khuáŋ',
    drB: 'kráŋ',
    AsB: 'priéŋ',
    dsB: 'kriéŋ',
    hsB: 'qriéŋ',
    VsB: 'sriéŋ',
    ksJ: 'xryéŋ',
    isJ: 'qhryéŋ',
    DsB: 'mriéŋ',
    dsJ: 'kryéŋ',
    jrB: 'xráŋ',
    DrB: 'mráŋ',
    drJ: 'kruáŋ',
    ArB: 'práŋ',
    JrB: 'tráŋ',
    LrB: 'dráŋ',
    hrJ: 'qruáŋ',
    IrB: 'lráŋ',
    jrJ: 'xruáŋ',
    CrB: 'bráŋ',
    erJ: 'khruáŋ',
    MrB: 'nráŋ',
    dtB: 'kréŋ',
    DtB: 'mréŋ',
    jtB: 'xréŋ',
    CtB: 'bréŋ',
    BtB: 'phréŋ',
    PuB: 'dziéŋ',
    XuB: 'tiéŋ',
    KuB: 'thriéŋ',
    luB: 'jiéŋ',
    fuB: 'giéŋ',
    luJ: 'jyéŋ',
    IuB: 'liéŋ',
    duB: 'kiéŋ',
    AuB: 'piéŋ',
    euJ: 'khyéŋ',
    NuB: 'ciéŋ',
    huB: 'qiéŋ',
    OuB: 'chiéŋ',
    QuB: 'siéŋ',
    DuB: 'miéŋ',
    jvJ: 'xuéŋ',
    dvJ: 'kuéŋ',
    DvB: 'méŋ',
    EvB: 'téŋ',
    GvB: 'déŋ',
    FvB: 'théŋ',
    PvB: 'dzéŋ',
    hvJ: 'quéŋ',
    evB: 'khéŋ',
    HvB: 'néŋ',
    jvB: 'xéŋ',
    QvB: 'séŋ',
    AvB: 'péŋ',
    evJ: 'khuéŋ',
    dvB: 'kéŋ',
    BvB: 'phéŋ',
    hvB: 'qéŋ',
    IvB: 'léŋ',
    CvB: 'béŋ',
    ivJ: 'qhuéŋ',
    gvB: 'ŋéŋ',
    XwB: 'tiə́ŋ',
    KwB: 'thriə́ŋ',
    fwB: 'giə́ŋ',
    VwB: 'sriə́ŋ',
    ExB: 'tə́ŋ',
    BxB: 'phə́ŋ',
    exB: 'khə́ŋ',
    HxB: 'nə́ŋ',
    kyB: 'xyú',
    IyB: 'lyú',
    MyB: 'nryú',
    KyB: 'thryú',
    JyB: 'tryú',
    iyB: 'qhyú',
    dyB: 'kyú',
    ayB: 'shyú',
    YyB: 'thyú',
    PyB: 'dzyú',
    CyB: 'byú',
    AyB: 'pyú',
    eyB: 'khyú',
    cyB: 'nyú',
    fyB: 'gyú',
    LyB: 'dryú',
    lyB: 'jyú',
    ZyB: 'dyú',
    QyB: 'syú',
    hyB: 'qyú',
    NyB: 'cyú',
    VyB: 'sryú',
    XyB: 'tyú',
    ByB: 'phyú',
    SyB: 'cryú',
    UyB: 'dzryú',
    TyB: 'chryú',
    jzB: 'xú',
    DzB: 'mú',
    CzB: 'bú',
    EzB: 'tú',
    FzB: 'thú',
    dzB: 'kú',
    gzB: 'ŋú',
    AzB: 'pú',
    HzB: 'nú',
    QzB: 'sú',
    izB: 'qhú',
    BzB: 'phú',
    hzB: 'qú',
    IzB: 'lú',
    NzB: 'cú',
    ezB: 'khú',
    GzB: 'dú',
    OzB: 'chú',
    h0B: 'qríu',
    d0B: 'kíu',
    f0B: 'gríu',
    O1B: 'chím',
    L1B: 'drím',
    e1F: 'khrím',
    I1B: 'lím',
    Q1B: 'sím',
    K1B: 'thrím',
    N1B: 'cím',
    c1B: 'ním',
    X1B: 'tím',
    a1B: 'shím',
    b1B: 'zhím',
    Z1B: 'dím',
    Y1B: 'thím',
    T1B: 'chrím',
    M1B: 'nrím',
    f1F: 'grím',
    d1F: 'krím',
    P1B: 'dzím',
    g1F: 'ŋrím',
    V1B: 'srím',
    A1F: 'prím',
    h1F: 'qrím',
    B1F: 'phrím',
    i1F: 'qhrím',
    J1B: 'trím',
    l1B: 'jím',
    U1B: 'dzrím',
    d2B: 'kə́m',
    G2B: 'də́m',
    h2B: 'qə́m',
    H2B: 'nə́m',
    F2B: 'thə́m',
    P2B: 'dzə́m',
    O2B: 'chə́m',
    g2B: 'ŋə́m',
    N2B: 'cə́m',
    Q2B: 'sə́m',
    e2B: 'khə́m',
    j2B: 'xə́m',
    I2B: 'lə́m',
    E2B: 'tə́m',
    i2B: 'qhə́m',
    d3B: 'kám',
    a3B: 'shьʼám',
    I3B: 'lám',
    F3B: 'thám',
    E3B: 'tám',
    G3B: 'dám',
    O3B: 'chám',
    D3B: 'mám',
    N3B: 'cám',
    P3B: 'dzám',
    i3B: 'qhám',
    h3B: 'qám',
    e3B: 'khám',
    l4B: 'jiém',
    I4B: 'liém',
    i4F: 'qhriém',
    A4F: 'priém',
    X4B: 'tiém',
    e4F: 'khriém',
    g4F: 'ŋriém',
    f4F: 'griém',
    d4F: 'kriém',
    h4B: 'qiém',
    c4B: 'niém',
    a4B: 'shiém',
    K4B: 'thriém',
    h4F: 'qriém',
    P4B: 'dziém',
    O4B: 'chiém',
    N4B: 'ciém',
    e4B: 'khiém',
    Z4B: 'diém',
    F5B: 'thém',
    H5B: 'ném',
    E5B: 'tém',
    G5B: 'dém',
    e5B: 'khém',
    I5B: 'lém',
    j5B: 'xém',
    O5B: 'chém',
    d5B: 'kém',
    D5B: 'mém',
    g8B: 'ŋiám',
    e8B: 'khiám',
    h8B: 'qiám',
    j6B: 'xrém',
    L6B: 'drém',
    e6B: 'khrém',
    d6B: 'krém',
    U6B: 'dzrém',
    I6B: 'lrém',
    S6B: 'crém',
    T6B: 'chrém',
    i6B: 'qhrém',
    V6B: 'srém',
    h6B: 'qrém',
    M6B: 'nrém',
    K6B: 'thrém',
    j7B: 'xrám',
    e7B: 'khrám',
    V7B: 'srám',
    h7B: 'qrám',
    T7B: 'chrám',
    i7B: 'qhrám',
    U7B: 'dzrám',
    C9B: 'byóm',
    D9B: 'myóm',
    A9B: 'pyóm',
    B9B: 'phyóm',
    QAC: 'sûŋ',
    CBC: 'byûŋ',
    dAC: 'kûŋ',
    IAC: 'lûŋ',
    EAC: 'tûŋ',
    eAC: 'khûŋ',
    NAC: 'cûŋ',
    hAC: 'qûŋ',
    OAC: 'chûŋ',
    GAC: 'dûŋ',
    FAC: 'thûŋ',
    LBC: 'dryûŋ',
    ABC: 'pyûŋ',
    eBC: 'khyûŋ',
    DBC: 'myûŋ',
    iBC: 'qhyûŋ',
    DAC: 'mûŋ',
    OBC: 'chyûŋ',
    BBC: 'phyûŋ',
    PAC: 'dzûŋ',
    JBC: 'tryûŋ',
    jAC: 'xûŋ',
    HAC: 'nûŋ',
    XBC: 'tyûŋ',
    YBC: 'thyûŋ',
    UBC: 'dzryûŋ',
    iAC: 'qhûŋ',
    QCC: 'sôŋ',
    NCC: 'côŋ',
    FCC: 'thôŋ',
    DCC: 'môŋ',
    jCC: 'xôŋ',
    lDC: 'jyôŋ',
    RDC: 'zyôŋ',
    CDC: 'byôŋ',
    fDC: 'gyôŋ',
    ADC: 'pyôŋ',
    dDC: 'kyôŋ',
    hDC: 'qyôŋ',
    JDC: 'tryôŋ',
    NDC: 'cyôŋ',
    KDC: 'thryôŋ',
    cDC: 'nyôŋ',
    XDC: 'tyôŋ',
    LDC: 'dryôŋ',
    IDC: 'lyôŋ',
    eDC: 'khyôŋ',
    PDC: 'dzyôŋ',
    MDC: 'nryôŋ',
    dEC: 'krôŋ',
    jEC: 'xrôŋ',
    JEC: 'trôŋ',
    LEC: 'drôŋ',
    KEC: 'thrôŋ',
    UEC: 'dzrôŋ',
    BEC: 'phrôŋ',
    TEC: 'chrôŋ',
    VEC: 'srôŋ',
    XFC: 'tiê',
    CFC: 'biê',
    XFK: 'tyê',
    IFC: 'liê',
    ZFC: 'diê',
    NFC: 'ciê',
    QFC: 'siê',
    kFK: 'xryê',
    dFO: 'kryê',
    BFG: 'phriê',
    AFG: 'priê',
    CFG: 'briê',
    IFK: 'lyê',
    dFG: 'kriê',
    AFC: 'piê',
    fFG: 'griê',
    OFC: 'chiê',
    lFC: 'jiê',
    gFG: 'ŋriê',
    BFC: 'phiê',
    PFC: 'dziê',
    JFC: 'triê',
    hFG: 'qriê',
    LFK: 'dryê',
    YFK: 'thyê',
    iFG: 'qhriê',
    eFC: 'khiê',
    hFC: 'qiê',
    aFC: 'shiê',
    VFC: 'sriê',
    eFK: 'khyê',
    hFO: 'qryê',
    gFO: 'ŋryê',
    iFO: 'qhryê',
    hFK: 'qyê',
    ZFK: 'dyê',
    dFC: 'kiê',
    cFK: 'nyê',
    YFC: 'thiê',
    QFK: 'syê',
    JFK: 'tryê',
    lFK: 'jyê',
    MFK: 'nryê',
    dFK: 'kyê',
    SFC: 'criê',
    iFK: 'qhyê',
    eFG: 'khriê',
    XGC: 'tî',
    kGK: 'xrŷi',
    DGG: 'mrŷi',
    RGK: 'zŷi',
    NGK: 'cŷi',
    QGK: 'sŷi',
    IGK: 'lŷi',
    AGG: 'prŷi',
    fGO: 'grŷi',
    BGG: 'phrŷi',
    CGG: 'brŷi',
    dGO: 'krŷi',
    VGK: 'srŷi',
    eGO: 'khrŷi',
    iGO: 'qhrŷi',
    ZGC: 'dî',
    IGC: 'lî',
    MGC: 'nrî',
    BGC: 'phî',
    gGG: 'ŋrî',
    JGC: 'trî',
    eGC: 'khî',
    LGC: 'drî',
    DGC: 'mî',
    KGC: 'thrî',
    dGG: 'krî',
    fGG: 'grî',
    fGK: 'gŷi',
    OGK: 'chŷi',
    cGC: 'nî',
    NGC: 'cî',
    OGC: 'chî',
    hGG: 'qrî',
    QGC: 'sî',
    eGG: 'khrî',
    dGK: 'kŷi',
    CGC: 'bî',
    iGK: 'qhŷi',
    AGC: 'pî',
    PGK: 'dzŷi',
    GGC: 'dʼî',
    iGG: 'qhrî',
    lGC: 'jî',
    bGC: 'zhî',
    PGC: 'dzî',
    LGK: 'drŷi',
    YGK: 'thŷi',
    lGK: 'jŷi',
    JGK: 'trŷi',
    YGC: 'thî',
    aGC: 'shî',
    TGK: 'chrŷi',
    aGK: 'shŷi',
    XHC: 'tiə̂',
    LHC: 'driə̂',
    RHC: 'ziə̂',
    OHC: 'chiə̂',
    QHC: 'siə̂',
    aHC: 'shiə̂',
    SHC: 'criə̂',
    IHC: 'liə̂',
    PHC: 'dziə̂',
    KHC: 'thriə̂',
    cHC: 'niə̂',
    VHC: 'sriə̂',
    THC: 'chriə̂',
    lHC: 'jiə̂',
    JHC: 'triə̂',
    ZHC: 'diə̂',
    UHC: 'dzriə̂',
    fHC: 'giə̂',
    YHC: 'thiə̂',
    hHC: 'qiə̂',
    dHC: 'kiə̂',
    iHC: 'qhiə̂',
    gHC: 'ŋiə̂',
    eHC: 'khiə̂',
    DIC: 'myûi',
    dIK: 'kyûi',
    kIK: 'xyûi',
    gIK: 'ŋyûi',
    AIC: 'pyûi',
    BIC: 'phyûi',
    eIK: 'khyûi',
    hIK: 'qyûi',
    iIK: 'qhyûi',
    CIC: 'byûi',
    dIC: 'kiə̂i',
    gIC: 'ŋiə̂i',
    eIC: 'khiə̂i',
    iIC: 'qhiə̂i',
    fIC: 'giə̂i',
    hIC: 'qiə̂i',
    gJC: 'ŋiô',
    IJC: 'liô',
    dJC: 'kiô',
    OJC: 'chiô',
    eJC: 'khiô',
    ZJC: 'diô',
    aJC: 'shiô',
    JJC: 'triô',
    XJC: 'tiô',
    VJC: 'sriô',
    hJC: 'qiô',
    LJC: 'driô',
    fJC: 'giô',
    QJC: 'siô',
    UJC: 'dzriô',
    NJC: 'ciô',
    SJC: 'criô',
    cJC: 'niô',
    lJC: 'jiô',
    iJC: 'qhiô',
    MJC: 'nriô',
    TJC: 'chriô',
    YJC: 'thiô',
    KJC: 'thriô',
    RJC: 'ziô',
    gKC: 'ŋyô',
    hKC: 'qyô',
    ZKC: 'dyô',
    LKC: 'dryô',
    CKC: 'byô',
    XKC: 'tyô',
    dKC: 'kyô',
    iKC: 'qhyô',
    aKC: 'shyô',
    lKC: 'jyô',
    cKC: 'nyô',
    BKC: 'phyô',
    DKC: 'myô',
    NKC: 'cyô',
    fKC: 'gyô',
    kKC: 'xyô',
    PKC: 'dzyô',
    VKC: 'sryô',
    AKC: 'pyô',
    OKC: 'chyô',
    JKC: 'tryô',
    eKC: 'khyô',
    TKC: 'chryô',
    KKC: 'thryô',
    QKC: 'syô',
    IKC: 'lyô',
    DLC: 'mô',
    GLC: 'dô',
    ILC: 'lô',
    ELC: 'tô',
    FLC: 'thô',
    dLC: 'kô',
    gLC: 'ŋô',
    jLC: 'xô',
    QLC: 'sô',
    PLC: 'dzô',
    HLC: 'nô',
    ALC: 'pô',
    hLC: 'qô',
    BLC: 'phô',
    OLC: 'chô',
    eLC: 'khô',
    CLC: 'bô',
    iLC: 'qhô',
    NLC: 'cô',
    NMC: 'cêi',
    EMC: 'têi',
    PMC: 'dzêi',
    FMC: 'thêi',
    GMC: 'dêi',
    OMC: 'chêi',
    QMC: 'sêi',
    gMC: 'ŋêi',
    dMC: 'kêi',
    jMC: 'xêi',
    eMC: 'khêi',
    hMC: 'qêi',
    DMC: 'mêi',
    AMC: 'pêi',
    jMK: 'xuêi',
    dMK: 'kuêi',
    iMK: 'qhuêi',
    BMC: 'phêi',
    CMC: 'bêi',
    IMC: 'lêi',
    HMC: 'nêi',
    iMC: 'qhêi',
    NNC: 'ciêi',
    QNK: 'syêi',
    kNK: 'xryêi',
    cNK: 'nyêi',
    XNK: 'tyêi',
    VNK: 'sryêi',
    ONK: 'chyêi',
    lNK: 'jyêi',
    JNK: 'tryêi',
    aNK: 'shyêi',
    CNC: 'biêi',
    TNK: 'chryêi',
    RNK: 'zyêi',
    ANC: 'piêi',
    dNO: 'kryêi',
    DNC: 'miêi',
    VNC: 'sriêi',
    YNC: 'thiêi',
    XNC: 'tiêi',
    ZNC: 'diêi',
    lNC: 'jiêi',
    hNG: 'qriêi',
    gNC: 'ŋiêi',
    LNC: 'driêi',
    INC: 'liêi',
    eNG: 'khriêi',
    aNC: 'shiêi',
    dNG: 'kriêi',
    JNC: 'triêi',
    fNG: 'griêi',
    ZNK: 'dyêi',
    KNC: 'thriêi',
    LNK: 'dryêi',
    gNG: 'ŋriêi',
    NNK: 'cyêi',
    BNC: 'phiêi',
    eUK: 'khyôi',
    iUK: 'qhyôi',
    FOC: 'thâi',
    dOC: 'kâi',
    gOC: 'ŋâi',
    hOC: 'qâi',
    HOC: 'nâi',
    GOC: 'dâi',
    jOC: 'xâi',
    EOC: 'tâi',
    AOC: 'pâi',
    BOC: 'phâi',
    jOK: 'xuâi',
    GOK: 'duâi',
    dOK: 'kuâi',
    NOK: 'cuâi',
    iOK: 'qhuâi',
    eOK: 'khuâi',
    IOK: 'luâi',
    gOK: 'ŋuâi',
    EOK: 'tuâi',
    hOK: 'quâi',
    POK: 'dzuâi',
    QOK: 'suâi',
    OOK: 'chuâi',
    COC: 'bâi',
    eOC: 'khâi',
    OOC: 'châi',
    IOC: 'lâi',
    iOC: 'qhâi',
    FOK: 'thuâi',
    DOC: 'mâi',
    dPK: 'kruê',
    dPC: 'krê',
    hPC: 'qrê',
    jPC: 'xrê',
    DPC: 'mrê',
    jPK: 'xruê',
    TPC: 'chrê',
    gPC: 'ŋrê',
    iPC: 'qhrê',
    APC: 'prê',
    CPC: 'brê',
    UPC: 'dzrê',
    BPC: 'phrê',
    SPC: 'crê',
    ePC: 'khrê',
    VPC: 'srê',
    iPK: 'qhruê',
    dQK: 'kruêi',
    hQC: 'qrêi',
    SQC: 'crêi',
    dQC: 'krêi',
    MQC: 'nrêi',
    iQC: 'qhrêi',
    jQC: 'xrêi',
    gQC: 'ŋrêi',
    eQK: 'khruêi',
    AQC: 'prêi',
    BQC: 'phrêi',
    jQK: 'xruêi',
    gQK: 'ŋruêi',
    CQC: 'brêi',
    DQC: 'mrêi',
    iQK: 'qhruêi',
    VQC: 'srêi',
    eQC: 'khrêi',
    dRK: 'kruâi',
    eRK: 'khruâi',
    DRC: 'mrâi',
    jRK: 'xruâi',
    CRC: 'brâi',
    hRK: 'qruâi',
    TRK: 'chruâi',
    dRC: 'krâi',
    KRC: 'thrâi',
    hRC: 'qrâi',
    VRC: 'srâi',
    LRC: 'drâi',
    iRC: 'qhrâi',
    iRK: 'qhruâi',
    ARC: 'prâi',
    URC: 'dzrâi',
    jRC: 'xrâi',
    GSK: 'dûi',
    CSC: 'bûi',
    DSC: 'mûi',
    BSC: 'phûi',
    iSK: 'qhûi',
    ESK: 'tûi',
    OSK: 'chûi',
    NSK: 'cûi',
    hSK: 'qûi',
    FSK: 'thûi',
    dSK: 'kûi',
    jSK: 'xûi',
    eSK: 'khûi',
    QSK: 'sûi',
    HSK: 'nûi',
    ISK: 'lûi',
    ASC: 'pûi',
    gSK: 'ŋûi',
    GTC: 'də̂i',
    NTC: 'cə̂i',
    QTC: 'sə̂i',
    FTC: 'thə̂i',
    dTC: 'kə̂i',
    eTC: 'khə̂i',
    gTC: 'ŋə̂i',
    hTC: 'qə̂i',
    jTC: 'xə̂i',
    HTC: 'nə̂i',
    ETC: 'tə̂i',
    ITC: 'lə̂i',
    OTC: 'chə̂i',
    PTC: 'dzə̂i',
    iTC: 'qhə̂i',
    AUC: 'pyôi',
    BUC: 'phyôi',
    hUK: 'qyôi',
    CUC: 'byôi',
    fUK: 'gyôi',
    gUC: 'ŋiâi',
    XVC: 'tîn',
    QVC: 'sîn',
    cVC: 'nîn',
    lVC: 'jîn',
    IVC: 'lîn',
    AVC: 'pîn',
    LVC: 'drîn',
    ZVC: 'dîn',
    aVC: 'shîn',
    eVC: 'khrîn',
    RVC: 'zîn',
    gVG: 'ŋrîn',
    NVC: 'cîn',
    iVG: 'qhrîn',
    JVC: 'trîn',
    fVG: 'grîn',
    TVC: 'chrîn',
    hVC: 'qîn',
    KVC: 'thrîn',
    BVC: 'phîn',
    OVC: 'chîn',
    dVK: 'kŷn',
    XVK: 'tŷn',
    QVK: 'sŷn',
    RVK: 'zŷn',
    NVK: 'cŷn',
    aVK: 'shŷn',
    cVK: 'nŷn',
    bVK: 'zhŷn',
    DXC: 'myûn',
    kXK: 'xyûn',
    iXK: 'qhyûn',
    BXC: 'phyûn',
    AXC: 'pyûn',
    hXK: 'qyûn',
    dXK: 'kyûn',
    fXK: 'gyûn',
    CXC: 'byûn',
    iYC: 'qhiə̂n',
    dYC: 'kiə̂n',
    fYC: 'giə̂n',
    hYC: 'qiə̂n',
    gYC: 'ŋiə̂n',
    gZK: 'ŋyôn',
    hZK: 'qyôn',
    AZC: 'pyôn',
    eZK: 'khyôn',
    DZC: 'myôn',
    CZC: 'byôn',
    BZC: 'phyôn',
    dZC: 'kiân',
    hZC: 'qiân',
    iZC: 'qhiân',
    iZK: 'qhyôn',
    fZC: 'giân',
    kZK: 'xyôn',
    gZC: 'ŋiân',
    fZK: 'gyôn',
    dZK: 'kyôn',
    jaK: 'xûn',
    EaK: 'tûn',
    QaK: 'sûn',
    eaK: 'khûn',
    HaK: 'nûn',
    haK: 'qûn',
    DaC: 'mûn',
    PaK: 'dzûn',
    daK: 'kûn',
    BaC: 'phûn',
    GaK: 'dûn',
    OaK: 'chûn',
    CaC: 'bûn',
    gaK: 'ŋûn',
    IaK: 'lûn',
    AaC: 'pûn',
    iaK: 'qhûn',
    NaK: 'cûn',
    jbC: 'xə̂n',
    dbC: 'kə̂n',
    gbC: 'ŋə̂n',
    hbC: 'qə̂n',
    jcC: 'xân',
    FcC: 'thân',
    hcC: 'qân',
    EcC: 'tân',
    GcC: 'dân',
    dcC: 'kân',
    gcC: 'ŋân',
    ecC: 'khân',
    icC: 'qhân',
    IcC: 'lân',
    HcC: 'nân',
    OcC: 'chân',
    QcC: 'sân',
    NcC: 'cân',
    PcC: 'dzân',
    jcK: 'xuân',
    NcK: 'cuân',
    hcK: 'quân',
    dcK: 'kuân',
    OcK: 'chuân',
    gcK: 'ŋuân',
    GcK: 'duân',
    IcK: 'luân',
    EcK: 'tuân',
    FcK: 'thuân',
    icK: 'qhuân',
    QcK: 'suân',
    DcC: 'mân',
    AcC: 'pân',
    BcC: 'phân',
    CcC: 'bân',
    HcK: 'nuân',
    PcK: 'dzuân',
    ecK: 'khuân',
    ddC: 'krân',
    gdC: 'ŋrân',
    hdC: 'qrân',
    VdC: 'srân',
    jdC: 'xrân',
    DdC: 'mrân',
    hdK: 'qruân',
    jdK: 'xruân',
    ddK: 'kruân',
    VdK: 'sruân',
    TdK: 'chruân',
    gdK: 'ŋruân',
    UdC: 'dzrân',
    BdC: 'phrân',
    MdK: 'nruân',
    TdC: 'chrân',
    KdC: 'thrân',
    deC: 'krên',
    jeC: 'xrên',
    CeC: 'brên',
    BeC: 'phrên',
    jeK: 'xruên',
    DeC: 'mrên',
    AeC: 'prên',
    deK: 'kruên',
    QgC: 'sên',
    OgC: 'chên',
    igK: 'qhuên',
    jgK: 'xuên',
    dgK: 'kuên',
    GgC: 'dên',
    FgC: 'thên',
    IgC: 'lên',
    dgC: 'kên',
    HgC: 'nên',
    egC: 'khên',
    jgC: 'xên',
    ggC: 'ŋên',
    hgC: 'qên',
    NgC: 'cên',
    DgC: 'mên',
    BgC: 'phên',
    PgC: 'dzên',
    hgK: 'quên',
    EgC: 'tên',
    igC: 'qhên',
    QfC: 'siên',
    XfC: 'tiên',
    ZfC: 'diên',
    gfG: 'ŋriên',
    efC: 'khiên',
    dfK: 'kyên',
    kfK: 'xryên',
    DfC: 'miên',
    YfK: 'thyên',
    lfK: 'jyên',
    JfC: 'triên',
    cfK: 'nyên',
    NfC: 'ciên',
    YfC: 'thiên',
    afC: 'shiên',
    dfO: 'kryên',
    ffO: 'gryên',
    IfK: 'lyên',
    KfK: 'thryên',
    AfG: 'pryên',
    VfK: 'sryên',
    OfK: 'chyên',
    CfG: 'bryên',
    RfK: 'zyên',
    QfK: 'syên',
    UfK: 'dzryên',
    SfK: 'cryên',
    LfK: 'dryên',
    PfC: 'dziên',
    RfC: 'ziên',
    BfC: 'phiên',
    ZfK: 'dyên',
    MfC: 'nriên',
    JfK: 'tryên',
    lfC: 'jiên',
    CfC: 'biên',
    LfC: 'driên',
    IfC: 'liên',
    efO: 'khryên',
    XfK: 'tyên',
    AgC: 'pên',
    QhC: 'sêu',
    FhC: 'thêu',
    EhC: 'têu',
    dhC: 'kêu',
    HhC: 'nêu',
    GhC: 'dêu',
    ehC: 'khêu',
    IhC: 'lêu',
    ghC: 'ŋêu',
    hhC: 'qêu',
    ihC: 'qhêu',
    QiC: 'siêu',
    XiC: 'tiêu',
    liC: 'jiêu',
    hiC: 'qiêu',
    LiC: 'driêu',
    ZiC: 'diêu',
    fiG: 'griêu',
    BiC: 'phiêu',
    PiC: 'dziêu',
    DiC: 'miêu',
    OiC: 'chiêu',
    IiC: 'liêu',
    eiG: 'khriêu',
    giG: 'ŋriêu',
    NiC: 'ciêu',
    DiG: 'mriêu',
    CiC: 'biêu',
    aiC: 'shiêu',
    KiC: 'thriêu',
    AiG: 'priêu',
    fiC: 'giêu',
    ciC: 'niêu',
    jjC: 'xrâu',
    djC: 'krâu',
    ijC: 'qhrâu',
    JjC: 'trâu',
    AjC: 'prâu',
    ejC: 'khrâu',
    DjC: 'mrâu',
    BjC: 'phrâu',
    KjC: 'thrâu',
    VjC: 'srâu',
    LjC: 'drâu',
    MjC: 'nrâu',
    SjC: 'crâu',
    CjC: 'brâu',
    TjC: 'chrâu',
    hjC: 'qrâu',
    gjC: 'ŋrâu',
    UjC: 'dzrâu',
    jkC: 'xâu',
    GkC: 'dâu',
    EkC: 'tâu',
    dkC: 'kâu',
    gkC: 'ŋâu',
    DkC: 'mâu',
    IkC: 'lâu',
    OkC: 'châu',
    CkC: 'bâu',
    AkC: 'pâu',
    PkC: 'dzâu',
    hkC: 'qâu',
    QkC: 'sâu',
    ekC: 'khâu',
    NkC: 'câu',
    ikC: 'qhâu',
    HkC: 'nâu',
    dlC: 'kâ',
    jlC: 'xâ',
    NlC: 'câ',
    ElC: 'tâ',
    IlC: 'lâ',
    elC: 'khâ',
    glC: 'ŋâ',
    GlC: 'dâ',
    HlC: 'nâ',
    QlC: 'sâ',
    ilC: 'qhâ',
    FlC: 'thâ',
    dlK: 'kuâ',
    jlK: 'xuâ',
    NlK: 'cuâ',
    elK: 'khuâ',
    FlK: 'thuâ',
    AlC: 'pâ',
    OlK: 'chuâ',
    DlC: 'mâ',
    HlK: 'nuâ',
    BlC: 'phâ',
    PlK: 'dzuâ',
    glK: 'ŋuâ',
    ilK: 'qhuâ',
    GlK: 'duâ',
    ClC: 'bâ',
    IlK: 'luâ',
    ElK: 'tuâ',
    QlK: 'suâ',
    hlC: 'qâ',
    hlK: 'quâ',
    DnC: 'mrâ',
    dnC: 'krâ',
    hnC: 'qrâ',
    inC: 'qhrâ',
    gnC: 'ŋrâ',
    KnC: 'thrâ',
    JnC: 'trâ',
    SnC: 'crâ',
    UnC: 'dzrâ',
    RoC: 'ziâ',
    enC: 'khrâ',
    jnC: 'xrâ',
    PoC: 'dziâ',
    loC: 'jiâ',
    YoC: 'thiâ',
    QoC: 'siâ',
    XoC: 'tiâ',
    NoC: 'ciâ',
    aoC: 'shiâ',
    boC: 'zhiâ',
    AnC: 'prâ',
    BnC: 'phrâ',
    jnK: 'xruâ',
    inK: 'qhruâ',
    enK: 'khruâ',
    VnK: 'sruâ',
    CnC: 'brâ',
    MnC: 'nrâ',
    OoC: 'chiâ',
    VnC: 'srâ',
    dnK: 'kruâ',
    LnC: 'drâ',
    gnK: 'ŋruâ',
    hnK: 'qruâ',
    lpC: 'jiâŋ',
    IpC: 'liâŋ',
    UpC: 'dzriâŋ',
    cpC: 'niâŋ',
    apC: 'shiâŋ',
    JpC: 'triâŋ',
    KpC: 'thriâŋ',
    ipC: 'qhiâŋ',
    LpC: 'driâŋ',
    MpC: 'nriâŋ',
    PpC: 'dziâŋ',
    XpC: 'tiâŋ',
    ZpC: 'diâŋ',
    SpC: 'criâŋ',
    hpC: 'qiâŋ',
    fpC: 'giâŋ',
    YpC: 'thiâŋ',
    TpC: 'chriâŋ',
    NpC: 'ciâŋ',
    gpC: 'ŋiâŋ',
    BpC: 'phyâŋ',
    DpC: 'myâŋ',
    ipK: 'qhyâŋ',
    dpK: 'kyâŋ',
    kpK: 'xyâŋ',
    ApC: 'pyâŋ',
    QpC: 'siâŋ',
    dpC: 'kiâŋ',
    epC: 'khiâŋ',
    OpC: 'chiâŋ',
    fpK: 'gyâŋ',
    CpC: 'byâŋ',
    GqC: 'dâŋ',
    IqC: 'lâŋ',
    jqC: 'xâŋ',
    hqC: 'qâŋ',
    gqC: 'ŋâŋ',
    NqC: 'câŋ',
    CqC: 'bâŋ',
    PqC: 'dzâŋ',
    EqC: 'tâŋ',
    eqC: 'khâŋ',
    AqC: 'pâŋ',
    FqC: 'thâŋ',
    eqK: 'khuâŋ',
    HqC: 'nâŋ',
    QqC: 'sâŋ',
    jqK: 'xuâŋ',
    dqK: 'kuâŋ',
    dqC: 'kâŋ',
    iqK: 'qhuâŋ',
    DqC: 'mâŋ',
    hqK: 'quâŋ',
    hsC: 'qriêŋ',
    dsC: 'kriêŋ',
    fsC: 'griêŋ',
    esC: 'khriêŋ',
    drC: 'krâŋ',
    DsC: 'mriêŋ',
    CsC: 'briêŋ',
    DrC: 'mrâŋ',
    jrK: 'xruâŋ',
    AsC: 'priêŋ',
    ksK: 'xryêŋ',
    jrC: 'xrâŋ',
    hrC: 'qrâŋ',
    TsC: 'chriêŋ',
    JrC: 'trâŋ',
    KrC: 'thrâŋ',
    ArC: 'prâŋ',
    LrC: 'drâŋ',
    VsC: 'sriêŋ',
    gsC: 'ŋriêŋ',
    CrC: 'brâŋ',
    irC: 'qhrâŋ',
    hrK: 'qruâŋ',
    StC: 'crêŋ',
    AtC: 'prêŋ',
    CtC: 'brêŋ',
    htC: 'qrêŋ',
    gtC: 'ŋrêŋ',
    itK: 'qhruêŋ',
    duC: 'kiêŋ',
    OuC: 'chiêŋ',
    XuC: 'tiêŋ',
    auC: 'shiêŋ',
    LuC: 'driêŋ',
    KuC: 'thriêŋ',
    QuC: 'siêŋ',
    IuC: 'liêŋ',
    BuC: 'phiêŋ',
    iuK: 'qhyêŋ',
    AuC: 'piêŋ',
    CuC: 'biêŋ',
    PuC: 'dziêŋ',
    ZuC: 'diêŋ',
    DuC: 'miêŋ',
    euC: 'khiêŋ',
    iuC: 'qhiêŋ',
    NuC: 'ciêŋ',
    dvC: 'kêŋ',
    HvC: 'nêŋ',
    QvC: 'sêŋ',
    jvC: 'xêŋ',
    GvC: 'dêŋ',
    EvC: 'têŋ',
    evC: 'khêŋ',
    FvC: 'thêŋ',
    OvC: 'chêŋ',
    DvC: 'mêŋ',
    IvC: 'lêŋ',
    XwC: 'tiə̂ŋ',
    lwC: 'jiə̂ŋ',
    bwC: 'zhiə̂ŋ',
    cwC: 'niə̂ŋ',
    hwC: 'qiə̂ŋ',
    NwC: 'ciə̂ŋ',
    iwC: 'qhiə̂ŋ',
    awC: 'shiə̂ŋ',
    LwC: 'driə̂ŋ',
    CwC: 'bryə̂ŋ',
    YwC: 'thiə̂ŋ',
    gwC: 'ŋiə̂ŋ',
    ZwC: 'diə̂ŋ',
    KwC: 'thriə̂ŋ',
    fwC: 'giə̂ŋ',
    ExC: 'tə̂ŋ',
    PxC: 'dzə̂ŋ',
    dxC: 'kə̂ŋ',
    OxC: 'chə̂ŋ',
    GxC: 'də̂ŋ',
    DxC: 'mə̂ŋ',
    AxC: 'pə̂ŋ',
    CxC: 'bə̂ŋ',
    IxC: 'lə̂ŋ',
    NxC: 'cə̂ŋ',
    QxC: 'sə̂ŋ',
    FxC: 'thə̂ŋ',
    kyC: 'xyû',
    dyC: 'kyû',
    LyC: 'dryû',
    JyC: 'tryû',
    ayC: 'shyû',
    YyC: 'thyû',
    RyC: 'zyû',
    iyC: 'qhyû',
    XyC: 'tyû',
    fyC: 'gyû',
    VyC: 'sryû',
    SyC: 'cryû',
    ByC: 'phyû',
    TyC: 'chryû',
    AyC: 'pyû',
    KyC: 'thryû',
    IyC: 'lyû',
    QyC: 'syû',
    NyC: 'cyû',
    UyC: 'dzryû',
    PyC: 'dzyû',
    MyC: 'nryû',
    CyC: 'byû',
    lyC: 'jyû',
    ZyC: 'dyû',
    cyC: 'nyû',
    DyC: 'myû',
    gyC: 'ŋyû',
    eyC: 'khyû',
    OyC: 'chyû',
    jzC: 'xû',
    ezC: 'khû',
    DzC: 'mû',
    BzC: 'phû',
    GzC: 'dû',
    EzC: 'tû',
    HzC: 'nû',
    QzC: 'sû',
    NzC: 'cû',
    FzC: 'thû',
    hzC: 'qû',
    dzC: 'kû',
    OzC: 'chû',
    IzC: 'lû',
    izC: 'qhû',
    CzC: 'bû',
    gzC: 'ŋû',
    PzC: 'dzû',
    h0C: 'qrîu',
    D0C: 'mrîu',
    e0C: 'khrîu',
    f0C: 'grîu',
    O1C: 'chîm',
    N1C: 'cîm',
    c1C: 'nîm',
    L1C: 'drîm',
    X1C: 'tîm',
    f1G: 'grîm',
    d1G: 'krîm',
    M1C: 'nrîm',
    h1G: 'qrîm',
    V1C: 'srîm',
    K1C: 'thrîm',
    S1C: 'crîm',
    T1C: 'chrîm',
    g1G: 'ŋrîm',
    J1C: 'trîm',
    I1C: 'lîm',
    Z1C: 'dîm',
    k1C: 'xrîm',
    a1C: 'shîm',
    e2C: 'khə̂m',
    d2C: 'kə̂m',
    j2C: 'xə̂m',
    h2C: 'qə̂m',
    F2C: 'thə̂m',
    Q2C: 'sə̂m',
    O2C: 'chə̂m',
    G2C: 'də̂m',
    g2C: 'ŋə̂m',
    E2C: 'tə̂m',
    H2C: 'nə̂m',
    I2C: 'lə̂m',
    i2C: 'qhə̂m',
    N2C: 'cə̂m',
    e3C: 'khâm',
    I3C: 'lâm',
    F3C: 'thâm',
    d3C: 'kâm',
    i3C: 'qhâm',
    j3C: 'xâm',
    G3C: 'dâm',
    P3C: 'dzâm',
    E3C: 'tâm',
    Q3C: 'sâm',
    l4C: 'jiêm',
    Z4C: 'diêm',
    c4C: 'niêm',
    h4C: 'qiêm',
    A4G: 'priêm',
    g4G: 'ŋriêm',
    a4C: 'shiêm',
    N4C: 'ciêm',
    O4C: 'chiêm',
    I4C: 'liêm',
    K4C: 'thriêm',
    Y4C: 'thiêm',
    h4G: 'qriêm',
    P4C: 'dziêm',
    X4C: 'tiêm',
    F5C: 'thêm',
    H5C: 'nêm',
    E5C: 'têm',
    Q5C: 'sêm',
    G5C: 'dêm',
    d5C: 'kêm',
    h5C: 'qêm',
    N5C: 'cêm',
    P5C: 'dzêm',
    e5C: 'khêm',
    I5C: 'lêm',
    g8C: 'ŋiâm',
    i8C: 'qhiâm',
    e8C: 'khiâm',
    D9C: 'myôm',
    j6C: 'xrêm',
    h6C: 'qrêm',
    S6C: 'crêm',
    J6C: 'trêm',
    e6C: 'khrêm',
    L6C: 'drêm',
    U6C: 'dzrêm',
    d6C: 'krêm',
    M6C: 'nrêm',
    g6C: 'ŋrêm',
    d7C: 'krâm',
    T7C: 'chrâm',
    S7C: 'crâm',
    V7C: 'srâm',
    i7C: 'qhrâm',
    C7C: 'brâm',
    j7C: 'xrâm',
    U7C: 'dzrâm',
    h7C: 'qrâm',
    C9C: 'byôm',
    B9C: 'phyôm',
    d8C: 'kiâm',
    h8C: 'qiâm',
    hAD: 'quk',
    GAD: 'duk',
    dAD: 'kuk',
    jAD: 'xuk',
    eAD: 'khuk',
    FAD: 'thuk',
    EAD: 'tuk',
    QAD: 'suk',
    IAD: 'luk',
    iAD: 'qhuk',
    PAD: 'dzuk',
    OAD: 'chuk',
    NAD: 'cuk',
    CAD: 'buk',
    BAD: 'phuk',
    AAD: 'puk',
    DAD: 'muk',
    ABD: 'pyuk',
    CBD: 'byuk',
    VBD: 'sryuk',
    IBD: 'lyuk',
    LBD: 'dryuk',
    dBD: 'kyuk',
    eBD: 'khyuk',
    ZBD: 'dyuk',
    YBD: 'thyuk',
    lBD: 'jyuk',
    fBD: 'gyuk',
    OBD: 'chyuk',
    cBD: 'nyuk',
    XBD: 'tyuk',
    aBD: 'shyuk',
    iBD: 'qhyuk',
    JBD: 'tryuk',
    NBD: 'cyuk',
    TBD: 'chryuk',
    MBD: 'nryuk',
    SBD: 'cryuk',
    BBD: 'phyuk',
    hBD: 'qyuk',
    QBD: 'syuk',
    DBD: 'myuk',
    kBD: 'xyuk',
    KBD: 'thryuk',
    gBD: 'ŋyuk',
    PBD: 'dzyuk',
    hCD: 'qok',
    GCD: 'dok',
    ECD: 'tok',
    eCD: 'khok',
    jCD: 'xok',
    CCD: 'bok',
    QCD: 'sok',
    dCD: 'kok',
    DCD: 'mok',
    iCD: 'qhok',
    HCD: 'nok',
    NCD: 'cok',
    ACD: 'pok',
    ICD: 'lok',
    gCD: 'ŋok',
    XDD: 'tyok',
    gDD: 'ŋyok',
    iDD: 'qhyok',
    dDD: 'kyok',
    fDD: 'gyok',
    ZDD: 'dyok',
    YDD: 'thyok',
    cDD: 'nyok',
    aDD: 'shyok',
    lDD: 'jyok',
    LDD: 'dryok',
    IDD: 'lyok',
    eDD: 'khyok',
    JDD: 'tryok',
    NDD: 'cyok',
    bDD: 'zhyok',
    CDD: 'byok',
    ODD: 'chyok',
    RDD: 'zyok',
    QDD: 'syok',
    ADD: 'pyok',
    KDD: 'thryok',
    dED: 'krok',
    gED: 'ŋrok',
    UED: 'dzrok',
    SED: 'crok',
    VED: 'srok',
    JED: 'trok',
    AED: 'prok',
    DED: 'mrok',
    CED: 'brok',
    BED: 'phrok',
    eED: 'khrok',
    LED: 'drok',
    hED: 'qrok',
    MED: 'nrok',
    KED: 'throk',
    IED: 'lrok',
    jED: 'xrok',
    iED: 'qhrok',
    TED: 'chrok',
    XVD: 'tit',
    cVD: 'nit',
    bVD: 'zhit',
    LVD: 'drit',
    QVD: 'sit',
    hVD: 'qit',
    OVD: 'chit',
    BVD: 'phit',
    dVD: 'kit',
    MVD: 'nrit',
    lVD: 'jit',
    eVD: 'khit',
    iVD: 'qhit',
    KVD: 'thrit',
    IVD: 'lit',
    JVD: 'trit',
    PVD: 'dzit',
    TVD: 'chrit',
    aVD: 'shit',
    NVD: 'cit',
    DVD: 'mit',
    AVD: 'pit',
    fVH: 'grit',
    CVD: 'bit',
    kVL: 'xryt',
    VVL: 'sryt',
    YVD: 'thit',
    UVD: 'dzrit',
    DVH: 'mryt',
    CVH: 'bryt',
    hVH: 'qrit',
    gVH: 'ŋrit',
    AVH: 'pryt',
    JVL: 'tryt',
    iVH: 'qhrit',
    dVH: 'krit',
    iVL: 'qhyt',
    bVL: 'zhyt',
    dVL: 'kyt',
    PVL: 'dzyt',
    lVL: 'jyt',
    NVL: 'cyt',
    QVL: 'syt',
    IVL: 'lyt',
    KVL: 'thryt',
    LVL: 'dryt',
    YVL: 'thyt',
    OVL: 'chyt',
    SVL: 'cryt',
    SWD: 'crit',
    VWD: 'srit',
    UWD: 'dzrit',
    DXD: 'myut',
    AXD: 'pyut',
    hXL: 'qyut',
    dXL: 'kyut',
    eXL: 'khyut',
    fXL: 'gyut',
    CXD: 'byut',
    iXL: 'qhyut',
    kXL: 'xyut',
    BXD: 'phyut',
    gXL: 'ŋyut',
    iYD: 'qhiət',
    dYD: 'kiət',
    gYD: 'ŋiət',
    fYD: 'giət',
    eYD: 'khiət',
    gZL: 'ŋyot',
    CZD: 'byot',
    kZL: 'xyot',
    dZL: 'kyot',
    hZL: 'qyot',
    fZL: 'gyot',
    eZL: 'khyot',
    AZD: 'pyot',
    DZD: 'myot',
    iZL: 'qhyot',
    hZD: 'qiat',
    iZD: 'qhiat',
    dZD: 'kiat',
    fZD: 'giat',
    BZD: 'phyot',
    gZD: 'ŋiat',
    DaD: 'mut',
    FaL: 'thut',
    daL: 'kut',
    CaD: 'but',
    EaL: 'tut',
    GaL: 'dut',
    haL: 'qut',
    iaL: 'qhut',
    gaL: 'ŋut',
    BaD: 'phut',
    IaL: 'lut',
    eaL: 'khut',
    HaL: 'nut',
    QaL: 'sut',
    OaL: 'chut',
    PaL: 'dzut',
    jbD: 'xət',
    jaL: 'xut',
    NaL: 'cut',
    jcD: 'xat',
    icD: 'qhat',
    EcD: 'tat',
    FcD: 'that',
    hcD: 'qat',
    IcD: 'lat',
    ecD: 'khat',
    GcD: 'dat',
    PcD: 'dzat',
    gcD: 'ŋat',
    dcD: 'kat',
    QcD: 'sat',
    OcD: 'chat',
    HcD: 'nat',
    DcD: 'mat',
    gcL: 'ŋuat',
    PcL: 'dzuat',
    AcD: 'pat',
    dcL: 'kuat',
    ecL: 'khuat',
    jcL: 'xuat',
    GcL: 'duat',
    icL: 'qhuat',
    hcL: 'quat',
    NcL: 'cuat',
    BcD: 'phat',
    FcL: 'thuat',
    IcL: 'luat',
    EcL: 'tuat',
    OcL: 'chuat',
    CcD: 'bat',
    jeD: 'xret',
    SeD: 'cret',
    CeD: 'bret',
    eeD: 'khret',
    jeL: 'xruet',
    AeD: 'pret',
    JeL: 'truet',
    heL: 'qruet',
    MeL: 'nruet',
    TeD: 'chret',
    deL: 'kruet',
    deD: 'kret',
    heD: 'qret',
    VeD: 'sret',
    DeD: 'mret',
    ieL: 'qhruet',
    MeD: 'nret',
    geL: 'ŋruet',
    eeL: 'khruet',
    BeD: 'phret',
    SeL: 'cruet',
    jdD: 'xrat',
    hdD: 'qrat',
    gdD: 'ŋrat',
    TdD: 'chrat',
    edD: 'khrat',
    idD: 'qhrat',
    KdD: 'thrat',
    ddL: 'kruat',
    jdL: 'xruat',
    KdL: 'thruat',
    JdL: 'truat',
    VdL: 'sruat',
    gdL: 'ŋruat',
    TdL: 'chruat',
    DdD: 'mrat',
    MdL: 'nruat',
    AdD: 'prat',
    ddD: 'krat',
    UdD: 'dzrat',
    JdD: 'trat',
    cdD: 'nьʼrat',
    QgD: 'set',
    OgD: 'chet',
    dgD: 'ket',
    NgD: 'cet',
    igL: 'qhuet',
    egL: 'khuet',
    dgL: 'kuet',
    jgL: 'xuet',
    hgL: 'quet',
    GgD: 'det',
    FgD: 'thet',
    jgD: 'xet',
    HgD: 'net',
    PgD: 'dzet',
    ggD: 'ŋet',
    DgD: 'met',
    AgD: 'pet',
    hgD: 'qet',
    egD: 'khet',
    igD: 'qhet',
    BgD: 'phet',
    CgD: 'bet',
    EgD: 'tet',
    IgD: 'let',
    QfD: 'siet',
    IfD: 'liet',
    JfD: 'triet',
    ffH: 'griet',
    cfD: 'niet',
    XfD: 'tiet',
    bfD: 'zhiet',
    ZfD: 'diet',
    gfH: 'ŋriet',
    DfD: 'miet',
    efH: 'khriet',
    AfD: 'piet',
    PfL: 'dzyet',
    NfL: 'cyet',
    QfL: 'syet',
    KfL: 'thryet',
    lfL: 'jyet',
    efL: 'khyet',
    hfP: 'qryet',
    cfL: 'nyet',
    afL: 'shyet',
    XfL: 'tyet',
    YfL: 'thyet',
    JfL: 'tryet',
    IfL: 'lyet',
    BfD: 'phiet',
    CfH: 'bryet',
    LfD: 'driet',
    AfH: 'pryet',
    VfL: 'sryet',
    afD: 'shiet',
    MfL: 'nryet',
    hfL: 'qyet',
    dfP: 'kryet',
    SfL: 'cryet',
    OfL: 'chyet',
    VfD: 'sriet',
    hfH: 'qriet',
    KfD: 'thriet',
    ifH: 'qhriet',
    RfL: 'zyet',
    YfD: 'thiet',
    lfD: 'jiet',
    UfD: 'dzriet',
    lpD: 'jiak',
    IpD: 'liak',
    dpD: 'kiak',
    XpD: 'tiak',
    apD: 'shiak',
    cpD: 'niak',
    YpD: 'thiak',
    hpD: 'qiak',
    epD: 'khiak',
    gpD: 'ŋiak',
    ZpD: 'diak',
    KpD: 'thriak',
    QpD: 'siak',
    SpD: 'criak',
    NpD: 'ciak',
    PpD: 'dziak',
    OpD: 'chiak',
    fpD: 'giak',
    hpL: 'qyak',
    CpD: 'byak',
    ipL: 'qhyak',
    kpL: 'xyak',
    dpL: 'kyak',
    JpD: 'triak',
    LpD: 'driak',
    epL: 'khyak',
    fpL: 'gyak',
    MpD: 'nriak',
    BpD: 'phyak',
    ipD: 'qhiak',
    GqD: 'dak',
    DqD: 'mak',
    IqD: 'lak',
    FqD: 'thak',
    NqD: 'cak',
    OqD: 'chak',
    dqD: 'kak',
    eqD: 'khak',
    gqD: 'ŋak',
    BqD: 'phak',
    hqD: 'qak',
    CqD: 'bak',
    iqD: 'qhak',
    QqD: 'sak',
    jqD: 'xak',
    PqD: 'dzak',
    AqD: 'pak',
    HqD: 'nak',
    iqL: 'qhuak',
    dqL: 'kuak',
    hqL: 'quak',
    jqL: 'xuak',
    eqL: 'khuak',
    gqL: 'ŋuak',
    IqL: 'luak',
    NqL: 'cuak',
    DrD: 'mrak',
    JrD: 'trak',
    CrD: 'brak',
    ArD: 'prak',
    fsD: 'griek',
    dsD: 'kriek',
    VsD: 'sriek',
    TsD: 'chriek',
    jrL: 'xruak',
    SrD: 'crak',
    UrD: 'dzrak',
    esD: 'khriek',
    grD: 'ŋrak',
    gsD: 'ŋriek',
    erD: 'khrak',
    hrD: 'qrak',
    isD: 'qhriek',
    KrD: 'thrak',
    BrD: 'phrak',
    irD: 'qhrak',
    jrD: 'xrak',
    drD: 'krak',
    irL: 'qhruak',
    LrD: 'drak',
    drL: 'kruak',
    hrL: 'qruak',
    erL: 'khruak',
    CsD: 'briek',
    DtD: 'mrek',
    jtL: 'xruek',
    dtL: 'kruek',
    AtD: 'prek',
    CtD: 'brek',
    UtD: 'dzrek',
    StD: 'crek',
    UtL: 'dzruek',
    TtD: 'chrek',
    etD: 'khrek',
    itL: 'qhruek',
    jtD: 'xrek',
    dtD: 'krek',
    JtD: 'trek',
    htD: 'qrek',
    VtD: 'srek',
    BtD: 'phrek',
    gtD: 'ŋrek',
    ItD: 'lrek',
    StL: 'cruek',
    VtL: 'sruek',
    ftL: 'gruek',
    MtD: 'nrek',
    QuD: 'siek',
    NuD: 'ciek',
    huD: 'qiek',
    luD: 'jiek',
    auD: 'shiek',
    YuD: 'thiek',
    ZuD: 'diek',
    XuD: 'tiek',
    LuD: 'driek',
    OuD: 'chiek',
    RuD: 'ziek',
    PuD: 'dziek',
    CuD: 'biek',
    luL: 'jyek',
    iuL: 'qhyek',
    AuD: 'piek',
    BuD: 'phiek',
    buD: 'zhiek',
    AuH: 'priek',
    JuD: 'triek',
    KuD: 'thriek',
    XuL: 'tyek',
    OuL: 'chyek',
    QvD: 'sek',
    dvD: 'kek',
    BvD: 'phek',
    IvD: 'lek',
    EvD: 'tek',
    jvD: 'xek',
    gvD: 'ŋek',
    GvD: 'dek',
    FvD: 'thek',
    NvD: 'cek',
    evD: 'khek',
    HvD: 'nek',
    PvD: 'dzek',
    DvD: 'mek',
    CvD: 'bek',
    AvD: 'pek',
    evL: 'khuek',
    dvL: 'kuek',
    OvD: 'chek',
    ivD: 'qhek',
    ivL: 'qhuek',
    XwD: 'tiək',
    LwD: 'driək',
    IwD: 'liək',
    KwD: 'thriək',
    JwD: 'triək',
    bwD: 'zhiək',
    QwD: 'siək',
    ZwD: 'diək',
    awD: 'shiək',
    iwD: 'qhiək',
    UwD: 'dzriək',
    fwD: 'giək',
    MwD: 'nriək',
    TwD: 'chriək',
    hwD: 'qiək',
    VwD: 'sriək',
    ewD: 'khiək',
    dwD: 'kiək',
    lwD: 'jiək',
    NwD: 'ciək',
    AwD: 'pryək',
    kwL: 'xryək',
    iwL: 'qhryək',
    BwD: 'phryək',
    SwD: 'criək',
    CwD: 'bryək',
    gwD: 'ŋiək',
    PwD: 'dziək',
    DwD: 'mryək',
    YwD: 'thiək',
    ExD: 'tək',
    NxD: 'cək',
    IxD: 'lək',
    FxD: 'thək',
    exD: 'khək',
    GxD: 'dək',
    ixD: 'qhək',
    DxD: 'mək',
    PxD: 'dzək',
    QxD: 'sək',
    AxD: 'pək',
    CxD: 'bək',
    jxL: 'xuək',
    dxL: 'kuək',
    hxD: 'qək',
    jxD: 'xək',
    HxD: 'nək',
    dxD: 'kək',
    OxD: 'chək',
    BxD: 'phək',
    ixL: 'qhuək',
    O1D: 'chip',
    Z1D: 'dip',
    X1D: 'tip',
    R1D: 'zip',
    P1D: 'dzip',
    c1D: 'nip',
    h1D: 'qip',
    a1D: 'ship',
    N1D: 'cip',
    f1H: 'grip',
    L1D: 'drip',
    J1D: 'trip',
    I1D: 'lip',
    d1H: 'krip',
    g1H: 'ŋrip',
    e1H: 'khrip',
    Q1D: 'sip',
    V1D: 'srip',
    i1H: 'qhrip',
    S1D: 'crip',
    h1H: 'qrip',
    K1D: 'thrip',
    k1D: 'xrip',
    l1D: 'jip',
    M1D: 'nrip',
    U1D: 'dzrip',
    C1H: 'brip',
    A1H: 'prip',
    T1D: 'chrip',
    Y1D: 'thip',
    j2D: 'xəp',
    d2D: 'kəp',
    E2D: 'təp',
    Q2D: 'səp',
    G2D: 'dəp',
    F2D: 'thəp',
    P2D: 'dzəp',
    N2D: 'cəp',
    I2D: 'ləp',
    H2D: 'nəp',
    e2D: 'khəp',
    h2D: 'qəp',
    i2D: 'qhəp',
    g2D: 'ŋəp',
    O2D: 'chəp',
    j3D: 'xap',
    I3D: 'lap',
    E3D: 'tap',
    F3D: 'thap',
    i3D: 'qhap',
    H3D: 'nap',
    G3D: 'dap',
    Q3D: 'sap',
    g3D: 'ŋap',
    P3D: 'dzap',
    d3D: 'kap',
    e3D: 'khap',
    h3D: 'qap',
    O3D: 'chap',
    X3D: 'tьʼap',
    l4D: 'jiep',
    N4D: 'ciep',
    a4D: 'shiep',
    Z4D: 'diep',
    I4D: 'liep',
    P4D: 'dziep',
    L4D: 'driep',
    h4H: 'qriep',
    M4D: 'nriep',
    Y4D: 'thiep',
    c4D: 'niep',
    X4D: 'tiep',
    O4D: 'chiep',
    K4D: 'thriep',
    f4H: 'griep',
    J4D: 'triep',
    k4D: 'xriep',
    V4D: 'sriep',
    d4H: 'kriep',
    h4D: 'qiep',
    F5D: 'thep',
    j5D: 'xep',
    d5D: 'kep',
    e5D: 'khep',
    G5D: 'dep',
    H5D: 'nep',
    Q5D: 'sep',
    I5D: 'lep',
    E5D: 'tep',
    P5D: 'dzep',
    N5D: 'cep',
    i5D: 'qhep',
    j6D: 'xrep',
    e6D: 'khrep',
    U6D: 'dzrep',
    d6D: 'krep',
    S6D: 'crep',
    T6D: 'chrep',
    M6D: 'nrep',
    i6D: 'qhrep',
    V6D: 'srep',
    J6D: 'trep',
    h6D: 'qrep',
    g6D: 'ŋrep',
    K6D: 'threp',
    j7D: 'xrap',
    h7D: 'qrap',
    d7D: 'krap',
    V7D: 'srap',
    i7D: 'qhrap',
    g8D: 'ŋiap',
    i8D: 'qhiap',
    e8D: 'khiap',
    d8D: 'kiap',
    h8D: 'qiap',
    l8D: 'jiap',
    f8D: 'giap',
    C9D: 'byop',
    A9D: 'pyop',
    B9D: 'phyop',
  }[音韻地位.編碼] || '?';
}

/**
 * 潘悟雲擬音
 *
 * 説明
 * 以下內容為生成潘悟雲擬音的函數體
 * 函數接受音韻地位，返回對應的潘悟雲擬音
 * 部分音資料暫缺，以問號表示
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的潘悟雲擬音
 */
export function panwuyun(音韻地位, 字頭) {
  return {
    EAA: 'tuŋ˧',
    GAA: 'duŋ˧',
    JBA: 'ʈiuŋ˧',
    LBA: 'ɖiuŋ˧',
    XBA: 'tɕiuŋ˧',
    KBA: 'ʈʰiuŋ˧',
    UBA: 'dʐiuŋ˧',
    QBA: 'siuŋ˧',
    cBA: 'ȵiuŋ˧',
    dBA: 'kiuŋ˧',
    lBA: 'jiuŋ˧',
    kBA: 'ɦiuŋ˧',
    DBA: 'miuŋ˧',
    eBA: 'kʰiuŋ˧',
    fBA: 'giuŋ˧',
    CBA: 'biuŋ˧',
    ABA: 'piuŋ˧',
    BBA: 'pʰiuŋ˧',
    YBA: 'tɕʰiuŋ˧',
    IBA: 'liuŋ˧',
    eAA: 'kʰuŋ˧',
    dAA: 'kuŋ˧',
    DAA: 'muŋ˧',
    IAA: 'luŋ˧',
    jAA: 'ɣuŋ˧',
    PAA: 'dzuŋ˧',
    hAA: 'ʔuŋ˧',
    OAA: 'tsʰuŋ˧',
    FAA: 'tʰuŋ˧',
    NAA: 'tsuŋ˧',
    CAA: 'buŋ˧',
    iAA: 'xuŋ˧',
    gAA: 'ŋuŋ˧',
    QAA: 'suŋ˧',
    ECA: 'tuoŋ˧',
    GCA: 'duoŋ˧',
    PCA: 'dzuoŋ˧',
    HCA: 'nuoŋ˧',
    dCA: 'kuoŋ˧',
    jCA: 'ɣuoŋ˧',
    ICA: 'luoŋ˧',
    NCA: 'tsuoŋ˧',
    QCA: 'suoŋ˧',
    FCA: 'tʰuoŋ˧',
    XDA: 'tɕioŋ˧',
    IDA: 'lioŋ˧',
    aDA: 'ɕioŋ˧',
    RDA: 'zioŋ˧',
    YDA: 'tɕʰioŋ˧',
    lDA: 'jioŋ˧',
    ADA: 'pioŋ˧',
    iDA: 'xioŋ˧',
    gDA: 'ŋioŋ˧',
    hDA: 'ʔioŋ˧',
    MDA: 'ɳioŋ˧',
    LDA: 'ɖioŋ˧',
    PDA: 'dzioŋ˧',
    KDA: 'ʈʰioŋ˧',
    CDA: 'bioŋ˧',
    BDA: 'pʰioŋ˧',
    NDA: 'tsioŋ˧',
    cDA: 'ȵioŋ˧',
    fDA: 'gioŋ˧',
    ZDA: 'dʑioŋ˧',
    dDA: 'kioŋ˧',
    QDA: 'sioŋ˧',
    ODA: 'tsʰioŋ˧',
    eDA: 'kʰioŋ˧',
    dEA: 'kɯɔŋ˧',
    DEA: 'mɯɔŋ˧',
    MEA: 'ɳɯɔŋ˧',
    TEA: 'tʂʰɯɔŋ˧',
    AEA: 'pɯɔŋ˧',
    jEA: 'ɣɯɔŋ˧',
    BEA: 'pʰɯɔŋ˧',
    IEA: 'lɯɔŋ˧',
    VEA: 'ʂɯɔŋ˧',
    CEA: 'bɯɔŋ˧',
    iEA: 'xɯɔŋ˧',
    hEA: 'ʔɯɔŋ˧',
    eEA: 'kʰɯɔŋ˧',
    LEA: 'ɖɯɔŋ˧',
    KEA: 'ʈʰɯɔŋ˧',
    JEA: 'ʈɯɔŋ˧',
    gEA: 'ŋɯɔŋ˧',
    UEA: 'dʐɯɔŋ˧',
    XFA: 'tɕiɛ˧',
    lFA: 'jiɛ˧',
    kFI: 'ɦʷiɛ˧',
    dFM: 'kʷɯiɛ˧',
    iFM: 'xʷɯiɛ˧',
    hFM: 'ʔʷɯiɛ˧',
    DFE: 'mɯiɛ˧',
    iFI: 'xʷiɛ˧',
    LFI: 'ɖʷiɛ˧',
    ZFI: 'dʑʷiɛ˧',
    IFI: 'lʷiɛ˧',
    YFI: 'tɕʰʷiɛ˧',
    BFE: 'pʰɯiɛ˧',
    AFE: 'pɯiɛ˧',
    RFI: 'zʷiɛ˧',
    eFM: 'kʰʷɯiɛ˧',
    eFI: 'kʰʷiɛ˧',
    fFE: 'gɯiɛ˧',
    fFA: 'giɛ˧',
    iFE: 'xɯiɛ˧',
    eFE: 'kʰɯiɛ˧',
    gFE: 'ŋɯiɛ˧',
    CFE: 'bɯiɛ˧',
    ZFA: 'dʑiɛ˧',
    cFA: 'ȵiɛ˧',
    IFA: 'liɛ˧',
    PFA: 'dziɛ˧',
    NFA: 'tsiɛ˧',
    dFE: 'kɯiɛ˧',
    AFA: 'piɛ˧',
    CFA: 'biɛ˧',
    aFA: 'ɕiɛ˧',
    QFA: 'siɛ˧',
    TFA: 'tʂʰiɛ˧',
    KFA: 'ʈʰiɛ˧',
    DFA: 'miɛ˧',
    OFA: 'tsʰiɛ˧',
    JFA: 'ʈiɛ˧',
    hFE: 'ʔɯiɛ˧',
    LFA: 'ɖiɛ˧',
    QFI: 'sʷiɛ˧',
    gFM: 'ŋʷɯiɛ˧',
    iFA: 'xiɛ˧',
    VFA: 'ʂiɛ˧',
    VFI: 'ʂʷiɛ˧',
    cFI: 'ȵʷiɛ˧',
    dFI: 'kʷiɛ˧',
    NFI: 'tsʷiɛ˧',
    TFI: 'tʂʰʷiɛ˧',
    JFI: 'ʈʷiɛ˧',
    YFA: 'tɕʰiɛ˧',
    SFA: 'tʂiɛ˧',
    BFA: 'pʰiɛ˧',
    lFI: 'jʷiɛ˧',
    UFA: 'dʐiɛ˧',
    XGA: 'tɕi˧',
    lGA: 'ji˧',
    VGA: 'ʂi˧',
    CGA: 'bi˧',
    NGA: 'tsi˧',
    dGE: 'kɯi˧',
    YGA: 'tɕʰi˧',
    KGA: 'ʈʰi˧',
    OGA: 'tsʰi˧',
    PGA: 'dzi˧',
    MGA: 'ɳi˧',
    LGA: 'ɖi˧',
    QGA: 'si˧',
    aGA: 'ɕi˧',
    fGE: 'gɯi˧',
    hGA: 'ʔi˧',
    IGA: 'li˧',
    fGI: 'gʷi˧',
    JGI: 'ʈʷi˧',
    dGM: 'kʷɯi˧',
    cGI: 'ȵʷi˧',
    VGI: 'ʂʷi˧',
    lGI: 'jʷi˧',
    IGI: 'lʷi˧',
    QGI: 'sʷi˧',
    fGM: 'gʷɯi˧',
    DGE: 'mɯi˧',
    AGE: 'pɯi˧',
    XGI: 'tɕʷi˧',
    ZGI: 'dʑʷi˧',
    kGI: 'ɦʷi˧',
    CGE: 'bɯi˧',
    BGE: 'pʰɯi˧',
    iGI: 'xʷi˧',
    LGI: 'ɖʷi˧',
    YGI: 'tɕʰʷi˧',
    JGA: 'ʈi˧',
    BGA: 'pʰi˧',
    NGI: 'tsʷi˧',
    eGM: 'kʰʷɯi˧',
    gGE: 'ŋɯi˧',
    iGA: 'xi˧',
    XHA: 'tɕɨ˧',
    lHA: 'jɨ˧',
    ZHA: 'dʑɨ˧',
    gHA: 'ŋɨ˧',
    QHA: 'sɨ˧',
    THA: 'tʂʰɨ˧',
    fHA: 'gɨ˧',
    aHA: 'ɕɨ˧',
    cHA: 'ȵɨ˧',
    eHA: 'kʰɨ˧',
    dHA: 'kɨ˧',
    RHA: 'zɨ˧',
    IHA: 'lɨ˧',
    SHA: 'tʂɨ˧',
    iHA: 'xɨ˧',
    hHA: 'ʔɨ˧',
    KHA: 'ʈʰɨ˧',
    LHA: 'ɖɨ˧',
    YHA: 'tɕʰɨ˧',
    PHA: 'dzɨ˧',
    NHA: 'tsɨ˧',
    UHA: 'dʐɨ˧',
    WHA: 'ʐɨ˧',
    DIA: 'mʷɨi˧',
    iII: 'xʷɨi˧',
    kII: 'ɦʷɨi˧',
    BIA: 'pʰʷɨi˧',
    AIA: 'pʷɨi˧',
    CIA: 'bʷɨi˧',
    hII: 'ʔʷɨi˧',
    fIA: 'gɨi˧',
    dIA: 'kɨi˧',
    iIA: 'xɨi˧',
    hIA: 'ʔɨi˧',
    gIA: 'ŋɨi˧',
    gII: 'ŋʷɨi˧',
    dII: 'kʷɨi˧',
    eII: 'kʰʷɨi˧',
    gJA: 'ŋiɔ˧',
    TJA: 'tʂʰiɔ˧',
    aJA: 'ɕiɔ˧',
    dJA: 'kiɔ˧',
    fJA: 'giɔ˧',
    lJA: 'jiɔ˧',
    QJA: 'siɔ˧',
    OJA: 'tsʰiɔ˧',
    UJA: 'dʐiɔ˧',
    KJA: 'ʈʰiɔ˧',
    VJA: 'ʂiɔ˧',
    iJA: 'xiɔ˧',
    RJA: 'ziɔ˧',
    hJA: 'ʔiɔ˧',
    JJA: 'ʈiɔ˧',
    IJA: 'liɔ˧',
    XJA: 'tɕiɔ˧',
    LJA: 'ɖiɔ˧',
    cJA: 'ȵiɔ˧',
    NJA: 'tsiɔ˧',
    eJA: 'kʰiɔ˧',
    SJA: 'tʂiɔ˧',
    ZJA: 'dʑiɔ˧',
    MJA: 'ɳiɔ˧',
    gKA: 'ŋio˧',
    TKA: 'tʂʰio˧',
    DKA: 'mio˧',
    kKA: 'ɦio˧',
    iKA: 'xio˧',
    fKA: 'gio˧',
    cKA: 'ȵio˧',
    QKA: 'sio˧',
    JKA: 'ʈio˧',
    KKA: 'ʈʰio˧',
    ZKA: 'dʑio˧',
    lKA: 'jio˧',
    eKA: 'kʰio˧',
    XKA: 'tɕio˧',
    OKA: 'tsʰio˧',
    IKA: 'lio˧',
    CKA: 'bio˧',
    UKA: 'dʐio˧',
    SKA: 'tʂio˧',
    BKA: 'pʰio˧',
    NKA: 'tsio˧',
    AKA: 'pio˧',
    hKA: 'ʔio˧',
    aKA: 'ɕio˧',
    YKA: 'tɕʰio˧',
    LKA: 'ɖio˧',
    dKA: 'kio˧',
    VKA: 'ʂio˧',
    DLA: 'muo˧',
    CLA: 'buo˧',
    jLA: 'ɣuo˧',
    dLA: 'kuo˧',
    GLA: 'duo˧',
    HLA: 'nuo˧',
    iLA: 'xuo˧',
    gLA: 'ŋuo˧',
    NLA: 'tsuo˧',
    ILA: 'luo˧',
    QLA: 'suo˧',
    PLA: 'dzuo˧',
    hLA: 'ʔuo˧',
    ALA: 'puo˧',
    eLA: 'kʰuo˧',
    OLA: 'tsʰuo˧',
    FLA: 'tʰuo˧',
    ELA: 'tuo˧',
    BLA: 'pʰuo˧',
    PMA: 'dzei˧',
    IMA: 'lei˧',
    OMA: 'tsʰei˧',
    EMA: 'tei˧',
    GMA: 'dei˧',
    AMA: 'pei˧',
    dMA: 'kei˧',
    jMA: 'ɣei˧',
    hMA: 'ʔei˧',
    gMA: 'ŋei˧',
    iMA: 'xei˧',
    QMA: 'sei˧',
    FMA: 'tʰei˧',
    CMA: 'bei˧',
    BMA: 'pʰei˧',
    NMA: 'tsei˧',
    DMA: 'mei˧',
    HMA: 'nei˧',
    eMA: 'kʰei˧',
    dMI: 'kʷei˧',
    eMI: 'kʰʷei˧',
    jMI: 'ɣʷei˧',
    cMA: 'ȵei˧',
    ZMA: 'dʑei˧',
    hMI: 'ʔʷei˧',
    iMI: 'xʷei˧',
    dPA: 'kɯæ˧',
    jPA: 'ɣɯæ˧',
    CPA: 'bɯæ˧',
    dPI: 'kʷɯæ˧',
    hPI: 'ʔʷɯæ˧',
    ePI: 'kʰʷɯæ˧',
    UPA: 'dʐɯæ˧',
    TPA: 'tʂʰɯæ˧',
    iPI: 'xʷɯæ˧',
    MPA: 'ɳɯæ˧',
    gPA: 'ŋɯæ˧',
    hPA: 'ʔɯæ˧',
    VPA: 'ʂɯæ˧',
    iPA: 'xɯæ˧',
    KPA: 'ʈʰɯæ˧',
    DPA: 'mɯæ˧',
    jPI: 'ɣʷɯæ˧',
    dQA: 'kɯæi˧',
    hQA: 'ʔɯæi˧',
    jQA: 'ɣɯæi˧',
    CQA: 'bɯæi˧',
    dQI: 'kʷɯæi˧',
    jQI: 'ɣʷɯæi˧',
    eQI: 'kʰʷɯæi˧',
    UQA: 'dʐɯæi˧',
    TQA: 'tʂʰɯæi˧',
    iQI: 'xʷɯæi˧',
    LQI: 'ɖʷɯæi˧',
    DQA: 'mɯæi˧',
    SQA: 'tʂɯæi˧',
    hQI: 'ʔʷɯæi˧',
    JQA: 'ʈɯæi˧',
    IQA: 'lɯæi˧',
    eQA: 'kʰɯæi˧',
    MQA: 'ɳɯæi˧',
    VQA: 'ʂɯæi˧',
    gQA: 'ŋɯæi˧',
    KQA: 'ʈʰɯæi˧',
    iQA: 'xɯæi˧',
    IQI: 'lʷɯæi˧',
    UQI: 'dʐʷɯæi˧',
    iSI: 'xuoi˧',
    eSI: 'kʰuoi˧',
    hSI: 'ʔuoi˧',
    jSI: 'ɣuoi˧',
    DSA: 'muoi˧',
    dSI: 'kuoi˧',
    ISI: 'luoi˧',
    GSI: 'duoi˧',
    OSI: 'tsʰuoi˧',
    ESI: 'tuoi˧',
    QSI: 'suoi˧',
    PSI: 'dzuoi˧',
    CSA: 'buoi˧',
    ASA: 'puoi˧',
    BSA: 'pʰuoi˧',
    gSI: 'ŋuoi˧',
    FSI: 'tʰuoi˧',
    HSI: 'nuoi˧',
    NSI: 'tsuoi˧',
    iTA: 'xəi˧',
    eTA: 'kʰəi˧',
    hTA: 'ʔəi˧',
    GTA: 'dəi˧',
    dTA: 'kəi˧',
    PTA: 'dzəi˧',
    ITA: 'ləi˧',
    NTA: 'tsəi˧',
    OTA: 'tsʰəi˧',
    CTA: 'bəi˧',
    FTA: 'tʰəi˧',
    jTA: 'ɣəi˧',
    QTA: 'səi˧',
    gTA: 'ŋəi˧',
    HTA: 'nəi˧',
    ETA: 'təi˧',
    YTA: 'tɕʰəi˧',
    BTA: 'pʰəi˧',
    XVA: 'tɕin˧',
    KVA: 'ʈʰin˧',
    hVA: 'ʔin˧',
    QVA: 'sin˧',
    ZVA: 'dʑin˧',
    cVA: 'ȵin˧',
    bVA: 'ʑin˧',
    OVA: 'tsʰin˧',
    kVA: 'ɦin˧',
    aVA: 'ɕin˧',
    AVA: 'pin˧',
    IVA: 'lin˧',
    fVE: 'gɯin˧',
    JVA: 'ʈin˧',
    LVA: 'ɖin˧',
    NVA: 'tsin˧',
    YVA: 'tɕʰin˧',
    PVA: 'dzin˧',
    lVA: 'jin˧',
    MVA: 'ɳin˧',
    BVA: 'pʰin˧',
    CVA: 'bin˧',
    gVE: 'ŋɯin˧',
    dVE: 'kɯin˧',
    dVM: 'kʷɯin˧',
    kVI: 'ɦʷin˧',
    eVM: 'kʰʷɯin˧',
    DVE: 'mɯin˧',
    CVE: 'bɯin˧',
    hVE: 'ʔɯin˧',
    hVM: 'ʔʷɯin˧',
    AVE: 'pɯin˧',
    DVA: 'min˧',
    XVI: 'tɕʷin˧',
    KVI: 'ʈʰʷin˧',
    LVI: 'ɖʷin˧',
    QVI: 'sʷin˧',
    ZVI: 'dʑʷin˧',
    cVI: 'ȵʷin˧',
    bVI: 'ʑʷin˧',
    IVI: 'lʷin˧',
    JVI: 'ʈʷin˧',
    OVI: 'tsʰʷin˧',
    NVI: 'tsʷin˧',
    YVI: 'tɕʰʷin˧',
    PVI: 'dzʷin˧',
    lVI: 'jʷin˧',
    RVI: 'zʷin˧',
    dVI: 'kʷin˧',
    fVA: 'gin˧',
    BVE: 'pʰɯin˧',
    SWA: 'tʂɪn˧',
    VWA: 'ʂɪn˧',
    UWA: 'dʐɪn˧',
    DXA: 'miun˧',
    kXI: 'ɦiun˧',
    hXI: 'ʔiun˧',
    CXA: 'biun˧',
    AXA: 'piun˧',
    fXI: 'giun˧',
    iXI: 'xiun˧',
    dXI: 'kiun˧',
    BXA: 'pʰiun˧',
    iYA: 'xɨn˧',
    hYA: 'ʔɨn˧',
    fYA: 'gɨn˧',
    dYA: 'kɨn˧',
    gYA: 'ŋɨn˧',
    gZI: 'ŋʷiɐn˧',
    kZI: 'ɦʷiɐn˧',
    CZA: 'bʷiɐn˧',
    BZA: 'pʰʷiɐn˧',
    iZI: 'xʷiɐn˧',
    hZI: 'ʔʷiɐn˧',
    gZA: 'ŋiɐn˧',
    eZA: 'kʰiɐn˧',
    iZA: 'xiɐn˧',
    dZA: 'kiɐn˧',
    hZA: 'ʔiɐn˧',
    AZA: 'pʷiɐn˧',
    fZA: 'giɐn˧',
    DZA: 'mʷiɐn˧',
    jaI: 'ɣuon˧',
    daI: 'kuon˧',
    haI: 'ʔuon˧',
    DaA: 'muon˧',
    QaI: 'suon˧',
    NaI: 'tsuon˧',
    PaI: 'dzuon˧',
    EaI: 'tuon˧',
    FaI: 'tʰuon˧',
    GaI: 'duon˧',
    OaI: 'tsʰuon˧',
    gaI: 'ŋuon˧',
    CaA: 'buon˧',
    AaA: 'puon˧',
    IaI: 'luon˧',
    eaI: 'kʰuon˧',
    iaI: 'xuon˧',
    BaA: 'pʰuon˧',
    HaI: 'nuon˧',
    jbA: 'ɣən˧',
    dbA: 'kən˧',
    hbA: 'ʔən˧',
    FbA: 'tʰən˧',
    gbA: 'ŋən˧',
    jcA: 'ɣɑn˧',
    gcA: 'ŋɑn˧',
    EcA: 'tɑn˧',
    hcA: 'ʔɑn˧',
    HcA: 'nɑn˧',
    OcA: 'tsʰɑn˧',
    FcA: 'tʰɑn˧',
    QcA: 'sɑn˧',
    GcA: 'dɑn˧',
    PcA: 'dzɑn˧',
    dcA: 'kɑn˧',
    IcA: 'lɑn˧',
    ecA: 'kʰɑn˧',
    icA: 'xɑn˧',
    HcI: 'nʷɑn˧',
    jcI: 'ɣʷɑn˧',
    gcI: 'ŋʷɑn˧',
    EcI: 'tʷɑn˧',
    hcI: 'ʔʷɑn˧',
    FcI: 'tʰʷɑn˧',
    QcI: 'sʷɑn˧',
    GcI: 'dʷɑn˧',
    PcI: 'dzʷɑn˧',
    dcI: 'kʷɑn˧',
    IcI: 'lʷɑn˧',
    icI: 'xʷɑn˧',
    ecI: 'kʰʷɑn˧',
    NcI: 'tsʷɑn˧',
    CcA: 'bʷɑn˧',
    DcA: 'mʷɑn˧',
    BcA: 'pʰʷɑn˧',
    AcA: 'pʷɑn˧',
    VdA: 'ʂɯan˧',
    ddI: 'kʷɯan˧',
    hdI: 'ʔʷɯan˧',
    VdI: 'ʂʷɯan˧',
    jdI: 'ɣʷɯan˧',
    AdA: 'pɯan˧',
    DdA: 'mɯan˧',
    gdA: 'ŋɯan˧',
    ddA: 'kɯan˧',
    BdA: 'pʰɯan˧',
    MdI: 'ɳʷɯan˧',
    gdI: 'ŋʷɯan˧',
    edA: 'kʰɯan˧',
    SdI: 'tʂʷɯan˧',
    VeA: 'ʂɯæn˧',
    deI: 'kʷɯæn˧',
    deA: 'kɯæn˧',
    jeA: 'ɣɯæn˧',
    eeA: 'kʰɯæn˧',
    UeA: 'dʐɯæn˧',
    ieA: 'xɯæn˧',
    geA: 'ŋɯæn˧',
    heA: 'ʔɯæn˧',
    AeA: 'pɯæn˧',
    IeA: 'lɯæn˧',
    MeA: 'ɳɯæn˧',
    YeA: 'tɕʰɯæn˧',
    heI: 'ʔʷɯæn˧',
    GeI: 'dʷɯæn˧',
    IeI: 'lʷɯæn˧',
    feI: 'gʷɯæn˧',
    JeA: 'ʈɯæn˧',
    LeA: 'ɖɯæn˧',
    jeI: 'ɣʷɯæn˧',
    PeA: 'dzɯæn˧',
    QgA: 'sen˧',
    PgA: 'dzen˧',
    OgA: 'tsʰen˧',
    NgA: 'tsen˧',
    FgA: 'tʰen˧',
    dgA: 'ken˧',
    jgA: 'ɣen˧',
    hgA: 'ʔen˧',
    IgA: 'len˧',
    GgA: 'den˧',
    HgA: 'nen˧',
    EgA: 'ten˧',
    egA: 'kʰen˧',
    ggA: 'ŋen˧',
    DgA: 'men˧',
    CgA: 'ben˧',
    hgI: 'ʔʷen˧',
    dgI: 'kʷen˧',
    igI: 'xʷen˧',
    AgA: 'pen˧',
    jgI: 'ɣʷen˧',
    igA: 'xen˧',
    UgI: 'dʐʷen˧',
    QfA: 'siɛn˧',
    PfA: 'dziɛn˧',
    OfA: 'tsʰiɛn˧',
    NfA: 'tsiɛn˧',
    cfA: 'ȵiɛn˧',
    lfA: 'jiɛn˧',
    XfA: 'tɕiɛn˧',
    dfA: 'kiɛn˧',
    JfA: 'ʈiɛn˧',
    UfA: 'dʐiɛn˧',
    afA: 'ɕiɛn˧',
    KfA: 'ʈʰiɛn˧',
    ZfA: 'dʑiɛn˧',
    LfA: 'ɖiɛn˧',
    ifA: 'xiɛn˧',
    IfA: 'liɛn˧',
    BfA: 'pʰiɛn˧',
    CfA: 'biɛn˧',
    DfA: 'miɛn˧',
    PfI: 'dzʷiɛn˧',
    QfI: 'sʷiɛn˧',
    NfI: 'tsʷiɛn˧',
    ifI: 'xʷiɛn˧',
    cfI: 'ȵʷiɛn˧',
    YfI: 'tɕʰʷiɛn˧',
    lfI: 'jʷiɛn˧',
    RfI: 'zʷiɛn˧',
    hfI: 'ʔʷiɛn˧',
    bfI: 'ʑʷiɛn˧',
    AfA: 'piɛn˧',
    RfA: 'ziɛn˧',
    OfI: 'tsʰʷiɛn˧',
    XfI: 'tɕʷiɛn˧',
    ZfI: 'dʑʷiɛn˧',
    kfI: 'ɦʷiɛn˧',
    SfI: 'tʂʷiɛn˧',
    VfI: 'ʂʷiɛn˧',
    KfI: 'ʈʰʷiɛn˧',
    ffE: 'gɯiɛn˧',
    efE: 'kʰɯiɛn˧',
    ffM: 'gʷɯiɛn˧',
    LfI: 'ɖʷiɛn˧',
    IfI: 'lʷiɛn˧',
    efM: 'kʰʷɯiɛn˧',
    hfE: 'ʔɯiɛn˧',
    kfA: 'ɦiɛn˧',
    hfM: 'ʔʷɯiɛn˧',
    YfA: 'tɕʰiɛn˧',
    JfI: 'ʈʷiɛn˧',
    dfM: 'kʷɯiɛn˧',
    QhA: 'seu˧',
    FhA: 'tʰeu˧',
    EhA: 'teu˧',
    GhA: 'deu˧',
    dhA: 'keu˧',
    IhA: 'leu˧',
    ghA: 'ŋeu˧',
    ihA: 'xeu˧',
    hhA: 'ʔeu˧',
    ehA: 'kʰeu˧',
    QiA: 'siɛu˧',
    KiA: 'ʈʰiɛu˧',
    JiA: 'ʈiɛu˧',
    LiA: 'ɖiɛu˧',
    iiE: 'xɯiɛu˧',
    PiA: 'dziɛu˧',
    diE: 'kɯiɛu˧',
    NiA: 'tsiɛu˧',
    ciA: 'ȵiɛu˧',
    aiA: 'ɕiɛu˧',
    liA: 'jiɛu˧',
    ZiA: 'dʑiɛu˧',
    XiA: 'tɕiɛu˧',
    AiA: 'piɛu˧',
    AiE: 'pɯiɛu˧',
    CiA: 'biɛu˧',
    DiA: 'miɛu˧',
    DiE: 'mɯiɛu˧',
    hiA: 'ʔiɛu˧',
    kiA: 'ɦiɛu˧',
    fiE: 'gɯiɛu˧',
    OiA: 'tsʰiɛu˧',
    hiE: 'ʔɯiɛu˧',
    eiA: 'kʰiɛu˧',
    YiA: 'tɕʰiɛu˧',
    BiA: 'pʰiɛu˧',
    fiA: 'giɛu˧',
    IiA: 'liɛu˧',
    eiE: 'kʰɯiɛu˧',
    jjA: 'ɣɯau˧',
    djA: 'kɯau˧',
    UjA: 'dʐɯau˧',
    MjA: 'ɳɯau˧',
    VjA: 'ʂɯau˧',
    DjA: 'mɯau˧',
    ijA: 'xɯau˧',
    AjA: 'pɯau˧',
    BjA: 'pʰɯau˧',
    ejA: 'kʰɯau˧',
    gjA: 'ŋɯau˧',
    SjA: 'tʂɯau˧',
    JjA: 'ʈɯau˧',
    TjA: 'tʂʰɯau˧',
    CjA: 'bɯau˧',
    hjA: 'ʔɯau˧',
    KjA: 'ʈʰɯau˧',
    IjA: 'lɯau˧',
    LjA: 'ɖɯau˧',
    jkA: 'ɣɑu˧',
    dkA: 'kɑu˧',
    IkA: 'lɑu˧',
    ikA: 'xɑu˧',
    DkA: 'mɑu˧',
    FkA: 'tʰɑu˧',
    EkA: 'tɑu˧',
    QkA: 'sɑu˧',
    CkA: 'bɑu˧',
    AkA: 'pɑu˧',
    GkA: 'dɑu˧',
    NkA: 'tsɑu˧',
    gkA: 'ŋɑu˧',
    PkA: 'dzɑu˧',
    hkA: 'ʔɑu˧',
    HkA: 'nɑu˧',
    ekA: 'kʰɑu˧',
    OkA: 'tsʰɑu˧',
    BkA: 'pʰɑu˧',
    dlA: 'kɑ˧',
    OlA: 'tsʰɑ˧',
    ElA: 'tɑ˧',
    QlA: 'sɑ˧',
    GlA: 'dɑ˧',
    PlA: 'dzɑ˧',
    glA: 'ŋɑ˧',
    FlA: 'tʰɑ˧',
    IlA: 'lɑ˧',
    HlA: 'nɑ˧',
    jlA: 'ɣɑ˧',
    ilA: 'xɑ˧',
    elA: 'kʰɑ˧',
    hlA: 'ʔɑ˧',
    dlI: 'kuɑ˧',
    OlI: 'tsʰuɑ˧',
    ElI: 'tuɑ˧',
    QlI: 'suɑ˧',
    ClA: 'buɑ˧',
    GlI: 'duɑ˧',
    DlA: 'muɑ˧',
    PlI: 'dzuɑ˧',
    glI: 'ŋuɑ˧',
    FlI: 'tʰuɑ˧',
    IlI: 'luɑ˧',
    HlI: 'nuɑ˧',
    AlA: 'puɑ˧',
    BlA: 'pʰuɑ˧',
    jlI: 'ɣuɑ˧',
    elI: 'kʰuɑ˧',
    hlI: 'ʔuɑ˧',
    imI: 'xʷiɑ˧',
    hmI: 'ʔʷiɑ˧',
    emI: 'kʰʷiɑ˧',
    fmA: 'giɑ˧',
    emA: 'kʰiɑ˧',
    dmA: 'kiɑ˧',
    OmI: 'tsʰʷiɑ˧',
    NmI: 'tsʷiɑ˧',
    fmI: 'gʷiɑ˧',
    ImI: 'lʷiɑ˧',
    DnA: 'mɯa˧',
    YoA: 'tɕʰia˧',
    aoA: 'ɕia˧',
    loA: 'jia˧',
    XoA: 'tɕia˧',
    NoA: 'tsia˧',
    boA: 'ʑia˧',
    jnI: 'ɣʷɯa˧',
    dnI: 'kʷɯa˧',
    inI: 'xʷɯa˧',
    enI: 'kʰʷɯa˧',
    MnA: 'ɳɯa˧',
    dnA: 'kɯa˧',
    jnA: 'ɣɯa˧',
    BnA: 'pʰɯa˧',
    hnA: 'ʔɯa˧',
    AnA: 'pɯa˧',
    TnA: 'tʂʰɯa˧',
    VnA: 'ʂɯa˧',
    gnA: 'ŋɯa˧',
    SnA: 'tʂɯa˧',
    LnA: 'ɖɯa˧',
    RoA: 'zia˧',
    ZoA: 'dʑia˧',
    hnI: 'ʔʷɯa˧',
    SnI: 'tʂʷɯa˧',
    JnI: 'ʈʷɯa˧',
    CnA: 'bɯa˧',
    UnA: 'dʐɯa˧',
    KnA: 'ʈʰɯa˧',
    JnA: 'ʈɯa˧',
    inA: 'xɯa˧',
    enA: 'kʰɯa˧',
    PoA: 'dzia˧',
    coA: 'ȵia˧',
    QoA: 'sia˧',
    JoA: 'ʈia˧',
    gnI: 'ŋʷɯa˧',
    lpA: 'jiɐŋ˧',
    RpA: 'ziɐŋ˧',
    IpA: 'liɐŋ˧',
    ipA: 'xiɐŋ˧',
    apA: 'ɕiɐŋ˧',
    CpA: 'bʷiɐŋ˧',
    XpA: 'tɕiɐŋ˧',
    YpA: 'tɕʰiɐŋ˧',
    epA: 'kʰiɐŋ˧',
    dpA: 'kiɐŋ˧',
    LpA: 'ɖiɐŋ˧',
    JpA: 'ʈiɐŋ˧',
    cpA: 'ȵiɐŋ˧',
    ApA: 'pʷiɐŋ˧',
    QpA: 'siɐŋ˧',
    NpA: 'tsiɐŋ˧',
    TpA: 'tʂʰiɐŋ˧',
    DpA: 'mʷiɐŋ˧',
    MpA: 'ɳiɐŋ˧',
    UpA: 'dʐiɐŋ˧',
    SpA: 'tʂiɐŋ˧',
    ZpA: 'dʑiɐŋ˧',
    VpA: 'ʂiɐŋ˧',
    PpA: 'dziɐŋ˧',
    OpA: 'tsʰiɐŋ˧',
    epI: 'kʰʷiɐŋ˧',
    kpI: 'ɦʷiɐŋ˧',
    hpA: 'ʔiɐŋ˧',
    fpA: 'giɐŋ˧',
    KpA: 'ʈʰiɐŋ˧',
    BpA: 'pʰʷiɐŋ˧',
    fpI: 'gʷiɐŋ˧',
    GqA: 'dɑŋ˧',
    IqA: 'lɑŋ˧',
    EqA: 'tɑŋ˧',
    OqA: 'tsʰɑŋ˧',
    dqA: 'kɑŋ˧',
    QqA: 'sɑŋ˧',
    eqA: 'kʰɑŋ˧',
    iqI: 'xʷɑŋ˧',
    jqI: 'ɣʷɑŋ˧',
    dqI: 'kʷɑŋ˧',
    FqA: 'tʰɑŋ˧',
    BqA: 'pʰɑŋ˧',
    hqI: 'ʔʷɑŋ˧',
    hqA: 'ʔɑŋ˧',
    iqA: 'xɑŋ˧',
    jqA: 'ɣɑŋ˧',
    DqA: 'mɑŋ˧',
    NqA: 'tsɑŋ˧',
    HqA: 'nɑŋ˧',
    CqA: 'bɑŋ˧',
    gqA: 'ŋɑŋ˧',
    PqA: 'dzɑŋ˧',
    eqI: 'kʰʷɑŋ˧',
    AqA: 'pɑŋ˧',
    drA: 'kɯaŋ˧',
    erA: 'kʰɯaŋ˧',
    DrA: 'mɯaŋ˧',
    jrI: 'ɣʷɯaŋ˧',
    ArA: 'pɯaŋ˧',
    irI: 'xʷɯaŋ˧',
    drI: 'kʷɯaŋ˧',
    CrA: 'bɯaŋ˧',
    irA: 'xɯaŋ˧',
    KrA: 'ʈʰɯaŋ˧',
    TrA: 'tʂʰɯaŋ˧',
    UrA: 'dʐɯaŋ˧',
    hsA: 'ʔɯiaŋ˧',
    BrA: 'pʰɯiaŋ˧',
    CsA: 'bɯiaŋ˧',
    dsA: 'kɯiaŋ˧',
    DsA: 'mɯiaŋ˧',
    LrA: 'ɖɯaŋ˧',
    JrA: 'ʈɯaŋ˧',
    ksI: 'ɦʷɯiaŋ˧',
    AsA: 'pɯiaŋ˧',
    isI: 'xʷɯiaŋ˧',
    esA: 'kʰɯiaŋ˧',
    VrA: 'ʂɯaŋ˧',
    fsA: 'gɯiaŋ˧',
    gsA: 'ŋɯiaŋ˧',
    jrA: 'ɣɯaŋ˧',
    MrA: 'ɳɯaŋ˧',
    dtA: 'kɯæŋ˧',
    etA: 'kʰɯæŋ˧',
    DtA: 'mɯæŋ˧',
    AtA: 'pɯæŋ˧',
    jtI: 'ɣʷɯæŋ˧',
    jtA: 'ɣɯæŋ˧',
    JtA: 'ʈɯæŋ˧',
    htA: 'ʔɯæŋ˧',
    UtA: 'dʐɯæŋ˧',
    TtA: 'tʂʰɯæŋ˧',
    MtA: 'ɳɯæŋ˧',
    BtA: 'pʰɯæŋ˧',
    itI: 'xʷɯæŋ˧',
    LtA: 'ɖɯæŋ˧',
    htI: 'ʔʷɯæŋ˧',
    CtA: 'bɯæŋ˧',
    gtA: 'ŋɯæŋ˧',
    StA: 'tʂɯæŋ˧',
    OuA: 'tsʰiɛŋ˧',
    PuA: 'dziɛŋ˧',
    NuA: 'tsiɛŋ˧',
    luA: 'jiɛŋ˧',
    luI: 'jʷiɛŋ˧',
    huA: 'ʔiɛŋ˧',
    JuA: 'ʈiɛŋ˧',
    KuA: 'ʈʰiɛŋ˧',
    ZuA: 'dʑiɛŋ˧',
    LuA: 'ɖiɛŋ˧',
    auA: 'ɕiɛŋ˧',
    XuA: 'tɕiɛŋ˧',
    euA: 'kʰiɛŋ˧',
    DuA: 'miɛŋ˧',
    IuA: 'liɛŋ˧',
    AuA: 'piɛŋ˧',
    euI: 'kʰʷiɛŋ˧',
    RuA: 'ziɛŋ˧',
    huI: 'ʔʷiɛŋ˧',
    fuI: 'gʷiɛŋ˧',
    QuI: 'sʷiɛŋ˧',
    fuA: 'giɛŋ˧',
    iuI: 'xʷiɛŋ˧',
    OvA: 'tsʰeŋ˧',
    dvA: 'keŋ˧',
    jvA: 'ɣeŋ˧',
    GvA: 'deŋ˧',
    EvA: 'teŋ˧',
    ivA: 'xeŋ˧',
    QvA: 'seŋ˧',
    BvA: 'pʰeŋ˧',
    IvA: 'leŋ˧',
    HvA: 'neŋ˧',
    FvA: 'tʰeŋ˧',
    DvA: 'meŋ˧',
    CvA: 'beŋ˧',
    jvI: 'ɣʷeŋ˧',
    dvI: 'kʷeŋ˧',
    XwA: 'tɕɨŋ˧',
    ZwA: 'dʑɨŋ˧',
    LwA: 'ɖɨŋ˧',
    IwA: 'lɨŋ˧',
    hwA: 'ʔɨŋ˧',
    CwA: 'bɨŋ˧',
    AwA: 'pɨŋ˧',
    lwA: 'jɨŋ˧',
    bwA: 'ʑɨŋ˧',
    awA: 'ɕɨŋ˧',
    cwA: 'ȵɨŋ˧',
    dwA: 'kɨŋ˧',
    JwA: 'ʈɨŋ˧',
    PwA: 'dzɨŋ˧',
    gwA: 'ŋɨŋ˧',
    iwA: 'xɨŋ˧',
    YwA: 'tɕʰɨŋ˧',
    fwA: 'gɨŋ˧',
    VwA: 'ʂɨŋ˧',
    ewA: 'kʰɨŋ˧',
    KwA: 'ʈʰɨŋ˧',
    UwA: 'dʐɨŋ˧',
    BwA: 'pʰɨŋ˧',
    ExA: 'təŋ˧',
    IxA: 'ləŋ˧',
    QxA: 'səŋ˧',
    AxA: 'pəŋ˧',
    NxA: 'tsəŋ˧',
    DxA: 'məŋ˧',
    PxA: 'dzəŋ˧',
    CxA: 'bəŋ˧',
    jxI: 'ɣʷəŋ˧',
    dxI: 'kʷəŋ˧',
    ixI: 'xʷəŋ˧',
    HxA: 'nəŋ˧',
    GxA: 'dəŋ˧',
    jxA: 'ɣəŋ˧',
    dxA: 'kəŋ˧',
    FxA: 'tʰəŋ˧',
    BxA: 'pʰəŋ˧',
    kyA: 'ɦiu˧',
    hyA: 'ʔiu˧',
    IyA: 'liu˧',
    OyA: 'tsʰiu˧',
    lyA: 'jiu˧',
    gyA: 'ŋiu˧',
    NyA: 'tsiu˧',
    PyA: 'dziu˧',
    QyA: 'siu˧',
    KyA: 'ʈʰiu˧',
    eyA: 'kʰiu˧',
    YyA: 'tɕʰiu˧',
    XyA: 'tɕiu˧',
    ZyA: 'dʑiu˧',
    cyA: 'ȵiu˧',
    ayA: 'ɕiu˧',
    ByA: 'pʰiu˧',
    dyA: 'kiu˧',
    AyA: 'piu˧',
    VyA: 'ʂiu˧',
    TyA: 'tʂʰiu˧',
    SyA: 'tʂiu˧',
    UyA: 'dʐiu˧',
    iyA: 'xiu˧',
    RyA: 'ziu˧',
    LyA: 'ɖiu˧',
    JyA: 'ʈiu˧',
    fyA: 'giu˧',
    CyA: 'biu˧',
    DyA: 'miu˧',
    jzA: 'ɣəu˧',
    hzA: 'ʔəu˧',
    HzA: 'nəu˧',
    IzA: 'ləu˧',
    QzA: 'səu˧',
    ezA: 'kʰəu˧',
    izA: 'xəu˧',
    NzA: 'tsəu˧',
    FzA: 'tʰəu˧',
    GzA: 'dəu˧',
    gzA: 'ŋəu˧',
    dzA: 'kəu˧',
    EzA: 'təu˧',
    PzA: 'dzəu˧',
    CzA: 'bəu˧',
    OzA: 'tsʰəu˧',
    DzA: 'məu˧',
    h0A: 'ʔɨu˧',
    f0A: 'gɨu˧',
    A0A: 'pɨu˧',
    I0A: 'lɨu˧',
    d0A: 'kɨu˧',
    C0A: 'bɨu˧',
    N0A: 'tsɨu˧',
    V0A: 'ʂɨu˧',
    g0A: 'ŋɨu˧',
    i0A: 'xɨu˧',
    D0A: 'mɨu˧',
    O1A: 'tsʰim˧',
    R1A: 'zim˧',
    I1A: 'lim˧',
    K1A: 'ʈʰim˧',
    X1A: 'tɕim˧',
    L1A: 'ɖim˧',
    J1A: 'ʈim˧',
    Z1A: 'dʑim˧',
    c1A: 'ȵim˧',
    a1A: 'ɕim˧',
    l1A: 'jim˧',
    Q1A: 'sim˧',
    h1A: 'ʔim˧',
    N1A: 'tsim˧',
    P1A: 'dzim˧',
    M1A: 'ɳim˧',
    f1E: 'gɯim˧',
    e1E: 'kʰɯim˧',
    g1E: 'ŋɯim˧',
    i1E: 'xɯim˧',
    d1E: 'kɯim˧',
    h1E: 'ʔɯim˧',
    V1A: 'ʂim˧',
    U1A: 'dʐim˧',
    S1A: 'tʂim˧',
    T1A: 'tʂʰim˧',
    Y1A: 'tɕʰim˧',
    G2A: 'dəm˧',
    O2A: 'tsʰəm˧',
    H2A: 'nəm˧',
    h2A: 'ʔəm˧',
    j2A: 'ɣəm˧',
    I2A: 'ləm˧',
    P2A: 'dzəm˧',
    N2A: 'tsəm˧',
    F2A: 'tʰəm˧',
    E2A: 'təm˧',
    e2A: 'kʰəm˧',
    i2A: 'xəm˧',
    Q2A: 'səm˧',
    d2A: 'kəm˧',
    g2A: 'ŋəm˧',
    G3A: 'dɑm˧',
    d3A: 'kɑm˧',
    E3A: 'tɑm˧',
    Q3A: 'sɑm˧',
    I3A: 'lɑm˧',
    e3A: 'kʰɑm˧',
    F3A: 'tʰɑm˧',
    P3A: 'dzɑm˧',
    j3A: 'ɣɑm˧',
    D3A: 'mɑm˧',
    i3A: 'xɑm˧',
    l4A: 'jiɛm˧',
    I4A: 'liɛm˧',
    A4E: 'pɯiɛm˧',
    Q4A: 'siɛm˧',
    O4A: 'tsʰiɛm˧',
    X4A: 'tɕiɛm˧',
    Z4A: 'dʑiɛm˧',
    a4A: 'ɕiɛm˧',
    Y4A: 'tɕʰiɛm˧',
    c4A: 'ȵiɛm˧',
    M4A: 'ɳiɛm˧',
    k4A: 'ɦiɛm˧',
    J4A: 'ʈiɛm˧',
    K4A: 'ʈʰiɛm˧',
    h4E: 'ʔɯiɛm˧',
    e4E: 'kʰɯiɛm˧',
    g4E: 'ŋɯiɛm˧',
    N4A: 'tsiɛm˧',
    P4A: 'dziɛm˧',
    f4E: 'gɯiɛm˧',
    h4A: 'ʔiɛm˧',
    R4A: 'ziɛm˧',
    V4A: 'ʂiɛm˧',
    L4A: 'ɖiɛm˧',
    f4A: 'giɛm˧',
    F5A: 'tʰem˧',
    E5A: 'tem˧',
    G5A: 'dem˧',
    I5A: 'lem˧',
    e5A: 'kʰem˧',
    d5A: 'kem˧',
    j5A: 'ɣem˧',
    H5A: 'nem˧',
    i5A: 'xem˧',
    j6A: 'ɣɯæm˧',
    d6A: 'kɯæm˧',
    V6A: 'ʂɯæm˧',
    h6A: 'ʔɯæm˧',
    g6A: 'ŋɯæm˧',
    i6A: 'xɯæm˧',
    J6A: 'ʈɯæm˧',
    M6A: 'ɳɯæm˧',
    U6A: 'dʐɯæm˧',
    e6A: 'kʰɯæm˧',
    j7A: 'ɣɯam˧',
    U7A: 'dʐɯam˧',
    g7A: 'ŋɯam˧',
    T7A: 'tʂʰɯam˧',
    V7A: 'ʂɯam˧',
    d7A: 'kɯam˧',
    C7A: 'bɯam˧',
    e7A: 'kʰɯam˧',
    g8A: 'ŋiɐm˧',
    i8A: 'xiɐm˧',
    h8A: 'ʔiɐm˧',
    e8A: 'kʰiɐm˧',
    C9A: 'biɐm˧',
    B9A: 'pʰiɐm˧',
    EAB: 'tuŋ˧˥',
    DAB: 'muŋ˧˥',
    eAB: 'kʰuŋ˧˥',
    QAB: 'suŋ˧˥',
    FAB: 'tʰuŋ˧˥',
    NAB: 'tsuŋ˧˥',
    jAB: 'ɣuŋ˧˥',
    hAB: 'ʔuŋ˧˥',
    HAB: 'nuŋ˧˥',
    AAB: 'puŋ˧˥',
    IAB: 'luŋ˧˥',
    iAB: 'xuŋ˧˥',
    GAB: 'duŋ˧˥',
    CAB: 'buŋ˧˥',
    XDB: 'tɕioŋ˧˥',
    KDB: 'ʈʰioŋ˧˥',
    IDB: 'lioŋ˧˥',
    hDB: 'ʔioŋ˧˥',
    cDB: 'ȵioŋ˧˥',
    LDB: 'ɖioŋ˧˥',
    JDB: 'ʈioŋ˧˥',
    CDB: 'bioŋ˧˥',
    BDB: 'pʰioŋ˧˥',
    lDB: 'jioŋ˧˥',
    eDB: 'kʰioŋ˧˥',
    ZDB: 'dʑioŋ˧˥',
    dDB: 'kioŋ˧˥',
    QDB: 'sioŋ˧˥',
    iDB: 'xioŋ˧˥',
    ADB: 'pioŋ˧˥',
    ECB: 'tuoŋ˧˥',
    DCB: 'muoŋ˧˥',
    fDB: 'gioŋ˧˥',
    YDB: 'tɕʰioŋ˧˥',
    NDB: 'tsioŋ˧˥',
    dEB: 'kɯɔŋ˧˥',
    CEB: 'bɯɔŋ˧˥',
    hEB: 'ʔɯɔŋ˧˥',
    DEB: 'mɯɔŋ˧˥',
    jEB: 'ɣɯɔŋ˧˥',
    AEB: 'pɯɔŋ˧˥',
    iEB: 'xɯɔŋ˧˥',
    XFB: 'tɕiɛ˧˥',
    ZFB: 'dʑiɛ˧˥',
    DFF: 'mɯiɛ˧˥',
    AFF: 'pɯiɛ˧˥',
    CFF: 'bɯiɛ˧˥',
    iFN: 'xʷɯiɛ˧˥',
    hFN: 'ʔʷɯiɛ˧˥',
    eFN: 'kʰʷɯiɛ˧˥',
    dFN: 'kʷɯiɛ˧˥',
    QFJ: 'sʷiɛ˧˥',
    IFJ: 'lʷiɛ˧˥',
    fFF: 'gɯiɛ˧˥',
    hFF: 'ʔɯiɛ˧˥',
    dFF: 'kɯiɛ˧˥',
    eFF: 'kʰɯiɛ˧˥',
    gFF: 'ŋɯiɛ˧˥',
    kFJ: 'ɦʷiɛ˧˥',
    NFJ: 'tsʷiɛ˧˥',
    cFJ: 'ȵʷiɛ˧˥',
    OFB: 'tsʰiɛ˧˥',
    LFB: 'ɖiɛ˧˥',
    QFB: 'siɛ˧˥',
    lFB: 'jiɛ˧˥',
    IFB: 'liɛ˧˥',
    VFB: 'ʂiɛ˧˥',
    AFB: 'piɛ˧˥',
    cFB: 'ȵiɛ˧˥',
    DFB: 'miɛ˧˥',
    CFB: 'biɛ˧˥',
    YFB: 'tɕʰiɛ˧˥',
    aFB: 'ɕiɛ˧˥',
    NFB: 'tsiɛ˧˥',
    XFJ: 'tɕʷiɛ˧˥',
    TFJ: 'tʂʰʷiɛ˧˥',
    RFJ: 'zʷiɛ˧˥',
    bFB: 'ʑiɛ˧˥',
    SFB: 'tʂiɛ˧˥',
    BFF: 'pʰɯiɛ˧˥',
    BFB: 'pʰiɛ˧˥',
    lFJ: 'jʷiɛ˧˥',
    PFJ: 'dzʷiɛ˧˥',
    eFJ: 'kʰʷiɛ˧˥',
    MFB: 'ɳiɛ˧˥',
    gFN: 'ŋʷɯiɛ˧˥',
    fFN: 'gʷɯiɛ˧˥',
    KFB: 'ʈʰiɛ˧˥',
    JFB: 'ʈiɛ˧˥',
    iFF: 'xɯiɛ˧˥',
    eFB: 'kʰiɛ˧˥',
    ZFJ: 'dʑʷiɛ˧˥',
    dFB: 'kiɛ˧˥',
    XGB: 'tɕi˧˥',
    ZGB: 'dʑi˧˥',
    DGF: 'mɯi˧˥',
    AGF: 'pɯi˧˥',
    RGB: 'zi˧˥',
    dGF: 'kɯi˧˥',
    NGB: 'tsi˧˥',
    AGB: 'pi˧˥',
    dGN: 'kʷɯi˧˥',
    kGJ: 'ɦʷi˧˥',
    aGB: 'ɕi˧˥',
    LGB: 'ɖi˧˥',
    QGB: 'si˧˥',
    CGB: 'bi˧˥',
    IGB: 'li˧˥',
    aGJ: 'ɕʷi˧˥',
    IGJ: 'lʷi˧˥',
    fGJ: 'gʷi˧˥',
    OGJ: 'tsʰʷi˧˥',
    MGB: 'ɳi˧˥',
    dGJ: 'kʷi˧˥',
    CGF: 'bɯi˧˥',
    PGJ: 'dzʷi˧˥',
    BGF: 'pʰɯi˧˥',
    cGJ: 'ȵʷi˧˥',
    lGJ: 'jʷi˧˥',
    hGF: 'ʔɯi˧˥',
    NGJ: 'tsʷi˧˥',
    JGB: 'ʈi˧˥',
    KGB: 'ʈʰi˧˥',
    iGJ: 'xʷi˧˥',
    fGN: 'gʷɯi˧˥',
    fGF: 'gɯi˧˥',
    eGN: 'kʰʷɯi˧˥',
    XHB: 'tɕɨ˧˥',
    ZHB: 'dʑɨ˧˥',
    JHB: 'ʈɨ˧˥',
    iHB: 'xɨ˧˥',
    dHB: 'kɨ˧˥',
    lHB: 'jɨ˧˥',
    RHB: 'zɨ˧˥',
    VHB: 'ʂɨ˧˥',
    cHB: 'ȵɨ˧˥',
    IHB: 'lɨ˧˥',
    QHB: 'sɨ˧˥',
    aHB: 'ɕɨ˧˥',
    LHB: 'ɖɨ˧˥',
    eHB: 'kʰɨ˧˥',
    UHB: 'dʐɨ˧˥',
    WHB: 'ʐɨ˧˥',
    NHB: 'tsɨ˧˥',
    kHB: 'ɦɨ˧˥',
    gHB: 'ŋɨ˧˥',
    YHB: 'tɕʰɨ˧˥',
    KHB: 'ʈʰɨ˧˥',
    THB: 'tʂʰɨ˧˥',
    SHB: 'tʂɨ˧˥',
    hHB: 'ʔɨ˧˥',
    MHB: 'ɳɨ˧˥',
    DIB: 'mʷɨi˧˥',
    hIB: 'ʔɨi˧˥',
    eIB: 'kʰɨi˧˥',
    dIB: 'kɨi˧˥',
    BIB: 'pʰʷɨi˧˥',
    AIB: 'pʷɨi˧˥',
    kIJ: 'ɦʷɨi˧˥',
    dIJ: 'kʷɨi˧˥',
    iIJ: 'xʷɨi˧˥',
    gIB: 'ŋɨi˧˥',
    iIB: 'xɨi˧˥',
    hIJ: 'ʔʷɨi˧˥',
    CIB: 'bʷɨi˧˥',
    gJB: 'ŋiɔ˧˥',
    IJB: 'liɔ˧˥',
    LJB: 'ɖiɔ˧˥',
    lJB: 'jiɔ˧˥',
    XJB: 'tɕiɔ˧˥',
    cJB: 'ȵiɔ˧˥',
    aJB: 'ɕiɔ˧˥',
    YJB: 'tɕʰiɔ˧˥',
    JJB: 'ʈiɔ˧˥',
    QJB: 'siɔ˧˥',
    KJB: 'ʈʰiɔ˧˥',
    MJB: 'ɳiɔ˧˥',
    iJB: 'xiɔ˧˥',
    fJB: 'giɔ˧˥',
    VJB: 'ʂiɔ˧˥',
    TJB: 'tʂʰiɔ˧˥',
    SJB: 'tʂiɔ˧˥',
    UJB: 'dʐiɔ˧˥',
    PJB: 'dziɔ˧˥',
    hJB: 'ʔiɔ˧˥',
    dJB: 'kiɔ˧˥',
    RJB: 'ziɔ˧˥',
    eJB: 'kʰiɔ˧˥',
    bJB: 'ʑiɔ˧˥',
    ZJB: 'dʑiɔ˧˥',
    OJB: 'tsʰiɔ˧˥',
    NJB: 'tsiɔ˧˥',
    gKB: 'ŋio˧˥',
    kKB: 'ɦio˧˥',
    PKB: 'dzio˧˥',
    AKB: 'pio˧˥',
    DKB: 'mio˧˥',
    CKB: 'bio˧˥',
    BKB: 'pʰio˧˥',
    LKB: 'ɖio˧˥',
    iKB: 'xio˧˥',
    ZKB: 'dʑio˧˥',
    lKB: 'jio˧˥',
    XKB: 'tɕio˧˥',
    hKB: 'ʔio˧˥',
    eKB: 'kʰio˧˥',
    JKB: 'ʈio˧˥',
    cKB: 'ȵio˧˥',
    fKB: 'gio˧˥',
    VKB: 'ʂio˧˥',
    dKB: 'kio˧˥',
    OKB: 'tsʰio˧˥',
    IKB: 'lio˧˥',
    QKB: 'sio˧˥',
    UKB: 'dʐio˧˥',
    DLB: 'muo˧˥',
    FLB: 'tʰuo˧˥',
    GLB: 'duo˧˥',
    ILB: 'luo˧˥',
    OLB: 'tsʰuo˧˥',
    ELB: 'tuo˧˥',
    dLB: 'kuo˧˥',
    gLB: 'ŋuo˧˥',
    CLB: 'buo˧˥',
    PLB: 'dzuo˧˥',
    NLB: 'tsuo˧˥',
    iLB: 'xuo˧˥',
    hLB: 'ʔuo˧˥',
    eLB: 'kʰuo˧˥',
    HLB: 'nuo˧˥',
    jLB: 'ɣuo˧˥',
    BLB: 'pʰuo˧˥',
    ALB: 'puo˧˥',
    PMB: 'dzei˧˥',
    IMB: 'lei˧˥',
    FMB: 'tʰei˧˥',
    BMB: 'pʰei˧˥',
    NMB: 'tsei˧˥',
    EMB: 'tei˧˥',
    GMB: 'dei˧˥',
    HMB: 'nei˧˥',
    QMB: 'sei˧˥',
    OMB: 'tsʰei˧˥',
    eMB: 'kʰei˧˥',
    jMB: 'ɣei˧˥',
    DMB: 'mei˧˥',
    CMB: 'bei˧˥',
    hMB: 'ʔei˧˥',
    gMB: 'ŋei˧˥',
    AMB: 'pei˧˥',
    jPB: 'ɣɯæ˧˥',
    DPB: 'mɯæ˧˥',
    ePB: 'kʰɯæ˧˥',
    LPB: 'ɖɯæ˧˥',
    MPB: 'ɳɯæ˧˥',
    CPB: 'bɯæ˧˥',
    hPB: 'ʔɯæ˧˥',
    APB: 'pɯæ˧˥',
    dPB: 'kɯæ˧˥',
    VPB: 'ʂɯæ˧˥',
    dPJ: 'kʷɯæ˧˥',
    LPJ: 'ɖʷɯæ˧˥',
    jPJ: 'ɣʷɯæ˧˥',
    iPJ: 'xʷɯæ˧˥',
    fPB: 'gɯæ˧˥',
    jQB: 'ɣɯæi˧˥',
    eQB: 'kʰɯæi˧˥',
    gQB: 'ŋɯæi˧˥',
    hQB: 'ʔɯæi˧˥',
    iSJ: 'xuoi˧˥',
    hSJ: 'ʔuoi˧˥',
    ISJ: 'luoi˧˥',
    GSJ: 'duoi˧˥',
    PSJ: 'dzuoi˧˥',
    DSB: 'muoi˧˥',
    FSJ: 'tʰuoi˧˥',
    jSJ: 'ɣuoi˧˥',
    eSJ: 'kʰuoi˧˥',
    ESJ: 'tuoi˧˥',
    HSJ: 'nuoi˧˥',
    JSJ: 'ʈuoi˧˥',
    gSJ: 'ŋuoi˧˥',
    OSJ: 'tsʰuoi˧˥',
    CSB: 'buoi˧˥',
    NSJ: 'tsuoi˧˥',
    kSJ: 'ɦuoi˧˥',
    iTB: 'xəi˧˥',
    eTB: 'kʰəi˧˥',
    NTB: 'tsəi˧˥',
    GTB: 'dəi˧˥',
    HTB: 'nəi˧˥',
    dTB: 'kəi˧˥',
    jTB: 'ɣəi˧˥',
    BTB: 'pʰəi˧˥',
    OTB: 'tsʰəi˧˥',
    YTB: 'tɕʰəi˧˥',
    ETB: 'təi˧˥',
    DTB: 'məi˧˥',
    PTB: 'dzəi˧˥',
    hTB: 'ʔəi˧˥',
    lTB: 'jəi˧˥',
    FTB: 'tʰəi˧˥',
    cTB: 'ȵəi˧˥',
    ITB: 'ləi˧˥',
    CTB: 'bəi˧˥',
    XVB: 'tɕin˧˥',
    KVB: 'ʈʰin˧˥',
    ZVB: 'dʑin˧˥',
    cVB: 'ȵin˧˥',
    aVB: 'ɕin˧˥',
    IVB: 'lin˧˥',
    LVB: 'ɖin˧˥',
    dVB: 'kin˧˥',
    PVB: 'dzin˧˥',
    NVB: 'tsin˧˥',
    CVB: 'bin˧˥',
    gVF: 'ŋɯin˧˥',
    OVB: 'tsʰin˧˥',
    fVN: 'gʷɯin˧˥',
    lVB: 'jin˧˥',
    DVF: 'mɯin˧˥',
    DVB: 'min˧˥',
    kVJ: 'ɦʷin˧˥',
    XVJ: 'tɕʷin˧˥',
    lVJ: 'jʷin˧˥',
    QVJ: 'sʷin˧˥',
    cVJ: 'ȵʷin˧˥',
    YVJ: 'tɕʰʷin˧˥',
    bVJ: 'ʑʷin˧˥',
    KVJ: 'ʈʰʷin˧˥',
    IVJ: 'lʷin˧˥',
    eVJ: 'kʰʷin˧˥',
    eVB: 'kʰin˧˥',
    aVJ: 'ɕʷin˧˥',
    iVB: 'xin˧˥',
    UVB: 'dʐin˧˥',
    JVB: 'ʈin˧˥',
    DXB: 'miun˧˥',
    AXB: 'piun˧˥',
    CXB: 'biun˧˥',
    BXB: 'pʰiun˧˥',
    hXJ: 'ʔiun˧˥',
    gXJ: 'ŋiun˧˥',
    kXJ: 'ɦiun˧˥',
    eXJ: 'kʰiun˧˥',
    hYB: 'ʔɨn˧˥',
    dYB: 'kɨn˧˥',
    SWB: 'tʂɪn˧˥',
    eYB: 'kʰɨn˧˥',
    fYB: 'gɨn˧˥',
    TWB: 'tʂʰɪn˧˥',
    gYB: 'ŋɨn˧˥',
    iYB: 'xɨn˧˥',
    gZJ: 'ŋʷiɐn˧˥',
    kZJ: 'ɦʷiɐn˧˥',
    hZB: 'ʔiɐn˧˥',
    dZB: 'kiɐn˧˥',
    fZB: 'giɐn˧˥',
    eZB: 'kʰiɐn˧˥',
    gZB: 'ŋiɐn˧˥',
    iZB: 'xiɐn˧˥',
    DZB: 'mʷiɐn˧˥',
    AZB: 'pʷiɐn˧˥',
    fZJ: 'gʷiɐn˧˥',
    hZJ: 'ʔʷiɐn˧˥',
    eZJ: 'kʰʷiɐn˧˥',
    iZJ: 'xʷiɐn˧˥',
    CZB: 'bʷiɐn˧˥',
    jaJ: 'ɣuon˧˥',
    BaB: 'pʰuon˧˥',
    OaJ: 'tsʰuon˧˥',
    AaB: 'puon˧˥',
    QaJ: 'suon˧˥',
    NaJ: 'tsuon˧˥',
    haJ: 'ʔuon˧˥',
    GaJ: 'duon˧˥',
    PaJ: 'dzuon˧˥',
    daJ: 'kuon˧˥',
    FaJ: 'tʰuon˧˥',
    eaJ: 'kʰuon˧˥',
    IaJ: 'luon˧˥',
    CaB: 'buon˧˥',
    DaB: 'muon˧˥',
    HaJ: 'nuon˧˥',
    iaJ: 'xuon˧˥',
    jbB: 'ɣən˧˥',
    ebB: 'kʰən˧˥',
    dbB: 'kən˧˥',
    jcB: 'ɣɑn˧˥',
    EcB: 'tɑn˧˥',
    FcB: 'tʰɑn˧˥',
    QcB: 'sɑn˧˥',
    GcB: 'dɑn˧˥',
    PcB: 'dzɑn˧˥',
    dcB: 'kɑn˧˥',
    IcB: 'lɑn˧˥',
    ecB: 'kʰɑn˧˥',
    icB: 'xɑn˧˥',
    NcB: 'tsɑn˧˥',
    jcJ: 'ɣʷɑn˧˥',
    EcJ: 'tʷɑn˧˥',
    hcJ: 'ʔʷɑn˧˥',
    FcJ: 'tʰʷɑn˧˥',
    QcJ: 'sʷɑn˧˥',
    dcJ: 'kʷɑn˧˥',
    IcJ: 'lʷɑn˧˥',
    ecJ: 'kʰʷɑn˧˥',
    HcJ: 'nʷɑn˧˥',
    NcJ: 'tsʷɑn˧˥',
    CcB: 'bʷɑn˧˥',
    DcB: 'mʷɑn˧˥',
    AcB: 'pʷɑn˧˥',
    RcJ: 'zʷɑn˧˥',
    GcJ: 'dʷɑn˧˥',
    HcB: 'nɑn˧˥',
    BcB: 'pʰʷɑn˧˥',
    VdB: 'ʂɯan˧˥',
    hdJ: 'ʔʷɯan˧˥',
    AdB: 'pɯan˧˥',
    SdB: 'tʂɯan˧˥',
    MdB: 'ɳɯan˧˥',
    jdB: 'ɣɯan˧˥',
    jdJ: 'ɣʷɯan˧˥',
    CdB: 'bɯan˧˥',
    DdB: 'mɯan˧˥',
    UdB: 'dʐɯan˧˥',
    gdB: 'ŋɯan˧˥',
    TdB: 'tʂʰɯan˧˥',
    BdB: 'pʰɯan˧˥',
    UdJ: 'dʐʷɯan˧˥',
    VeB: 'ʂɯæn˧˥',
    jeB: 'ɣɯæn˧˥',
    DeB: 'mɯæn˧˥',
    deB: 'kɯæn˧˥',
    TeB: 'tʂʰɯæn˧˥',
    UeB: 'dʐɯæn˧˥',
    geB: 'ŋɯæn˧˥',
    SeB: 'tʂɯæn˧˥',
    TeJ: 'tʂʰʷɯæn˧˥',
    eeB: 'kʰɯæn˧˥',
    QgB: 'sen˧˥',
    FgB: 'tʰen˧˥',
    EgB: 'ten˧˥',
    hgB: 'ʔen˧˥',
    GgB: 'den˧˥',
    dgB: 'ken˧˥',
    jgB: 'ɣen˧˥',
    igB: 'xen˧˥',
    DgB: 'men˧˥',
    HgB: 'nen˧˥',
    AgB: 'pen˧˥',
    jgJ: 'ɣʷen˧˥',
    dgJ: 'kʷen˧˥',
    CgB: 'ben˧˥',
    egJ: 'kʰʷen˧˥',
    egB: 'kʰen˧˥',
    ggB: 'ŋen˧˥',
    QfB: 'siɛn˧˥',
    lfB: 'jiɛn˧˥',
    PfB: 'dziɛn˧˥',
    JfB: 'ʈiɛn˧˥',
    XfB: 'tɕiɛn˧˥',
    MfB: 'ɳiɛn˧˥',
    OfB: 'tsʰiɛn˧˥',
    YfB: 'tɕʰiɛn˧˥',
    efB: 'kʰiɛn˧˥',
    dfF: 'kɯiɛn˧˥',
    ZfB: 'dʑiɛn˧˥',
    NfB: 'tsiɛn˧˥',
    cfB: 'ȵiɛn˧˥',
    RfB: 'ziɛn˧˥',
    IfB: 'liɛn˧˥',
    gfF: 'ŋɯiɛn˧˥',
    ffF: 'gɯiɛn˧˥',
    CfF: 'bɯiɛn˧˥',
    DfB: 'miɛn˧˥',
    AfB: 'piɛn˧˥',
    NfJ: 'tsʷiɛn˧˥',
    PfJ: 'dzʷiɛn˧˥',
    AfF: 'pɯiɛn˧˥',
    lfJ: 'jʷiɛn˧˥',
    IfJ: 'lʷiɛn˧˥',
    JfJ: 'ʈʷiɛn˧˥',
    dfN: 'kʷɯiɛn˧˥',
    ffN: 'gʷɯiɛn˧˥',
    cfJ: 'ȵʷiɛn˧˥',
    YfJ: 'tɕʰʷiɛn˧˥',
    ZfJ: 'dʑʷiɛn˧˥',
    LfJ: 'ɖʷiɛn˧˥',
    XfJ: 'tɕʷiɛn˧˥',
    QfJ: 'sʷiɛn˧˥',
    UfJ: 'dʐʷiɛn˧˥',
    ffJ: 'gʷiɛn˧˥',
    ifJ: 'xʷiɛn˧˥',
    CfB: 'biɛn˧˥',
    DfF: 'mɯiɛn˧˥',
    KfB: 'ʈʰiɛn˧˥',
    BfF: 'pʰɯiɛn˧˥',
    afB: 'ɕiɛn˧˥',
    hfF: 'ʔɯiɛn˧˥',
    LfB: 'ɖiɛn˧˥',
    UfB: 'dʐiɛn˧˥',
    QhB: 'seu˧˥',
    dhB: 'keu˧˥',
    EhB: 'teu˧˥',
    IhB: 'leu˧˥',
    FhB: 'tʰeu˧˥',
    ihB: 'xeu˧˥',
    hhB: 'ʔeu˧˥',
    HhB: 'neu˧˥',
    jhB: 'ɣeu˧˥',
    GhB: 'deu˧˥',
    ehB: 'kʰeu˧˥',
    NhB: 'tseu˧˥',
    QiB: 'siɛu˧˥',
    LiB: 'ɖiɛu˧˥',
    XiB: 'tɕiɛu˧˥',
    hiF: 'ʔɯiɛu˧˥',
    KiB: 'ʈʰiɛu˧˥',
    aiB: 'ɕiɛu˧˥',
    ciB: 'ȵiɛu˧˥',
    CiB: 'biɛu˧˥',
    YiB: 'tɕʰiɛu˧˥',
    BiB: 'pʰiɛu˧˥',
    DiB: 'miɛu˧˥',
    ZiB: 'dʑiɛu˧˥',
    diF: 'kɯiɛu˧˥',
    AiF: 'pɯiɛu˧˥',
    AiB: 'piɛu˧˥',
    CiF: 'bɯiɛu˧˥',
    liB: 'jiɛu˧˥',
    OiB: 'tsʰiɛu˧˥',
    NiB: 'tsiɛu˧˥',
    fiF: 'gɯiɛu˧˥',
    IiB: 'liɛu˧˥',
    BiF: 'pʰɯiɛu˧˥',
    hiB: 'ʔiɛu˧˥',
    ejB: 'kʰɯau˧˥',
    jjB: 'ɣɯau˧˥',
    AjB: 'pɯau˧˥',
    MjB: 'ɳɯau˧˥',
    DjB: 'mɯau˧˥',
    djB: 'kɯau˧˥',
    SjB: 'tʂɯau˧˥',
    hjB: 'ʔɯau˧˥',
    CjB: 'bɯau˧˥',
    gjB: 'ŋɯau˧˥',
    UjB: 'dʐɯau˧˥',
    TjB: 'tʂʰɯau˧˥',
    JjB: 'ʈɯau˧˥',
    VjB: 'ʂɯau˧˥',
    jkB: 'ɣɑu˧˥',
    CkB: 'bɑu˧˥',
    IkB: 'lɑu˧˥',
    FkB: 'tʰɑu˧˥',
    GkB: 'dɑu˧˥',
    HkB: 'nɑu˧˥',
    QkB: 'sɑu˧˥',
    EkB: 'tɑu˧˥',
    OkB: 'tsʰɑu˧˥',
    NkB: 'tsɑu˧˥',
    PkB: 'dzɑu˧˥',
    dkB: 'kɑu˧˥',
    ikB: 'xɑu˧˥',
    DkB: 'mɑu˧˥',
    AkB: 'pɑu˧˥',
    hkB: 'ʔɑu˧˥',
    ekB: 'kʰɑu˧˥',
    gkB: 'ŋɑu˧˥',
    dlB: 'kɑ˧˥',
    OlB: 'tsʰɑ˧˥',
    ElB: 'tɑ˧˥',
    QlB: 'sɑ˧˥',
    GlB: 'dɑ˧˥',
    glB: 'ŋɑ˧˥',
    FlB: 'tʰɑ˧˥',
    IlB: 'lɑ˧˥',
    HlB: 'nɑ˧˥',
    jlB: 'ɣɑ˧˥',
    ilB: 'xɑ˧˥',
    elB: 'kʰɑ˧˥',
    hlB: 'ʔɑ˧˥',
    NlB: 'tsɑ˧˥',
    dlJ: 'kuɑ˧˥',
    ElJ: 'tuɑ˧˥',
    QlJ: 'suɑ˧˥',
    GlJ: 'duɑ˧˥',
    FlJ: 'tʰuɑ˧˥',
    DlB: 'muɑ˧˥',
    PlJ: 'dzuɑ˧˥',
    glJ: 'ŋuɑ˧˥',
    IlJ: 'luɑ˧˥',
    hlJ: 'ʔuɑ˧˥',
    HlJ: 'nuɑ˧˥',
    AlB: 'puɑ˧˥',
    BlB: 'pʰuɑ˧˥',
    jlJ: 'ɣuɑ˧˥',
    ilJ: 'xuɑ˧˥',
    elJ: 'kʰuɑ˧˥',
    ClB: 'bɑ˧˥',
    OlJ: 'tsʰuɑ˧˥',
    NlJ: 'tsuɑ˧˥',
    DnB: 'mɯa˧˥',
    XoB: 'tɕia˧˥',
    loB: 'jia˧˥',
    gnB: 'ŋɯa˧˥',
    dnB: 'kɯa˧˥',
    VnB: 'ʂɯa˧˥',
    hnB: 'ʔɯa˧˥',
    RoB: 'zia˧˥',
    jnB: 'ɣɯa˧˥',
    QoB: 'sia˧˥',
    NmJ: 'tsʷiɑ˧˥',
    DoB: 'mia˧˥',
    OoB: 'tsʰia˧˥',
    inB: 'xɯa˧˥',
    ZoB: 'dʑia˧˥',
    enB: 'kʰɯa˧˥',
    CnB: 'bɯa˧˥',
    aoB: 'ɕia˧˥',
    NoB: 'tsia˧˥',
    AnB: 'pɯa˧˥',
    jnJ: 'ɣʷɯa˧˥',
    dnJ: 'kʷɯa˧˥',
    gnJ: 'ŋʷɯa˧˥',
    coB: 'ȵia˧˥',
    SnB: 'tʂɯa˧˥',
    JnB: 'ʈɯa˧˥',
    UnB: 'dʐɯa˧˥',
    YoB: 'tɕʰia˧˥',
    MnB: 'ɳɯa˧˥',
    enJ: 'kʰʷɯa˧˥',
    TnJ: 'tʂʰʷɯa˧˥',
    KnJ: 'ʈʰʷɯa˧˥',
    VnJ: 'ʂʷɯa˧˥',
    KnB: 'ʈʰɯa˧˥',
    InB: 'lɯa˧˥',
    lpB: 'jiɐŋ˧˥',
    RpB: 'ziɐŋ˧˥',
    NpB: 'tsiɐŋ˧˥',
    IpB: 'liɐŋ˧˥',
    hpB: 'ʔiɐŋ˧˥',
    fpB: 'giɐŋ˧˥',
    gpB: 'ŋiɐŋ˧˥',
    TpB: 'tʂʰiɐŋ˧˥',
    QpB: 'siɐŋ˧˥',
    XpB: 'tɕiɐŋ˧˥',
    VpB: 'ʂiɐŋ˧˥',
    ipB: 'xiɐŋ˧˥',
    YpB: 'tɕʰiɐŋ˧˥',
    dpB: 'kiɐŋ˧˥',
    LpB: 'ɖiɐŋ˧˥',
    KpB: 'ʈʰiɐŋ˧˥',
    dpJ: 'kʷiɐŋ˧˥',
    cpB: 'ȵiɐŋ˧˥',
    apB: 'ɕiɐŋ˧˥',
    BpB: 'pʰʷiɐŋ˧˥',
    DpB: 'mʷiɐŋ˧˥',
    ApB: 'pʷiɐŋ˧˥',
    hpJ: 'ʔʷiɐŋ˧˥',
    kpJ: 'ɦʷiɐŋ˧˥',
    ipJ: 'xʷiɐŋ˧˥',
    OpB: 'tsʰiɐŋ˧˥',
    JpB: 'ʈiɐŋ˧˥',
    ZpB: 'dʑiɐŋ˧˥',
    fpJ: 'gʷiɐŋ˧˥',
    CpB: 'bʷiɐŋ˧˥',
    GqB: 'dɑŋ˧˥',
    QqB: 'sɑŋ˧˥',
    dqJ: 'kʷɑŋ˧˥',
    AqB: 'pɑŋ˧˥',
    NqB: 'tsɑŋ˧˥',
    HqB: 'nɑŋ˧˥',
    jqB: 'ɣɑŋ˧˥',
    FqB: 'tʰɑŋ˧˥',
    DqB: 'mɑŋ˧˥',
    EqB: 'tɑŋ˧˥',
    IqB: 'lɑŋ˧˥',
    hqB: 'ʔɑŋ˧˥',
    eqB: 'kʰɑŋ˧˥',
    hqJ: 'ʔʷɑŋ˧˥',
    jqJ: 'ɣʷɑŋ˧˥',
    BqB: 'pʰɑŋ˧˥',
    iqJ: 'xʷɑŋ˧˥',
    dqB: 'kɑŋ˧˥',
    gqB: 'ŋɑŋ˧˥',
    PqB: 'dzɑŋ˧˥',
    OqB: 'tsʰɑŋ˧˥',
    iqB: 'xɑŋ˧˥',
    eqJ: 'kʰʷɑŋ˧˥',
    drB: 'kɯaŋ˧˥',
    AsB: 'pɯiaŋ˧˥',
    dsB: 'kɯiaŋ˧˥',
    hsB: 'ʔɯiaŋ˧˥',
    VsB: 'ʂɯiaŋ˧˥',
    ksJ: 'ɦʷɯiaŋ˧˥',
    isJ: 'xʷɯiaŋ˧˥',
    DsB: 'mɯiaŋ˧˥',
    dsJ: 'kʷɯiaŋ˧˥',
    jrB: 'ɣɯaŋ˧˥',
    DrB: 'mɯaŋ˧˥',
    drJ: 'kʷɯaŋ˧˥',
    ArB: 'pɯaŋ˧˥',
    JrB: 'ʈɯaŋ˧˥',
    LrB: 'ɖɯaŋ˧˥',
    hrJ: 'ʔʷɯaŋ˧˥',
    ErB: 'tɯaŋ˧˥',
    IrB: 'lɯaŋ˧˥',
    jrJ: 'ɣʷɯaŋ˧˥',
    CrB: 'bɯaŋ˧˥',
    erJ: 'kʰʷɯaŋ˧˥',
    MrB: 'ɳɯaŋ˧˥',
    dtB: 'kɯæŋ˧˥',
    DtB: 'mɯæŋ˧˥',
    jtB: 'ɣɯæŋ˧˥',
    CtB: 'bɯæŋ˧˥',
    BtB: 'pʰɯæŋ˧˥',
    PuB: 'dziɛŋ˧˥',
    XuB: 'tɕiɛŋ˧˥',
    KuB: 'ʈʰiɛŋ˧˥',
    luB: 'jiɛŋ˧˥',
    fuB: 'giɛŋ˧˥',
    luJ: 'jʷiɛŋ˧˥',
    IuB: 'liɛŋ˧˥',
    duB: 'kiɛŋ˧˥',
    AuB: 'piɛŋ˧˥',
    euJ: 'kʰʷiɛŋ˧˥',
    NuB: 'tsiɛŋ˧˥',
    huB: 'ʔiɛŋ˧˥',
    OuB: 'tsʰiɛŋ˧˥',
    QuB: 'siɛŋ˧˥',
    DuB: 'miɛŋ˧˥',
    LuB: 'ɖiɛŋ˧˥',
    jvJ: 'ɣʷeŋ˧˥',
    dvJ: 'kʷeŋ˧˥',
    DvB: 'meŋ˧˥',
    EvB: 'teŋ˧˥',
    GvB: 'deŋ˧˥',
    FvB: 'tʰeŋ˧˥',
    PvB: 'dzeŋ˧˥',
    hvJ: 'ʔʷeŋ˧˥',
    evB: 'kʰeŋ˧˥',
    HvB: 'neŋ˧˥',
    jvB: 'ɣeŋ˧˥',
    QvB: 'seŋ˧˥',
    AvB: 'peŋ˧˥',
    evJ: 'kʰʷeŋ˧˥',
    dvB: 'keŋ˧˥',
    BvB: 'pʰeŋ˧˥',
    hvB: 'ʔeŋ˧˥',
    IvB: 'leŋ˧˥',
    CvB: 'beŋ˧˥',
    ivJ: 'xʷeŋ˧˥',
    gvB: 'ŋeŋ˧˥',
    XwB: 'tɕɨŋ˧˥',
    KwB: 'ʈʰɨŋ˧˥',
    fwB: 'gɨŋ˧˥',
    VwB: 'ʂɨŋ˧˥',
    ExB: 'təŋ˧˥',
    BxB: 'pʰəŋ˧˥',
    exB: 'kʰəŋ˧˥',
    HxB: 'nəŋ˧˥',
    kyB: 'ɦiu˧˥',
    IyB: 'liu˧˥',
    MyB: 'ɳiu˧˥',
    KyB: 'ʈʰiu˧˥',
    JyB: 'ʈiu˧˥',
    iyB: 'xiu˧˥',
    dyB: 'kiu˧˥',
    ayB: 'ɕiu˧˥',
    YyB: 'tɕʰiu˧˥',
    PyB: 'dziu˧˥',
    CyB: 'biu˧˥',
    AyB: 'piu˧˥',
    eyB: 'kʰiu˧˥',
    cyB: 'ȵiu˧˥',
    fyB: 'giu˧˥',
    LyB: 'ɖiu˧˥',
    lyB: 'jiu˧˥',
    ZyB: 'dʑiu˧˥',
    QyB: 'siu˧˥',
    hyB: 'ʔiu˧˥',
    NyB: 'tsiu˧˥',
    VyB: 'ʂiu˧˥',
    XyB: 'tɕiu˧˥',
    ByB: 'pʰiu˧˥',
    SyB: 'tʂiu˧˥',
    UyB: 'dʐiu˧˥',
    TyB: 'tʂʰiu˧˥',
    jzB: 'ɣəu˧˥',
    DzB: 'məu˧˥',
    CzB: 'bəu˧˥',
    EzB: 'təu˧˥',
    FzB: 'tʰəu˧˥',
    dzB: 'kəu˧˥',
    gzB: 'ŋəu˧˥',
    AzB: 'pəu˧˥',
    HzB: 'nəu˧˥',
    QzB: 'səu˧˥',
    izB: 'xəu˧˥',
    BzB: 'pʰəu˧˥',
    hzB: 'ʔəu˧˥',
    IzB: 'ləu˧˥',
    NzB: 'tsəu˧˥',
    ezB: 'kʰəu˧˥',
    GzB: 'dəu˧˥',
    UzB: 'dʐəu˧˥',
    OzB: 'tsʰəu˧˥',
    h0B: 'ʔɨu˧˥',
    d0B: 'kɨu˧˥',
    f0B: 'gɨu˧˥',
    O1B: 'tsʰim˧˥',
    L1B: 'ɖim˧˥',
    e1B: 'kʰim˧˥',
    I1B: 'lim˧˥',
    Q1B: 'sim˧˥',
    K1B: 'ʈʰim˧˥',
    N1B: 'tsim˧˥',
    c1B: 'ȵim˧˥',
    X1B: 'tɕim˧˥',
    a1B: 'ɕim˧˥',
    b1B: 'ʑim˧˥',
    Z1B: 'dʑim˧˥',
    Y1B: 'tɕʰim˧˥',
    T1B: 'tʂʰim˧˥',
    M1B: 'ɳim˧˥',
    f1F: 'gɯim˧˥',
    d1F: 'kɯim˧˥',
    P1B: 'dzim˧˥',
    g1F: 'ŋɯim˧˥',
    V1B: 'ʂim˧˥',
    A1F: 'pɯim˧˥',
    h1F: 'ʔɯim˧˥',
    B1F: 'pʰɯim˧˥',
    e1F: 'kʰɯim˧˥',
    i1F: 'xɯim˧˥',
    J1B: 'ʈim˧˥',
    l1B: 'jim˧˥',
    U1B: 'dʐim˧˥',
    d2B: 'kəm˧˥',
    G2B: 'dəm˧˥',
    h2B: 'ʔəm˧˥',
    H2B: 'nəm˧˥',
    F2B: 'tʰəm˧˥',
    P2B: 'dzəm˧˥',
    O2B: 'tsʰəm˧˥',
    g2B: 'ŋəm˧˥',
    N2B: 'tsəm˧˥',
    Q2B: 'səm˧˥',
    e2B: 'kʰəm˧˥',
    j2B: 'ɣəm˧˥',
    I2B: 'ləm˧˥',
    E2B: 'təm˧˥',
    i2B: 'xəm˧˥',
    d3B: 'kɑm˧˥',
    a3B: 'ɕɑm˧˥',
    I3B: 'lɑm˧˥',
    F3B: 'tʰɑm˧˥',
    E3B: 'tɑm˧˥',
    G3B: 'dɑm˧˥',
    O3B: 'tsʰɑm˧˥',
    D3B: 'mɑm˧˥',
    N3B: 'tsɑm˧˥',
    P3B: 'dzɑm˧˥',
    i3B: 'xɑm˧˥',
    h3B: 'ʔɑm˧˥',
    e3B: 'kʰɑm˧˥',
    l4B: 'jiɛm˧˥',
    I4B: 'liɛm˧˥',
    i4F: 'xɯiɛm˧˥',
    A4F: 'pɯiɛm˧˥',
    X4B: 'tɕiɛm˧˥',
    e4F: 'kʰɯiɛm˧˥',
    g4F: 'ŋɯiɛm˧˥',
    f4F: 'gɯiɛm˧˥',
    d4F: 'kɯiɛm˧˥',
    h4B: 'ʔiɛm˧˥',
    c4B: 'ȵiɛm˧˥',
    a4B: 'ɕiɛm˧˥',
    K4B: 'ʈʰiɛm˧˥',
    h4F: 'ʔɯiɛm˧˥',
    P4B: 'dziɛm˧˥',
    O4B: 'tsʰiɛm˧˥',
    N4B: 'tsiɛm˧˥',
    e4B: 'kʰiɛm˧˥',
    Z4B: 'dʑiɛm˧˥',
    F5B: 'tʰem˧˥',
    H5B: 'nem˧˥',
    E5B: 'tem˧˥',
    G5B: 'dem˧˥',
    e5B: 'kʰem˧˥',
    I5B: 'lem˧˥',
    j5B: 'ɣem˧˥',
    O5B: 'tsʰem˧˥',
    d5B: 'kem˧˥',
    D5B: 'mem˧˥',
    g8B: 'ŋiɐm˧˥',
    e8B: 'kʰiɐm˧˥',
    h8B: 'ʔiɐm˧˥',
    j6B: 'ɣɯæm˧˥',
    L6B: 'ɖɯæm˧˥',
    e6B: 'kʰɯæm˧˥',
    d6B: 'kɯæm˧˥',
    U6B: 'dʐɯæm˧˥',
    I6B: 'lɯæm˧˥',
    S6B: 'tʂɯæm˧˥',
    T6B: 'tʂʰɯæm˧˥',
    i6B: 'xɯæm˧˥',
    V6B: 'ʂɯæm˧˥',
    h6B: 'ʔɯæm˧˥',
    M6B: 'ɳɯæm˧˥',
    K6B: 'ʈʰɯæm˧˥',
    j7B: 'ɣɯam˧˥',
    e7B: 'kʰɯam˧˥',
    V7B: 'ʂɯam˧˥',
    h7B: 'ʔɯam˧˥',
    T7B: 'tʂʰɯam˧˥',
    i7B: 'xɯam˧˥',
    U7B: 'dʐɯam˧˥',
    C9B: 'biɐm˧˥',
    D9B: 'miɐm˧˥',
    A9B: 'piɐm˧˥',
    e9J: 'kʰiɐm˧˥',
    B9B: 'pʰiɐm˧˥',
    K9J: 'ʈʰiɐm˧˥',
    QAC: 'suŋ˥˨',
    CBC: 'biuŋ˥˨',
    dAC: 'kuŋ˥˨',
    IAC: 'luŋ˥˨',
    EAC: 'tuŋ˥˨',
    eAC: 'kʰuŋ˥˨',
    NAC: 'tsuŋ˥˨',
    hAC: 'ʔuŋ˥˨',
    OAC: 'tsʰuŋ˥˨',
    GAC: 'duŋ˥˨',
    FAC: 'tʰuŋ˥˨',
    LBC: 'ɖiuŋ˥˨',
    ABC: 'piuŋ˥˨',
    eBC: 'kʰiuŋ˥˨',
    DBC: 'miuŋ˥˨',
    iBC: 'xiuŋ˥˨',
    DAC: 'muŋ˥˨',
    OBC: 'tsʰiuŋ˥˨',
    BBC: 'pʰiuŋ˥˨',
    PAC: 'dzuŋ˥˨',
    JBC: 'ʈiuŋ˥˨',
    jAC: 'ɣuŋ˥˨',
    HAC: 'nuŋ˥˨',
    XBC: 'tɕiuŋ˥˨',
    YBC: 'tɕʰiuŋ˥˨',
    UBC: 'dʐiuŋ˥˨',
    iAC: 'xuŋ˥˨',
    QCC: 'suoŋ˥˨',
    NCC: 'tsuoŋ˥˨',
    FCC: 'tʰuoŋ˥˨',
    DCC: 'muoŋ˥˨',
    jCC: 'ɣuoŋ˥˨',
    lDC: 'jioŋ˥˨',
    RDC: 'zioŋ˥˨',
    CDC: 'bioŋ˥˨',
    fDC: 'gioŋ˥˨',
    ADC: 'pioŋ˥˨',
    dDC: 'kioŋ˥˨',
    hDC: 'ʔioŋ˥˨',
    JDC: 'ʈioŋ˥˨',
    NDC: 'tsioŋ˥˨',
    KDC: 'ʈʰioŋ˥˨',
    cDC: 'ȵioŋ˥˨',
    XDC: 'tɕioŋ˥˨',
    LDC: 'ɖioŋ˥˨',
    IDC: 'lioŋ˥˨',
    eDC: 'kʰioŋ˥˨',
    PDC: 'dzioŋ˥˨',
    MDC: 'ɳioŋ˥˨',
    dEC: 'kɯɔŋ˥˨',
    jEC: 'ɣɯɔŋ˥˨',
    JEC: 'ʈɯɔŋ˥˨',
    LEC: 'ɖɯɔŋ˥˨',
    KEC: 'ʈʰɯɔŋ˥˨',
    UEC: 'dʐɯɔŋ˥˨',
    BEC: 'pʰɯɔŋ˥˨',
    TEC: 'tʂʰɯɔŋ˥˨',
    VEC: 'ʂɯɔŋ˥˨',
    XFC: 'tɕiɛ˥˨',
    CFC: 'biɛ˥˨',
    XFK: 'tɕʷiɛ˥˨',
    IFC: 'liɛ˥˨',
    ZFC: 'dʑiɛ˥˨',
    NFC: 'tsiɛ˥˨',
    QFC: 'siɛ˥˨',
    kFK: 'ɦʷiɛ˥˨',
    dFO: 'kʷɯiɛ˥˨',
    BFG: 'pʰɯiɛ˥˨',
    AFG: 'pɯiɛ˥˨',
    CFG: 'bɯiɛ˥˨',
    IFK: 'lʷiɛ˥˨',
    dFG: 'kɯiɛ˥˨',
    AFC: 'piɛ˥˨',
    fFG: 'gɯiɛ˥˨',
    OFC: 'tsʰiɛ˥˨',
    lFC: 'jiɛ˥˨',
    gFG: 'ŋɯiɛ˥˨',
    BFC: 'pʰiɛ˥˨',
    PFC: 'dziɛ˥˨',
    JFC: 'ʈiɛ˥˨',
    hFG: 'ʔɯiɛ˥˨',
    LFK: 'ɖʷiɛ˥˨',
    YFK: 'tɕʰʷiɛ˥˨',
    iFG: 'xɯiɛ˥˨',
    eFC: 'kʰiɛ˥˨',
    hFC: 'ʔiɛ˥˨',
    aFC: 'ɕiɛ˥˨',
    VFC: 'ʂiɛ˥˨',
    eFK: 'kʰʷiɛ˥˨',
    hFO: 'ʔʷɯiɛ˥˨',
    gFO: 'ŋʷɯiɛ˥˨',
    iFO: 'xʷɯiɛ˥˨',
    hFK: 'ʔʷiɛ˥˨',
    ZFK: 'dʑʷiɛ˥˨',
    dFC: 'kiɛ˥˨',
    cFK: 'ȵʷiɛ˥˨',
    YFC: 'tɕʰiɛ˥˨',
    QFK: 'sʷiɛ˥˨',
    JFK: 'ʈʷiɛ˥˨',
    lFK: 'jʷiɛ˥˨',
    MFK: 'ɳʷiɛ˥˨',
    dFK: 'kʷiɛ˥˨',
    SFC: 'tʂiɛ˥˨',
    iFK: 'xʷiɛ˥˨',
    eFG: 'kʰɯiɛ˥˨',
    XGC: 'tɕi˥˨',
    kGK: 'ɦʷi˥˨',
    DGG: 'mɯi˥˨',
    RGK: 'zʷi˥˨',
    NGK: 'tsʷi˥˨',
    QGK: 'sʷi˥˨',
    IGK: 'lʷi˥˨',
    AGG: 'pɯi˥˨',
    fGO: 'gʷɯi˥˨',
    BGG: 'pʰɯi˥˨',
    CGG: 'bɯi˥˨',
    dGO: 'kʷɯi˥˨',
    VGK: 'ʂʷi˥˨',
    eGO: 'kʰʷɯi˥˨',
    iGO: 'xʷɯi˥˨',
    ZGC: 'dʑi˥˨',
    IGC: 'li˥˨',
    MGC: 'ɳi˥˨',
    BGC: 'pʰi˥˨',
    gGG: 'ŋɯi˥˨',
    JGC: 'ʈi˥˨',
    eGC: 'kʰi˥˨',
    LGC: 'ɖi˥˨',
    DGC: 'mi˥˨',
    KGC: 'ʈʰi˥˨',
    dGG: 'kɯi˥˨',
    fGG: 'gɯi˥˨',
    fGK: 'gʷi˥˨',
    OGK: 'tsʰʷi˥˨',
    cGC: 'ȵi˥˨',
    NGC: 'tsi˥˨',
    OGC: 'tsʰi˥˨',
    hGG: 'ʔɯi˥˨',
    QGC: 'si˥˨',
    eGG: 'kʰɯi˥˨',
    dGK: 'kʷi˥˨',
    CGC: 'bi˥˨',
    iGK: 'xʷi˥˨',
    AGC: 'pi˥˨',
    PGK: 'dzʷi˥˨',
    GGC: 'di˥˨',
    iGG: 'xɯi˥˨',
    lGC: 'ji˥˨',
    bGC: 'ʑi˥˨',
    PGC: 'dzi˥˨',
    LGK: 'ɖʷi˥˨',
    YGK: 'tɕʰʷi˥˨',
    lGK: 'jʷi˥˨',
    JGK: 'ʈʷi˥˨',
    YGC: 'tɕʰi˥˨',
    aGC: 'ɕi˥˨',
    TGK: 'tʂʰʷi˥˨',
    aGK: 'ɕʷi˥˨',
    XHC: 'tɕɨ˥˨',
    LHC: 'ɖɨ˥˨',
    RHC: 'zɨ˥˨',
    OHC: 'tsʰɨ˥˨',
    QHC: 'sɨ˥˨',
    aHC: 'ɕɨ˥˨',
    SHC: 'tʂɨ˥˨',
    IHC: 'lɨ˥˨',
    PHC: 'dzɨ˥˨',
    KHC: 'ʈʰɨ˥˨',
    cHC: 'ȵɨ˥˨',
    VHC: 'ʂɨ˥˨',
    THC: 'tʂʰɨ˥˨',
    lHC: 'jɨ˥˨',
    JHC: 'ʈɨ˥˨',
    ZHC: 'dʑɨ˥˨',
    UHC: 'dʐɨ˥˨',
    fHC: 'gɨ˥˨',
    YHC: 'tɕʰɨ˥˨',
    hHC: 'ʔɨ˥˨',
    dHC: 'kɨ˥˨',
    iHC: 'xɨ˥˨',
    gHC: 'ŋɨ˥˨',
    eHC: 'kʰɨ˥˨',
    DIC: 'mʷɨi˥˨',
    dIK: 'kʷɨi˥˨',
    kIK: 'ɦʷɨi˥˨',
    gIK: 'ŋʷɨi˥˨',
    AIC: 'pʷɨi˥˨',
    BIC: 'pʰʷɨi˥˨',
    eIK: 'kʰʷɨi˥˨',
    hIK: 'ʔʷɨi˥˨',
    iIK: 'xʷɨi˥˨',
    CIC: 'bʷɨi˥˨',
    dIC: 'kɨi˥˨',
    gIC: 'ŋɨi˥˨',
    eIC: 'kʰɨi˥˨',
    iIC: 'xɨi˥˨',
    fIC: 'gɨi˥˨',
    hIC: 'ʔɨi˥˨',
    gJC: 'ŋiɔ˥˨',
    IJC: 'liɔ˥˨',
    dJC: 'kiɔ˥˨',
    OJC: 'tsʰiɔ˥˨',
    eJC: 'kʰiɔ˥˨',
    ZJC: 'dʑiɔ˥˨',
    aJC: 'ɕiɔ˥˨',
    JJC: 'ʈiɔ˥˨',
    XJC: 'tɕiɔ˥˨',
    VJC: 'ʂiɔ˥˨',
    hJC: 'ʔiɔ˥˨',
    LJC: 'ɖiɔ˥˨',
    fJC: 'giɔ˥˨',
    QJC: 'siɔ˥˨',
    UJC: 'dʐiɔ˥˨',
    NJC: 'tsiɔ˥˨',
    SJC: 'tʂiɔ˥˨',
    cJC: 'ȵiɔ˥˨',
    lJC: 'jiɔ˥˨',
    iJC: 'xiɔ˥˨',
    MJC: 'ɳiɔ˥˨',
    TJC: 'tʂʰiɔ˥˨',
    YJC: 'tɕʰiɔ˥˨',
    KJC: 'ʈʰiɔ˥˨',
    RJC: 'ziɔ˥˨',
    gKC: 'ŋio˥˨',
    hKC: 'ʔio˥˨',
    ZKC: 'dʑio˥˨',
    LKC: 'ɖio˥˨',
    CKC: 'bio˥˨',
    XKC: 'tɕio˥˨',
    dKC: 'kio˥˨',
    iKC: 'xio˥˨',
    aKC: 'ɕio˥˨',
    lKC: 'jio˥˨',
    cKC: 'ȵio˥˨',
    BKC: 'pʰio˥˨',
    DKC: 'mio˥˨',
    NKC: 'tsio˥˨',
    fKC: 'gio˥˨',
    kKC: 'ɦio˥˨',
    PKC: 'dzio˥˨',
    VKC: 'ʂio˥˨',
    AKC: 'pio˥˨',
    OKC: 'tsʰio˥˨',
    JKC: 'ʈio˥˨',
    eKC: 'kʰio˥˨',
    TKC: 'tʂʰio˥˨',
    KKC: 'ʈʰio˥˨',
    QKC: 'sio˥˨',
    IKC: 'lio˥˨',
    DLC: 'muo˥˨',
    GLC: 'duo˥˨',
    ILC: 'luo˥˨',
    ELC: 'tuo˥˨',
    FLC: 'tʰuo˥˨',
    dLC: 'kuo˥˨',
    gLC: 'ŋuo˥˨',
    jLC: 'ɣuo˥˨',
    QLC: 'suo˥˨',
    PLC: 'dzuo˥˨',
    HLC: 'nuo˥˨',
    ALC: 'puo˥˨',
    hLC: 'ʔuo˥˨',
    BLC: 'pʰuo˥˨',
    OLC: 'tsʰuo˥˨',
    eLC: 'kʰuo˥˨',
    CLC: 'buo˥˨',
    iLC: 'xuo˥˨',
    NLC: 'tsuo˥˨',
    NMC: 'tsei˥˨',
    EMC: 'tei˥˨',
    PMC: 'dzei˥˨',
    FMC: 'tʰei˥˨',
    GMC: 'dei˥˨',
    OMC: 'tsʰei˥˨',
    QMC: 'sei˥˨',
    gMC: 'ŋei˥˨',
    dMC: 'kei˥˨',
    jMC: 'ɣei˥˨',
    eMC: 'kʰei˥˨',
    hMC: 'ʔei˥˨',
    DMC: 'mei˥˨',
    AMC: 'pei˥˨',
    jMK: 'ɣʷei˥˨',
    dMK: 'kʷei˥˨',
    iMK: 'xʷei˥˨',
    BMC: 'pʰei˥˨',
    CMC: 'bei˥˨',
    IMC: 'lei˥˨',
    KMC: 'ʈʰei˥˨',
    HMC: 'nei˥˨',
    iMC: 'xei˥˨',
    NNC: 'tsiɛi˥˨',
    QNK: 'sʷiei˥˨',
    kNK: 'ɦʷiei˥˨',
    cNK: 'ȵʷiei˥˨',
    XNK: 'tɕʷiei˥˨',
    VNK: 'ʂʷiei˥˨',
    ONK: 'tsʰʷiei˥˨',
    lNK: 'jʷiei˥˨',
    JNK: 'ʈʷiei˥˨',
    aNK: 'ɕʷiei˥˨',
    CNC: 'biɛi˥˨',
    TNK: 'tʂʰʷiei˥˨',
    RNK: 'zʷiei˥˨',
    ANC: 'piɛi˥˨',
    dNO: 'kʷɯiei˥˨',
    DNC: 'miɛi˥˨',
    VNC: 'ʂiɛi˥˨',
    YNC: 'tɕʰiɛi˥˨',
    XNC: 'tɕiɛi˥˨',
    ZNC: 'dʑiɛi˥˨',
    lNC: 'jiɛi˥˨',
    hNG: 'ʔɯiɛi˥˨',
    gNC: 'ŋiɛi˥˨',
    LNC: 'ɖiɛi˥˨',
    INC: 'liɛi˥˨',
    eNG: 'kʰɯiɛi˥˨',
    aNC: 'ɕiɛi˥˨',
    dNG: 'kɯiɛi˥˨',
    JNC: 'ʈiɛi˥˨',
    fNG: 'gɯiɛi˥˨',
    ZNK: 'dʑʷiei˥˨',
    KNC: 'ʈʰiɛi˥˨',
    LNK: 'ɖʷiei˥˨',
    gNG: 'ŋɯiɛi˥˨',
    NNK: 'tsʷiei˥˨',
    BNC: 'pʰiɛi˥˨',
    eUK: 'kʰʷiɐi˥˨',
    iUK: 'xʷiɐi˥˨',
    FOC: 'tʰɑi˥˨',
    dOC: 'kɑi˥˨',
    gOC: 'ŋɑi˥˨',
    hOC: 'ʔɑi˥˨',
    HOC: 'nɑi˥˨',
    GOC: 'dɑi˥˨',
    jOC: 'ɣɑi˥˨',
    EOC: 'tɑi˥˨',
    AOC: 'pɑi˥˨',
    BOC: 'pʰɑi˥˨',
    jOK: 'ɣʷɑi˥˨',
    GOK: 'dʷɑi˥˨',
    dOK: 'kʷɑi˥˨',
    NOK: 'tsʷɑi˥˨',
    iOK: 'xʷɑi˥˨',
    eOK: 'kʰʷɑi˥˨',
    IOK: 'lʷɑi˥˨',
    gOK: 'ŋʷɑi˥˨',
    EOK: 'tʷɑi˥˨',
    hOK: 'ʔʷɑi˥˨',
    POK: 'dzʷɑi˥˨',
    QOK: 'sʷɑi˥˨',
    OOK: 'tsʰʷɑi˥˨',
    COC: 'bɑi˥˨',
    eOC: 'kʰɑi˥˨',
    OOC: 'tsʰɑi˥˨',
    IOC: 'lɑi˥˨',
    iOC: 'xɑi˥˨',
    FOK: 'tʰʷɑi˥˨',
    DOC: 'mɑi˥˨',
    dPK: 'kʷɯæ˥˨',
    dPC: 'kɯæ˥˨',
    hPC: 'ʔɯæ˥˨',
    jPC: 'ɣɯæ˥˨',
    DPC: 'mɯæ˥˨',
    jPK: 'ɣʷɯæ˥˨',
    TPC: 'tʂʰɯæ˥˨',
    gPC: 'ŋɯæ˥˨',
    iPC: 'xɯæ˥˨',
    APC: 'pɯæ˥˨',
    CPC: 'bɯæ˥˨',
    UPC: 'dʐɯæ˥˨',
    BPC: 'pʰɯæ˥˨',
    SPC: 'tʂɯæ˥˨',
    ePC: 'kʰɯæ˥˨',
    VPC: 'ʂɯæ˥˨',
    iPK: 'xʷɯæ˥˨',
    JPK: 'ʈʷɯæ˥˨',
    dQK: 'kʷɯæi˥˨',
    hQC: 'ʔɯæi˥˨',
    SQC: 'tʂɯæi˥˨',
    dQC: 'kɯæi˥˨',
    MQC: 'ɳɯæi˥˨',
    iQC: 'xɯæi˥˨',
    jQC: 'ɣɯæi˥˨',
    gQC: 'ŋɯæi˥˨',
    eQK: 'kʰʷɯæi˥˨',
    AQC: 'pɯæi˥˨',
    BQC: 'pʰɯæi˥˨',
    jQK: 'ɣʷɯæi˥˨',
    gQK: 'ŋʷɯæi˥˨',
    CQC: 'bɯæi˥˨',
    DQC: 'mɯæi˥˨',
    iQK: 'xʷɯæi˥˨',
    VQC: 'ʂɯæi˥˨',
    eQC: 'kʰɯæi˥˨',
    KQK: 'ʈʰʷɯæi˥˨',
    dRK: 'kʷɯai˥˨',
    eRK: 'kʰʷɯai˥˨',
    DRC: 'mɯai˥˨',
    jRK: 'ɣʷɯai˥˨',
    CRC: 'bɯai˥˨',
    hRK: 'ʔʷɯai˥˨',
    TRK: 'tʂʰʷɯai˥˨',
    dRC: 'kɯai˥˨',
    KRC: 'ʈʰɯai˥˨',
    hRC: 'ʔɯai˥˨',
    VRC: 'ʂɯai˥˨',
    LRC: 'ɖɯai˥˨',
    iRC: 'xɯai˥˨',
    iRK: 'xʷɯai˥˨',
    ARC: 'pɯai˥˨',
    ORK: 'tsʰʷɯai˥˨',
    URC: 'dʐɯai˥˨',
    jRC: 'ɣɯai˥˨',
    GSK: 'duoi˥˨',
    CSC: 'buoi˥˨',
    DSC: 'muoi˥˨',
    BSC: 'pʰuoi˥˨',
    iSK: 'xuoi˥˨',
    ESK: 'tuoi˥˨',
    OSK: 'tsʰuoi˥˨',
    NSK: 'tsuoi˥˨',
    hSK: 'ʔuoi˥˨',
    FSK: 'tʰuoi˥˨',
    dSK: 'kuoi˥˨',
    jSK: 'ɣuoi˥˨',
    eSK: 'kʰuoi˥˨',
    QSK: 'suoi˥˨',
    HSK: 'nuoi˥˨',
    ISK: 'luoi˥˨',
    ASC: 'puoi˥˨',
    gSK: 'ŋuoi˥˨',
    GTC: 'dəi˥˨',
    NTC: 'tsəi˥˨',
    DTC: 'məi˥˨',
    QTC: 'səi˥˨',
    FTC: 'tʰəi˥˨',
    dTC: 'kəi˥˨',
    eTC: 'kʰəi˥˨',
    gTC: 'ŋəi˥˨',
    hTC: 'ʔəi˥˨',
    jTC: 'ɣəi˥˨',
    HTC: 'nəi˥˨',
    ETC: 'təi˥˨',
    ITC: 'ləi˥˨',
    OTC: 'tsʰəi˥˨',
    PTC: 'dzəi˥˨',
    iTC: 'xəi˥˨',
    AUC: 'pʷiɐi˥˨',
    BUC: 'pʰʷiɐi˥˨',
    hUK: 'ʔʷiɐi˥˨',
    CUC: 'bʷiɐi˥˨',
    fUK: 'gʷiɐi˥˨',
    gUC: 'ŋiɐi˥˨',
    XVC: 'tɕin˥˨',
    QVC: 'sin˥˨',
    cVC: 'ȵin˥˨',
    lVC: 'jin˥˨',
    IVC: 'lin˥˨',
    AVC: 'pin˥˨',
    LVC: 'ɖin˥˨',
    ZVC: 'dʑin˥˨',
    aVC: 'ɕin˥˨',
    eVG: 'kʰɯin˥˨',
    RVC: 'zin˥˨',
    gVG: 'ŋɯin˥˨',
    NVC: 'tsin˥˨',
    iVG: 'xɯin˥˨',
    JVC: 'ʈin˥˨',
    fVG: 'gɯin˥˨',
    TWC: 'tʂʰɪn˥˨',
    hVC: 'ʔin˥˨',
    KVC: 'ʈʰin˥˨',
    BVC: 'pʰin˥˨',
    OVC: 'tsʰin˥˨',
    eVC: 'kʰin˥˨',
    dVK: 'kʷin˥˨',
    XVK: 'tɕʷin˥˨',
    QVK: 'sʷin˥˨',
    RVK: 'zʷin˥˨',
    NVK: 'tsʷin˥˨',
    aVK: 'ɕʷin˥˨',
    cVK: 'ȵʷin˥˨',
    bVK: 'ʑʷin˥˨',
    DXC: 'miun˥˨',
    kXK: 'ɦiun˥˨',
    iXK: 'xiun˥˨',
    BXC: 'pʰiun˥˨',
    AXC: 'piun˥˨',
    hXK: 'ʔiun˥˨',
    dXK: 'kiun˥˨',
    fXK: 'giun˥˨',
    CXC: 'biun˥˨',
    iYC: 'xɨn˥˨',
    dYC: 'kɨn˥˨',
    fYC: 'gɨn˥˨',
    hYC: 'ʔɨn˥˨',
    gYC: 'ŋɨn˥˨',
    gZK: 'ŋʷiɐn˥˨',
    hZK: 'ʔʷiɐn˥˨',
    AZC: 'pʷiɐn˥˨',
    eZK: 'kʰʷiɐn˥˨',
    DZC: 'mʷiɐn˥˨',
    CZC: 'bʷiɐn˥˨',
    BZC: 'pʰʷiɐn˥˨',
    dZC: 'kiɐn˥˨',
    hZC: 'ʔiɐn˥˨',
    iZC: 'xiɐn˥˨',
    iZK: 'xʷiɐn˥˨',
    fZC: 'giɐn˥˨',
    kZK: 'ɦʷiɐn˥˨',
    gZC: 'ŋiɐn˥˨',
    fZK: 'gʷiɐn˥˨',
    dZK: 'kʷiɐn˥˨',
    jaK: 'ɣuon˥˨',
    EaK: 'tuon˥˨',
    QaK: 'suon˥˨',
    eaK: 'kʰuon˥˨',
    HaK: 'nuon˥˨',
    haK: 'ʔuon˥˨',
    DaC: 'muon˥˨',
    PaK: 'dzuon˥˨',
    daK: 'kuon˥˨',
    BaC: 'pʰuon˥˨',
    GaK: 'duon˥˨',
    OaK: 'tsʰuon˥˨',
    CaC: 'buon˥˨',
    gaK: 'ŋuon˥˨',
    IaK: 'luon˥˨',
    AaC: 'puon˥˨',
    iaK: 'xuon˥˨',
    NaK: 'tsuon˥˨',
    jbC: 'ɣən˥˨',
    dbC: 'kən˥˨',
    gbC: 'ŋən˥˨',
    hbC: 'ʔən˥˨',
    jcC: 'ɣɑn˥˨',
    FcC: 'tʰɑn˥˨',
    hcC: 'ʔɑn˥˨',
    EcC: 'tɑn˥˨',
    GcC: 'dɑn˥˨',
    dcC: 'kɑn˥˨',
    gcC: 'ŋɑn˥˨',
    ecC: 'kʰɑn˥˨',
    icC: 'xɑn˥˨',
    IcC: 'lɑn˥˨',
    HcC: 'nɑn˥˨',
    OcC: 'tsʰɑn˥˨',
    QcC: 'sɑn˥˨',
    NcC: 'tsɑn˥˨',
    PcC: 'dzɑn˥˨',
    jcK: 'ɣʷɑn˥˨',
    NcK: 'tsʷɑn˥˨',
    hcK: 'ʔʷɑn˥˨',
    dcK: 'kʷɑn˥˨',
    OcK: 'tsʰʷɑn˥˨',
    gcK: 'ŋʷɑn˥˨',
    GcK: 'dʷɑn˥˨',
    IcK: 'lʷɑn˥˨',
    EcK: 'tʷɑn˥˨',
    FcK: 'tʰʷɑn˥˨',
    icK: 'xʷɑn˥˨',
    QcK: 'sʷɑn˥˨',
    DcC: 'mʷɑn˥˨',
    AcC: 'pʷɑn˥˨',
    BcC: 'pʰʷɑn˥˨',
    CcC: 'bʷɑn˥˨',
    HcK: 'nʷɑn˥˨',
    PcK: 'dzʷɑn˥˨',
    ecK: 'kʰʷɑn˥˨',
    ddC: 'kɯan˥˨',
    gdC: 'ŋɯan˥˨',
    hdC: 'ʔɯan˥˨',
    VdC: 'ʂɯan˥˨',
    jdC: 'ɣɯan˥˨',
    DdC: 'mɯan˥˨',
    hdK: 'ʔʷɯan˥˨',
    jdK: 'ɣʷɯan˥˨',
    ddK: 'kʷɯan˥˨',
    VdK: 'ʂʷɯan˥˨',
    TdK: 'tʂʰʷɯan˥˨',
    gdK: 'ŋʷɯan˥˨',
    UdC: 'dʐɯan˥˨',
    BdC: 'pʰɯan˥˨',
    MdK: 'ɳʷɯan˥˨',
    TdC: 'tʂʰɯan˥˨',
    KdC: 'ʈʰɯan˥˨',
    deC: 'kɯæn˥˨',
    jeC: 'ɣɯæn˥˨',
    CeC: 'bɯæn˥˨',
    BeC: 'pʰɯæn˥˨',
    jeK: 'ɣʷɯæn˥˨',
    DeC: 'mɯæn˥˨',
    LeC: 'ɖɯæn˥˨',
    AeC: 'pɯæn˥˨',
    deK: 'kʷɯæn˥˨',
    QgC: 'sen˥˨',
    OgC: 'tsʰen˥˨',
    igK: 'xʷen˥˨',
    jgK: 'ɣʷen˥˨',
    dgK: 'kʷen˥˨',
    GgC: 'den˥˨',
    FgC: 'tʰen˥˨',
    IgC: 'len˥˨',
    dgC: 'ken˥˨',
    HgC: 'nen˥˨',
    egC: 'kʰen˥˨',
    jgC: 'ɣen˥˨',
    ggC: 'ŋen˥˨',
    hgC: 'ʔen˥˨',
    NgC: 'tsen˥˨',
    DgC: 'men˥˨',
    BgC: 'pʰen˥˨',
    PgC: 'dzen˥˨',
    hgK: 'ʔʷen˥˨',
    EgC: 'ten˥˨',
    igC: 'xen˥˨',
    QfC: 'siɛn˥˨',
    XfC: 'tɕiɛn˥˨',
    ZfC: 'dʑiɛn˥˨',
    gfG: 'ŋɯiɛn˥˨',
    efC: 'kʰiɛn˥˨',
    dfK: 'kʷiɛn˥˨',
    kfK: 'ɦʷiɛn˥˨',
    DfC: 'miɛn˥˨',
    YfK: 'tɕʰʷiɛn˥˨',
    lfK: 'jʷiɛn˥˨',
    JfC: 'ʈiɛn˥˨',
    cfK: 'ȵʷiɛn˥˨',
    NfC: 'tsiɛn˥˨',
    YfC: 'tɕʰiɛn˥˨',
    afC: 'ɕiɛn˥˨',
    hfC: 'ʔiɛn˥˨',
    dfO: 'kʷɯiɛn˥˨',
    ffO: 'gʷɯiɛn˥˨',
    IfK: 'lʷiɛn˥˨',
    KfK: 'ʈʰʷiɛn˥˨',
    AfG: 'pɯiɛn˥˨',
    VfK: 'ʂʷiɛn˥˨',
    OfK: 'tsʰʷiɛn˥˨',
    CfG: 'bɯiɛn˥˨',
    RfK: 'zʷiɛn˥˨',
    QfK: 'sʷiɛn˥˨',
    UfK: 'dʐʷiɛn˥˨',
    SfK: 'tʂʷiɛn˥˨',
    LfK: 'ɖʷiɛn˥˨',
    PfC: 'dziɛn˥˨',
    RfC: 'ziɛn˥˨',
    BfC: 'pʰiɛn˥˨',
    ZfK: 'dʑʷiɛn˥˨',
    MfC: 'ɳiɛn˥˨',
    JfK: 'ʈʷiɛn˥˨',
    lfC: 'jiɛn˥˨',
    CfC: 'biɛn˥˨',
    LfC: 'ɖiɛn˥˨',
    IfC: 'liɛn˥˨',
    efO: 'kʰʷɯiɛn˥˨',
    XfK: 'tɕʷiɛn˥˨',
    AgC: 'pen˥˨',
    QhC: 'seu˥˨',
    FhC: 'tʰeu˥˨',
    EhC: 'teu˥˨',
    dhC: 'keu˥˨',
    HhC: 'neu˥˨',
    GhC: 'deu˥˨',
    ehC: 'kʰeu˥˨',
    IhC: 'leu˥˨',
    ghC: 'ŋeu˥˨',
    hhC: 'ʔeu˥˨',
    ihC: 'xeu˥˨',
    QiC: 'siɛu˥˨',
    XiC: 'tɕiɛu˥˨',
    liC: 'jiɛu˥˨',
    hiC: 'ʔiɛu˥˨',
    LiC: 'ɖiɛu˥˨',
    ZiC: 'dʑiɛu˥˨',
    fiG: 'gɯiɛu˥˨',
    BiC: 'pʰiɛu˥˨',
    PiC: 'dziɛu˥˨',
    DiC: 'miɛu˥˨',
    OiC: 'tsʰiɛu˥˨',
    IiC: 'liɛu˥˨',
    eiG: 'kʰɯiɛu˥˨',
    giG: 'ŋɯiɛu˥˨',
    NiC: 'tsiɛu˥˨',
    DiG: 'mɯiɛu˥˨',
    CiC: 'biɛu˥˨',
    aiC: 'ɕiɛu˥˨',
    KiC: 'ʈʰiɛu˥˨',
    AiG: 'pɯiɛu˥˨',
    fiC: 'giɛu˥˨',
    ciC: 'ȵiɛu˥˨',
    jjC: 'ɣɯau˥˨',
    djC: 'kɯau˥˨',
    ijC: 'xɯau˥˨',
    JjC: 'ʈɯau˥˨',
    AjC: 'pɯau˥˨',
    ejC: 'kʰɯau˥˨',
    DjC: 'mɯau˥˨',
    BjC: 'pʰɯau˥˨',
    KjC: 'ʈʰɯau˥˨',
    VjC: 'ʂɯau˥˨',
    LjC: 'ɖɯau˥˨',
    MjC: 'ɳɯau˥˨',
    SjC: 'tʂɯau˥˨',
    CjC: 'bɯau˥˨',
    TjC: 'tʂʰɯau˥˨',
    hjC: 'ʔɯau˥˨',
    gjC: 'ŋɯau˥˨',
    UjC: 'dʐɯau˥˨',
    jkC: 'ɣɑu˥˨',
    GkC: 'dɑu˥˨',
    EkC: 'tɑu˥˨',
    dkC: 'kɑu˥˨',
    gkC: 'ŋɑu˥˨',
    DkC: 'mɑu˥˨',
    IkC: 'lɑu˥˨',
    OkC: 'tsʰɑu˥˨',
    CkC: 'bɑu˥˨',
    AkC: 'pɑu˥˨',
    PkC: 'dzɑu˥˨',
    hkC: 'ʔɑu˥˨',
    QkC: 'sɑu˥˨',
    ekC: 'kʰɑu˥˨',
    NkC: 'tsɑu˥˨',
    ikC: 'xɑu˥˨',
    HkC: 'nɑu˥˨',
    dlC: 'kɑ˥˨',
    jlC: 'ɣɑ˥˨',
    NlC: 'tsɑ˥˨',
    ElC: 'tɑ˥˨',
    IlC: 'lɑ˥˨',
    elC: 'kʰɑ˥˨',
    glC: 'ŋɑ˥˨',
    GlC: 'dɑ˥˨',
    HlC: 'nɑ˥˨',
    QlC: 'sɑ˥˨',
    ilC: 'xɑ˥˨',
    FlC: 'tʰɑ˥˨',
    dlK: 'kuɑ˥˨',
    jlK: 'ɣuɑ˥˨',
    NlK: 'tsuɑ˥˨',
    elK: 'kʰuɑ˥˨',
    FlK: 'tʰuɑ˥˨',
    AlC: 'puɑ˥˨',
    OlK: 'tsʰuɑ˥˨',
    DlC: 'muɑ˥˨',
    HlK: 'nuɑ˥˨',
    BlC: 'pʰuɑ˥˨',
    PlK: 'dzuɑ˥˨',
    glK: 'ŋuɑ˥˨',
    ilK: 'xuɑ˥˨',
    GlK: 'duɑ˥˨',
    ClC: 'buɑ˥˨',
    IlK: 'luɑ˥˨',
    ElK: 'tuɑ˥˨',
    QlK: 'suɑ˥˨',
    hlC: 'ʔɑ˥˨',
    hlK: 'ʔuɑ˥˨',
    DnC: 'mɯa˥˨',
    dnC: 'kɯa˥˨',
    hnC: 'ʔɯa˥˨',
    inC: 'xɯa˥˨',
    gnC: 'ŋɯa˥˨',
    KnC: 'ʈʰɯa˥˨',
    JnC: 'ʈɯa˥˨',
    SnC: 'tʂɯa˥˨',
    UnC: 'dʐɯa˥˨',
    RoC: 'zia˥˨',
    enC: 'kʰɯa˥˨',
    jnC: 'ɣɯa˥˨',
    PoC: 'dzia˥˨',
    loC: 'jia˥˨',
    YoC: 'tɕʰia˥˨',
    QoC: 'sia˥˨',
    XoC: 'tɕia˥˨',
    NoC: 'tsia˥˨',
    aoC: 'ɕia˥˨',
    boC: 'ʑia˥˨',
    AnC: 'pɯa˥˨',
    BnC: 'pʰɯa˥˨',
    jnK: 'ɣʷɯa˥˨',
    inK: 'xʷɯa˥˨',
    enK: 'kʰʷɯa˥˨',
    VnK: 'ʂʷɯa˥˨',
    CnC: 'bɯa˥˨',
    HnC: 'nɯa˥˨',
    OoC: 'tsʰia˥˨',
    VnC: 'ʂɯa˥˨',
    dnK: 'kʷɯa˥˨',
    LnC: 'ɖɯa˥˨',
    gnK: 'ŋʷɯa˥˨',
    hnK: 'ʔʷɯa˥˨',
    lpC: 'jiɐŋ˥˨',
    IpC: 'liɐŋ˥˨',
    UpC: 'dʐiɐŋ˥˨',
    cpC: 'ȵiɐŋ˥˨',
    apC: 'ɕiɐŋ˥˨',
    JpC: 'ʈiɐŋ˥˨',
    KpC: 'ʈʰiɐŋ˥˨',
    ipC: 'xiɐŋ˥˨',
    LpC: 'ɖiɐŋ˥˨',
    MpC: 'ɳiɐŋ˥˨',
    PpC: 'dziɐŋ˥˨',
    XpC: 'tɕiɐŋ˥˨',
    ZpC: 'dʑiɐŋ˥˨',
    SpC: 'tʂiɐŋ˥˨',
    hpC: 'ʔiɐŋ˥˨',
    fpC: 'giɐŋ˥˨',
    YpC: 'tɕʰiɐŋ˥˨',
    TpC: 'tʂʰiɐŋ˥˨',
    NpC: 'tsiɐŋ˥˨',
    gpC: 'ŋiɐŋ˥˨',
    BpC: 'pʰʷiɐŋ˥˨',
    DpC: 'mʷiɐŋ˥˨',
    ipK: 'xʷiɐŋ˥˨',
    dpK: 'kʷiɐŋ˥˨',
    kpK: 'ɦʷiɐŋ˥˨',
    ApC: 'pʷiɐŋ˥˨',
    QpC: 'siɐŋ˥˨',
    dpC: 'kiɐŋ˥˨',
    epC: 'kʰiɐŋ˥˨',
    OpC: 'tsʰiɐŋ˥˨',
    fpK: 'gʷiɐŋ˥˨',
    CpC: 'bʷiɐŋ˥˨',
    GqC: 'dɑŋ˥˨',
    IqC: 'lɑŋ˥˨',
    jqC: 'ɣɑŋ˥˨',
    hqC: 'ʔɑŋ˥˨',
    gqC: 'ŋɑŋ˥˨',
    NqC: 'tsɑŋ˥˨',
    CqC: 'bɑŋ˥˨',
    PqC: 'dzɑŋ˥˨',
    EqC: 'tɑŋ˥˨',
    eqC: 'kʰɑŋ˥˨',
    AqC: 'pɑŋ˥˨',
    FqC: 'tʰɑŋ˥˨',
    eqK: 'kʰʷɑŋ˥˨',
    HqC: 'nɑŋ˥˨',
    QqC: 'sɑŋ˥˨',
    jqK: 'ɣʷɑŋ˥˨',
    dqK: 'kʷɑŋ˥˨',
    dqC: 'kɑŋ˥˨',
    iqK: 'xʷɑŋ˥˨',
    DqC: 'mɑŋ˥˨',
    hqK: 'ʔʷɑŋ˥˨',
    hsC: 'ʔɯiaŋ˥˨',
    dsC: 'kɯiaŋ˥˨',
    fsC: 'gɯiaŋ˥˨',
    esC: 'kʰɯiaŋ˥˨',
    drC: 'kɯaŋ˥˨',
    DsC: 'mɯiaŋ˥˨',
    CsC: 'bɯiaŋ˥˨',
    DrC: 'mɯaŋ˥˨',
    jrK: 'ɣʷɯaŋ˥˨',
    AsC: 'pɯiaŋ˥˨',
    ksK: 'ɦʷɯiaŋ˥˨',
    jrC: 'ɣɯaŋ˥˨',
    hrC: 'ʔɯaŋ˥˨',
    TrC: 'tʂʰɯaŋ˥˨',
    JrC: 'ʈɯaŋ˥˨',
    KrC: 'ʈʰɯaŋ˥˨',
    ArC: 'pɯaŋ˥˨',
    LrC: 'ɖɯaŋ˥˨',
    VrC: 'ʂɯaŋ˥˨',
    gsC: 'ŋɯiaŋ˥˨',
    CrC: 'bɯaŋ˥˨',
    irC: 'xɯaŋ˥˨',
    hrK: 'ʔʷɯaŋ˥˨',
    StC: 'tʂɯæŋ˥˨',
    AtC: 'pɯæŋ˥˨',
    CtC: 'bɯæŋ˥˨',
    htC: 'ʔɯæŋ˥˨',
    gtC: 'ŋɯæŋ˥˨',
    itK: 'xʷɯæŋ˥˨',
    duC: 'kiɛŋ˥˨',
    OuC: 'tsʰiɛŋ˥˨',
    XuC: 'tɕiɛŋ˥˨',
    auC: 'ɕiɛŋ˥˨',
    LuC: 'ɖiɛŋ˥˨',
    KuC: 'ʈʰiɛŋ˥˨',
    QuC: 'siɛŋ˥˨',
    IuC: 'liɛŋ˥˨',
    BuC: 'pʰiɛŋ˥˨',
    iuK: 'xʷiɛŋ˥˨',
    AuC: 'piɛŋ˥˨',
    CuC: 'biɛŋ˥˨',
    PuC: 'dziɛŋ˥˨',
    ZuC: 'dʑiɛŋ˥˨',
    DuC: 'miɛŋ˥˨',
    euC: 'kʰiɛŋ˥˨',
    iuC: 'xiɛŋ˥˨',
    NuC: 'tsiɛŋ˥˨',
    dvC: 'keŋ˥˨',
    HvC: 'neŋ˥˨',
    QvC: 'seŋ˥˨',
    jvC: 'ɣeŋ˥˨',
    GvC: 'deŋ˥˨',
    EvC: 'teŋ˥˨',
    evC: 'kʰeŋ˥˨',
    FvC: 'tʰeŋ˥˨',
    OvC: 'tsʰeŋ˥˨',
    DvC: 'meŋ˥˨',
    hvK: 'ʔʷeŋ˥˨',
    IvC: 'leŋ˥˨',
    XwC: 'tɕɨŋ˥˨',
    lwC: 'jɨŋ˥˨',
    bwC: 'ʑɨŋ˥˨',
    cwC: 'ȵɨŋ˥˨',
    hwC: 'ʔɨŋ˥˨',
    NwC: 'tsɨŋ˥˨',
    iwC: 'xɨŋ˥˨',
    awC: 'ɕɨŋ˥˨',
    LwC: 'ɖɨŋ˥˨',
    IwC: 'lɨŋ˥˨',
    CwC: 'bɨŋ˥˨',
    YwC: 'tɕʰɨŋ˥˨',
    gwC: 'ŋɨŋ˥˨',
    ZwC: 'dʑɨŋ˥˨',
    KwC: 'ʈʰɨŋ˥˨',
    fwC: 'gɨŋ˥˨',
    ExC: 'təŋ˥˨',
    PxC: 'dzəŋ˥˨',
    dxC: 'kəŋ˥˨',
    OxC: 'tsʰəŋ˥˨',
    GxC: 'dəŋ˥˨',
    DxC: 'məŋ˥˨',
    AxC: 'pəŋ˥˨',
    CxC: 'bəŋ˥˨',
    IxC: 'ləŋ˥˨',
    NxC: 'tsəŋ˥˨',
    QxC: 'səŋ˥˨',
    FxC: 'tʰəŋ˥˨',
    kyC: 'ɦiu˥˨',
    dyC: 'kiu˥˨',
    LyC: 'ɖiu˥˨',
    JyC: 'ʈiu˥˨',
    ayC: 'ɕiu˥˨',
    YyC: 'tɕʰiu˥˨',
    RyC: 'ziu˥˨',
    iyC: 'xiu˥˨',
    XyC: 'tɕiu˥˨',
    fyC: 'giu˥˨',
    VyC: 'ʂiu˥˨',
    SyC: 'tʂiu˥˨',
    ByC: 'pʰiu˥˨',
    TyC: 'tʂʰiu˥˨',
    AyC: 'piu˥˨',
    KyC: 'ʈʰiu˥˨',
    IyC: 'liu˥˨',
    QyC: 'siu˥˨',
    NyC: 'tsiu˥˨',
    UyC: 'dʐiu˥˨',
    PyC: 'dziu˥˨',
    MyC: 'ɳiu˥˨',
    CyC: 'biu˥˨',
    lyC: 'jiu˥˨',
    ZyC: 'dʑiu˥˨',
    cyC: 'ȵiu˥˨',
    DyC: 'miu˥˨',
    gyC: 'ŋiu˥˨',
    eyC: 'kʰiu˥˨',
    OyC: 'tsʰiu˥˨',
    jzC: 'ɣəu˥˨',
    ezC: 'kʰəu˥˨',
    DzC: 'məu˥˨',
    BzC: 'pʰəu˥˨',
    GzC: 'dəu˥˨',
    EzC: 'təu˥˨',
    HzC: 'nəu˥˨',
    QzC: 'səu˥˨',
    NzC: 'tsəu˥˨',
    FzC: 'tʰəu˥˨',
    hzC: 'ʔəu˥˨',
    dzC: 'kəu˥˨',
    OzC: 'tsʰəu˥˨',
    IzC: 'ləu˥˨',
    izC: 'xəu˥˨',
    CzC: 'bəu˥˨',
    gzC: 'ŋəu˥˨',
    PzC: 'dzəu˥˨',
    h0C: 'ʔɨu˥˨',
    D0C: 'mɨu˥˨',
    e0C: 'kʰɨu˥˨',
    f0C: 'gɨu˥˨',
    O1C: 'tsʰim˥˨',
    N1C: 'tsim˥˨',
    c1C: 'ȵim˥˨',
    L1C: 'ɖim˥˨',
    X1C: 'tɕim˥˨',
    f1G: 'gɯim˥˨',
    d1G: 'kɯim˥˨',
    M1C: 'ɳim˥˨',
    h1G: 'ʔɯim˥˨',
    V1C: 'ʂim˥˨',
    K1C: 'ʈʰim˥˨',
    S1C: 'tʂim˥˨',
    T1C: 'tʂʰim˥˨',
    g1G: 'ŋɯim˥˨',
    J1C: 'ʈim˥˨',
    I1C: 'lim˥˨',
    Z1C: 'dʑim˥˨',
    k1C: 'ɦim˥˨',
    a1C: 'ɕim˥˨',
    e2C: 'kʰəm˥˨',
    d2C: 'kəm˥˨',
    j2C: 'ɣəm˥˨',
    h2C: 'ʔəm˥˨',
    F2C: 'tʰəm˥˨',
    Q2C: 'səm˥˨',
    O2C: 'tsʰəm˥˨',
    G2C: 'dəm˥˨',
    g2C: 'ŋəm˥˨',
    E2C: 'təm˥˨',
    H2C: 'nəm˥˨',
    I2C: 'ləm˥˨',
    i2C: 'xəm˥˨',
    N2C: 'tsəm˥˨',
    e3C: 'kʰɑm˥˨',
    I3C: 'lɑm˥˨',
    F3C: 'tʰɑm˥˨',
    d3C: 'kɑm˥˨',
    i3C: 'xɑm˥˨',
    j3C: 'ɣɑm˥˨',
    G3C: 'dɑm˥˨',
    P3C: 'dzɑm˥˨',
    E3C: 'tɑm˥˨',
    Q3C: 'sɑm˥˨',
    l4C: 'jiɛm˥˨',
    Z4C: 'dʑiɛm˥˨',
    c4C: 'ȵiɛm˥˨',
    h4C: 'ʔiɛm˥˨',
    A4G: 'pɯiɛm˥˨',
    g4G: 'ŋɯiɛm˥˨',
    a4C: 'ɕiɛm˥˨',
    N4C: 'tsiɛm˥˨',
    O4C: 'tsʰiɛm˥˨',
    I4C: 'liɛm˥˨',
    K4C: 'ʈʰiɛm˥˨',
    Y4C: 'tɕʰiɛm˥˨',
    h4G: 'ʔɯiɛm˥˨',
    P4C: 'dziɛm˥˨',
    X4C: 'tɕiɛm˥˨',
    F5C: 'tʰem˥˨',
    H5C: 'nem˥˨',
    E5C: 'tem˥˨',
    Q5C: 'sem˥˨',
    G5C: 'dem˥˨',
    d5C: 'kem˥˨',
    h5C: 'ʔem˥˨',
    N5C: 'tsem˥˨',
    P5C: 'dzem˥˨',
    e5C: 'kʰem˥˨',
    I5C: 'lem˥˨',
    g8C: 'ŋiɐm˥˨',
    i8C: 'xiɐm˥˨',
    e8C: 'kʰiɐm˥˨',
    D9C: 'miɐm˥˨',
    j6C: 'ɣɯæm˥˨',
    h6C: 'ʔɯæm˥˨',
    S6C: 'tʂɯæm˥˨',
    J6C: 'ʈɯæm˥˨',
    e6C: 'kʰɯæm˥˨',
    L6C: 'ɖɯæm˥˨',
    U6C: 'dʐɯæm˥˨',
    d6C: 'kɯæm˥˨',
    M6C: 'ɳɯæm˥˨',
    g6C: 'ŋɯæm˥˨',
    d7C: 'kɯam˥˨',
    T7C: 'tʂʰɯam˥˨',
    S7C: 'tʂɯam˥˨',
    V7C: 'ʂɯam˥˨',
    i7C: 'xɯam˥˨',
    C7C: 'bɯam˥˨',
    j7C: 'ɣɯam˥˨',
    U7C: 'dʐɯam˥˨',
    h7C: 'ʔɯam˥˨',
    C9C: 'biɐm˥˨',
    B9C: 'pʰiɐm˥˨',
    d8C: 'kiɐm˥˨',
    h8C: 'ʔiɐm˥˨',
    hAD: 'ʔuk꜊',
    GAD: 'duk꜊',
    dAD: 'kuk꜊',
    jAD: 'ɣuk꜊',
    eAD: 'kʰuk꜊',
    FAD: 'tʰuk꜊',
    EAD: 'tuk꜊',
    QAD: 'suk꜊',
    IAD: 'luk꜊',
    iAD: 'xuk꜊',
    PAD: 'dzuk꜊',
    OAD: 'tsʰuk꜊',
    NAD: 'tsuk꜊',
    CAD: 'buk꜊',
    BAD: 'pʰuk꜊',
    AAD: 'puk꜊',
    DAD: 'muk꜊',
    ABD: 'piuk꜊',
    CBD: 'biuk꜊',
    VBD: 'ʂiuk꜊',
    IBD: 'liuk꜊',
    LBD: 'ɖiuk꜊',
    dBD: 'kiuk꜊',
    eBD: 'kʰiuk꜊',
    ZBD: 'dʑiuk꜊',
    YBD: 'tɕʰiuk꜊',
    lBD: 'jiuk꜊',
    fBD: 'giuk꜊',
    OBD: 'tsʰiuk꜊',
    cBD: 'ȵiuk꜊',
    XBD: 'tɕiuk꜊',
    aBD: 'ɕiuk꜊',
    iBD: 'xiuk꜊',
    JBD: 'ʈiuk꜊',
    NBD: 'tsiuk꜊',
    TBD: 'tʂʰiuk꜊',
    MBD: 'ɳiuk꜊',
    SBD: 'tʂiuk꜊',
    BBD: 'pʰiuk꜊',
    hBD: 'ʔiuk꜊',
    QBD: 'siuk꜊',
    DBD: 'miuk꜊',
    kBD: 'ɦiuk꜊',
    KBD: 'ʈʰiuk꜊',
    gBD: 'ŋiuk꜊',
    PBD: 'dziuk꜊',
    hCD: 'ʔuok꜊',
    GCD: 'duok꜊',
    ECD: 'tuok꜊',
    eCD: 'kʰuok꜊',
    jCD: 'ɣuok꜊',
    CCD: 'buok꜊',
    QCD: 'suok꜊',
    dCD: 'kuok꜊',
    DCD: 'muok꜊',
    iCD: 'xuok꜊',
    HCD: 'nuok꜊',
    NCD: 'tsuok꜊',
    ACD: 'puok꜊',
    ICD: 'luok꜊',
    gCD: 'ŋuok꜊',
    XDD: 'tɕiok꜊',
    gDD: 'ŋiok꜊',
    iDD: 'xiok꜊',
    dDD: 'kiok꜊',
    fDD: 'giok꜊',
    ZDD: 'dʑiok꜊',
    YDD: 'tɕʰiok꜊',
    cDD: 'ȵiok꜊',
    aDD: 'ɕiok꜊',
    lDD: 'jiok꜊',
    LDD: 'ɖiok꜊',
    IDD: 'liok꜊',
    eDD: 'kʰiok꜊',
    JDD: 'ʈiok꜊',
    NDD: 'tsiok꜊',
    bDD: 'ʑiok꜊',
    CDD: 'biok꜊',
    ODD: 'tsʰiok꜊',
    RDD: 'ziok꜊',
    QDD: 'siok꜊',
    ADD: 'piok꜊',
    KDD: 'ʈʰiok꜊',
    dED: 'kɯɔk꜊',
    gED: 'ŋɯɔk꜊',
    UED: 'dʐɯɔk꜊',
    SED: 'tʂɯɔk꜊',
    VED: 'ʂɯɔk꜊',
    JED: 'ʈɯɔk꜊',
    AED: 'pɯɔk꜊',
    DED: 'mɯɔk꜊',
    CED: 'bɯɔk꜊',
    BED: 'pʰɯɔk꜊',
    eED: 'kʰɯɔk꜊',
    LED: 'ɖɯɔk꜊',
    hED: 'ʔɯɔk꜊',
    MED: 'ɳɯɔk꜊',
    KED: 'ʈʰɯɔk꜊',
    IED: 'lɯɔk꜊',
    jED: 'ɣɯɔk꜊',
    iED: 'xɯɔk꜊',
    TED: 'tʂʰɯɔk꜊',
    XVD: 'tɕit꜊',
    cVD: 'ȵit꜊',
    bVD: 'ʑit꜊',
    LVD: 'ɖit꜊',
    QVD: 'sit꜊',
    hVD: 'ʔit꜊',
    OVD: 'tsʰit꜊',
    BVD: 'pʰit꜊',
    dVD: 'kit꜊',
    MVD: 'ɳit꜊',
    lVD: 'jit꜊',
    eVD: 'kʰit꜊',
    iVD: 'xit꜊',
    KVD: 'ʈʰit꜊',
    IVD: 'lit꜊',
    JVD: 'ʈit꜊',
    PVD: 'dzit꜊',
    TWD: 'tʂʰɪt꜊',
    aVD: 'ɕit꜊',
    NVD: 'tsit꜊',
    DVD: 'mit꜊',
    AVD: 'pit꜊',
    fVH: 'gɯit꜊',
    CVD: 'bit꜊',
    kVL: 'ɦʷit꜊',
    VVL: 'ʂʷit꜊',
    YVD: 'tɕʰit꜊',
    UWD: 'dʐɪt꜊',
    DVH: 'mɯit꜊',
    CVH: 'bɯit꜊',
    hVH: 'ʔɯit꜊',
    gVH: 'ŋɯit꜊',
    AVH: 'pɯit꜊',
    JVL: 'ʈʷit꜊',
    iVH: 'xɯit꜊',
    dVH: 'kɯit꜊',
    iVL: 'xʷit꜊',
    bVL: 'ʑʷit꜊',
    dVL: 'kʷit꜊',
    PVL: 'dzʷit꜊',
    lVL: 'jʷit꜊',
    NVL: 'tsʷit꜊',
    QVL: 'sʷit꜊',
    IVL: 'lʷit꜊',
    KVL: 'ʈʰʷit꜊',
    LVL: 'ɖʷit꜊',
    YVL: 'tɕʰʷit꜊',
    OVL: 'tsʰʷit꜊',
    SVL: 'tʂʷit꜊',
    iVP: 'xʷɯit꜊',
    SWD: 'tʂɪt꜊',
    VWD: 'ʂɪt꜊',
    DXD: 'miut꜊',
    AXD: 'piut꜊',
    hXL: 'ʔiut꜊',
    dXL: 'kiut꜊',
    eXL: 'kʰiut꜊',
    fXL: 'giut꜊',
    CXD: 'biut꜊',
    iXL: 'xiut꜊',
    kXL: 'ɦiut꜊',
    BXD: 'pʰiut꜊',
    gXL: 'ŋiut꜊',
    iYD: 'xɨt꜊',
    dYD: 'kɨt꜊',
    gYD: 'ŋɨt꜊',
    fYD: 'gɨt꜊',
    eYD: 'kʰɨt꜊',
    gZL: 'ŋʷiɐt꜊',
    CZD: 'bʷiɐt꜊',
    kZL: 'ɦʷiɐt꜊',
    dZL: 'kʷiɐt꜊',
    hZL: 'ʔʷiɐt꜊',
    fZL: 'gʷiɐt꜊',
    eZL: 'kʰʷiɐt꜊',
    AZD: 'pʷiɐt꜊',
    DZD: 'mʷiɐt꜊',
    iZL: 'xʷiɐt꜊',
    hZD: 'ʔiɐt꜊',
    iZD: 'xiɐt꜊',
    dZD: 'kiɐt꜊',
    fZD: 'giɐt꜊',
    BZD: 'pʰʷiɐt꜊',
    gZD: 'ŋiɐt꜊',
    DaD: 'muot꜊',
    FaL: 'tʰuot꜊',
    daL: 'kuot꜊',
    CaD: 'buot꜊',
    EaL: 'tuot꜊',
    GaL: 'duot꜊',
    haL: 'ʔuot꜊',
    iaL: 'xuot꜊',
    gaL: 'ŋuot꜊',
    BaD: 'pʰuot꜊',
    IaL: 'luot꜊',
    eaL: 'kʰuot꜊',
    HaL: 'nuot꜊',
    QaL: 'suot꜊',
    OaL: 'tsʰuot꜊',
    PaL: 'dzuot꜊',
    jbD: 'ɣət꜊',
    jaL: 'ɣuot꜊',
    NaL: 'tsuot꜊',
    jcD: 'ɣɑt꜊',
    icD: 'xɑt꜊',
    EcD: 'tɑt꜊',
    FcD: 'tʰɑt꜊',
    hcD: 'ʔɑt꜊',
    IcD: 'lɑt꜊',
    ecD: 'kʰɑt꜊',
    GcD: 'dɑt꜊',
    PcD: 'dzɑt꜊',
    gcD: 'ŋɑt꜊',
    dcD: 'kɑt꜊',
    QcD: 'sɑt꜊',
    OcD: 'tsʰɑt꜊',
    HcD: 'nɑt꜊',
    DcD: 'mʷɑt꜊',
    gcL: 'ŋʷɑt꜊',
    PcL: 'dzʷɑt꜊',
    AcD: 'pʷɑt꜊',
    NcD: 'tsɑt꜊',
    dcL: 'kʷɑt꜊',
    ecL: 'kʰʷɑt꜊',
    jcL: 'ɣʷɑt꜊',
    GcL: 'dʷɑt꜊',
    icL: 'xʷɑt꜊',
    hcL: 'ʔʷɑt꜊',
    NcL: 'tsʷɑt꜊',
    BcD: 'pʰʷɑt꜊',
    FcL: 'tʰʷɑt꜊',
    IcL: 'lʷɑt꜊',
    EcL: 'tʷɑt꜊',
    OcL: 'tsʰʷɑt꜊',
    CcD: 'bʷɑt꜊',
    jeD: 'ɣɯæt꜊',
    SeD: 'tʂɯæt꜊',
    CeD: 'bɯæt꜊',
    eeD: 'kʰɯæt꜊',
    jeL: 'ɣʷɯæt꜊',
    AeD: 'pɯæt꜊',
    JeL: 'ʈʷɯæt꜊',
    heL: 'ʔʷɯæt꜊',
    MeL: 'ɳʷɯæt꜊',
    TeD: 'tʂʰɯæt꜊',
    deL: 'kʷɯæt꜊',
    deD: 'kɯæt꜊',
    heD: 'ʔɯæt꜊',
    VeD: 'ʂɯæt꜊',
    DeD: 'mɯæt꜊',
    ieL: 'xʷɯæt꜊',
    MeD: 'ɳɯæt꜊',
    geL: 'ŋʷɯæt꜊',
    eeL: 'kʰʷɯæt꜊',
    BeD: 'pʰɯæt꜊',
    SeL: 'tʂʷɯæt꜊',
    jdD: 'ɣɯat꜊',
    hdD: 'ʔɯat꜊',
    gdD: 'ŋɯat꜊',
    TdD: 'tʂʰɯat꜊',
    edD: 'kʰɯat꜊',
    idD: 'xɯat꜊',
    KdD: 'ʈʰɯat꜊',
    ddL: 'kʷɯat꜊',
    jdL: 'ɣʷɯat꜊',
    KdL: 'ʈʰʷɯat꜊',
    JdL: 'ʈʷɯat꜊',
    VdL: 'ʂʷɯat꜊',
    gdL: 'ŋʷɯat꜊',
    TdL: 'tʂʰʷɯat꜊',
    DdD: 'mɯat꜊',
    MdL: 'ɳʷɯat꜊',
    AdD: 'pɯat꜊',
    ddD: 'kɯat꜊',
    UdD: 'dʐɯat꜊',
    JdD: 'ʈɯat꜊',
    cdD: 'ȵɯat꜊',
    QgD: 'set꜊',
    OgD: 'tsʰet꜊',
    dgD: 'ket꜊',
    NgD: 'tset꜊',
    igL: 'xʷet꜊',
    egL: 'kʰʷet꜊',
    dgL: 'kʷet꜊',
    jgL: 'ɣʷet꜊',
    hgL: 'ʔʷet꜊',
    GgD: 'det꜊',
    FgD: 'tʰet꜊',
    jgD: 'ɣet꜊',
    HgD: 'net꜊',
    PgD: 'dzet꜊',
    ggD: 'ŋet꜊',
    DgD: 'met꜊',
    AgD: 'pet꜊',
    hgD: 'ʔet꜊',
    egD: 'kʰet꜊',
    igD: 'xet꜊',
    BgD: 'pʰet꜊',
    CgD: 'bet꜊',
    EgD: 'tet꜊',
    IgD: 'let꜊',
    QfD: 'siɛt꜊',
    IfD: 'liɛt꜊',
    JfD: 'ʈiɛt꜊',
    ffH: 'gɯiɛt꜊',
    cfD: 'ȵiɛt꜊',
    XfD: 'tɕiɛt꜊',
    bfD: 'ʑiɛt꜊',
    ZfD: 'dʑiɛt꜊',
    gfH: 'ŋɯiɛt꜊',
    DfD: 'miɛt꜊',
    efH: 'kʰɯiɛt꜊',
    AfD: 'piɛt꜊',
    PfL: 'dzʷiet꜊',
    NfL: 'tsʷiet꜊',
    QfL: 'sʷiet꜊',
    KfL: 'ʈʰʷiet꜊',
    lfL: 'jʷiet꜊',
    efL: 'kʰʷiɛt꜊',
    hfP: 'ʔʷɯiet꜊',
    cfL: 'ȵʷiet꜊',
    afL: 'ɕʷiet꜊',
    XfL: 'tɕʷiet꜊',
    YfL: 'tɕʰʷiet꜊',
    JfL: 'ʈʷiet꜊',
    IfL: 'lʷiet꜊',
    BfD: 'pʰiɛt꜊',
    CfH: 'bɯiɛt꜊',
    LfD: 'ɖiɛt꜊',
    AfH: 'pɯiɛt꜊',
    VfL: 'ʂʷiet꜊',
    dfH: 'kɯiɛt꜊',
    afD: 'ɕiɛt꜊',
    MfL: 'ɳʷiet꜊',
    ifP: 'xʷɯiet꜊',
    hfL: 'ʔʷiɛt꜊',
    dfP: 'kʷɯiet꜊',
    SfL: 'tʂʷiet꜊',
    OfL: 'tsʰʷiet꜊',
    NfD: 'tsiɛt꜊',
    VfD: 'ʂiɛt꜊',
    hfH: 'ʔɯiɛt꜊',
    KfD: 'ʈʰiɛt꜊',
    ifH: 'xɯiɛt꜊',
    TfD: 'tʂʰiɛt꜊',
    RfL: 'zʷiet꜊',
    YfD: 'tɕʰiɛt꜊',
    lfD: 'jiɛt꜊',
    UfD: 'dʐiɛt꜊',
    lpD: 'jiɐk꜊',
    IpD: 'liɐk꜊',
    dpD: 'kiɐk꜊',
    XpD: 'tɕiɐk꜊',
    apD: 'ɕiɐk꜊',
    cpD: 'ȵiɐk꜊',
    YpD: 'tɕʰiɐk꜊',
    hpD: 'ʔiɐk꜊',
    epD: 'kʰiɐk꜊',
    gpD: 'ŋiɐk꜊',
    ZpD: 'dʑiɐk꜊',
    KpD: 'ʈʰiɐk꜊',
    QpD: 'siɐk꜊',
    SpD: 'tʂiɐk꜊',
    NpD: 'tsiɐk꜊',
    PpD: 'dziɐk꜊',
    OpD: 'tsʰiɐk꜊',
    fpD: 'giɐk꜊',
    hpL: 'ʔʷiɐk꜊',
    CpD: 'bʷiɐk꜊',
    ipL: 'xʷiɐk꜊',
    kpL: 'ɦʷiɐk꜊',
    dpL: 'kʷiɐk꜊',
    JpD: 'ʈiɐk꜊',
    LpD: 'ɖiɐk꜊',
    epL: 'kʰʷiɐk꜊',
    fpL: 'gʷiɐk꜊',
    MpD: 'ɳiɐk꜊',
    BpD: 'pʰʷiɐk꜊',
    ipD: 'xiɐk꜊',
    GqD: 'dɑk꜊',
    DqD: 'mɑk꜊',
    IqD: 'lɑk꜊',
    FqD: 'tʰɑk꜊',
    NqD: 'tsɑk꜊',
    OqD: 'tsʰɑk꜊',
    dqD: 'kɑk꜊',
    eqD: 'kʰɑk꜊',
    gqD: 'ŋɑk꜊',
    BqD: 'pʰɑk꜊',
    hqD: 'ʔɑk꜊',
    CqD: 'bɑk꜊',
    iqD: 'xɑk꜊',
    QqD: 'sɑk꜊',
    jqD: 'ɣɑk꜊',
    PqD: 'dzɑk꜊',
    AqD: 'pɑk꜊',
    HqD: 'nɑk꜊',
    iqL: 'xʷɑk꜊',
    dqL: 'kʷɑk꜊',
    hqL: 'ʔʷɑk꜊',
    jqL: 'ɣʷɑk꜊',
    eqL: 'kʰʷɑk꜊',
    gqL: 'ŋʷɑk꜊',
    IqL: 'lʷɑk꜊',
    NqL: 'tsʷɑk꜊',
    DrD: 'mɯak꜊',
    JrD: 'ʈɯak꜊',
    CrD: 'bɯak꜊',
    ArD: 'pɯak꜊',
    fsD: 'gɯiak꜊',
    dsD: 'kɯiak꜊',
    VsD: 'ʂɯiak꜊',
    TsD: 'tʂʰɯiak꜊',
    jrL: 'ɣʷɯak꜊',
    SrD: 'tʂɯak꜊',
    UrD: 'dʐɯak꜊',
    esD: 'kʰɯiak꜊',
    grD: 'ŋɯak꜊',
    gsD: 'ŋɯiak꜊',
    erD: 'kʰɯak꜊',
    hrD: 'ʔɯak꜊',
    isD: 'xɯiak꜊',
    KrD: 'ʈʰɯak꜊',
    hrL: 'ʔʷɯak꜊',
    BrD: 'pʰɯak꜊',
    irD: 'xɯak꜊',
    jrD: 'ɣɯak꜊',
    drD: 'kɯak꜊',
    irL: 'xʷɯak꜊',
    LrD: 'ɖɯak꜊',
    drL: 'kʷɯak꜊',
    hsL: 'ʔʷɯiak꜊',
    MrD: 'ɳɯak꜊',
    erL: 'kʰʷɯak꜊',
    CsD: 'bɯiak꜊',
    DtD: 'mɯæk꜊',
    jtL: 'ɣʷɯæk꜊',
    dtL: 'kʷɯæk꜊',
    AtD: 'pɯæk꜊',
    CtD: 'bɯæk꜊',
    UtD: 'dʐɯæk꜊',
    StD: 'tʂɯæk꜊',
    UtL: 'dʐʷɯæk꜊',
    TtD: 'tʂʰɯæk꜊',
    etD: 'kʰɯæk꜊',
    itL: 'xʷɯæk꜊',
    jtD: 'ɣɯæk꜊',
    dtD: 'kɯæk꜊',
    JtD: 'ʈɯæk꜊',
    htD: 'ʔɯæk꜊',
    VtD: 'ʂɯæk꜊',
    BtD: 'pʰɯæk꜊',
    gtD: 'ŋɯæk꜊',
    ItD: 'lɯæk꜊',
    StL: 'tʂʷɯæk꜊',
    VtL: 'ʂʷɯæk꜊',
    ftL: 'gʷɯæk꜊',
    MtD: 'ɳɯæk꜊',
    QuD: 'siɛk꜊',
    NuD: 'tsiɛk꜊',
    huD: 'ʔiɛk꜊',
    luD: 'jiɛk꜊',
    auD: 'ɕiɛk꜊',
    YuD: 'tɕʰiɛk꜊',
    ZuD: 'dʑiɛk꜊',
    XuD: 'tɕiɛk꜊',
    LuD: 'ɖiɛk꜊',
    OuD: 'tsʰiɛk꜊',
    RuD: 'ziɛk꜊',
    PuD: 'dziɛk꜊',
    CuD: 'biɛk꜊',
    luL: 'jʷiɛk꜊',
    iuL: 'xʷiɛk꜊',
    AuD: 'piɛk꜊',
    BuD: 'pʰiɛk꜊',
    buD: 'ʑiɛk꜊',
    AsD: 'pɯiak꜊',
    JuD: 'ʈiɛk꜊',
    KuD: 'ʈʰiɛk꜊',
    XuL: 'tɕʷiɛk꜊',
    OuL: 'tsʰʷiɛk꜊',
    QvD: 'sek꜊',
    dvD: 'kek꜊',
    BvD: 'pʰek꜊',
    IvD: 'lek꜊',
    EvD: 'tek꜊',
    jvD: 'ɣek꜊',
    gvD: 'ŋek꜊',
    GvD: 'dek꜊',
    FvD: 'tʰek꜊',
    NvD: 'tsek꜊',
    evD: 'kʰek꜊',
    HvD: 'nek꜊',
    PvD: 'dzek꜊',
    DvD: 'mek꜊',
    CvD: 'bek꜊',
    AvD: 'pek꜊',
    evL: 'kʰʷek꜊',
    dvL: 'kʷek꜊',
    OvD: 'tsʰek꜊',
    ivD: 'xek꜊',
    ivL: 'xʷek꜊',
    KvD: 'ʈʰek꜊',
    XwD: 'tɕɨk꜊',
    LwD: 'ɖɨk꜊',
    IwD: 'lɨk꜊',
    KwD: 'ʈʰɨk꜊',
    JwD: 'ʈɨk꜊',
    bwD: 'ʑɨk꜊',
    QwD: 'sɨk꜊',
    ZwD: 'dʑɨk꜊',
    awD: 'ɕɨk꜊',
    iwD: 'xɨk꜊',
    UwD: 'dʐɨk꜊',
    fwD: 'gɨk꜊',
    MwD: 'ɳɨk꜊',
    TwD: 'tʂʰɨk꜊',
    hwD: 'ʔɨk꜊',
    VwD: 'ʂɨk꜊',
    ewD: 'kʰɨk꜊',
    dwD: 'kɨk꜊',
    lwD: 'jɨk꜊',
    NwD: 'tsɨk꜊',
    AwD: 'pɨk꜊',
    kwL: 'ɦʷɨk꜊',
    iwL: 'xʷɨk꜊',
    BwD: 'pʰɨk꜊',
    SwD: 'tʂɨk꜊',
    CwD: 'bɨk꜊',
    gwD: 'ŋɨk꜊',
    PwD: 'dzɨk꜊',
    DwD: 'mɨk꜊',
    YwD: 'tɕʰɨk꜊',
    ExD: 'tək꜊',
    NxD: 'tsək꜊',
    IxD: 'lək꜊',
    FxD: 'tʰək꜊',
    exD: 'kʰək꜊',
    GxD: 'dək꜊',
    ixD: 'xək꜊',
    DxD: 'mək꜊',
    PxD: 'dzək꜊',
    QxD: 'sək꜊',
    AxD: 'pək꜊',
    CxD: 'bək꜊',
    jxL: 'ɣʷək꜊',
    dxL: 'kʷək꜊',
    hxD: 'ʔək꜊',
    jxD: 'ɣək꜊',
    HxD: 'nək꜊',
    dxD: 'kək꜊',
    OxD: 'tsʰək꜊',
    BxD: 'pʰək꜊',
    ixL: 'xʷək꜊',
    O1D: 'tsʰip꜊',
    Z1D: 'dʑip꜊',
    X1D: 'tɕip꜊',
    R1D: 'zip꜊',
    P1D: 'dzip꜊',
    c1D: 'ȵip꜊',
    h1D: 'ʔip꜊',
    a1D: 'ɕip꜊',
    N1D: 'tsip꜊',
    f1H: 'gɯip꜊',
    L1D: 'ɖip꜊',
    J1D: 'ʈip꜊',
    I1D: 'lip꜊',
    d1H: 'kɯip꜊',
    g1H: 'ŋɯip꜊',
    e1H: 'kʰɯip꜊',
    Q1D: 'sip꜊',
    V1D: 'ʂip꜊',
    i1H: 'xɯip꜊',
    S1D: 'tʂip꜊',
    h1H: 'ʔɯip꜊',
    K1D: 'ʈʰip꜊',
    k1D: 'ɦip꜊',
    l1D: 'jip꜊',
    M1D: 'ɳip꜊',
    U1D: 'dʐip꜊',
    C1H: 'bɯip꜊',
    A1H: 'pɯip꜊',
    T1D: 'tʂʰip꜊',
    Y1D: 'tɕʰip꜊',
    j2D: 'ɣəp꜊',
    d2D: 'kəp꜊',
    E2D: 'təp꜊',
    Q2D: 'səp꜊',
    G2D: 'dəp꜊',
    F2D: 'tʰəp꜊',
    P2D: 'dzəp꜊',
    N2D: 'tsəp꜊',
    I2D: 'ləp꜊',
    H2D: 'nəp꜊',
    e2D: 'kʰəp꜊',
    h2D: 'ʔəp꜊',
    i2D: 'xəp꜊',
    g2D: 'ŋəp꜊',
    O2D: 'tsʰəp꜊',
    Z2D: 'dʑəp꜊',
    j3D: 'ɣɑp꜊',
    I3D: 'lɑp꜊',
    E3D: 'tɑp꜊',
    F3D: 'tʰɑp꜊',
    i3D: 'xɑp꜊',
    H3D: 'nɑp꜊',
    G3D: 'dɑp꜊',
    Q3D: 'sɑp꜊',
    g3D: 'ŋɑp꜊',
    P3D: 'dzɑp꜊',
    d3D: 'kɑp꜊',
    e3D: 'kʰɑp꜊',
    h3D: 'ʔɑp꜊',
    O3D: 'tsʰɑp꜊',
    X3D: 'tɕɑp꜊',
    l4D: 'jiɛp꜊',
    N4D: 'tsiɛp꜊',
    a4D: 'ɕiɛp꜊',
    Z4D: 'dʑiɛp꜊',
    I4D: 'liɛp꜊',
    P4D: 'dziɛp꜊',
    L4D: 'ɖiɛp꜊',
    h4H: 'ʔɯiɛp꜊',
    M4D: 'ɳiɛp꜊',
    Y4D: 'tɕʰiɛp꜊',
    c4D: 'ȵiɛp꜊',
    X4D: 'tɕiɛp꜊',
    O4D: 'tsʰiɛp꜊',
    K4D: 'ʈʰiɛp꜊',
    f4H: 'gɯiɛp꜊',
    J4D: 'ʈiɛp꜊',
    k4D: 'ɦiɛp꜊',
    e4D: 'kʰiɛp꜊',
    V4D: 'ʂiɛp꜊',
    d4H: 'kɯiɛp꜊',
    h4D: 'ʔiɛp꜊',
    F5D: 'tʰep꜊',
    j5D: 'ɣep꜊',
    d5D: 'kep꜊',
    e5D: 'kʰep꜊',
    G5D: 'dep꜊',
    H5D: 'nep꜊',
    Q5D: 'sep꜊',
    I5D: 'lep꜊',
    E5D: 'tep꜊',
    P5D: 'dzep꜊',
    N5D: 'tsep꜊',
    i5D: 'xep꜊',
    j6D: 'ɣɯæp꜊',
    e6D: 'kʰɯæp꜊',
    U6D: 'dʐɯæp꜊',
    d6D: 'kɯæp꜊',
    S6D: 'tʂɯæp꜊',
    T6D: 'tʂʰɯæp꜊',
    M6D: 'ɳɯæp꜊',
    i6D: 'xɯæp꜊',
    V6D: 'ʂɯæp꜊',
    J6D: 'ʈɯæp꜊',
    h6D: 'ʔɯæp꜊',
    g6D: 'ŋɯæp꜊',
    K6D: 'ʈʰɯæp꜊',
    j7D: 'ɣɯap꜊',
    L7D: 'ɖɯap꜊',
    h7D: 'ʔɯap꜊',
    d7D: 'kɯap꜊',
    V7D: 'ʂɯap꜊',
    i7D: 'xɯap꜊',
    g8D: 'ŋiɐp꜊',
    i8D: 'xiɐp꜊',
    e8D: 'kʰiɐp꜊',
    d8D: 'kiɐp꜊',
    h8D: 'ʔiɐp꜊',
    l8D: 'jiɐp꜊',
    f8D: 'giɐp꜊',
    C9D: 'biɐp꜊',
    A9D: 'piɐp꜊',
    B9D: 'pʰiɐp꜊',
    e9L: 'kʰiɐp꜊',
    M9L: 'ɳiɐp꜊',
    K9L: 'ʈʰiɐp꜊',
    fGA: 'gi˧',
    LeI: 'ɖʷɯæn˧',
    ifE: 'xɯiɛn˧',
    VsA: 'ʂɯiaŋ˧',
    N3A: 'tsɑm˧',
    BSB: 'pʰuoi˧˥',
    eVN: 'kʰʷɯin˧˥',
    iVF: 'xɯin˧˥',
    SYB: 'tʂɨn˧˥',
    TYB: 'tʂʰɨn˧˥',
    SnJ: 'tʂʷɯa˧˥',
    PzB: 'dzəu˧˥',
    K8B: 'ʈʰiɐm˧˥',
    JPC: 'ʈɯæ˥˨',
    JQK: 'ʈʷɯæi˥˨',
    TVC: 'tʂʰin˥˨',
    TZK: 'tʂʰʷiɐn˥˨',
    OlC: 'tsʰɑ˥˨',
    MnC: 'ɳɯa˥˨',
    TsC: 'tʂʰɯiaŋ˥˨',
    VsC: 'ʂɯiaŋ˥˨',
    hvC: 'ʔeŋ˥˨',
    TVD: 'tʂʰit꜊',
    UVD: 'dʐit꜊',
    lcD: 'jɑt꜊',
    dfD: 'kiɛt꜊',
    ifL: 'xʷiɛt꜊',
    ZfL: 'dʑʷiɛt꜊',
    TfL: 'tʂʰʷiɛt꜊',
    AuH: 'pɯiɛk꜊',
    M8D: 'ɳiɐp꜊',
    K8D: 'ʈʰiɐp꜊',
  }[音韻地位.編碼] || '?';
}

/**
 * unt 切韻朗讀音
 * https://zhuanlan.zhihu.com/p/58227457
 *
 * 説明
 * 以下內容為生成 unt 切韻朗讀音的函數體
 * 函數接受音韻地位，返回對應的 unt 切韻朗讀音
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的unt 切韻朗讀音
 */
export function unt(音韻地位, 字頭) {
  const is = (x) => 音韻地位.屬於(x);

  function 聲母規則() {
    if (is('幫母')) return 'p';
    if (is('滂母')) return 'pʰ';
    if (is('並母')) return 'b';
    if (is('明母')) return 'm';
    if (is('端母')) return 't';
    if (is('透母')) return 'tʰ';
    if (is('定母')) return 'd';
    if (is('泥母')) return 'n';
    if (is('知母')) return 'ʈ';
    if (is('徹母')) return 'ʈʰ';
    if (is('澄母')) return 'ɖ';
    if (is('孃母')) return 'ɳ';
    if (is('精母')) return 't͡s';
    if (is('清母')) return 't͡sʰ';
    if (is('從母')) return 'd͡z';
    if (is('心母')) return 's';
    if (is('邪母')) return 'z';
    if (is('莊母')) return 't͡ʂ';
    if (is('初母')) return 't͡ʂʰ';
    if (is('崇母')) return 'd͡ʐ';
    if (is('生母')) return 'ʂ';
    if (is('俟母')) return 'ʐ';
    if (is('章母')) return 't͡ɕ';
    if (is('昌母')) return 't͡ɕʰ';
    if (is('常母')) return 'd͡ʑ';
    if (is('書母')) return 'ɕ';
    if (is('船母')) return 'ʑ';
    if (is('見母')) return is('三等') ? 'k' : 'q';
    if (is('溪母')) return is('三等') ? 'kʰ' : 'qʰ';
    if (is('羣母')) return is('三等') ? 'ɡ' : 'ɢ';
    if (is('疑母')) return is('三等') ? 'ŋ' : 'ɴ';
    if (is('影母')) return 'ʔ';
    if (is('曉母')) return is('三等') ? 'h' : 'χ';
    if (is('匣母')) return is('三等') ? 'ɦ' : 'ʁ';
    if (is('云母')) return '';
    if (is('以母')) return 'j';
    if (is('來母')) return 'l';
    if (is('日母')) return 'ɲ';
    throw new Error('無聲母規則');
  }

  function 韻母規則() {
    // 通攝
    if (is('東韻 一等')) return 'uŋʷ';
    if (is('東韻 三等')) return 'ẅuŋʷ';
    if (is('冬韻')) return 'oŋʷ';
    if (is('鍾韻')) return 'ẅɔŋʷ';
    // 江攝
    if (is('江韻')) return 'ɻæŋʷ';
    // 止攝
    if (is('支韻 合口')) return 'ʳɥɛ';
    if (is('支韻')) return 'ʳjɛ';
    if (is('脂韻 合口')) return 'ʳɥi';
    if (is('脂韻')) return 'ʳi';
    if (is('之韻')) return 'ɨ';
    if (is('微韻 開口')) return 'ɨj';
    if (is('微韻')) return 'ẅɨj';
    // 遇攝
    if (is('魚韻')) return 'j̈ə';
    if (is('虞韻')) return 'ẅɔ';
    if (is('模韻')) return 'o';
    // 蟹攝
    if (is('齊韻 合口')) return 'wej';
    if (is('齊韻')) return 'ej';
    if (is('祭韻 合口')) return 'ʳɥɛjɕ';
    if (is('祭韻')) return 'ʳjɛjɕ';
    if (is('泰韻 合口')) return 'wɑjɕ';
    if (is('泰韻')) return 'ɑjɕ';
    if (is('佳韻 合口')) return 'wɻæ';
    if (is('佳韻')) return 'ɻæ';
    if (is('皆韻 合口')) return 'wɻæj';
    if (is('皆韻')) return 'ɻæj';
    if (is('夬韻 合口')) return 'wɻajɕ';
    if (is('夬韻')) return 'ɻajɕ';
    if (is('咍韻')) return 'ɐj';
    if (is('灰韻')) return 'wɔ̞j';
    if (is('廢韻 開口')) return 'j̈əjɕ';
    if (is('廢韻')) return 'ẅəjɕ';
    // 臻攝
    if (is('眞韻 合口')) return 'ʳɥin';
    if (is('眞韻')) return 'ʳin';
    if (is('臻韻')) { // [^1]
      if (is('平入聲')) return 'ɻi˞n';
      if (is('上聲')) return 'ɨn';
      if (is('去聲')) return 'ɻin';
    }
    if (is('欣韻')) return 'ɨn';
    if (is('文韻')) return 'ẅun';
    if (is('元韻 開口')) return 'j̈ən';
    if (is('元韻')) return 'ẅən';
    if (is('痕韻')) return 'ɘn';
    if (is('魂韻')) return 'won';
    // 山攝
    if (is('寒韻 開口')) return 'ɑn';
    if (is('寒韻')) return 'wɑn';
    if (is('刪韻 合口')) return 'wɻan';
    if (is('刪韻')) return 'ɻan';
    if (is('山韻 合口')) return 'wɻæn';
    if (is('山韻')) return 'ɻæn';
    if (is('仙韻 合口')) return 'ʳɥɛn';
    if (is('仙韻')) return 'ʳjɛn';
    if (is('先韻 合口')) return 'wen';
    if (is('先韻')) return 'en';
    // 效攝
    if (is('蕭韻')) return 'ew';
    if (is('宵韻')) return 'ʳjɛw';
    if (is('肴韻')) return 'ɻaw';
    if (is('豪韻')) return 'ɑw';
    // 果攝
    if (is('歌韻 一等 開口')) return 'ɑ';
    if (is('歌韻 一等')) return 'wɑ';
    if (is('歌韻 三等 開口')) return 'j̈ɑ';
    if (is('歌韻 三等')) return 'ẅɑ';
    // 假攝
    if (is('麻韻 二等 合口')) return 'wɻa';
    if (is('麻韻 二等')) return 'ɻa';
    if (is('麻韻 三等')) return 'ja';
    // 宕攝
    if (is('陽韻 開口')) return 'j̈ɐŋ';
    if (is('陽韻')) return 'ẅɐŋ';
    if (is('唐韻 合口')) return 'wɑŋ';
    if (is('唐韻')) return 'ɑŋ';
    // 梗攝
    if (is('庚韻 二等 合口')) return 'wɻaɲ';
    if (is('庚韻 二等')) return 'ɻaɲ';
    if (is('庚韻 三等 合口')) return 'ɻɥaɲ';
    if (is('庚韻 三等')) return 'ɻjaɲ';
    if (is('耕韻 合口')) return 'wɻæɲ';
    if (is('耕韻')) return 'ɻæɲ';
    if (is('清韻 合口')) return 'ɥɛɲ';
    if (is('清韻')) return 'ʳjɛɲ'; // [^2]
    if (is('青韻 合口')) return 'weɲ';
    if (is('青韻')) return 'eɲ';
    // 曾攝
    if (is('蒸韻 合口')) return 'ʳɥiŋ'; // [^2]
    if (is('蒸韻')) return 'ʳiŋ'; // [^2]
    if (is('登韻 合口')) return 'wɘŋ';
    if (is('登韻')) return 'ɘŋ';
    // 流攝
    if (is('尤韻')) return 'ɥ̈u';
    if (is('侯韻')) { // 根據設定，侯韻幫組為 u，其他為 ɘu
      if (is('幫組')) return 'u';
      return 'ɘu';
    }
    if (is('幽韻')) return 'ɥÿ';
    // 深攝
    if (is('侵韻')) return 'ʳim';
    // 咸攝
    if (is('覃韻')) return 'ɐm';
    if (is('談韻')) return 'ɑm';
    if (is('鹽韻')) return 'ʳjɛm';
    if (is('添韻')) return 'em';
    if (is('咸韻')) return 'ɻæm';
    if (is('銜韻')) return 'ɻam';
    if (is('嚴韻')) return 'j̈əm';
    if (is('凡韻')) return 'ẅɞm';
    throw new Error('無韻母規則');
  }

  function 聲調規則() {
    if (is('平聲')) return is('全清 或 次清') ? '˦' : '˨˩';
    if (is('上聲')) return !is('全濁') ? '˦˦˥' : '˨˨˧';
    if (is('去聲')) return is('全清 或 次清') ? '˥˩' : '˧˩˨';
    if (is('入聲')) return !is('全濁') ? '˥' : '˨˩';
    throw new Error('無聲調規則');
  }

  let 聲母 = 聲母規則();
  let 韻母 = 韻母規則();
  let 聲調 = 聲調規則();

  // 處理重紐 [^3]
  if (韻母.startsWith('ʳ')) {
    if (is('知莊組')) {
      韻母 = `ɻ${韻母.slice(1)}`; // 無重紐對立，認定為重紐B類
    } else if (is('端精章組 或 以來日母')) {
      韻母 = 韻母.slice(1); // 無重紐對立，認定為重紐A類
    } else if (is('重紐A類')) {
      韻母 = 韻母.slice(1); // 有重紐對立，默認為重紐A類
    } else {
      韻母 = `ɻ${韻母.slice(1)}`; // 有重紐對立，重紐B類
    }
  }

  // 處理脣音
  if (is('幫組')) {
    // 除三等C类外，其他韵母与帮组p相拼时均为开口（不含 [ʷ] 成分）。
    if (韻母.startsWith('ɥ') && !韻母.startsWith('ɥ̈')) {
      韻母 = `j${韻母.slice(1)}`;
    } else if (韻母.startsWith('w') && !韻母.startsWith('ẅ')) {
      韻母 = 韻母.slice(1);
    // 三等C类与帮组p相拼时是ɥ̈，即 [ÿ] 对应的半元音。[^4]
    } if (is('三等') && 韻母.startsWith('ẅ')) {
      韻母 = `ɥ̈${韻母.slice(2)}`;
    }
  }

  // 調整 ɻ、w 的顺序 [^5]
  if (韻母.startsWith('wɻ') && is('知莊組')) {
    韻母 = `ɻw${韻母.slice(2)}`;
  }

  // 以母簡化拼式
  // 總結：以母三A/三B，以 j 或 ɥ 起始時，省略聲母
  if (is('以母 三等 支脂祭眞仙宵諄臻麻清庚蒸侵鹽韻') && (韻母.startsWith('j') || 韻母.startsWith('ɥ'))) {
    聲母 = '';
  }

  // 處理入聲韻
  if (is('入聲')) {
    if (韻母.endsWith('m')) {
      韻母 = `${韻母.slice(0, -1)}p`;
    } else if (韻母.endsWith('n')) {
      韻母 = `${韻母.slice(0, -1)}t`;
    } else if (韻母.endsWith('ɲ')) {
      韻母 = `${韻母.slice(0, -1)}c`;
    } else if (韻母.endsWith('ŋ')) {
      韻母 = `${韻母.slice(0, -1)}k`;
    } else if (韻母.endsWith('ŋʷ')) {
      韻母 = `${韻母.slice(0, -2)}kʷ`;
    }
  }

  // 當没有聲母，韻母以 ɨ 起始時，添加聲母 j̈
  if (聲母 === '' && 韻母 === 'ɨ') {
    聲母 = 'j̈';
  }

  return 聲母 + 韻母 + 聲調;

  /*
  [^1]: 臻韵ɻi˞n的韵腹i˞带有r色彩，同时像三等B类介音ɻj一样，这个i˞没有那么靠前（见图2）。这就像大埔方言（客家话）的“真”[tʃ˞i˞n]（[t͡ʃ] = [t͡ʂ̻]）[20]一样。臻韵ɻi˞n只出现于庄组t͡ʂ的平声和入声后，与上声和去声的真B开ɻin、欣韵ɨn互补，实际上它的韵腹只是i在卷舌声母和舌冠（coronal）韵尾n的共同影响下产生的强r色彩变体，而且可以看作一种声调变韵；但根据元音分韵原则，我们必须承认臻韵ɻi˞n的独立性。这里的“ɻ”对标音来说完全是多余的（见介音第8c条），但保留“ɻ”能使它的前两个字符“ɻi”符合三等B类韵母的拼写，因此保留。
  庄组t͡ʂ上声和去声的真B开ɻin、欣韵ɨn（包括“𧤛”t͡ʂɨn˦˦˥、“龀”t͡ʂʰɨn˦˦˥、“浕”d͡ʐɻin˨˨˧、“榇”t͡ʂʰɻin˥˩ 4个小韵。两个韵母在庄组t͡ʂ后没有对立）可以遵照《广韵》的归韵，不一定也要读成臻韵ɻi˞n。

  [^2]: 谆韵ɥin、清开jɛɲ、蒸韵ɻiŋ虽无重纽对立（表2中无下划线），但实际上有跨A、B两类的情况：谆韵ɥin、清开jɛɲ与知ʈ、庄t͡ʂ组相拼时是B类（如“椿”ʈʰɻɥin˦、“贞”ʈɻjɛɲ˦），蒸韵ɻiŋ与精t͡s、章t͡ɕ组和日ɲ、以j母相拼时是A类（如“仍”ɲiŋ˨˩）。

  [^3]:  “ʳ”代表该韵有重纽对立：B类在“ʳ”的位置有ɻ介音，A类无ɻ介音。
  为节省空间，A、B两类韵母写在同一格中。

  [^4]: [ÿ]（= [ɨᵝ]）是央高不突出圆唇元音，与央高突出圆唇元音 [ʉ] 的圆唇类型不同。突出圆唇（protruded rounding）元音的双唇外凸，露出嘴唇的内侧，由内唇（endolabial）参与调音；不突出圆唇（compressed rounding）元音的双唇收敛，不露出内唇，由外唇（exolabial）参与调音，可视为元音与双唇近音 [β̞] 的双重调音。
  ［这个不突出唇化的近音ɥ̈与帮组p调音部位相同，在后世导致声母（塞）擦化，即轻唇化（见声母第4b条）。］

  [^5]: 卷舌音后的二等合口介音写成ɻw，体现ɻ和卷舌声母的一体性
  卷舌音（即知ʈ、庄t͡ʂ组）
  */
}

/**
 * unt 切韵拟音 J
 * 音系规则和代码实现 - https://zhuanlan.zhihu.com/p/305516512
 * 语音描写和拟音说明 - https://zhuanlan.zhihu.com/p/313005024
 *
 * 帮助
 * 以下内容为生成 unt 切韵拟音 J 的函数体
 * 函数接受音韵地位，返回对应的 unt 切韵拟音 J
 * J 为拟音的版本号
 *
 * 之前的 unt 切韵朗读音推导方案已归档，请查看 https://github.com/nk2028/qieyun-examples/blob/main/unt.js
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的unt 切韵拟音 J
 */
export function unt_j(音韻地位, 字頭) {
  const is = (x) => 音韻地位.屬於(x);

  /** 目录

  一、流程控制开关
  二、音节结构
  三、音韵地位对应音位（及其代码实现）和区别性特征
  四、音系规则（及其代码实现）
  五、后处理的代码实现

  下面语音学术语对应的音韵学术语将用全角尖括号〈〉在语音学术语后注明。
  */

  /** 一、流程控制开关

  开关用半角尖括号 < > 表明。
  */

  const switches = {};
  // 音韵地位对应音位开关
  switches.肴豪韵韵核归为低元音   = false; // 关闭：e͇w əw（实现为 œ͇w ʌw），打开：a͇w aw

  // 音系规则开关
  switches.祭泰夬废韵尾推导为ɹ    = true;  // 音系规则 (4)。关闭：-j，打开：-ɹ
  switches.要推导松元音           = true;  // 音系规则 (5)。普通用户请保持打开，否则后面一些推导可能出错
  switches.要推导二合元音         = true;  // 音系规则 (6)。关闭：iə ɨə ia ɨa，打开：iɛ ɨɜ iæ ɨɐ
  switches.要推导a                = true;  // 音系规则 (7)。关闭：a，打开：a ɑ
  switches.庄组臻摄开口推导为ɹ̩    = true;  // 音系规则 (9)。关闭：in，打开：ɹ̩n。不包括合口“率”小韵
  switches.豪覃韵韵核推导为ʌ      = true;  // 音系规则 (10)。关闭：əw əm，打开：ʌw ʌm
  switches.肴凡韵韵核推导为œ      = true;  // 音系规则 (11)。关闭：e͇w βəm，打开：œ͇w βœm
  switches.精三寅合口介音推导为ɥ  = false; // 音系规则 (12)。关闭：sʷɹ-，打开：sʷɥ-。关闭
  switches.蒸幽韵合口增生ɹ滑音    = true;  // 音系规则 (13)。“冰”，关闭：pîŋ，打开：pɹîŋ
  switches.云母推导为ɹ            = true;  // 音系规则 (14)。关闭：ɣɹ- ɣj̈-，打开：ɹ-。不论三 B 还是三 C
  switches.见系非三推导为软腭后音 = true;  // 音系规则 (15)。包含匣母
  switches.晓母非三推导为软腭后音 = true;  // 音系规则 (16)。按常理，需要 <见系非三推导为软腭后音> 打开才能打开
  switches.通江宕摄推导为软腭后音 = true;  // 音系规则 (17)。如果 <要推导a> 没有打开，那么不推导宕摄韵尾
  switches.侯韵裂化为ɘu           = true;  // 音系规则 (18)。关闭：u，打开：ɘu
  switches.精组非后高元音省略介音 = true;  // 音系规则 (21)。“踪”，关闭：tsɹʉ̂ɜɴ，打开：tsʉ̂ɜɴ

  // 后处理开关
  switches.知组写成卷舌塞音       = false; // 关闭：tɹ，打开：二等 ʈ、三等 ʈɹ
  switches.中元音写成半低元音     = false; // 关闭：e o，打开：ɛ ɔ。不转换 ə
  switches.ɑ写成a                 = true;  // 默认打开
  switches.半元音介音写成元音     = false;
  switches.半元音韵尾写成元音     = false;
  switches.二等元音写成r音钩      = false;
  switches.二等元音写成双下横线   = false; // 关闭：下等号 a͇（U+0347），打开：双下横线 a̳（U+0333）。一些字体（如 Times New Roman、Arial）把这两个附加符号弄反了，为了显示的效果要打开。只在 <二等元音写成r音钩> 关闭时有效

  // 声调开关
  switches.声调分阴阳             = true;  // 默认打开
  switches.声调写成五度标记       = false;
  switches.声调附加符号写在音节前 = false; // 只在 <声调写成五度标记> 关闭时有效

  /** 二、音节结构

  切韵拟音 J 的音节结构是 CʷGVCᵀ。

  Cʷ：辅音或唇化辅音，作为声母（initial）
  G：滑音（glide），作为介音（medial）
  V：单元音或二合元音，作为韵核（nucleus）。成音节辅音 [ɹ̩] 也可以作韵核〈臻韵〉
  C：辅音，作为韵尾（coda）。韵核和韵尾加在一起叫作韵基（rime）
  ᵀ：声调（tone）

  介音和韵尾可有可无，声母、韵核、声调是必须出现的。
  */

  /** 三、音韵地位对应音位（及其代码实现）和区别性特征（distinctive feature）

  1. 辅音的特征

  发声态和调音方式：

  [±voi]:   带声（voice）
  [±sg]:    展声门（spread glottis），即辅音送气。注意 [h]〈晓母三等〉按照理论是 [+sg]，但本文为了和“全清”对应，算作 [−sg]
  [±son]:   响音（sonorant）性，包括鼻音和近音（本文的响音不包含元音）。相反的 [−son] 是阻音（obstruent）
  [±stop]:  塞音性，包括鼻塞音（即鼻音）和口塞音（即爆发音和塞擦音）。包含鼻音的“塞音”严格来说应该叫 occlusive 而非 stop，本文从简直接用 [±stop]
  [±fric]:  擦音（fricative）性。本文将塞擦音也算入 [+fric]，不使用现代音系学常用的 [±delayed release]（延缓除阻）

  调音部位：

  [LAB]:    唇（labial）
    [±rnd]: 圆唇（round），包括唇化辅音和圆唇元音
  [COR]:    舌冠（coronal），即锐音。本文为了简便将硬腭辅音也算入舌冠音。本文从简仍然使用“锐音、钝音”的叫法，但不用 [∓grave]
    [±ant]: 前部（anterior）。前部锐音包括齿–龈、龈，后部锐音包括龈后、卷舌、龈–腭等
    [±r]:   r 色彩
  [DOR]:    舌面（dorsal）
    [±high]:高，对辅音而言 [DOR, +high] 是软腭音，[DOR, −high] 是小舌音，正好符合三等、非三等之分。
            本文将软腭音分为软腭前音（prevelar）、软腭后音（postvelar）两组，分别用软腭音和小舌音的记号表示，
            通过 <见系非三推导为软腭后音>、<晓母非三推导为软腭后音>、<通江宕摄推导为软腭后音> 控制。

  2. 音韵地位对应辅音音位

  详见下面代码实现。
  */

  // 函数：将声母的音韵地位转换为音位，不含开合信息
  // 介音音位的特征也在下面用 /* */ 注释列出，以说明其区别性特征
  // 同样地，条件变体在下面用括号标注列出
  function getInitialWithoutRounding() {
    switch (音韻地位.母) {
    // 不送气清阻音〈全清〉    送气清阻音〈次清〉       浊阻音〈全浊〉           浊响音〈次浊〉
    // −voi, −sg, −son         −voi, +sg, −son          +voi, −sg, −son          +voi, −sg, +son
      case '幫': return 'p'  ; case '滂': return 'pʰ' ; case '並': return 'b'  ; case '明': return 'm'  ; // +stop, −fric; LAB            双唇塞音  〈帮组/唇音〉
                             ;                        ;                        ; /*(帮组三 C 介音   β)*/; // −stop, −fric; LAB            双唇近音
      case '端': return 't'  ; case '透': return 'tʰ' ; case '定': return 'd'  ; case '泥': return 'n'  ; // +stop, −fric; COR, +ant      齿龈塞音  〈端组/舌头音〉
      case '精': return 'ts' ; case '清': return 'tsʰ'; case '從': return 'dz' ;                        ; // +stop, +fric; COR, +ant      齿龈塞擦音〈精组/齿头音〉
      case '心': return 's'  ;                        ; case '邪': return 'z'  ;                        ; // −stop, +fric; COR, +ant      齿龈擦音  〈精组/齿头音〉
                             ;                        ;                        ; case '來': return 'l'  ; // −stop, −fric; COR, +ant      齿龈近音  〈来母/半舌音〉
      case '知': return 'tɹ' ; case '徹': return 'tɹʰ'; case '澄': return 'dɹ' ; case '孃': return 'nɹ' ; // +stop, −fric; COR, −ant, +r  卷舌塞音  〈知组/舌上音〉
      case '莊': return 'tʂ' ; case '初': return 'tʂʰ'; case '崇': return 'dʐ' ;                        ; // +stop, +fric; COR, −ant, +r  卷舌塞擦音〈庄组/正齿音〉
      case '生': return 'ʂ'  ;                        ; case '俟': return 'ʐ'  ;                        ; // −stop, +fric; COR, −ant, +r  卷舌擦音  〈庄组/正齿音〉
                             ;                        ;                        ; /* 钝音三 B 介音   ɹ */; // −stop, −fric; COR, −ant, +r  龈后近音
                             ;                        ;                        ; case '日': return 'ɲ'  ; // +stop, −fric; COR, −ant, −r  龈腭塞音  〈日母/半齿音〉
      case '章': return 'tɕ' ; case '昌': return 'tɕʰ'; case '常': return 'dʑ' ;                        ; // +stop, +fric; COR, −ant, −r  龈腭塞擦音〈章组/正齿音〉
      case '書': return 'ɕ'  ;                        ; case '船': return 'ʑ'  ;                        ; // −stop, +fric; COR, −ant, −r  龈腭擦音  〈章组/正齿音〉
                             ;                        ;                        ; case '以': return 'j'  ; // −stop, −fric; COR, −ant, −r  硬腭近音  〈以母/喉音〉
      case '見': return 'k'  ; case '溪': return 'kʰ' ; case '羣': return 'ɡ'  ; case '疑': return 'ŋ'  ; // +stop, −fric; DOR (+high)    软腭前塞音〈见组/牙音〉
                             ;                        ; case '匣': return 'ɣ'  ; case '云': return 'ɣ'  ; // −stop, +fric; DOR (+high)    软腭前擦音〈影组/喉音〉
                             ;                        ;                        ; /* 见系三 C 介音   j̈ */; // −stop, −fric; DOR (+high)    软腭前近音
      /*(见_非三等       q)  ;   (溪_非三等       qʰ) ;                        ;   (疑_非三等       ɴ)*/; // +stop, −fric; DOR (−high)    软腭后塞音
      /*(晓_非三等       χ)  ;                        ;   (匣              ʁ)*/;                        ; // −stop, +fric; DOR (−high)    软腭后擦音
      case '影': return 'ʔ'  ;                        ;                        ;                        ; // +stop, −fric                 喉塞音    〈影组/喉音〉
      case '曉': return 'h'  ;                        ;                        ;                        ; // −stop, +fric                 喉擦音    〈影组/喉音〉
    }
    throw new Error('无辅音规则');
  }

  const is全清 = is('幫端精心知莊生章書見影曉母'); // [−voi, −sg, −son]
  const is次清 = is('滂透清　徹初　昌　溪　　母'); // [−voi, +sg, −son]
  const is全浊 = is('並定從邪澄崇俟常船羣　匣母'); // [+voi, −sg, −son]
  const is次浊 = is('明泥來　孃　　日以疑云　母'); // [+voi, −sg, +son]
  const is清 = is全清 || is次清;
  // 云母已按推导后的结果 [ɹ] 算入次浊

  const is锐前 = is('端精組 或 來母 一二四等'); // [COR, +ant]
  const is锐后 = is('知莊章組 或 日以母');      // [COR, −ant]
  const is锐 = is锐前 || is锐后 || is('來母');  // [COR]
  // 来母按推导后只有非三等 [l] 算入前部锐音，但来母三等 [lɹ] 不算前部也不算后部

  // 函数：将声母的音韵地位转换为音位，包含开合信息
  function getInitial() {
    let result = getInitialWithoutRounding();
    // 音韵学术语开合对应 [±rnd]。如果主要调音部位就是 [LAB]〈帮组〉，那么本文一律视为 [−rnd]
    // 没有开合对立的韵母一般视为开口，但虞韵本文视为鱼韵对应的合口；平行地，钟韵也视为合口
    // 𠑆𦑣䎎小韵算合口
    if (is('合口 或 虞鍾韻 或 知組 嚴凡韻') && !is('幫組')) { // [+rnd]
      result += 'ʷ';
      result = result.replace('ʰʷ', 'ʷʰ');
    } // else [−rnd]
    return result;
  }

  // 函数：将软腭前音转换为软腭后音
  function velarToUvular(consonant) {
    switch (consonant[0]) {
      case 'k': return 'q' + consonant.substring(1);
      case 'ɡ': return 'ɢ' + consonant.substring(1);
      case 'ŋ': return 'ɴ' + consonant.substring(1);
      case 'ɣ': return 'ʁ' + consonant.substring(1);
    }
    return consonant;
  }

  // 函数：将知组转换为卷舌塞音
  function retroflexToStop(consonant) {
    switch (consonant.substring(0, 2)) {
      case 'tɹ': return 'ʈ' + consonant.substring(2);
      case 'dɹ': return 'ɖ' + consonant.substring(2);
      case 'nɹ': return 'ɳ' + consonant.substring(2);
    }
    return consonant;
  }

  // 函数：将介音的音韵地位转换为音位，不含开合信息
  function getGlide() {
    if (is('云母 灰韻')) return 'ɹ'; // “倄”小韵

    // 一二四等无介音
    if (!is('三等')) return '';

    // 锐音声母三等介音一律用 /ɹ/
    if (is锐) return 'ɹ';

    // 钝音声母分三 A、B、C
    if (is('重紐B類 或 庚臻韻')) return 'ɹ';
    if ('抑𡊁烋'.includes(字頭)) return 'ɹ'; // 蒸韵“抑𡊁”二字、幽韵“烋”字归三 B
    if (is('云母 支脂祭眞臻仙宵麻庚清蒸幽侵鹽韻')) return 'ɹ'; // 云母前元音韵归三 B
    if (is('重紐A類 或 麻蒸清幽韻')) return 'j'; // 三 A
    return 'j̈'; // 三 C
  }

  // 函数：将韵尾的音韵地位转换为音位
  function getCoda() {
    if (is('通江宕梗曾攝')) return is('入聲') ? 'k' : 'ŋ';
    if (is('深咸攝'))       return is('入聲') ? 'p' : 'm';
    if (is('臻山攝'))       return is('入聲') ? 't' : 'n';
    if (is('佳韻'))         return ''; // 从蟹摄中排除无韵尾的佳韵
    if (is('微韻 或 蟹攝')) return 'j';
    if (is('幽韻 或 效攝')) return 'w';
    return '';
  }

  /**
  3. 元音的特征

  [±high]:  高
  [±low]:   低
  [±front]: 前
  [±back]:  后
  [±rnd]:   圆唇（round）
  [±tense]: 紧。三子韵及其对应的三寅韵是松元音，三丑韵及其对应的三寅韵是紧元音，非三等韵都是紧元音
  [±divII]: 二等（division-II）。本文不指定它的具体语音实现

  列表如下。[ʉ] 是音位变体，也加入下表

               i  ɨ (ʉ) u  ɪ  +high, −low
         e͇  œ͇  e  ə     o  ɜ  −high, −low
         a͇        a           −high, +low
  front        +  −  −  −
  back         −  −  −  +
  rnd    −  +  −  −  +  +
  tense  +  +  +  +  +  +  −
  divII  +  +  −  −  −  −  −

  4. 音韵地位对应元音音位

  详见下面代码实现。
  */

  // 函数：将韵核的音韵地位转换为音位
  function getNucleus() {
    // 松元音
    // 韵尾: m     j   n       w
    if (is('侵　　微　眞臻欣文　韻')) return 'ɪ'; // +high, −tense
    if (is('鹽嚴凡祭廢仙　元　宵韻')) return 'ɜ'; // −high, −tense

    // 紧元音
    if (switches.肴豪韵韵核归为低元音) {
      if (is('肴韻')) return 'a͇';
      if (is('豪韻')) return 'a';
    }
    // 脂韵、尤韵的韵基也可分别视为 /ɪj/、/ɪw/，本文从简直接视为紧元音 /i/、/u/
    // 韵尾:   ŋ m j n w
    if (is('脂蒸　　　幽韻')) return 'i'; // +high, −low, +front, −back, −rnd, +tense
    if (is('之　　　　　韻')) return 'ɨ'; // +high, −low, −front, −back, −rnd, +tense
    if (is('尤東　　　侯韻')) return 'u'; // +high, −low, −front, +back, +rnd, +tense
    if (is('佳耕咸皆山肴韻')) return 'e͇'; // −high, −low, +divII,        −rnd, +tense
    if (is('　江　　　　韻')) return 'œ͇'; // −high, −low, +divII,        +rnd, +tense
    if (is('　青添齊先蕭韻')) return 'e'; // −high, −low, +front, −back, −rnd, +tense
    if (is('　登覃咍痕豪韻')) return 'ə'; // −high, −low, +front, −back, −rnd, +tense
    if (is('模冬　灰魂　韻')) return 'o'; // −high, −low, +front, −back, −rnd, +tense
    if (is('麻庚銜夬刪　韻 二等')) return 'a͇';
                                          // −high, +low, +divII,        −rnd, +tense
    if (is('麻庚　　　　韻 三等') ||
        is('歌唐談泰寒　韻')) return 'a'; // −high, +low, −front, −back, −rnd, +tense

    // 二合元音
    if (is('支韻'))     return 'iə'; // +front, −back, −rnd, +tense
    if (is('魚虞鍾韻')) return 'ɨə'; // −front, −back, −rnd, +tense
    if (is('清韻'))     return 'ia'; // +front, −back, −rnd, +tense
    if (is('陽韻'))     return 'ɨa'; // +front, −back, −rnd, +tense

    throw new Error('无元音规则');
  }

  // 函数：将半元音转换为元音
  function semivowelToVowel(consonant) {
    switch (consonant) {
      case 'j': return 'i';
      case 'ɥ': return 'y';
      case 'j̈': return 'ɨ';
      case 'ɥ̈': return 'ʉ';
      case 'w': return 'u';
    }
    return consonant;
  }

  /**
  5. 声调

  本文从简，声调无视音系层级范式，直接转换为最终形式。
  <声调写成五度标记>、<声调分阴阳>、<声调附加符号写在音节前> 也在这里应用。详见下面代码实现。
  */

  // 函数：将声调的音韵地位转换为语音
  function getTone() {
    if (switches.声调写成五度标记) {
      if (switches.声调分阴阳) {
        if (is('平聲')) return is清    ? '˦˧' : '˨˩';
        if (is('上聲')) return !is全浊 ? '˦˥' : '˨˧';
        if (is('去聲')) return is清    ? '˥˦˥' : '˧˨˧';
        if (is('入聲')) return !is全浊 ? '˥' : '˨˩';
      } else {
        if (is('平聲')) return '˧˩';
        if (is('上聲')) return '˧˥';
        if (is('去聲')) return '˦˨˦';
        if (is('入聲')) return '˧';
      }
    } else {
      if (switches.声调分阴阳) {
        if (is('平聲')) return is清    ? '̂' : '̏';
        if (is('上聲')) return !is全浊 ? '̋' : '̌';
        if (is('去聲')) return is清    ? '᷇' : '᷅';
        if (is('入聲')) return !is全浊 ? '́' : '̀';
      } else {
        if (is('平聲')) return switches.声调附加符号写在音节前 ? 'ˋ' : '̀'; // 写在音节前时直接用独立的附加符号
        if (is('上聲')) return switches.声调附加符号写在音节前 ? 'ˊ' : '́';
        if (is('去聲')) return switches.声调附加符号写在音节前 ? 'ˉ' : '̄';
        if (is('入聲')) return '';
      }
    }
    throw new Error('无声调规则');
  }

  /** 四、音系规则（及其代码实现）
  */

  // 获取音节的各部分
  let initial = getInitial();
  let glide = getGlide();
  let nucleus = getNucleus();
  let coda = getCoda();
  let tone = getTone();

  /**
  (1)  介音在后部锐音后被声母吸收而删除
       G -> ∅ / [COR, −ant]__
  */
  if (is('知莊章組 或 日以母')) glide = '';

  /**
  (2)  舌面介音被唇音或唇化声母同化〈帮组或合口三 A、C〉
       j -> ɥ / [+rnd]__
       j̈ -> ɥ̈ / [+rnd]__
            β / [LAB, −rnd]__
  */
  if (initial.includes('ʷ')) {
    glide = glide.replace('j', 'ɥ');
  } else if (is('幫組')) {
    if (glide == 'j̈') glide = 'β';
  }

  /**
  (3)  唇化的声母 j 实现为 ɥ〈以母合口〉
       jʷ -> ɥ
  */
  if (initial == 'jʷ') initial = 'ɥ';

  /**
  (4)  j 韵尾在低元音和中松元音后〈祭泰夬废〉实现为 ɹ
       j -> ɹ / {[+low], [−high, −tense]}__
  */
  if (switches.祭泰夬废韵尾推导为ɹ) {
    if (nucleus.includes('a') || nucleus == 'ɜ') {
      if (coda == 'j') coda = 'ɹ';
    }
  }

  /**
  (5)  松元音的前后被前接辅音的锐钝同化
       ɪ -> i / [COR]__
            ɨ / 其他环境
       ɜ -> e / [COR]__
            ə / 其他环境
  */
  if (switches.要推导松元音) {
    if ([...'ɹjɥ'].includes(glide) || is锐) { // 不包含 glide 为零的情况，所以用 [...'ɹjɥ']
      if (nucleus == 'ɪ') nucleus = 'i';
      if (nucleus == 'ɜ') nucleus = 'e';
    } else {
      if (nucleus == 'ɪ') nucleus = 'ɨ';
      if (nucleus == 'ɜ') nucleus = 'ə';
    }
  }

  /**
  (6)  二合元音的后滑音（off-glide）部分被元音的前后同化
       ə -> ɛ / i__
         -> ɜ / ɨ__
       a -> æ / i__
         -> ɐ / ɨ__
  */
  if (switches.要推导二合元音) {
    if (nucleus == 'iə') nucleus = 'iɛ';
    if (nucleus == 'ɨə') nucleus = 'ɨɜ';
    if (nucleus == 'ia') nucleus = 'iæ';
    if (nucleus == 'ɨa') nucleus = 'ɨɐ';
  }

  /**
  (7)  一等韵的韵核 /a/ 实现为 [ɑ]
       三等韵的韵核 /a/ 在锐音后实现为 [a]，但锐音是唇化〈歌三合〉的除外；在钝音后〈歌韵〉实现为 [ɑ]
       a -> a / {[COR, +ant, −rnd]G, [COR, −ant]}__
         -> ɑ / 其他环境
  */
  if (switches.要推导a && nucleus == 'a') {
    nucleus = 'ɑ';
    if (is锐 && glide && !initial.includes('ʷ') || is锐后 || !is锐 && [...'ɹjɥ'].includes(glide)) {
      // 音系规则本来不限制韵尾，但章组谈韵有“㶒譫”两小韵，需要归到 ɑ，所以在这里过滤
      if ('ŋk'.includes(coda)) nucleus = 'a'; // 注意 'ŋk'.includes(coda) 包含的是 ŋ、k 和零韵尾这 3 种
    }
  }

  /**
  (8)  央高元音被唇音或唇化声母同化（包括二合元音 ɨɜ -> ʉɜ〈虞钟阳韵〉）
       ɨ -> ʉ / [LAB]__
  */
  if (initial.includes('ʷ') || initial == 'ɥ' || glide == 'β') {
    nucleus = nucleus.replace('ɨ', 'ʉ');
  }

  /**
  (9)  i 在卷舌咝音和龈韵尾之间〈庄组真臻欣韵开口〉舌冠化为 ɹ̩
       i -> ɹ̩ / [COR, −ant, +r, +fric, −rnd]__[COR]
  */
  if (switches.庄组臻摄开口推导为ɹ̩ && is('莊組') && !initial.includes('ʷ') && [...'nt'].includes(coda)) {
    if (nucleus == 'i') nucleus = 'ɹ̩';
  }

  /**
  (10) 零介音、唇音或唇化韵尾前的 ə〈豪覃韵〉实现为 ʌ
       ə -> ʌ / 非G__[LAB]
  */
  if (switches.豪覃韵韵核推导为ʌ && !glide && [...'mpw'].includes(coda)) {
    if (nucleus == 'ə') nucleus = 'ʌ';
  }

  /**
  (11) e͇w〈肴韵〉、βəm〈凡韵〉的韵核实现为圆唇元音
       e͇      -> œ͇ / __w
       {e, ə} -> œ / [LAB]G__m
  */
  // 先转换𠑆𦑣䎎小韵，它们属于凡韵
  if (is锐 && initial.includes('ʷ') && [...'mp'].includes(coda)) {
    if (nucleus == 'e') nucleus = 'ə';
  }
  if (switches.肴凡韵韵核推导为œ) {
    if (nucleus == 'e͇' && coda == 'w') nucleus = 'œ͇';
    if (nucleus == 'ə' && (initial.includes('ʷ') || is('幫組')) && [...'mp'].includes(coda)) nucleus = 'œ';
  }

  /**
  (12) 齿龈阻音〈端精组〉后的介音接前元音时被同化（圆唇时可选）
       G -> j / [COR, +ant, −son, −rnd]__[+front]
            ɥ / [COR, +ant, −son, +rnd]__[+front]
  */
  if (is锐前 && 'ieæa'.includes(nucleus[0])) {
    if (!initial.includes('ʷ')) {
      if (glide) glide = 'j';
    } else if (switches.精三寅合口介音推导为ɥ) {
      if (glide) glide = 'ɥ';
    }
  }

  /**
  (13) i 在唇音或唇化声母和软腭韵尾之间〈蒸幽韵〉增生 ɹ 滑音
       G -> ɹ / [LAB]__i[DOR]
  */
  if (switches.蒸幽韵合口增生ɹ滑音 && (initial.includes('ʷ') || is('幫組')) && nucleus == 'i' && [...'ŋkw'].includes(coda)) {
    glide = 'ɹ';
  }

  /**
  (14) 接介音的 ɣ〈云母〉实现为 ɹ
       ɣG -> ɹ
       ɣʷG -> ɹʷ
  */
  if (switches.云母推导为ɹ && initial.includes('ɣ') && glide) {
    initial = initial.replace('ɣ', 'ɹ'); // ɹ 视为声母
    glide = '';
  }

  /**
  (15) 软腭音直接后接元音时〈见系和匣母非三等〉实现为软腭后音
       [DOR] -> [−high] / __V
  */
  if (switches.见系非三推导为软腭后音 && !glide) {
    initial = velarToUvular(initial);
  }

  /**
  (16) h 直接后接元音时〈晓母非三等〉实现为软腭后音
       h -> χ / __V
  */
  if (switches.晓母非三推导为软腭后音 && !glide) {
    initial = initial.replace('h', 'χ');
  }

  /**
  (17) 圆唇元音和低非前元音后的软腭韵尾〈通江宕摄〉实现为软腭后音
       [DOR] -> [+back] / {[+round], [+low, −front]}__
  */
  if (switches.通江宕摄推导为软腭后音) {
    if ('ʉuoœɑ'.includes(nucleus[0]) || (nucleus.includes('ɐ') && switches.要推导a)) {
      coda = velarToUvular(coda);
    }
  }

  /**
  (18) u 在钝音声母和无介音齿龈声母后〈侯韵〉裂化
       u -> u / [COR, −ant]__#
            ɘu / 其他__#
  */
  if (switches.侯韵裂化为ɘu && nucleus == 'u' && !coda) {
    nucleus = 'ɘu';
    if (is锐后 || is('云母') || glide) nucleus = 'u';
  }

  /**
  (19) 高元音 + 半元音韵尾〈微幽韵〉实现为二合元音
       j -> i / [+high]__
       w -> u / [+high]__
  */
  if ('iɨʉuɪ'.includes(nucleus)) {
    if (coda == 'j') coda = 'i';
    if (coda == 'w') coda = 'u';
  }

  /**
  (20) 移除与韵核同部位介音
       [DOR, +son] -> ∅ / C__[+high, −back]
  */
  if ('iɨʉ'.includes(nucleus[0])) {
    if (['j', 'ɥ', 'j̈', 'ɥ̈'].includes(glide)) glide = '';
  }

  /**
  (21) 齿龈音〈端精组〉接非后高元音时省略介音
       {j, ɥ} -> ∅ / [COR, +ant, −son]__[+high, +front, −back]
       G      -> ∅ / [COR, +ant, −son]__[+high, −front, −back]
  */
  if (switches.精组非后高元音省略介音 && is锐前) {
    if (nucleus[0] == 'i' && [...'jɥ'].includes(glide)) glide = '';
    if ('ɨʉ'.includes(nucleus[0])) glide = '';
  }

  /** 五、后处理的代码实现
  */

  if (switches.知组写成卷舌塞音 && is('知組 或 來母')) {
    initial = retroflexToStop(initial);
    if (is('知組 三等')) glide = 'ɹ'; // 还原出三等介音
    if ('iɨʉ'.includes(nucleus[0])) glide = ''; // 再次应用音系规则 (21)。平行地，也要应用给来母
  }

  if (switches.中元音写成半低元音) {
    nucleus = nucleus.replace('e', 'ɛ');
    nucleus = nucleus.replace('o', 'ɔ');
  }

  if (switches.ɑ写成a) {
    nucleus = nucleus.replace('ɑ', 'a');
  }

  if (switches.半元音介音写成元音) {
    glide = semivowelToVowel(glide);
  }

  if (switches.半元音韵尾写成元音) {
    coda = semivowelToVowel(coda);
  }

  if (is('二等')) {
    if (switches.二等元音写成r音钩) {
      nucleus = nucleus.replace('͇', '˞');
    } else if (switches.二等元音写成双下横线) {
      nucleus = nucleus.replace('͇', '̳');
    }
  }

  if (switches.声调写成五度标记) return initial + glide + nucleus + coda + tone;

  if (switches.声调附加符号写在音节前) {
    if (switches.声调分阴阳) return '\xA0' + tone + initial + glide + nucleus + coda; // 需要用无中断空格（U+00A0），否则位于行首的空格可能被浏览器忽略
    return tone + initial + glide + nucleus + coda;
  }

  // 声调附加符号写在韵核主体上
  if (nucleus.includes('͇') || nucleus.includes('̳') || nucleus == 'ɘu' || nucleus == 'ɹ̩') return initial + glide + nucleus + tone + coda;
  return initial + glide + nucleus[0] + tone + nucleus.substring(1) + coda;
}

/**
 * msoeg 中古拟音 V8
 * https://zhuanlan.zhihu.com/p/145409852
 *
 * 説明
 * 以下內容為生成 msoeg 中古拟音 V8 的函數體
 * 函數接受音韻地位，返回對應的 msoeg 中古拟音 V8
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的msoeg 中古拟音 V8
 */
export function msoeg_v8(音韻地位, 字頭) {
  return {
    EAA: 'tᵒuɴ',
    GAA: 'dᵒuɴ',
    JBA: 'ʈuɴ',
    LBA: 'ɖuɴ',
    XBA: 'tçuɴ',
    KBA: 'ʈʰuɴ',
    UBA: 'dʐɻuɴ',
    QBA: 'suɴ',
    cBA: 'ɲuɴ',
    dBA: 'kuɴ',
    lBA: 'ɥuɴ',
    kBA: 'wuɴ',
    DBA: 'muɴ',
    eBA: 'kʰuɴ',
    fBA: 'ɡuɴ',
    CBA: 'buɴ',
    ABA: 'puɴ',
    BBA: 'pʰuɴ',
    YBA: 'tçʰuɴ',
    IBA: 'ɭuɴ',
    eAA: 'qʰᵒuɴ',
    dAA: 'qᵒuɴ',
    DAA: 'mᵒuɴ',
    IAA: 'lᵒuɴ',
    jAA: 'ʁᵒuɴ',
    PAA: 'dzᵒuɴ',
    hAA: 'ʔᵒuɴ',
    OAA: 'tsʰᵒuɴ',
    FAA: 'tʰᵒuɴ',
    NAA: 'tsᵒuɴ',
    CAA: 'bᵒuɴ',
    iAA: 'χᵒuɴ',
    gAA: 'ɴᵒuɴ',
    QAA: 'sᵒuɴ',
    ECA: 'tɔɴ',
    GCA: 'dɔɴ',
    PCA: 'dzɔɴ',
    HCA: 'nɔɴ',
    dCA: 'qɔɴ',
    jCA: 'ʁɔɴ',
    ICA: 'lɔɴ',
    NCA: 'tsɔɴ',
    QCA: 'sɔɴ',
    FCA: 'tʰɔɴ',
    XDA: 'tçuoɴ',
    IDA: 'ɭuoɴ',
    aDA: 'çuoɴ',
    RDA: 'zuoɴ',
    YDA: 'tçʰuoɴ',
    lDA: 'ɥuoɴ',
    ADA: 'puoɴ',
    iDA: 'xuoɴ',
    gDA: 'ŋuoɴ',
    hDA: 'ʔuoɴ',
    MDA: 'ɳuoɴ',
    LDA: 'ɖuoɴ',
    PDA: 'dzuoɴ',
    KDA: 'ʈʰuoɴ',
    CDA: 'buoɴ',
    BDA: 'pʰuoɴ',
    NDA: 'tsuoɴ',
    cDA: 'ɲuoɴ',
    fDA: 'ɡuoɴ',
    ZDA: 'dʝuoɴ',
    dDA: 'kuoɴ',
    QDA: 'suoɴ',
    ODA: 'tsʰuoɴ',
    eDA: 'kʰuoɴ',
    dEA: 'qɞ˞ɴ',
    DEA: 'mɞ˞ɴ',
    MEA: 'ɳɞ˞ɴ',
    TEA: 'tʂʰɻɞ˞ɴ',
    AEA: 'pɞ˞ɴ',
    jEA: 'ʁɞ˞ɴ',
    BEA: 'pʰɞ˞ɴ',
    IEA: 'ɭɞ˞ɴ',
    VEA: 'ʂɻɞ˞ɴ',
    CEA: 'bɞ˞ɴ',
    iEA: 'χɞ˞ɴ',
    hEA: 'ʔɞ˞ɴ',
    eEA: 'qʰɞ˞ɴ',
    LEA: 'ɖɞ˞ɴ',
    KEA: 'ʈʰɞ˞ɴ',
    JEA: 'ʈɞ˞ɴ',
    gEA: 'ɴɞ˞ɴ',
    UEA: 'dʐɻɞ˞ɴ',
    XFA: 'tçie',
    lFA: 'jie',
    kFI: 'wue',
    dFM: 'kue',
    iFM: 'xue',
    hFM: 'ʔue',
    DFE: 'mɯe',
    iFI: 'xʷie',
    LFI: 'ɖʷie',
    ZFI: 'dʝʷie',
    IFI: 'ɭʷie',
    YFI: 'tçʰʷie',
    BFE: 'pʰɯe',
    AFE: 'pɯe',
    RFI: 'zʷie',
    eFM: 'kʰue',
    eFI: 'kʰʷie',
    fFE: 'ɡɯe',
    fFA: 'ɡie',
    iFE: 'xɯe',
    eFE: 'kʰɯe',
    gFE: 'ŋɯe',
    CFE: 'bɯe',
    ZFA: 'dʝie',
    cFA: 'ɲie',
    IFA: 'ɭie',
    PFA: 'dzie',
    NFA: 'tsie',
    dFE: 'kɯe',
    AFA: 'pie',
    CFA: 'bie',
    aFA: 'çie',
    QFA: 'sie',
    TFA: 'tʂʰɻe',
    KFA: 'ʈʰie',
    DFA: 'mie',
    OFA: 'tsʰie',
    JFA: 'ʈie',
    hFE: 'ʔɯe',
    LFA: 'ɖie',
    QFI: 'sʷie',
    gFM: 'ŋue',
    iFA: 'xie',
    VFA: 'ʂɻe',
    VFI: 'ʂɻʷe',
    cFI: 'ɲʷie',
    NFA: 'tsie',
    dFI: 'kʷie',
    NFI: 'tsʷie',
    TFI: 'tʂʰɻʷe',
    JFI: 'ʈʷie',
    NFI: 'tsʷie',
    YFA: 'tçʰie',
    SFA: 'tʂɻe',
    BFA: 'pʰie',
    lFI: 'ɥie',
    UFA: 'dʐɻe',
    XGA: 'tçi',
    lGA: 'ji',
    VGA: 'ʂɻi',
    CGA: 'bi',
    NGA: 'tsi',
    dGE: 'ki˞i',
    YGA: 'tçʰi',
    KGA: 'ʈʰi',
    OGA: 'tsʰi',
    PGA: 'dzi',
    MGA: 'ɳi',
    LGA: 'ɖi',
    QGA: 'si',
    aGA: 'çi',
    fGE: 'ɡi˞i',
    hGA: 'ʔi',
    IGA: 'ɭi',
    fGI: 'ɡʷi',
    JGI: 'ʈʷi',
    dGM: 'kʷi˞i',
    cGI: 'ɲʷi',
    VGI: 'ʂɻʷi',
    lGI: 'ɥi',
    IGI: 'ɭʷi',
    QGI: 'sʷi',
    fGM: 'ɡʷi˞i',
    DGE: 'mi˞i',
    AGE: 'pi˞i',
    XGI: 'tçʷi',
    ZGI: 'dʝʷi',
    kGI: 'wi˞i',
    CGE: 'bi˞i',
    BGE: 'pʰi˞i',
    iGI: 'xʷi',
    LGI: 'ɖʷi',
    YGI: 'tçʰʷi',
    JGA: 'ʈi',
    BGA: 'pʰi',
    NGI: 'tsʷi',
    eGM: 'kʰʷi˞i',
    gGE: 'ŋi˞i',
    iGA: 'xi',
    XHA: 'tçiɯ',
    lHA: 'jiɯ',
    ZHA: 'dʝiɯ',
    gHA: 'ŋɯ',
    QHA: 'siɯ',
    THA: 'tʂʰɻɯ',
    fHA: 'ɡɯ',
    aHA: 'çiɯ',
    cHA: 'ɲiɯ',
    eHA: 'kʰɯ',
    dHA: 'kɯ',
    RHA: 'ziɯ',
    IHA: 'ɭiɯ',
    SHA: 'tʂɻɯ',
    iHA: 'xɯ',
    hHA: 'ʔɯ',
    KHA: 'ʈʰiɯ',
    LHA: 'ɖiɯ',
    YHA: 'tçʰiɯ',
    PHA: 'dziɯ',
    NHA: 'tsiɯ',
    UHA: 'dʐɻɯ',
    WHA: 'ʐɻɯ',
    eHA: 'kʰɯ',
    aHA: 'çiɯ',
    DIA: 'mui',
    iII: 'xui',
    kII: 'wui',
    BIA: 'pʰui',
    AIA: 'pui',
    CIA: 'bui',
    hII: 'ʔui',
    fIA: 'ɡɯi',
    dIA: 'kɯi',
    iIA: 'xɯi',
    hIA: 'ʔɯi',
    gIA: 'ŋɯi',
    gII: 'ŋui',
    dII: 'kui',
    eII: 'kʰui',
    gJA: 'ŋɯɤ',
    TJA: 'tʂʰɻɯɤ',
    aJA: 'çɯɤ',
    dJA: 'kɯɤ',
    fJA: 'ɡɯɤ',
    lJA: 'jɯɤ',
    QJA: 'sɯɤ',
    OJA: 'tsʰɯɤ',
    UJA: 'dʐɻɯɤ',
    KJA: 'ʈʰɯɤ',
    VJA: 'ʂɻɯɤ',
    iJA: 'xɯɤ',
    RJA: 'zɯɤ',
    hJA: 'ʔɯɤ',
    JJA: 'ʈɯɤ',
    IJA: 'ɭɯɤ',
    XJA: 'tçɯɤ',
    LJA: 'ɖɯɤ',
    cJA: 'ɲɯɤ',
    NJA: 'tsɯɤ',
    eJA: 'kʰɯɤ',
    SJA: 'tʂɻɯɤ',
    ZJA: 'dʝɯɤ',
    MJA: 'ɳɯɤ',
    gKA: 'ŋuo',
    TKA: 'tʂʰɻuo',
    DKA: 'muo',
    kKA: 'wuo',
    iKA: 'xuo',
    fKA: 'ɡuo',
    cKA: 'ɲuo',
    QKA: 'suo',
    JKA: 'ʈuo',
    KKA: 'ʈʰuo',
    ZKA: 'dʝuo',
    lKA: 'ɥuo',
    eKA: 'kʰuo',
    XKA: 'tçuo',
    OKA: 'tsʰuo',
    IKA: 'ɭuo',
    CKA: 'buo',
    UKA: 'dʐɻuo',
    SKA: 'tʂɻuo',
    BKA: 'pʰuo',
    NKA: 'tsuo',
    AKA: 'puo',
    hKA: 'ʔuo',
    aKA: 'çuo',
    YKA: 'tçʰuo',
    LKA: 'ɖuo',
    dKA: 'kuo',
    VKA: 'ʂɻuo',
    DLA: 'mo',
    CLA: 'bo',
    jLA: 'ʁo',
    dLA: 'qo',
    GLA: 'do',
    HLA: 'no',
    iLA: 'χo',
    gLA: 'ɴo',
    NLA: 'tso',
    ILA: 'lo',
    QLA: 'so',
    PLA: 'dzo',
    hLA: 'ʔo',
    ALA: 'po',
    eLA: 'qʰo',
    OLA: 'tsʰo',
    FLA: 'tʰo',
    ELA: 'to',
    BLA: 'pʰo',
    PMA: 'dzɛi',
    IMA: 'lɛi',
    OMA: 'tsʰɛi',
    EMA: 'tɛi',
    GMA: 'dɛi',
    AMA: 'pɛi',
    dMA: 'qɛi',
    jMA: 'ʁɛi',
    hMA: 'ʔɛi',
    gMA: 'ɴɛi',
    iMA: 'χɛi',
    QMA: 'sɛi',
    FMA: 'tʰɛi',
    CMA: 'bɛi',
    BMA: 'pʰɛi',
    NMA: 'tsɛi',
    DMA: 'mɛi',
    HMA: 'nɛi',
    eMA: 'qʰɛi',
    dMI: 'qʷɛi',
    eMI: 'qʰʷɛi',
    jMI: 'ʁʷɛi',
    cMA: 'ɲɛi',
    ZMA: 'dʝɛi',
    hMI: 'ʔʷɛi',
    iMI: 'χʷɛi',
    dPA: 'qɜ˞',
    jPA: 'ʁɜ˞',
    CPA: 'bɜ˞',
    dPI: 'qʷɜ˞',
    hPI: 'ʔʷɜ˞',
    ePI: 'qʰʷɜ˞',
    UPA: 'dʐɻɜ˞',
    TPA: 'tʂʰɻɜ˞',
    iPI: 'χʷɜ˞',
    MPA: 'ɳɜ˞',
    gPA: 'ɴɜ˞',
    hPA: 'ʔɜ˞',
    VPA: 'ʂɻɜ˞',
    iPA: 'χɜ˞',
    KPA: 'ʈʰɜ˞',
    DPA: 'mɜ˞',
    jPI: 'ʁʷɜ˞',
    dQA: 'qɜ˞i',
    hQA: 'ʔɜ˞i',
    jQA: 'ʁɜ˞i',
    CQA: 'bɜ˞i',
    dQI: 'qʷɜ˞i',
    jQI: 'ʁʷɜ˞i',
    eQI: 'qʰʷɜ˞i',
    UQA: 'dʐɻɜ˞i',
    TQA: 'tʂʰɻɜ˞i',
    iQI: 'χʷɜ˞i',
    LQI: 'ɖʷɜ˞i',
    DQA: 'mɜ˞i',
    SQA: 'tʂɻɜ˞i',
    hQI: 'ʔʷɜ˞i',
    JQA: 'ʈɜ˞i',
    IQA: 'ɭɜ˞i',
    eQA: 'qʰɜ˞i',
    MQA: 'ɳɜ˞i',
    VQA: 'ʂɻɜ˞i',
    gQA: 'ɴɜ˞i',
    KQA: 'ʈʰɜ˞i',
    iQA: 'χɜ˞i',
    IQI: 'ɭʷɜ˞i',
    UQI: 'dʐɻʷɜ˞i',
    iSI: 'χʷʌi',
    eSI: 'qʰʷʌi',
    hSI: 'ʔʷʌi',
    jSI: 'ʁʷʌi',
    DSA: 'mʌi',
    dSI: 'qʷʌi',
    ISI: 'lʷʌi',
    GSI: 'dʷʌi',
    OSI: 'tsʰʷʌi',
    ESI: 'tʷʌi',
    QSI: 'sʷʌi',
    PSI: 'dzʷʌi',
    CSA: 'bʌi',
    ASA: 'pʌi',
    BSA: 'pʰʌi',
    gSI: 'ɴʷʌi',
    FSI: 'tʰʷʌi',
    HSI: 'nʷʌi',
    NSI: 'tsʷʌi',
    iTA: 'χʌi',
    eTA: 'qʰʌi',
    hTA: 'ʔʌi',
    GTA: 'dʌi',
    dTA: 'qʌi',
    PTA: 'dzʌi',
    ITA: 'lʌi',
    NTA: 'tsʌi',
    OTA: 'tsʰʌi',
    CTA: 'bʌi',
    FTA: 'tʰʌi',
    jTA: 'ʁʌi',
    QTA: 'sʌi',
    gTA: 'ɴʌi',
    HTA: 'nʌi',
    ETA: 'tʌi',
    YTA: 'tçʰʌi',
    BTA: 'pʰʌi',
    XVA: 'tçin',
    KVA: 'ʈʰin',
    hVA: 'ʔin',
    QVA: 'sin',
    ZVA: 'dʝin',
    cVA: 'ɲin',
    bVA: 'ʝin',
    OVA: 'tsʰin',
    kVA: 'ɰi˞n',
    aVA: 'çin',
    AVA: 'pin',
    IVA: 'ɭin',
    fVE: 'ɡi˞n',
    JVA: 'ʈin',
    LVA: 'ɖin',
    NVA: 'tsin',
    YVA: 'tçʰin',
    PVA: 'dzin',
    lVA: 'jin',
    MVA: 'ɳin',
    BVA: 'pʰin',
    CVA: 'bin',
    gVE: 'ŋi˞n',
    dVE: 'ki˞n',
    dVM: 'kʷi˞n',
    kVI: 'wi˞n',
    eVM: 'kʰʷi˞n',
    DVE: 'mi˞n',
    CVE: 'bi˞n',
    hVE: 'ʔi˞n',
    hVM: 'ʔʷi˞n',
    AVE: 'pi˞n',
    DVA: 'min',
    XVI: 'tçʷin',
    KVI: 'ʈʰʷin',
    LVI: 'ɖʷin',
    QVI: 'sʷin',
    ZVI: 'dʝʷin',
    cVI: 'ɲʷin',
    bVI: 'ʝʷin',
    IVI: 'ɭʷin',
    JVI: 'ʈʷin',
    OVI: 'tsʰʷin',
    NVI: 'tsʷin',
    YVI: 'tçʰʷin',
    PVI: 'dzʷin',
    lVI: 'ɥin',
    RVI: 'zʷin',
    dVI: 'kʷin',
    fVA: 'ɡin',
    BVE: 'pʰi˞n',
    SWA: 'tʂɻn',
    VWA: 'ʂɻn',
    UWA: 'dʐɻn',
    DXA: 'mun',
    kXI: 'wun',
    hXI: 'ʔun',
    CXA: 'bun',
    AXA: 'pun',
    fXI: 'ɡun',
    iXI: 'xun',
    dXI: 'kun',
    BXA: 'pʰun',
    iYA: 'xɯn',
    hYA: 'ʔɯn',
    fYA: 'ɡɯn',
    dYA: 'kɯn',
    gYA: 'ŋɯn',
    gZI: 'ŋuʌn',
    kZI: 'wuʌn',
    CZA: 'buʌn',
    BZA: 'pʰuʌn',
    iZI: 'xuʌn',
    hZI: 'ʔuʌn',
    gZA: 'ŋɯʌn',
    eZA: 'kʰɯʌn',
    iZA: 'xɯʌn',
    dZA: 'kɯʌn',
    hZA: 'ʔɯʌn',
    AZA: 'puʌn',
    fZA: 'ɡɯʌn',
    DZA: 'muʌn',
    jaI: 'ʁʷʌn',
    daI: 'qʷʌn',
    haI: 'ʔʷʌn',
    DaA: 'mʌn',
    QaI: 'sʷʌn',
    NaI: 'tsʷʌn',
    PaI: 'dzʷʌn',
    EaI: 'tʷʌn',
    FaI: 'tʰʷʌn',
    GaI: 'dʷʌn',
    OaI: 'tsʰʷʌn',
    gaI: 'ɴʷʌn',
    CaA: 'bʌn',
    AaA: 'pʌn',
    IaI: 'lʷʌn',
    eaI: 'qʰʷʌn',
    iaI: 'χʷʌn',
    BaA: 'pʰʌn',
    HaI: 'nʷʌn',
    jbA: 'ʁʌn',
    dbA: 'qʌn',
    hbA: 'ʔʌn',
    FbA: 'tʰʌn',
    gbA: 'ɴʌn',
    jcA: 'ʁɑn',
    gcA: 'ɴɑn',
    EcA: 'tɑn',
    hcA: 'ʔɑn',
    HcA: 'nɑn',
    OcA: 'tsʰɑn',
    FcA: 'tʰɑn',
    QcA: 'sɑn',
    GcA: 'dɑn',
    PcA: 'dzɑn',
    dcA: 'qɑn',
    IcA: 'lɑn',
    ecA: 'qʰɑn',
    icA: 'χɑn',
    HcI: 'nʷɑn',
    jcI: 'ʁʷɑn',
    gcI: 'ɴʷɑn',
    EcI: 'tʷɑn',
    hcI: 'ʔʷɑn',
    FcI: 'tʰʷɑn',
    QcI: 'sʷɑn',
    GcI: 'dʷɑn',
    PcI: 'dzʷɑn',
    dcI: 'qʷɑn',
    IcI: 'lʷɑn',
    icI: 'χʷɑn',
    ecI: 'qʰʷɑn',
    NcI: 'tsʷɑn',
    CcA: 'bɑn',
    DcA: 'mɑn',
    BcA: 'pʰɑn',
    AcA: 'pɑn',
    VdA: 'ʂɻa˞n',
    ddI: 'qʷa˞n',
    hdI: 'ʔʷa˞n',
    VdI: 'ʂɻʷa˞n',
    jdI: 'ʁʷa˞n',
    AdA: 'pa˞n',
    DdA: 'ma˞n',
    gdA: 'ɴa˞n',
    ddA: 'qa˞n',
    BdA: 'pʰa˞n',
    MdI: 'ɳʷa˞n',
    gdI: 'ɴʷa˞n',
    edA: 'qʰa˞n',
    edA: 'qʰa˞n',
    SdI: 'tʂɻʷa˞n',
    VeA: 'ʂɻɜ˞n',
    deI: 'qʷɜ˞n',
    deA: 'qɜ˞n',
    jeA: 'ʁɜ˞n',
    eeA: 'qʰɜ˞n',
    UeA: 'dʐɻɜ˞n',
    ieA: 'χɜ˞n',
    geA: 'ɴɜ˞n',
    heA: 'ʔɜ˞n',
    AeA: 'pɜ˞n',
    IeA: 'ɭɜ˞n',
    MeA: 'ɳɜ˞n',
    YeA: 'tçʰɜ˞n',
    heI: 'ʔʷɜ˞n',
    GeI: 'dʷɜ˞n',
    IeI: 'ɭʷɜ˞n',
    feI: 'ɢʷɜ˞n',
    JeA: 'ʈɜ˞n',
    LeA: 'ɖɜ˞n',
    jeI: 'ʁʷɜ˞n',
    PeA: 'dzɜ˞n',
    QgA: 'sɛn',
    PgA: 'dzɛn',
    OgA: 'tsʰɛn',
    NgA: 'tsɛn',
    FgA: 'tʰɛn',
    dgA: 'qɛn',
    jgA: 'ʁɛn',
    hgA: 'ʔɛn',
    IgA: 'lɛn',
    GgA: 'dɛn',
    HgA: 'nɛn',
    EgA: 'tɛn',
    egA: 'qʰɛn',
    ggA: 'ɴɛn',
    DgA: 'mɛn',
    CgA: 'bɛn',
    hgI: 'ʔʷɛn',
    dgI: 'qʷɛn',
    igI: 'χʷɛn',
    AgA: 'pɛn',
    jgI: 'ʁʷɛn',
    igA: 'χɛn',
    UgI: 'dʐɻʷɛn',
    QfA: 'siɛn',
    PfA: 'dziɛn',
    OfA: 'tsʰiɛn',
    NfA: 'tsiɛn',
    cfA: 'ɲiɛn',
    lfA: 'jiɛn',
    XfA: 'tçiɛn',
    dfA: 'kiɛn',
    JfA: 'ʈiɛn',
    UfA: 'dʐɻɛn',
    afA: 'çiɛn',
    KfA: 'ʈʰiɛn',
    ZfA: 'dʝiɛn',
    LfA: 'ɖiɛn',
    ifA: 'xiɛn',
    IfA: 'ɭiɛn',
    BfA: 'pʰiɛn',
    CfA: 'biɛn',
    DfA: 'miɛn',
    PfI: 'dzʷiɛn',
    QfI: 'sʷiɛn',
    NfI: 'tsʷiɛn',
    ifI: 'xʷiɛn',
    cfI: 'ɲʷiɛn',
    YfI: 'tçʰʷiɛn',
    lfI: 'ɥiɛn',
    RfI: 'zʷiɛn',
    hfI: 'ʔʷiɛn',
    bfI: 'ʝʷiɛn',
    AfA: 'piɛn',
    RfA: 'ziɛn',
    OfI: 'tsʰʷiɛn',
    XfI: 'tçʷiɛn',
    ZfI: 'dʝʷiɛn',
    kfI: 'wi˞ɛn',
    SfI: 'tʂɻʷɛn',
    VfI: 'ʂɻʷɛn',
    KfI: 'ʈʰʷiɛn',
    ffE: 'ɡi˞ɛn',
    efE: 'kʰi˞ɛn',
    ffM: 'ɡʷi˞ɛn',
    LfI: 'ɖʷiɛn',
    IfI: 'ɭʷiɛn',
    efM: 'kʰʷi˞ɛn',
    hfE: 'ʔi˞ɛn',
    kfA: 'ɰi˞ɛn',
    hfM: 'ʔʷi˞ɛn',
    YfA: 'tçʰiɛn',
    JfI: 'ʈʷiɛn',
    dfM: 'kʷi˞ɛn',
    QhA: 'sɛu',
    FhA: 'tʰɛu',
    EhA: 'tɛu',
    GhA: 'dɛu',
    dhA: 'qɛu',
    IhA: 'lɛu',
    ghA: 'ɴɛu',
    ihA: 'χɛu',
    hhA: 'ʔɛu',
    ehA: 'qʰɛu',
    QiA: 'siɛu',
    KiA: 'ʈʰiɛu',
    JiA: 'ʈiɛu',
    LiA: 'ɖiɛu',
    iiE: 'xɯɛu',
    PiA: 'dziɛu',
    diE: 'kɯɛu',
    NiA: 'tsiɛu',
    ciA: 'ɲiɛu',
    aiA: 'çiɛu',
    liA: 'jiɛu',
    ZiA: 'dʝiɛu',
    XiA: 'tçiɛu',
    AiA: 'piɛu',
    AiE: 'pɯɛu',
    CiA: 'biɛu',
    DiA: 'miɛu',
    DiE: 'mɯɛu',
    hiA: 'ʔiɛu',
    kiA: 'ɰɯɛu',
    fiE: 'ɡɯɛu',
    OiA: 'tsʰiɛu',
    hiE: 'ʔɯɛu',
    eiA: 'kʰiɛu',
    YiA: 'tçʰiɛu',
    BiA: 'pʰiɛu',
    fiA: 'ɡiɛu',
    IiA: 'ɭiɛu',
    eiE: 'kʰɯɛu',
    jjA: 'ʁa˞u',
    djA: 'qa˞u',
    UjA: 'dʐɻa˞u',
    MjA: 'ɳa˞u',
    VjA: 'ʂɻa˞u',
    DjA: 'ma˞u',
    ijA: 'χa˞u',
    AjA: 'pa˞u',
    BjA: 'pʰa˞u',
    ejA: 'qʰa˞u',
    gjA: 'ɴa˞u',
    SjA: 'tʂɻa˞u',
    JjA: 'ʈa˞u',
    TjA: 'tʂʰɻa˞u',
    CjA: 'ba˞u',
    hjA: 'ʔa˞u',
    KjA: 'ʈʰa˞u',
    IjA: 'ɭa˞u',
    LjA: 'ɖa˞u',
    jkA: 'ʁɑu',
    dkA: 'qɑu',
    IkA: 'lɑu',
    ikA: 'χɑu',
    DkA: 'mɑu',
    FkA: 'tʰɑu',
    EkA: 'tɑu',
    QkA: 'sɑu',
    CkA: 'bɑu',
    AkA: 'pɑu',
    GkA: 'dɑu',
    NkA: 'tsɑu',
    gkA: 'ɴɑu',
    PkA: 'dzɑu',
    hkA: 'ʔɑu',
    HkA: 'nɑu',
    ekA: 'qʰɑu',
    OkA: 'tsʰɑu',
    BkA: 'pʰɑu',
    dlA: 'qɑ',
    OlA: 'tsʰɑ',
    ElA: 'tɑ',
    QlA: 'sɑ',
    GlA: 'dɑ',
    PlA: 'dzɑ',
    glA: 'ɴɑ',
    FlA: 'tʰɑ',
    IlA: 'lɑ',
    HlA: 'nɑ',
    jlA: 'ʁɑ',
    ilA: 'χɑ',
    elA: 'qʰɑ',
    hlA: 'ʔɑ',
    dlI: 'qʷɑ',
    OlI: 'tsʰʷɑ',
    ElI: 'tʷɑ',
    QlI: 'sʷɑ',
    ClA: 'bɑ',
    GlI: 'dʷɑ',
    DlA: 'mɑ',
    PlI: 'dzʷɑ',
    glI: 'ɴʷɑ',
    FlI: 'tʰʷɑ',
    IlI: 'lʷɑ',
    HlI: 'nʷɑ',
    AlA: 'pɑ',
    BlA: 'pʰɑ',
    jlI: 'ʁʷɑ',
    elI: 'qʰʷɑ',
    hlI: 'ʔʷɑ',
    imI: 'xuɑ',
    hmI: 'ʔuɑ',
    emI: 'kʰuɑ',
    fmA: 'ɡɯɑ',
    emA: 'kʰɯɑ',
    dmA: 'kɯɑ',
    OmI: 'tsʰuɑ',
    NmI: 'tsuɑ',
    fmI: 'ɡuɑ',
    ImI: 'ɭuɑ',
    DnA: 'ma˞',
    YoA: 'tçʰia',
    aoA: 'çia',
    loA: 'jia',
    XoA: 'tçia',
    NoA: 'tsia',
    boA: 'ʝia',
    jnI: 'ʁʷa˞',
    dnI: 'qʷa˞',
    inI: 'χʷa˞',
    enI: 'qʰʷa˞',
    MnA: 'ɳa˞',
    dnA: 'qa˞',
    jnA: 'ʁa˞',
    BnA: 'pʰa˞',
    hnA: 'ʔa˞',
    AnA: 'pa˞',
    TnA: 'tʂʰɻa˞',
    VnA: 'ʂɻa˞',
    gnA: 'ɴa˞',
    SnA: 'tʂɻa˞',
    LnA: 'ɖa˞',
    RoA: 'zia',
    ZoA: 'dʝia',
    hnI: 'ʔʷa˞',
    SnI: 'tʂɻʷa˞',
    JnI: 'ʈʷa˞',
    CnA: 'ba˞',
    UnA: 'dʐɻa˞',
    KnA: 'ʈʰa˞',
    JnA: 'ʈa˞',
    inA: 'χa˞',
    enA: 'qʰa˞',
    PoA: 'dzia',
    coA: 'ɲia',
    QoA: 'sia',
    JoA: 'ʈia',
    gnI: 'ɴʷa˞',
    enA: 'qʰa˞',
    lpA: 'jɯɑɴ',
    RpA: 'zɯɑɴ',
    IpA: 'ɭɯɑɴ',
    ipA: 'xɯɑɴ',
    apA: 'çɯɑɴ',
    CpA: 'buɑɴ',
    XpA: 'tçɯɑɴ',
    YpA: 'tçʰɯɑɴ',
    epA: 'kʰɯɑɴ',
    dpA: 'kɯɑɴ',
    LpA: 'ɖɯɑɴ',
    JpA: 'ʈɯɑɴ',
    cpA: 'ɲɯɑɴ',
    ApA: 'puɑɴ',
    QpA: 'sɯɑɴ',
    NpA: 'tsɯɑɴ',
    TpA: 'tʂʰɻɑɴ',
    DpA: 'muɑɴ',
    MpA: 'ɳɯɑɴ',
    UpA: 'dʐɻɑɴ',
    SpA: 'tʂɻɑɴ',
    ZpA: 'dʝɯɑɴ',
    VpA: 'ʂɻɑɴ',
    PpA: 'dzɯɑɴ',
    OpA: 'tsʰɯɑɴ',
    epI: 'kʰuɑɴ',
    kpI: 'wuɑɴ',
    hpA: 'ʔɯɑɴ',
    fpA: 'ɡɯɑɴ',
    KpA: 'ʈʰɯɑɴ',
    BpA: 'pʰuɑɴ',
    fpI: 'ɡuɑɴ',
    GqA: 'dɑɴ',
    IqA: 'lɑɴ',
    EqA: 'tɑɴ',
    OqA: 'tsʰɑɴ',
    dqA: 'qɑɴ',
    QqA: 'sɑɴ',
    eqA: 'qʰɑɴ',
    iqI: 'χʷɑɴ',
    jqI: 'ʁʷɑɴ',
    dqI: 'qʷɑɴ',
    FqA: 'tʰɑɴ',
    BqA: 'pʰɑɴ',
    hqI: 'ʔʷɑɴ',
    hqA: 'ʔɑɴ',
    iqA: 'χɑɴ',
    jqA: 'ʁɑɴ',
    DqA: 'mɑɴ',
    NqA: 'tsɑɴ',
    HqA: 'nɑɴ',
    CqA: 'bɑɴ',
    gqA: 'ɴɑɴ',
    PqA: 'dzɑɴ',
    eqI: 'qʰʷɑɴ',
    AqA: 'pɑɴ',
    drA: 'qa˞ŋ',
    erA: 'qʰa˞ŋ',
    DrA: 'ma˞ŋ',
    jrI: 'ʁʷa˞ŋ',
    ArA: 'pa˞ŋ',
    irI: 'χʷa˞ŋ',
    drI: 'qʷa˞ŋ',
    CrA: 'ba˞ŋ',
    irA: 'χa˞ŋ',
    KrA: 'ʈʰa˞ŋ',
    TrA: 'tʂʰɻa˞ŋ',
    UrA: 'dʐɻa˞ŋ',
    hsA: 'ʔi˞aŋ',
    BrA: 'pʰa˞ŋ',
    CsA: 'bi˞aŋ',
    dsA: 'ki˞aŋ',
    DsA: 'mi˞aŋ',
    LrA: 'ɖa˞ŋ',
    JrA: 'ʈa˞ŋ',
    ksI: 'wi˞aŋ',
    AsA: 'pi˞aŋ',
    isI: 'xʷi˞aŋ',
    esA: 'kʰi˞aŋ',
    VrA: 'ʂɻa˞ŋ',
    fsA: 'ɡi˞aŋ',
    gsA: 'ŋi˞aŋ',
    jrA: 'ʁa˞ŋ',
    MrA: 'ɳa˞ŋ',
    dtA: 'qɜ˞ŋ',
    etA: 'qʰɜ˞ŋ',
    DtA: 'mɜ˞ŋ',
    AtA: 'pɜ˞ŋ',
    jtI: 'ʁʷɜ˞ŋ',
    jtA: 'ʁɜ˞ŋ',
    JtA: 'ʈɜ˞ŋ',
    htA: 'ʔɜ˞ŋ',
    UtA: 'dʐɻɜ˞ŋ',
    TtA: 'tʂʰɻɜ˞ŋ',
    MtA: 'ɳɜ˞ŋ',
    BtA: 'pʰɜ˞ŋ',
    itI: 'χʷɜ˞ŋ',
    AtA: 'pɜ˞ŋ',
    LtA: 'ɖɜ˞ŋ',
    htI: 'ʔʷɜ˞ŋ',
    CtA: 'bɜ˞ŋ',
    gtA: 'ɴɜ˞ŋ',
    StA: 'tʂɻɜ˞ŋ',
    OuA: 'tsʰiaŋ',
    PuA: 'dziaŋ',
    NuA: 'tsiaŋ',
    luA: 'jiaŋ',
    luI: 'ɥiaŋ',
    huA: 'ʔiaŋ',
    JuA: 'ʈiaŋ',
    KuA: 'ʈʰiaŋ',
    ZuA: 'dʝiaŋ',
    LuA: 'ɖiaŋ',
    auA: 'çiaŋ',
    XuA: 'tçiaŋ',
    euA: 'kʰiaŋ',
    DuA: 'miaŋ',
    IuA: 'ɭiaŋ',
    AuA: 'piaŋ',
    euI: 'kʰʷiaŋ',
    RuA: 'ziaŋ',
    huI: 'ʔʷiaŋ',
    fuI: 'ɡʷiaŋ',
    QuI: 'sʷiaŋ',
    fuA: 'ɡiaŋ',
    iuI: 'xʷiaŋ',
    OvA: 'tsʰɛŋ',
    dvA: 'qɛŋ',
    jvA: 'ʁɛŋ',
    GvA: 'dɛŋ',
    EvA: 'tɛŋ',
    ivA: 'χɛŋ',
    QvA: 'sɛŋ',
    BvA: 'pʰɛŋ',
    IvA: 'lɛŋ',
    HvA: 'nɛŋ',
    FvA: 'tʰɛŋ',
    DvA: 'mɛŋ',
    CvA: 'bɛŋ',
    jvI: 'ʁʷɛŋ',
    dvI: 'qʷɛŋ',
    XwA: 'tçiɯŋ',
    ZwA: 'dʝiɯŋ',
    LwA: 'ɖiɯŋ',
    IwA: 'ɭiɯŋ',
    hwA: 'ʔɯŋ',
    CwA: 'bi˞ɯŋ',
    AwA: 'pi˞ɯŋ',
    lwA: 'jiɯŋ',
    bwA: 'ʝiɯŋ',
    awA: 'çiɯŋ',
    cwA: 'ɲiɯŋ',
    dwA: 'kɯŋ',
    JwA: 'ʈiɯŋ',
    PwA: 'dziɯŋ',
    gwA: 'ŋɯŋ',
    iwA: 'xɯŋ',
    YwA: 'tçʰiɯŋ',
    fwA: 'ɡɯŋ',
    VwA: 'ʂɻɯŋ',
    ewA: 'kʰɯŋ',
    KwA: 'ʈʰiɯŋ',
    UwA: 'dʐɻɯŋ',
    BwA: 'pʰi˞ɯŋ',
    ExA: 'tʌŋ',
    IxA: 'lʌŋ',
    QxA: 'sʌŋ',
    AxA: 'pʌŋ',
    NxA: 'tsʌŋ',
    DxA: 'mʌŋ',
    PxA: 'dzʌŋ',
    CxA: 'bʌŋ',
    jxI: 'ʁʷʌŋ',
    dxI: 'qʷʌŋ',
    ixI: 'χʷʌŋ',
    HxA: 'nʌŋ',
    GxA: 'dʌŋ',
    jxA: 'ʁʌŋ',
    dxA: 'qʌŋ',
    FxA: 'tʰʌŋ',
    BxA: 'pʰʌŋ',
    kyA: 'wu',
    hyA: 'ʔu',
    IyA: 'ɭu',
    OyA: 'tsʰu',
    lyA: 'ju',
    gyA: 'ŋu',
    NyA: 'tsu',
    PyA: 'dzu',
    QyA: 'su',
    KyA: 'ʈʰu',
    eyA: 'kʰu',
    YyA: 'tçʰu',
    XyA: 'tçu',
    ZyA: 'dʝu',
    cyA: 'ɲu',
    ayA: 'çu',
    eyA: 'kʰu',
    ByA: 'pʰu',
    dyA: 'ku',
    AyA: 'pu',
    VyA: 'ʂɻu',
    TyA: 'tʂʰɻu',
    SyA: 'tʂɻu',
    UyA: 'dʐɻu',
    iyA: 'xu',
    RyA: 'zu',
    LyA: 'ɖu',
    JyA: 'ʈu',
    fyA: 'ɡu',
    CyA: 'bu',
    DyA: 'mu',
    jzA: 'ʁᵒu',
    hzA: 'ʔᵒu',
    HzA: 'nᵒu',
    IzA: 'lᵒu',
    QzA: 'sᵒu',
    ezA: 'qʰᵒu',
    izA: 'χᵒu',
    NzA: 'tsᵒu',
    FzA: 'tʰᵒu',
    GzA: 'dᵒu',
    gzA: 'ɴᵒu',
    dzA: 'qᵒu',
    EzA: 'tᵒu',
    PzA: 'dzᵒu',
    CzA: 'bᵒu',
    OzA: 'tsʰᵒu',
    DzA: 'mᵒu',
    h0A: 'ʔiu',
    f0A: 'ɡiu',
    A0A: 'piu',
    I0A: 'ɭiu',
    d0A: 'kiu',
    C0A: 'biu',
    N0A: 'tsiu',
    V0A: 'ʂɻu',
    g0A: 'ŋiu',
    i0A: 'xiu',
    D0A: 'miu',
    O1A: 'tsʰim',
    R1A: 'zim',
    I1A: 'ɭim',
    K1A: 'ʈʰim',
    X1A: 'tçim',
    L1A: 'ɖim',
    J1A: 'ʈim',
    Z1A: 'dʝim',
    c1A: 'ɲim',
    a1A: 'çim',
    l1A: 'jim',
    Q1A: 'sim',
    h1A: 'ʔim',
    N1A: 'tsim',
    P1A: 'dzim',
    M1A: 'ɳim',
    f1E: 'ɡɯm',
    e1E: 'kʰɯm',
    g1E: 'ŋɯm',
    i1E: 'xɯm',
    d1E: 'kɯm',
    h1E: 'ʔɯm',
    V1A: 'ʂɻm',
    U1A: 'dʐɻm',
    S1A: 'tʂɻm',
    T1A: 'tʂʰɻm',
    Y1A: 'tçʰim',
    G2A: 'dɔm',
    O2A: 'tsʰɔm',
    H2A: 'nɔm',
    h2A: 'ʔɔm',
    j2A: 'ʁɔm',
    I2A: 'lɔm',
    P2A: 'dzɔm',
    N2A: 'tsɔm',
    F2A: 'tʰɔm',
    E2A: 'tɔm',
    e2A: 'qʰɔm',
    i2A: 'χɔm',
    Q2A: 'sɔm',
    d2A: 'qɔm',
    g2A: 'ɴɔm',
    G3A: 'dɑm',
    d3A: 'qɑm',
    E3A: 'tɑm',
    Q3A: 'sɑm',
    I3A: 'lɑm',
    e3A: 'qʰɑm',
    F3A: 'tʰɑm',
    P3A: 'dzɑm',
    j3A: 'ʁɑm',
    D3A: 'mɑm',
    P3A: 'dzɑm',
    i3A: 'χɑm',
    l4A: 'jiɛm',
    I4A: 'ɭiɛm',
    A4E: 'pi˞ɛm',
    Q4A: 'siɛm',
    O4A: 'tsʰiɛm',
    X4A: 'tçiɛm',
    Z4A: 'dʝiɛm',
    a4A: 'çiɛm',
    Y4A: 'tçʰiɛm',
    c4A: 'ɲiɛm',
    M4A: 'ɳiɛm',
    k4A: 'wi˞ɛm',
    J4A: 'ʈiɛm',
    K4A: 'ʈʰiɛm',
    h4E: 'ʔi˞ɛm',
    e4E: 'kʰi˞ɛm',
    g4E: 'ŋi˞ɛm',
    N4A: 'tsiɛm',
    P4A: 'dziɛm',
    f4E: 'ɡi˞ɛm',
    h4A: 'ʔiɛm',
    R4A: 'ziɛm',
    V4A: 'ʂɻɛm',
    L4A: 'ɖiɛm',
    f4A: 'ɡiɛm',
    F5A: 'tʰɛm',
    E5A: 'tɛm',
    G5A: 'dɛm',
    I5A: 'lɛm',
    e5A: 'qʰɛm',
    d5A: 'qɛm',
    j5A: 'ʁɛm',
    H5A: 'nɛm',
    i5A: 'χɛm',
    j6A: 'ʁɜ˞m',
    d6A: 'qɜ˞m',
    V6A: 'ʂɻɜ˞m',
    h6A: 'ʔɜ˞m',
    g6A: 'ɴɜ˞m',
    i6A: 'χɜ˞m',
    J6A: 'ʈɜ˞m',
    M6A: 'ɳɜ˞m',
    U6A: 'dʐɻɜ˞m',
    e6A: 'qʰɜ˞m',
    j7A: 'ʁa˞m',
    U7A: 'dʐɻa˞m',
    g7A: 'ɴa˞m',
    T7A: 'tʂʰɻa˞m',
    V7A: 'ʂɻa˞m',
    d7A: 'qa˞m',
    C7A: 'ba˞m',
    e7A: 'qʰa˞m',
    g8A: 'ŋɯʌm',
    i8A: 'xɯʌm',
    h8A: 'ʔɯʌm',
    e8A: 'kʰɯʌm',
    C9A: 'buʌm',
    B9A: 'pʰuʌm',
    EAB: 'tᵒuɴˀ',
    DAB: 'mᵒuɴˀ',
    eAB: 'qʰᵒuɴˀ',
    QAB: 'sᵒuɴˀ',
    FAB: 'tʰᵒuɴˀ',
    NAB: 'tsᵒuɴˀ',
    jAB: 'ʁᵒuɴˀ',
    hAB: 'ʔᵒuɴˀ',
    HAB: 'nᵒuɴˀ',
    AAB: 'pᵒuɴˀ',
    IAB: 'lᵒuɴˀ',
    iAB: 'χᵒuɴˀ',
    GAB: 'dᵒuɴˀ',
    CAB: 'bᵒuɴˀ',
    XDB: 'tçuoɴˀ',
    KDB: 'ʈʰuoɴˀ',
    IDB: 'ɭuoɴˀ',
    hDB: 'ʔuoɴˀ',
    cDB: 'ɲuoɴˀ',
    LDB: 'ɖuoɴˀ',
    JDB: 'ʈuoɴˀ',
    CDB: 'buoɴˀ',
    BDB: 'pʰuoɴˀ',
    lDB: 'ɥuoɴˀ',
    eDB: 'kʰuoɴˀ',
    ZDB: 'dʝuoɴˀ',
    dDB: 'kuoɴˀ',
    QDB: 'suoɴˀ',
    iDB: 'xuoɴˀ',
    ADB: 'puoɴˀ',
    ECB: 'tɔɴˀ',
    DCB: 'mɔɴˀ',
    fDB: 'ɡuoɴˀ',
    YDB: 'tçʰuoɴˀ',
    XDB: 'tçuoɴˀ',
    NDB: 'tsuoɴˀ',
    dEB: 'qɞ˞ɴˀ',
    CEB: 'bɞ˞ɴˀ',
    hEB: 'ʔɞ˞ɴˀ',
    DEB: 'mɞ˞ɴˀ',
    jEB: 'ʁɞ˞ɴˀ',
    AEB: 'pɞ˞ɴˀ',
    iEB: 'χɞ˞ɴˀ',
    XFB: 'tçieˀ',
    ZFB: 'dʝieˀ',
    DFF: 'mɯeˀ',
    AFF: 'pɯeˀ',
    CFF: 'bɯeˀ',
    iFN: 'xueˀ',
    hFN: 'ʔueˀ',
    eFN: 'kʰueˀ',
    dFN: 'kueˀ',
    QFJ: 'sʷieˀ',
    IFJ: 'ɭʷieˀ',
    fFF: 'ɡɯeˀ',
    hFF: 'ʔɯeˀ',
    dFF: 'kɯeˀ',
    eFF: 'kʰɯeˀ',
    gFF: 'ŋɯeˀ',
    kFJ: 'wueˀ',
    NFJ: 'tsʷieˀ',
    cFJ: 'ɲʷieˀ',
    OFB: 'tsʰieˀ',
    LFB: 'ɖieˀ',
    QFB: 'sieˀ',
    lFB: 'jieˀ',
    IFB: 'ɭieˀ',
    VFB: 'ʂɻeˀ',
    AFB: 'pieˀ',
    cFB: 'ɲieˀ',
    DFB: 'mieˀ',
    CFB: 'bieˀ',
    YFB: 'tçʰieˀ',
    aFB: 'çieˀ',
    NFB: 'tsieˀ',
    XFJ: 'tçʷieˀ',
    TFJ: 'tʂʰɻʷeˀ',
    RFJ: 'zʷieˀ',
    bFB: 'ʝieˀ',
    SFB: 'tʂɻeˀ',
    BFF: 'pʰɯeˀ',
    BFB: 'pʰieˀ',
    lFJ: 'ɥieˀ',
    PFJ: 'dzʷieˀ',
    eFJ: 'kʰʷieˀ',
    MFB: 'ɳieˀ',
    gFN: 'ŋueˀ',
    fFN: 'ɡueˀ',
    KFB: 'ʈʰieˀ',
    JFB: 'ʈieˀ',
    iFF: 'xɯeˀ',
    eFB: 'kʰieˀ',
    ZFJ: 'dʝʷieˀ',
    dFB: 'kieˀ',
    XGB: 'tçiˀ',
    ZGB: 'dʝiˀ',
    DGF: 'mi˞iˀ',
    AGF: 'pi˞iˀ',
    RGB: 'ziˀ',
    dGF: 'ki˞iˀ',
    NGB: 'tsiˀ',
    AGB: 'piˀ',
    dGN: 'kʷi˞iˀ',
    kGJ: 'wi˞iˀ',
    aGB: 'çiˀ',
    LGB: 'ɖiˀ',
    QGB: 'siˀ',
    CGB: 'biˀ',
    IGB: 'ɭiˀ',
    aGJ: 'çʷiˀ',
    IGJ: 'ɭʷiˀ',
    fGJ: 'ɡʷiˀ',
    OGJ: 'tsʰʷiˀ',
    MGB: 'ɳiˀ',
    dGJ: 'kʷiˀ',
    CGF: 'bi˞iˀ',
    PGJ: 'dzʷiˀ',
    BGF: 'pʰi˞iˀ',
    cGJ: 'ɲʷiˀ',
    lGJ: 'ɥiˀ',
    hGF: 'ʔi˞iˀ',
    NGJ: 'tsʷiˀ',
    JGB: 'ʈiˀ',
    KGB: 'ʈʰiˀ',
    XGB: 'tçiˀ',
    iGJ: 'xʷiˀ',
    fGN: 'ɡʷi˞iˀ',
    fGF: 'ɡi˞iˀ',
    eGN: 'kʰʷi˞iˀ',
    XHB: 'tçiɯˀ',
    ZHB: 'dʝiɯˀ',
    JHB: 'ʈiɯˀ',
    iHB: 'xɯˀ',
    dHB: 'kɯˀ',
    lHB: 'jiɯˀ',
    RHB: 'ziɯˀ',
    VHB: 'ʂɻɯˀ',
    cHB: 'ɲiɯˀ',
    IHB: 'ɭiɯˀ',
    QHB: 'siɯˀ',
    aHB: 'çiɯˀ',
    LHB: 'ɖiɯˀ',
    eHB: 'kʰɯˀ',
    UHB: 'dʐɻɯˀ',
    WHB: 'ʐɻɯˀ',
    NHB: 'tsiɯˀ',
    kHB: 'ɰɯˀ',
    gHB: 'ŋɯˀ',
    YHB: 'tçʰiɯˀ',
    KHB: 'ʈʰiɯˀ',
    THB: 'tʂʰɻɯˀ',
    SHB: 'tʂɻɯˀ',
    hHB: 'ʔɯˀ',
    MHB: 'ɳiɯˀ',
    DIB: 'muiˀ',
    hIB: 'ʔɯiˀ',
    eIB: 'kʰɯiˀ',
    dIB: 'kɯiˀ',
    BIB: 'pʰuiˀ',
    AIB: 'puiˀ',
    kIJ: 'wuiˀ',
    dIJ: 'kuiˀ',
    iIJ: 'xuiˀ',
    gIB: 'ŋɯiˀ',
    iIB: 'xɯiˀ',
    hIJ: 'ʔuiˀ',
    CIB: 'buiˀ',
    gJB: 'ŋɯɤˀ',
    IJB: 'ɭɯɤˀ',
    LJB: 'ɖɯɤˀ',
    lJB: 'jɯɤˀ',
    XJB: 'tçɯɤˀ',
    cJB: 'ɲɯɤˀ',
    aJB: 'çɯɤˀ',
    YJB: 'tçʰɯɤˀ',
    JJB: 'ʈɯɤˀ',
    QJB: 'sɯɤˀ',
    KJB: 'ʈʰɯɤˀ',
    MJB: 'ɳɯɤˀ',
    iJB: 'xɯɤˀ',
    fJB: 'ɡɯɤˀ',
    VJB: 'ʂɻɯɤˀ',
    TJB: 'tʂʰɻɯɤˀ',
    SJB: 'tʂɻɯɤˀ',
    UJB: 'dʐɻɯɤˀ',
    PJB: 'dzɯɤˀ',
    hJB: 'ʔɯɤˀ',
    dJB: 'kɯɤˀ',
    RJB: 'zɯɤˀ',
    eJB: 'kʰɯɤˀ',
    bJB: 'ʝɯɤˀ',
    ZJB: 'dʝɯɤˀ',
    OJB: 'tsʰɯɤˀ',
    NJB: 'tsɯɤˀ',
    gKB: 'ŋuoˀ',
    kKB: 'wuoˀ',
    PKB: 'dzuoˀ',
    AKB: 'puoˀ',
    DKB: 'muoˀ',
    CKB: 'buoˀ',
    BKB: 'pʰuoˀ',
    LKB: 'ɖuoˀ',
    iKB: 'xuoˀ',
    ZKB: 'dʝuoˀ',
    lKB: 'ɥuoˀ',
    XKB: 'tçuoˀ',
    hKB: 'ʔuoˀ',
    eKB: 'kʰuoˀ',
    JKB: 'ʈuoˀ',
    cKB: 'ɲuoˀ',
    fKB: 'ɡuoˀ',
    VKB: 'ʂɻuoˀ',
    dKB: 'kuoˀ',
    OKB: 'tsʰuoˀ',
    IKB: 'ɭuoˀ',
    QKB: 'suoˀ',
    UKB: 'dʐɻuoˀ',
    DLB: 'moˀ',
    FLB: 'tʰoˀ',
    GLB: 'doˀ',
    ILB: 'loˀ',
    OLB: 'tsʰoˀ',
    ELB: 'toˀ',
    dLB: 'qoˀ',
    gLB: 'ɴoˀ',
    CLB: 'boˀ',
    PLB: 'dzoˀ',
    NLB: 'tsoˀ',
    iLB: 'χoˀ',
    hLB: 'ʔoˀ',
    eLB: 'qʰoˀ',
    HLB: 'noˀ',
    jLB: 'ʁoˀ',
    BLB: 'pʰoˀ',
    ALB: 'poˀ',
    PMB: 'dzɛiˀ',
    IMB: 'lɛiˀ',
    FMB: 'tʰɛiˀ',
    BMB: 'pʰɛiˀ',
    NMB: 'tsɛiˀ',
    EMB: 'tɛiˀ',
    GMB: 'dɛiˀ',
    HMB: 'nɛiˀ',
    QMB: 'sɛiˀ',
    OMB: 'tsʰɛiˀ',
    eMB: 'qʰɛiˀ',
    jMB: 'ʁɛiˀ',
    DMB: 'mɛiˀ',
    CMB: 'bɛiˀ',
    hMB: 'ʔɛiˀ',
    gMB: 'ɴɛiˀ',
    AMB: 'pɛiˀ',
    jPB: 'ʁɜ˞ˀ',
    DPB: 'mɜ˞ˀ',
    ePB: 'qʰɜ˞ˀ',
    LPB: 'ɖɜ˞ˀ',
    MPB: 'ɳɜ˞ˀ',
    CPB: 'bɜ˞ˀ',
    hPB: 'ʔɜ˞ˀ',
    APB: 'pɜ˞ˀ',
    dPB: 'qɜ˞ˀ',
    VPB: 'ʂɻɜ˞ˀ',
    dPJ: 'qʷɜ˞ˀ',
    LPJ: 'ɖʷɜ˞ˀ',
    jPJ: 'ʁʷɜ˞ˀ',
    iPJ: 'χʷɜ˞ˀ',
    fPB: 'ɢɜ˞ˀ',
    jQB: 'ʁɜ˞iˀ',
    eQB: 'qʰɜ˞iˀ',
    gQB: 'ɴɜ˞iˀ',
    hQB: 'ʔɜ˞iˀ',
    iSJ: 'χʷʌiˀ',
    hSJ: 'ʔʷʌiˀ',
    ISJ: 'lʷʌiˀ',
    GSJ: 'dʷʌiˀ',
    PSJ: 'dzʷʌiˀ',
    DSB: 'mʌiˀ',
    FSJ: 'tʰʷʌiˀ',
    jSJ: 'ʁʷʌiˀ',
    eSJ: 'qʰʷʌiˀ',
    ESJ: 'tʷʌiˀ',
    HSJ: 'nʷʌiˀ',
    JSJ: 'ʈʷʌiˀ',
    gSJ: 'ɴʷʌiˀ',
    OSJ: 'tsʰʷʌiˀ',
    CSB: 'bʌiˀ',
    NSJ: 'tsʷʌiˀ',
    kSJ: 'wʌiˀ',
    iTB: 'χʌiˀ',
    eTB: 'qʰʌiˀ',
    NTB: 'tsʌiˀ',
    GTB: 'dʌiˀ',
    HTB: 'nʌiˀ',
    dTB: 'qʌiˀ',
    jTB: 'ʁʌiˀ',
    BTB: 'pʰʌiˀ',
    OTB: 'tsʰʌiˀ',
    YTB: 'tçʰʌiˀ',
    ETB: 'tʌiˀ',
    DTB: 'mʌiˀ',
    PTB: 'dzʌiˀ',
    BTB: 'pʰʌiˀ',
    hTB: 'ʔʌiˀ',
    lTB: 'jʌiˀ',
    FTB: 'tʰʌiˀ',
    cTB: 'ɲʌiˀ',
    ITB: 'lʌiˀ',
    lTB: 'jʌiˀ',
    CTB: 'bʌiˀ',
    XVB: 'tçinˀ',
    KVB: 'ʈʰinˀ',
    ZVB: 'dʝinˀ',
    cVB: 'ɲinˀ',
    aVB: 'çinˀ',
    IVB: 'ɭinˀ',
    LVB: 'ɖinˀ',
    dVB: 'kinˀ',
    PVB: 'dzinˀ',
    NVB: 'tsinˀ',
    CVB: 'binˀ',
    gVF: 'ŋi˞nˀ',
    OVB: 'tsʰinˀ',
    fVN: 'ɡʷi˞nˀ',
    lVB: 'jinˀ',
    DVF: 'mi˞nˀ',
    DVB: 'minˀ',
    kVJ: 'wi˞nˀ',
    XVJ: 'tçʷinˀ',
    lVJ: 'ɥinˀ',
    QVJ: 'sʷinˀ',
    cVJ: 'ɲʷinˀ',
    YVJ: 'tçʰʷinˀ',
    bVJ: 'ʝʷinˀ',
    KVJ: 'ʈʰʷinˀ',
    IVJ: 'ɭʷinˀ',
    cVJ: 'ɲʷinˀ',
    eVJ: 'kʰʷinˀ',
    eVB: 'kʰinˀ',
    aVJ: 'çʷinˀ',
    iVB: 'xinˀ',
    UVB: 'dʐɻnˀ',
    JVB: 'ʈinˀ',
    DXB: 'munˀ',
    AXB: 'punˀ',
    CXB: 'bunˀ',
    BXB: 'pʰunˀ',
    hXJ: 'ʔunˀ',
    gXJ: 'ŋunˀ',
    kXJ: 'wunˀ',
    eXJ: 'kʰunˀ',
    hYB: 'ʔɯnˀ',
    dYB: 'kɯnˀ',
    SWB: 'tʂɻnˀ',
    eYB: 'kʰɯnˀ',
    fYB: 'ɡɯnˀ',
    TWB: 'tʂʰɻnˀ',
    gYB: 'ŋɯnˀ',
    iYB: 'xɯnˀ',
    gZJ: 'ŋuʌnˀ',
    kZJ: 'wuʌnˀ',
    hZB: 'ʔɯʌnˀ',
    dZB: 'kɯʌnˀ',
    fZB: 'ɡɯʌnˀ',
    eZB: 'kʰɯʌnˀ',
    gZB: 'ŋɯʌnˀ',
    iZB: 'xɯʌnˀ',
    DZB: 'muʌnˀ',
    AZB: 'puʌnˀ',
    fZJ: 'ɡuʌnˀ',
    hZJ: 'ʔuʌnˀ',
    eZJ: 'kʰuʌnˀ',
    iZJ: 'xuʌnˀ',
    CZB: 'buʌnˀ',
    jaJ: 'ʁʷʌnˀ',
    BaB: 'pʰʌnˀ',
    OaJ: 'tsʰʷʌnˀ',
    AaB: 'pʌnˀ',
    QaJ: 'sʷʌnˀ',
    NaJ: 'tsʷʌnˀ',
    haJ: 'ʔʷʌnˀ',
    GaJ: 'dʷʌnˀ',
    PaJ: 'dzʷʌnˀ',
    daJ: 'qʷʌnˀ',
    FaJ: 'tʰʷʌnˀ',
    eaJ: 'qʰʷʌnˀ',
    IaJ: 'lʷʌnˀ',
    CaB: 'bʌnˀ',
    DaB: 'mʌnˀ',
    HaJ: 'nʷʌnˀ',
    iaJ: 'χʷʌnˀ',
    jbB: 'ʁʌnˀ',
    ebB: 'qʰʌnˀ',
    dbB: 'qʌnˀ',
    jcB: 'ʁɑnˀ',
    EcB: 'tɑnˀ',
    FcB: 'tʰɑnˀ',
    QcB: 'sɑnˀ',
    GcB: 'dɑnˀ',
    PcB: 'dzɑnˀ',
    dcB: 'qɑnˀ',
    IcB: 'lɑnˀ',
    ecB: 'qʰɑnˀ',
    icB: 'χɑnˀ',
    NcB: 'tsɑnˀ',
    jcJ: 'ʁʷɑnˀ',
    EcJ: 'tʷɑnˀ',
    hcJ: 'ʔʷɑnˀ',
    FcJ: 'tʰʷɑnˀ',
    QcJ: 'sʷɑnˀ',
    dcJ: 'qʷɑnˀ',
    IcJ: 'lʷɑnˀ',
    ecJ: 'qʰʷɑnˀ',
    HcJ: 'nʷɑnˀ',
    NcJ: 'tsʷɑnˀ',
    CcB: 'bɑnˀ',
    DcB: 'mɑnˀ',
    AcB: 'pɑnˀ',
    RcJ: 'zʷɑnˀ',
    GcJ: 'dʷɑnˀ',
    HcB: 'nɑnˀ',
    BcB: 'pʰɑnˀ',
    VdB: 'ʂɻa˞nˀ',
    hdJ: 'ʔʷa˞nˀ',
    AdB: 'pa˞nˀ',
    SdB: 'tʂɻa˞nˀ',
    MdB: 'ɳa˞nˀ',
    jdB: 'ʁa˞nˀ',
    jdJ: 'ʁʷa˞nˀ',
    CdB: 'ba˞nˀ',
    DdB: 'ma˞nˀ',
    UdB: 'dʐɻa˞nˀ',
    gdB: 'ɴa˞nˀ',
    TdB: 'tʂʰɻa˞nˀ',
    BdB: 'pʰa˞nˀ',
    UdJ: 'dʐɻʷa˞nˀ',
    VeB: 'ʂɻɜ˞nˀ',
    jeB: 'ʁɜ˞nˀ',
    DeB: 'mɜ˞nˀ',
    deB: 'qɜ˞nˀ',
    TeB: 'tʂʰɻɜ˞nˀ',
    UeB: 'dʐɻɜ˞nˀ',
    geB: 'ɴɜ˞nˀ',
    SeB: 'tʂɻɜ˞nˀ',
    TeJ: 'tʂʰɻʷɜ˞nˀ',
    eeB: 'qʰɜ˞nˀ',
    QgB: 'sɛnˀ',
    FgB: 'tʰɛnˀ',
    EgB: 'tɛnˀ',
    hgB: 'ʔɛnˀ',
    GgB: 'dɛnˀ',
    dgB: 'qɛnˀ',
    jgB: 'ʁɛnˀ',
    igB: 'χɛnˀ',
    DgB: 'mɛnˀ',
    HgB: 'nɛnˀ',
    AgB: 'pɛnˀ',
    jgJ: 'ʁʷɛnˀ',
    dgJ: 'qʷɛnˀ',
    CgB: 'bɛnˀ',
    egJ: 'qʰʷɛnˀ',
    egB: 'qʰɛnˀ',
    ggB: 'ɴɛnˀ',
    QfB: 'siɛnˀ',
    lfB: 'jiɛnˀ',
    PfB: 'dziɛnˀ',
    JfB: 'ʈiɛnˀ',
    XfB: 'tçiɛnˀ',
    MfB: 'ɳiɛnˀ',
    OfB: 'tsʰiɛnˀ',
    YfB: 'tçʰiɛnˀ',
    efB: 'kʰiɛnˀ',
    dfF: 'ki˞ɛnˀ',
    ZfB: 'dʝiɛnˀ',
    NfB: 'tsiɛnˀ',
    cfB: 'ɲiɛnˀ',
    RfB: 'ziɛnˀ',
    IfB: 'ɭiɛnˀ',
    gfF: 'ŋi˞ɛnˀ',
    ffF: 'ɡi˞ɛnˀ',
    CfF: 'bi˞ɛnˀ',
    DfB: 'miɛnˀ',
    AfB: 'piɛnˀ',
    NfJ: 'tsʷiɛnˀ',
    PfJ: 'dzʷiɛnˀ',
    AfF: 'pi˞ɛnˀ',
    lfJ: 'ɥiɛnˀ',
    IfJ: 'ɭʷiɛnˀ',
    JfJ: 'ʈʷiɛnˀ',
    dfN: 'kʷi˞ɛnˀ',
    ffN: 'ɡʷi˞ɛnˀ',
    cfJ: 'ɲʷiɛnˀ',
    YfJ: 'tçʰʷiɛnˀ',
    ZfJ: 'dʝʷiɛnˀ',
    LfJ: 'ɖʷiɛnˀ',
    XfJ: 'tçʷiɛnˀ',
    QfJ: 'sʷiɛnˀ',
    UfJ: 'dʐɻʷɛnˀ',
    ffJ: 'ɡʷiɛnˀ',
    ifJ: 'xʷiɛnˀ',
    CfB: 'biɛnˀ',
    DfF: 'mi˞ɛnˀ',
    KfB: 'ʈʰiɛnˀ',
    BfF: 'pʰi˞ɛnˀ',
    afB: 'çiɛnˀ',
    hfF: 'ʔi˞ɛnˀ',
    LfB: 'ɖiɛnˀ',
    UfB: 'dʐɻɛnˀ',
    QhB: 'sɛuˀ',
    dhB: 'qɛuˀ',
    EhB: 'tɛuˀ',
    IhB: 'lɛuˀ',
    FhB: 'tʰɛuˀ',
    ihB: 'χɛuˀ',
    hhB: 'ʔɛuˀ',
    HhB: 'nɛuˀ',
    jhB: 'ʁɛuˀ',
    GhB: 'dɛuˀ',
    ehB: 'qʰɛuˀ',
    NhB: 'tsɛuˀ',
    QiB: 'siɛuˀ',
    LiB: 'ɖiɛuˀ',
    XiB: 'tçiɛuˀ',
    hiF: 'ʔɯɛuˀ',
    KiB: 'ʈʰiɛuˀ',
    aiB: 'çiɛuˀ',
    ciB: 'ɲiɛuˀ',
    CiB: 'biɛuˀ',
    YiB: 'tçʰiɛuˀ',
    BiB: 'pʰiɛuˀ',
    DiB: 'miɛuˀ',
    ZiB: 'dʝiɛuˀ',
    diF: 'kɯɛuˀ',
    AiF: 'pɯɛuˀ',
    AiB: 'piɛuˀ',
    CiF: 'bɯɛuˀ',
    liB: 'jiɛuˀ',
    OiB: 'tsʰiɛuˀ',
    NiB: 'tsiɛuˀ',
    fiF: 'ɡɯɛuˀ',
    IiB: 'ɭiɛuˀ',
    BiF: 'pʰɯɛuˀ',
    hiB: 'ʔiɛuˀ',
    ejB: 'qʰa˞uˀ',
    jjB: 'ʁa˞uˀ',
    AjB: 'pa˞uˀ',
    MjB: 'ɳa˞uˀ',
    DjB: 'ma˞uˀ',
    djB: 'qa˞uˀ',
    SjB: 'tʂɻa˞uˀ',
    hjB: 'ʔa˞uˀ',
    CjB: 'ba˞uˀ',
    gjB: 'ɴa˞uˀ',
    UjB: 'dʐɻa˞uˀ',
    TjB: 'tʂʰɻa˞uˀ',
    JjB: 'ʈa˞uˀ',
    VjB: 'ʂɻa˞uˀ',
    jkB: 'ʁɑuˀ',
    CkB: 'bɑuˀ',
    IkB: 'lɑuˀ',
    FkB: 'tʰɑuˀ',
    GkB: 'dɑuˀ',
    HkB: 'nɑuˀ',
    QkB: 'sɑuˀ',
    EkB: 'tɑuˀ',
    OkB: 'tsʰɑuˀ',
    NkB: 'tsɑuˀ',
    PkB: 'dzɑuˀ',
    dkB: 'qɑuˀ',
    ikB: 'χɑuˀ',
    DkB: 'mɑuˀ',
    AkB: 'pɑuˀ',
    hkB: 'ʔɑuˀ',
    ekB: 'qʰɑuˀ',
    gkB: 'ɴɑuˀ',
    dlB: 'qɑˀ',
    OlB: 'tsʰɑˀ',
    ElB: 'tɑˀ',
    QlB: 'sɑˀ',
    GlB: 'dɑˀ',
    glB: 'ɴɑˀ',
    FlB: 'tʰɑˀ',
    IlB: 'lɑˀ',
    HlB: 'nɑˀ',
    jlB: 'ʁɑˀ',
    ilB: 'χɑˀ',
    elB: 'qʰɑˀ',
    hlB: 'ʔɑˀ',
    NlB: 'tsɑˀ',
    dlJ: 'qʷɑˀ',
    ElJ: 'tʷɑˀ',
    QlJ: 'sʷɑˀ',
    GlJ: 'dʷɑˀ',
    FlJ: 'tʰʷɑˀ',
    DlB: 'mɑˀ',
    PlJ: 'dzʷɑˀ',
    glJ: 'ɴʷɑˀ',
    IlJ: 'lʷɑˀ',
    hlJ: 'ʔʷɑˀ',
    HlJ: 'nʷɑˀ',
    AlB: 'pɑˀ',
    BlB: 'pʰɑˀ',
    jlJ: 'ʁʷɑˀ',
    ilJ: 'χʷɑˀ',
    elJ: 'qʰʷɑˀ',
    ClB: 'bɑˀ',
    OlJ: 'tsʰʷɑˀ',
    NlJ: 'tsʷɑˀ',
    DnB: 'ma˞ˀ',
    XoB: 'tçiaˀ',
    loB: 'jiaˀ',
    gnB: 'ɴa˞ˀ',
    dnB: 'qa˞ˀ',
    VnB: 'ʂɻa˞ˀ',
    hnB: 'ʔa˞ˀ',
    RoB: 'ziaˀ',
    jnB: 'ʁa˞ˀ',
    QoB: 'siaˀ',
    NmJ: 'tsuɑˀ',
    DoB: 'miaˀ',
    OoB: 'tsʰiaˀ',
    inB: 'χa˞ˀ',
    ZoB: 'dʝiaˀ',
    enB: 'qʰa˞ˀ',
    CnB: 'ba˞ˀ',
    aoB: 'çiaˀ',
    NoB: 'tsiaˀ',
    AnB: 'pa˞ˀ',
    jnJ: 'ʁʷa˞ˀ',
    dnJ: 'qʷa˞ˀ',
    gnJ: 'ɴʷa˞ˀ',
    coB: 'ɲiaˀ',
    SnB: 'tʂɻa˞ˀ',
    JnB: 'ʈa˞ˀ',
    UnB: 'dʐɻa˞ˀ',
    YoB: 'tçʰiaˀ',
    JnB: 'ʈa˞ˀ',
    MnB: 'ɳa˞ˀ',
    enJ: 'qʰʷa˞ˀ',
    TnJ: 'tʂʰɻʷa˞ˀ',
    KnJ: 'ʈʰʷa˞ˀ',
    VnJ: 'ʂɻʷa˞ˀ',
    KnB: 'ʈʰa˞ˀ',
    InB: 'ɭa˞ˀ',
    lpB: 'jɯɑɴˀ',
    RpB: 'zɯɑɴˀ',
    NpB: 'tsɯɑɴˀ',
    IpB: 'ɭɯɑɴˀ',
    hpB: 'ʔɯɑɴˀ',
    fpB: 'ɡɯɑɴˀ',
    gpB: 'ŋɯɑɴˀ',
    TpB: 'tʂʰɻɑɴˀ',
    QpB: 'sɯɑɴˀ',
    XpB: 'tçɯɑɴˀ',
    VpB: 'ʂɻɑɴˀ',
    ipB: 'xɯɑɴˀ',
    YpB: 'tçʰɯɑɴˀ',
    dpB: 'kɯɑɴˀ',
    LpB: 'ɖɯɑɴˀ',
    KpB: 'ʈʰɯɑɴˀ',
    dpJ: 'kuɑɴˀ',
    cpB: 'ɲɯɑɴˀ',
    apB: 'çɯɑɴˀ',
    BpB: 'pʰuɑɴˀ',
    DpB: 'muɑɴˀ',
    ApB: 'puɑɴˀ',
    hpJ: 'ʔuɑɴˀ',
    kpJ: 'wuɑɴˀ',
    ipJ: 'xuɑɴˀ',
    OpB: 'tsʰɯɑɴˀ',
    JpB: 'ʈɯɑɴˀ',
    ZpB: 'dʝɯɑɴˀ',
    fpJ: 'ɡuɑɴˀ',
    dpJ: 'kuɑɴˀ',
    CpB: 'buɑɴˀ',
    TpB: 'tʂʰɻɑɴˀ',
    GqB: 'dɑɴˀ',
    QqB: 'sɑɴˀ',
    dqJ: 'qʷɑɴˀ',
    AqB: 'pɑɴˀ',
    NqB: 'tsɑɴˀ',
    HqB: 'nɑɴˀ',
    jqB: 'ʁɑɴˀ',
    FqB: 'tʰɑɴˀ',
    DqB: 'mɑɴˀ',
    EqB: 'tɑɴˀ',
    IqB: 'lɑɴˀ',
    hqB: 'ʔɑɴˀ',
    eqB: 'qʰɑɴˀ',
    hqJ: 'ʔʷɑɴˀ',
    jqJ: 'ʁʷɑɴˀ',
    BqB: 'pʰɑɴˀ',
    iqJ: 'χʷɑɴˀ',
    dqB: 'qɑɴˀ',
    gqB: 'ɴɑɴˀ',
    PqB: 'dzɑɴˀ',
    OqB: 'tsʰɑɴˀ',
    iqB: 'χɑɴˀ',
    eqJ: 'qʰʷɑɴˀ',
    drB: 'qa˞ŋˀ',
    AsB: 'pi˞aŋˀ',
    dsB: 'ki˞aŋˀ',
    hsB: 'ʔi˞aŋˀ',
    VsB: 'ʂɻaŋˀ',
    ksJ: 'wi˞aŋˀ',
    isJ: 'xʷi˞aŋˀ',
    DsB: 'mi˞aŋˀ',
    dsJ: 'kʷi˞aŋˀ',
    jrB: 'ʁa˞ŋˀ',
    DrB: 'ma˞ŋˀ',
    drJ: 'qʷa˞ŋˀ',
    ArB: 'pa˞ŋˀ',
    JrB: 'ʈa˞ŋˀ',
    LrB: 'ɖa˞ŋˀ',
    hrJ: 'ʔʷa˞ŋˀ',
    ErB: 'taŋˀ',
    IrB: 'laŋˀ',
    jrJ: 'ʁʷa˞ŋˀ',
    CrB: 'ba˞ŋˀ',
    erJ: 'qʰʷa˞ŋˀ',
    MrB: 'ɳa˞ŋˀ',
    dtB: 'qɜ˞ŋˀ',
    DtB: 'mɜ˞ŋˀ',
    jtB: 'ʁɜ˞ŋˀ',
    CtB: 'bɜ˞ŋˀ',
    BtB: 'pʰɜ˞ŋˀ',
    PuB: 'dziaŋˀ',
    XuB: 'tçiaŋˀ',
    KuB: 'ʈʰiaŋˀ',
    luB: 'jiaŋˀ',
    fuB: 'ɡiaŋˀ',
    luJ: 'ɥiaŋˀ',
    IuB: 'ɭiaŋˀ',
    duB: 'kiaŋˀ',
    AuB: 'piaŋˀ',
    euJ: 'kʰʷiaŋˀ',
    NuB: 'tsiaŋˀ',
    huB: 'ʔiaŋˀ',
    OuB: 'tsʰiaŋˀ',
    QuB: 'siaŋˀ',
    DuB: 'miaŋˀ',
    LuB: 'ɖiaŋˀ',
    jvJ: 'ʁʷɛŋˀ',
    dvJ: 'qʷɛŋˀ',
    DvB: 'mɛŋˀ',
    EvB: 'tɛŋˀ',
    GvB: 'dɛŋˀ',
    FvB: 'tʰɛŋˀ',
    PvB: 'dzɛŋˀ',
    hvJ: 'ʔʷɛŋˀ',
    evB: 'qʰɛŋˀ',
    HvB: 'nɛŋˀ',
    jvB: 'ʁɛŋˀ',
    QvB: 'sɛŋˀ',
    AvB: 'pɛŋˀ',
    evJ: 'qʰʷɛŋˀ',
    dvB: 'qɛŋˀ',
    BvB: 'pʰɛŋˀ',
    hvB: 'ʔɛŋˀ',
    IvB: 'lɛŋˀ',
    CvB: 'bɛŋˀ',
    ivJ: 'χʷɛŋˀ',
    gvB: 'ɴɛŋˀ',
    XwB: 'tçiɯŋˀ',
    KwB: 'ʈʰiɯŋˀ',
    fwB: 'ɡɯŋˀ',
    VwB: 'ʂɻɯŋˀ',
    ExB: 'tʌŋˀ',
    BxB: 'pʰʌŋˀ',
    exB: 'qʰʌŋˀ',
    HxB: 'nʌŋˀ',
    kyB: 'wuˀ',
    IyB: 'ɭuˀ',
    MyB: 'ɳuˀ',
    KyB: 'ʈʰuˀ',
    JyB: 'ʈuˀ',
    iyB: 'xuˀ',
    dyB: 'kuˀ',
    ayB: 'çuˀ',
    YyB: 'tçʰuˀ',
    PyB: 'dzuˀ',
    CyB: 'buˀ',
    AyB: 'puˀ',
    eyB: 'kʰuˀ',
    cyB: 'ɲuˀ',
    fyB: 'ɡuˀ',
    LyB: 'ɖuˀ',
    lyB: 'juˀ',
    ZyB: 'dʝuˀ',
    QyB: 'suˀ',
    hyB: 'ʔuˀ',
    NyB: 'tsuˀ',
    VyB: 'ʂɻuˀ',
    XyB: 'tçuˀ',
    ByB: 'pʰuˀ',
    SyB: 'tʂɻuˀ',
    UyB: 'dʐɻuˀ',
    TyB: 'tʂʰɻuˀ',
    ByB: 'pʰuˀ',
    jzB: 'ʁᵒuˀ',
    DzB: 'mᵒuˀ',
    CzB: 'bᵒuˀ',
    EzB: 'tᵒuˀ',
    FzB: 'tʰᵒuˀ',
    dzB: 'qᵒuˀ',
    gzB: 'ɴᵒuˀ',
    AzB: 'pᵒuˀ',
    HzB: 'nᵒuˀ',
    QzB: 'sᵒuˀ',
    izB: 'χᵒuˀ',
    BzB: 'pʰᵒuˀ',
    hzB: 'ʔᵒuˀ',
    IzB: 'lᵒuˀ',
    NzB: 'tsᵒuˀ',
    ezB: 'qʰᵒuˀ',
    GzB: 'dᵒuˀ',
    UzB: 'dʐɻᵒuˀ',
    OzB: 'tsʰᵒuˀ',
    h0B: 'ʔiuˀ',
    d0B: 'kiuˀ',
    f0B: 'ɡiuˀ',
    O1B: 'tsʰimˀ',
    L1B: 'ɖimˀ',
    e1B: 'kʰimˀ',
    I1B: 'ɭimˀ',
    Q1B: 'simˀ',
    K1B: 'ʈʰimˀ',
    N1B: 'tsimˀ',
    c1B: 'ɲimˀ',
    X1B: 'tçimˀ',
    a1B: 'çimˀ',
    b1B: 'ʝimˀ',
    Z1B: 'dʝimˀ',
    Y1B: 'tçʰimˀ',
    T1B: 'tʂʰɻmˀ',
    M1B: 'ɳimˀ',
    f1F: 'ɡɯmˀ',
    d1F: 'kɯmˀ',
    P1B: 'dzimˀ',
    g1F: 'ŋɯmˀ',
    V1B: 'ʂɻmˀ',
    A1F: 'pi˞mˀ',
    h1F: 'ʔɯmˀ',
    B1F: 'pʰi˞mˀ',
    e1F: 'kʰɯmˀ',
    i1F: 'xɯmˀ',
    J1B: 'ʈimˀ',
    l1B: 'jimˀ',
    U1B: 'dʐɻmˀ',
    d2B: 'qɔmˀ',
    G2B: 'dɔmˀ',
    h2B: 'ʔɔmˀ',
    H2B: 'nɔmˀ',
    F2B: 'tʰɔmˀ',
    P2B: 'dzɔmˀ',
    O2B: 'tsʰɔmˀ',
    g2B: 'ɴɔmˀ',
    N2B: 'tsɔmˀ',
    Q2B: 'sɔmˀ',
    e2B: 'qʰɔmˀ',
    j2B: 'ʁɔmˀ',
    I2B: 'lɔmˀ',
    E2B: 'tɔmˀ',
    i2B: 'χɔmˀ',
    d3B: 'qɑmˀ',
    a3B: 'çɑmˀ',
    I3B: 'lɑmˀ',
    F3B: 'tʰɑmˀ',
    E3B: 'tɑmˀ',
    G3B: 'dɑmˀ',
    O3B: 'tsʰɑmˀ',
    D3B: 'mɑmˀ',
    N3B: 'tsɑmˀ',
    P3B: 'dzɑmˀ',
    i3B: 'χɑmˀ',
    h3B: 'ʔɑmˀ',
    e3B: 'qʰɑmˀ',
    l4B: 'jiɛmˀ',
    I4B: 'ɭiɛmˀ',
    i4F: 'xi˞ɛmˀ',
    A4F: 'pi˞ɛmˀ',
    X4B: 'tçiɛmˀ',
    e4F: 'kʰi˞ɛmˀ',
    g4F: 'ŋi˞ɛmˀ',
    f4F: 'ɡi˞ɛmˀ',
    d4F: 'ki˞ɛmˀ',
    h4B: 'ʔiɛmˀ',
    c4B: 'ɲiɛmˀ',
    a4B: 'çiɛmˀ',
    K4B: 'ʈʰiɛmˀ',
    h4F: 'ʔi˞ɛmˀ',
    P4B: 'dziɛmˀ',
    O4B: 'tsʰiɛmˀ',
    N4B: 'tsiɛmˀ',
    e4B: 'kʰiɛmˀ',
    Z4B: 'dʝiɛmˀ',
    F5B: 'tʰɛmˀ',
    H5B: 'nɛmˀ',
    E5B: 'tɛmˀ',
    G5B: 'dɛmˀ',
    e5B: 'qʰɛmˀ',
    I5B: 'lɛmˀ',
    j5B: 'ʁɛmˀ',
    O5B: 'tsʰɛmˀ',
    d5B: 'qɛmˀ',
    D5B: 'mɛmˀ',
    g8B: 'ŋɯʌmˀ',
    e8B: 'kʰɯʌmˀ',
    h8B: 'ʔɯʌmˀ',
    j6B: 'ʁɜ˞mˀ',
    L6B: 'ɖɜ˞mˀ',
    e6B: 'qʰɜ˞mˀ',
    d6B: 'qɜ˞mˀ',
    U6B: 'dʐɻɜ˞mˀ',
    I6B: 'ɭɜ˞mˀ',
    S6B: 'tʂɻɜ˞mˀ',
    T6B: 'tʂʰɻɜ˞mˀ',
    i6B: 'χɜ˞mˀ',
    V6B: 'ʂɻɜ˞mˀ',
    h6B: 'ʔɜ˞mˀ',
    i6B: 'χɜ˞mˀ',
    M6B: 'ɳɜ˞mˀ',
    K6B: 'ʈʰɜ˞mˀ',
    j7B: 'ʁa˞mˀ',
    e7B: 'qʰa˞mˀ',
    V7B: 'ʂɻa˞mˀ',
    h7B: 'ʔa˞mˀ',
    T7B: 'tʂʰɻa˞mˀ',
    i7B: 'χa˞mˀ',
    U7B: 'dʐɻa˞mˀ',
    C9B: 'buʌmˀ',
    D9B: 'muʌmˀ',
    A9B: 'puʌmˀ',
    e9J: 'kʰuʌmˀ',
    B9B: 'pʰuʌmˀ',
    K9J: 'ʈʰuʌmˀ',
    QAC: 'sᵒuɴʰ',
    CBC: 'buɴʰ',
    dAC: 'qᵒuɴʰ',
    IAC: 'lᵒuɴʰ',
    EAC: 'tᵒuɴʰ',
    eAC: 'qʰᵒuɴʰ',
    NAC: 'tsᵒuɴʰ',
    hAC: 'ʔᵒuɴʰ',
    OAC: 'tsʰᵒuɴʰ',
    GAC: 'dᵒuɴʰ',
    FAC: 'tʰᵒuɴʰ',
    LBC: 'ɖuɴʰ',
    ABC: 'puɴʰ',
    eBC: 'kʰuɴʰ',
    DBC: 'muɴʰ',
    iBC: 'xuɴʰ',
    DAC: 'mᵒuɴʰ',
    OBC: 'tsʰuɴʰ',
    BBC: 'pʰuɴʰ',
    PAC: 'dzᵒuɴʰ',
    JBC: 'ʈuɴʰ',
    jAC: 'ʁᵒuɴʰ',
    HAC: 'nᵒuɴʰ',
    XBC: 'tçuɴʰ',
    YBC: 'tçʰuɴʰ',
    UBC: 'dʐɻuɴʰ',
    iAC: 'χᵒuɴʰ',
    QCC: 'sɔɴʰ',
    NCC: 'tsɔɴʰ',
    FCC: 'tʰɔɴʰ',
    DCC: 'mɔɴʰ',
    jCC: 'ʁɔɴʰ',
    lDC: 'ɥuoɴʰ',
    RDC: 'zuoɴʰ',
    CDC: 'buoɴʰ',
    fDC: 'ɡuoɴʰ',
    ADC: 'puoɴʰ',
    dDC: 'kuoɴʰ',
    hDC: 'ʔuoɴʰ',
    JDC: 'ʈuoɴʰ',
    NDC: 'tsuoɴʰ',
    KDC: 'ʈʰuoɴʰ',
    cDC: 'ɲuoɴʰ',
    XDC: 'tçuoɴʰ',
    LDC: 'ɖuoɴʰ',
    IDC: 'ɭuoɴʰ',
    eDC: 'kʰuoɴʰ',
    PDC: 'dzuoɴʰ',
    MDC: 'ɳuoɴʰ',
    dEC: 'qɞ˞ɴʰ',
    jEC: 'ʁɞ˞ɴʰ',
    JEC: 'ʈɞ˞ɴʰ',
    LEC: 'ɖɞ˞ɴʰ',
    KEC: 'ʈʰɞ˞ɴʰ',
    UEC: 'dʐɻɞ˞ɴʰ',
    BEC: 'pʰɞ˞ɴʰ',
    TEC: 'tʂʰɻɞ˞ɴʰ',
    VEC: 'ʂɻɞ˞ɴʰ',
    XFC: 'tçieʰ',
    CFC: 'bieʰ',
    XFK: 'tçʷieʰ',
    IFC: 'ɭieʰ',
    ZFC: 'dʝieʰ',
    NFC: 'tsieʰ',
    QFC: 'sieʰ',
    kFK: 'wueʰ',
    dFO: 'kueʰ',
    BFG: 'pʰɯeʰ',
    AFG: 'pɯeʰ',
    CFG: 'bɯeʰ',
    IFK: 'ɭʷieʰ',
    dFG: 'kɯeʰ',
    AFC: 'pieʰ',
    fFG: 'ɡɯeʰ',
    OFC: 'tsʰieʰ',
    lFC: 'jieʰ',
    gFG: 'ŋɯeʰ',
    BFC: 'pʰieʰ',
    PFC: 'dzieʰ',
    JFC: 'ʈieʰ',
    hFG: 'ʔɯeʰ',
    LFK: 'ɖʷieʰ',
    YFK: 'tçʰʷieʰ',
    iFG: 'xɯeʰ',
    eFC: 'kʰieʰ',
    hFC: 'ʔieʰ',
    aFC: 'çieʰ',
    VFC: 'ʂɻeʰ',
    eFK: 'kʰʷieʰ',
    hFO: 'ʔueʰ',
    gFO: 'ŋueʰ',
    iFO: 'xueʰ',
    hFK: 'ʔʷieʰ',
    ZFK: 'dʝʷieʰ',
    dFC: 'kieʰ',
    cFK: 'ɲʷieʰ',
    YFC: 'tçʰieʰ',
    QFK: 'sʷieʰ',
    JFK: 'ʈʷieʰ',
    lFK: 'ɥieʰ',
    MFK: 'ɳʷieʰ',
    dFK: 'kʷieʰ',
    SFC: 'tʂɻeʰ',
    iFK: 'xʷieʰ',
    eFG: 'kʰɯeʰ',
    XGC: 'tçiʰ',
    kGK: 'wi˞iʰ',
    DGG: 'mi˞iʰ',
    RGK: 'zʷiʰ',
    NGK: 'tsʷiʰ',
    QGK: 'sʷiʰ',
    IGK: 'ɭʷiʰ',
    AGG: 'pi˞iʰ',
    fGO: 'ɡʷi˞iʰ',
    BGG: 'pʰi˞iʰ',
    CGG: 'bi˞iʰ',
    dGO: 'kʷi˞iʰ',
    VGK: 'ʂɻʷiʰ',
    eGO: 'kʰʷi˞iʰ',
    iGO: 'xʷi˞iʰ',
    ZGC: 'dʝiʰ',
    IGC: 'ɭiʰ',
    MGC: 'ɳiʰ',
    BGC: 'pʰiʰ',
    gGG: 'ŋi˞iʰ',
    JGC: 'ʈiʰ',
    eGC: 'kʰiʰ',
    LGC: 'ɖiʰ',
    DGC: 'miʰ',
    KGC: 'ʈʰiʰ',
    dGG: 'ki˞iʰ',
    fGG: 'ɡi˞iʰ',
    fGK: 'ɡʷiʰ',
    OGK: 'tsʰʷiʰ',
    cGC: 'ɲiʰ',
    NGC: 'tsiʰ',
    OGC: 'tsʰiʰ',
    hGG: 'ʔi˞iʰ',
    QGC: 'siʰ',
    eGG: 'kʰi˞iʰ',
    dGK: 'kʷiʰ',
    CGC: 'biʰ',
    iGK: 'xʷiʰ',
    AGC: 'piʰ',
    PGK: 'dzʷiʰ',
    GGC: 'diʰ',
    iGG: 'xi˞iʰ',
    lGC: 'jiʰ',
    bGC: 'ʝiʰ',
    PGC: 'dziʰ',
    LGK: 'ɖʷiʰ',
    YGK: 'tçʰʷiʰ',
    lGK: 'ɥiʰ',
    iGK: 'xʷiʰ',
    JGK: 'ʈʷiʰ',
    YGC: 'tçʰiʰ',
    aGC: 'çiʰ',
    TGK: 'tʂʰɻʷiʰ',
    aGK: 'çʷiʰ',
    XHC: 'tçiɯʰ',
    LHC: 'ɖiɯʰ',
    RHC: 'ziɯʰ',
    OHC: 'tsʰiɯʰ',
    QHC: 'siɯʰ',
    aHC: 'çiɯʰ',
    SHC: 'tʂɻɯʰ',
    IHC: 'ɭiɯʰ',
    PHC: 'dziɯʰ',
    KHC: 'ʈʰiɯʰ',
    cHC: 'ɲiɯʰ',
    VHC: 'ʂɻɯʰ',
    THC: 'tʂʰɻɯʰ',
    lHC: 'jiɯʰ',
    JHC: 'ʈiɯʰ',
    ZHC: 'dʝiɯʰ',
    UHC: 'dʐɻɯʰ',
    fHC: 'ɡɯʰ',
    YHC: 'tçʰiɯʰ',
    hHC: 'ʔɯʰ',
    dHC: 'kɯʰ',
    iHC: 'xɯʰ',
    gHC: 'ŋɯʰ',
    eHC: 'kʰɯʰ',
    DIC: 'muiʰ',
    dIK: 'kuiʰ',
    kIK: 'wuiʰ',
    gIK: 'ŋuiʰ',
    AIC: 'puiʰ',
    BIC: 'pʰuiʰ',
    eIK: 'kʰuiʰ',
    hIK: 'ʔuiʰ',
    iIK: 'xuiʰ',
    CIC: 'buiʰ',
    dIC: 'kɯiʰ',
    gIC: 'ŋɯiʰ',
    eIC: 'kʰɯiʰ',
    iIC: 'xɯiʰ',
    fIC: 'ɡɯiʰ',
    hIC: 'ʔɯiʰ',
    gJC: 'ŋɯɤʰ',
    IJC: 'ɭɯɤʰ',
    dJC: 'kɯɤʰ',
    OJC: 'tsʰɯɤʰ',
    eJC: 'kʰɯɤʰ',
    ZJC: 'dʝɯɤʰ',
    aJC: 'çɯɤʰ',
    JJC: 'ʈɯɤʰ',
    XJC: 'tçɯɤʰ',
    VJC: 'ʂɻɯɤʰ',
    hJC: 'ʔɯɤʰ',
    LJC: 'ɖɯɤʰ',
    fJC: 'ɡɯɤʰ',
    QJC: 'sɯɤʰ',
    UJC: 'dʐɻɯɤʰ',
    NJC: 'tsɯɤʰ',
    SJC: 'tʂɻɯɤʰ',
    cJC: 'ɲɯɤʰ',
    lJC: 'jɯɤʰ',
    iJC: 'xɯɤʰ',
    MJC: 'ɳɯɤʰ',
    TJC: 'tʂʰɻɯɤʰ',
    YJC: 'tçʰɯɤʰ',
    KJC: 'ʈʰɯɤʰ',
    RJC: 'zɯɤʰ',
    gKC: 'ŋuoʰ',
    hKC: 'ʔuoʰ',
    ZKC: 'dʝuoʰ',
    LKC: 'ɖuoʰ',
    CKC: 'buoʰ',
    XKC: 'tçuoʰ',
    dKC: 'kuoʰ',
    iKC: 'xuoʰ',
    aKC: 'çuoʰ',
    lKC: 'ɥuoʰ',
    cKC: 'ɲuoʰ',
    BKC: 'pʰuoʰ',
    DKC: 'muoʰ',
    NKC: 'tsuoʰ',
    fKC: 'ɡuoʰ',
    kKC: 'wuoʰ',
    PKC: 'dzuoʰ',
    VKC: 'ʂɻuoʰ',
    AKC: 'puoʰ',
    OKC: 'tsʰuoʰ',
    JKC: 'ʈuoʰ',
    eKC: 'kʰuoʰ',
    TKC: 'tʂʰɻuoʰ',
    KKC: 'ʈʰuoʰ',
    QKC: 'suoʰ',
    IKC: 'ɭuoʰ',
    DLC: 'moʰ',
    GLC: 'doʰ',
    ILC: 'loʰ',
    ELC: 'toʰ',
    FLC: 'tʰoʰ',
    dLC: 'qoʰ',
    gLC: 'ɴoʰ',
    jLC: 'ʁoʰ',
    QLC: 'soʰ',
    PLC: 'dzoʰ',
    HLC: 'noʰ',
    ALC: 'poʰ',
    hLC: 'ʔoʰ',
    BLC: 'pʰoʰ',
    OLC: 'tsʰoʰ',
    eLC: 'qʰoʰ',
    CLC: 'boʰ',
    iLC: 'χoʰ',
    NLC: 'tsoʰ',
    NMC: 'tsɛiʰ',
    EMC: 'tɛiʰ',
    PMC: 'dzɛiʰ',
    FMC: 'tʰɛiʰ',
    GMC: 'dɛiʰ',
    OMC: 'tsʰɛiʰ',
    QMC: 'sɛiʰ',
    gMC: 'ɴɛiʰ',
    dMC: 'qɛiʰ',
    jMC: 'ʁɛiʰ',
    eMC: 'qʰɛiʰ',
    hMC: 'ʔɛiʰ',
    DMC: 'mɛiʰ',
    AMC: 'pɛiʰ',
    jMK: 'ʁʷɛiʰ',
    dMK: 'qʷɛiʰ',
    iMK: 'χʷɛiʰ',
    BMC: 'pʰɛiʰ',
    CMC: 'bɛiʰ',
    IMC: 'lɛiʰ',
    KMC: 'ʈʰɛiʰ',
    HMC: 'nɛiʰ',
    iMC: 'χɛiʰ',
    NNC: 'tsiɛiʰ',
    QNK: 'sʷiɛiʰ',
    kNK: 'wi˞ɛiʰ',
    cNK: 'ɲʷiɛiʰ',
    XNK: 'tçʷiɛiʰ',
    VNK: 'ʂɻʷɛiʰ',
    ONK: 'tsʰʷiɛiʰ',
    lNK: 'ɥiɛiʰ',
    JNK: 'ʈʷiɛiʰ',
    aNK: 'çʷiɛiʰ',
    CNC: 'biɛiʰ',
    TNK: 'tʂʰɻʷɛiʰ',
    RNK: 'zʷiɛiʰ',
    ANC: 'piɛiʰ',
    dNO: 'kʷi˞ɛiʰ',
    DNC: 'miɛiʰ',
    VNC: 'ʂɻɛiʰ',
    YNC: 'tçʰiɛiʰ',
    XNC: 'tçiɛiʰ',
    ZNC: 'dʝiɛiʰ',
    lNC: 'jiɛiʰ',
    hNG: 'ʔi˞ɛiʰ',
    gNC: 'ŋiɛiʰ',
    LNC: 'ɖiɛiʰ',
    INC: 'ɭiɛiʰ',
    eNG: 'kʰi˞ɛiʰ',
    aNC: 'çiɛiʰ',
    dNG: 'ki˞ɛiʰ',
    JNC: 'ʈiɛiʰ',
    fNG: 'ɡi˞ɛiʰ',
    ZNK: 'dʝʷiɛiʰ',
    KNC: 'ʈʰiɛiʰ',
    LNK: 'ɖʷiɛiʰ',
    gNG: 'ŋi˞ɛiʰ',
    NNK: 'tsʷiɛiʰ',
    BNC: 'pʰiɛiʰ',
    eUK: 'kʰuʌiʰ',
    iUK: 'xuʌiʰ',
    FOC: 'tʰɑiʰ',
    dOC: 'qɑiʰ',
    gOC: 'ɴɑiʰ',
    hOC: 'ʔɑiʰ',
    HOC: 'nɑiʰ',
    GOC: 'dɑiʰ',
    jOC: 'ʁɑiʰ',
    EOC: 'tɑiʰ',
    AOC: 'pɑiʰ',
    BOC: 'pʰɑiʰ',
    jOK: 'ʁʷɑiʰ',
    GOK: 'dʷɑiʰ',
    dOK: 'qʷɑiʰ',
    NOK: 'tsʷɑiʰ',
    iOK: 'χʷɑiʰ',
    eOK: 'qʰʷɑiʰ',
    IOK: 'lʷɑiʰ',
    gOK: 'ɴʷɑiʰ',
    EOK: 'tʷɑiʰ',
    hOK: 'ʔʷɑiʰ',
    POK: 'dzʷɑiʰ',
    QOK: 'sʷɑiʰ',
    OOK: 'tsʰʷɑiʰ',
    COC: 'bɑiʰ',
    eOC: 'qʰɑiʰ',
    OOC: 'tsʰɑiʰ',
    IOC: 'lɑiʰ',
    iOC: 'χɑiʰ',
    FOK: 'tʰʷɑiʰ',
    DOC: 'mɑiʰ',
    OOK: 'tsʰʷɑiʰ',
    dPK: 'qʷɜ˞ʰ',
    dPC: 'qɜ˞ʰ',
    hPC: 'ʔɜ˞ʰ',
    jPC: 'ʁɜ˞ʰ',
    DPC: 'mɜ˞ʰ',
    jPK: 'ʁʷɜ˞ʰ',
    TPC: 'tʂʰɻɜ˞ʰ',
    gPC: 'ɴɜ˞ʰ',
    iPC: 'χɜ˞ʰ',
    APC: 'pɜ˞ʰ',
    CPC: 'bɜ˞ʰ',
    UPC: 'dʐɻɜ˞ʰ',
    BPC: 'pʰɜ˞ʰ',
    SPC: 'tʂɻɜ˞ʰ',
    ePC: 'qʰɜ˞ʰ',
    VPC: 'ʂɻɜ˞ʰ',
    iPK: 'χʷɜ˞ʰ',
    APC: 'pɜ˞ʰ',
    JPK: 'ʈʷɜ˞ʰ',
    dQK: 'qʷɜ˞iʰ',
    hQC: 'ʔɜ˞iʰ',
    SQC: 'tʂɻɜ˞iʰ',
    dQC: 'qɜ˞iʰ',
    MQC: 'ɳɜ˞iʰ',
    iQC: 'χɜ˞iʰ',
    jQC: 'ʁɜ˞iʰ',
    gQC: 'ɴɜ˞iʰ',
    eQK: 'qʰʷɜ˞iʰ',
    AQC: 'pɜ˞iʰ',
    BQC: 'pʰɜ˞iʰ',
    jQK: 'ʁʷɜ˞iʰ',
    gQK: 'ɴʷɜ˞iʰ',
    CQC: 'bɜ˞iʰ',
    DQC: 'mɜ˞iʰ',
    iQK: 'χʷɜ˞iʰ',
    VQC: 'ʂɻɜ˞iʰ',
    eQC: 'qʰɜ˞iʰ',
    KQK: 'ʈʰʷɜ˞iʰ',
    dRK: 'qʷa˞iʰ',
    eRK: 'qʰʷa˞iʰ',
    DRC: 'ma˞iʰ',
    jRK: 'ʁʷa˞iʰ',
    CRC: 'ba˞iʰ',
    hRK: 'ʔʷa˞iʰ',
    TRK: 'tʂʰɻʷa˞iʰ',
    dRC: 'qa˞iʰ',
    KRC: 'ʈʰa˞iʰ',
    hRC: 'ʔa˞iʰ',
    VRC: 'ʂɻa˞iʰ',
    LRC: 'ɖa˞iʰ',
    iRC: 'χa˞iʰ',
    iRK: 'χʷa˞iʰ',
    ARC: 'pa˞iʰ',
    ORK: 'tsʰʷa˞iʰ',
    URC: 'dʐɻa˞iʰ',
    jRC: 'ʁa˞iʰ',
    GSK: 'dʷʌiʰ',
    CSC: 'bʌiʰ',
    DSC: 'mʌiʰ',
    BSC: 'pʰʌiʰ',
    iSK: 'χʷʌiʰ',
    ESK: 'tʷʌiʰ',
    OSK: 'tsʰʷʌiʰ',
    NSK: 'tsʷʌiʰ',
    hSK: 'ʔʷʌiʰ',
    FSK: 'tʰʷʌiʰ',
    dSK: 'qʷʌiʰ',
    jSK: 'ʁʷʌiʰ',
    eSK: 'qʰʷʌiʰ',
    QSK: 'sʷʌiʰ',
    HSK: 'nʷʌiʰ',
    ISK: 'lʷʌiʰ',
    ASC: 'pʌiʰ',
    gSK: 'ɴʷʌiʰ',
    jSK: 'ʁʷʌiʰ',
    GTC: 'dʌiʰ',
    NTC: 'tsʌiʰ',
    DTC: 'mʌiʰ',
    QTC: 'sʌiʰ',
    FTC: 'tʰʌiʰ',
    dTC: 'qʌiʰ',
    eTC: 'qʰʌiʰ',
    gTC: 'ɴʌiʰ',
    hTC: 'ʔʌiʰ',
    jTC: 'ʁʌiʰ',
    HTC: 'nʌiʰ',
    ETC: 'tʌiʰ',
    ITC: 'lʌiʰ',
    OTC: 'tsʰʌiʰ',
    PTC: 'dzʌiʰ',
    iTC: 'χʌiʰ',
    AUC: 'puʌiʰ',
    BUC: 'pʰuʌiʰ',
    hUK: 'ʔuʌiʰ',
    CUC: 'buʌiʰ',
    iUK: 'xuʌiʰ',
    fUK: 'ɡuʌiʰ',
    gUC: 'ŋɯʌiʰ',
    XVC: 'tçinʰ',
    QVC: 'sinʰ',
    cVC: 'ɲinʰ',
    lVC: 'jinʰ',
    IVC: 'ɭinʰ',
    AVC: 'pinʰ',
    LVC: 'ɖinʰ',
    ZVC: 'dʝinʰ',
    aVC: 'çinʰ',
    eVG: 'kʰi˞nʰ',
    RVC: 'zinʰ',
    gVG: 'ŋi˞nʰ',
    NVC: 'tsinʰ',
    iVG: 'xi˞nʰ',
    JVC: 'ʈinʰ',
    fVG: 'ɡi˞nʰ',
    TWC: 'tʂʰɻnʰ',
    hVC: 'ʔinʰ',
    KVC: 'ʈʰinʰ',
    BVC: 'pʰinʰ',
    OVC: 'tsʰinʰ',
    eVC: 'kʰinʰ',
    dVK: 'kʷinʰ',
    XVK: 'tçʷinʰ',
    QVK: 'sʷinʰ',
    RVK: 'zʷinʰ',
    NVK: 'tsʷinʰ',
    aVK: 'çʷinʰ',
    cVK: 'ɲʷinʰ',
    bVK: 'ʝʷinʰ',
    DXC: 'munʰ',
    kXK: 'wunʰ',
    iXK: 'xunʰ',
    BXC: 'pʰunʰ',
    AXC: 'punʰ',
    hXK: 'ʔunʰ',
    dXK: 'kunʰ',
    fXK: 'ɡunʰ',
    CXC: 'bunʰ',
    iYC: 'xɯnʰ',
    dYC: 'kɯnʰ',
    fYC: 'ɡɯnʰ',
    hYC: 'ʔɯnʰ',
    gYC: 'ŋɯnʰ',
    gZK: 'ŋuʌnʰ',
    hZK: 'ʔuʌnʰ',
    AZC: 'puʌnʰ',
    eZK: 'kʰuʌnʰ',
    DZC: 'muʌnʰ',
    CZC: 'buʌnʰ',
    BZC: 'pʰuʌnʰ',
    dZC: 'kɯʌnʰ',
    hZC: 'ʔɯʌnʰ',
    iZC: 'xɯʌnʰ',
    iZK: 'xuʌnʰ',
    fZC: 'ɡɯʌnʰ',
    BZC: 'pʰuʌnʰ',
    kZK: 'wuʌnʰ',
    gZC: 'ŋɯʌnʰ',
    fZK: 'ɡuʌnʰ',
    dZK: 'kuʌnʰ',
    jaK: 'ʁʷʌnʰ',
    EaK: 'tʷʌnʰ',
    QaK: 'sʷʌnʰ',
    eaK: 'qʰʷʌnʰ',
    HaK: 'nʷʌnʰ',
    haK: 'ʔʷʌnʰ',
    DaC: 'mʌnʰ',
    PaK: 'dzʷʌnʰ',
    daK: 'qʷʌnʰ',
    BaC: 'pʰʌnʰ',
    GaK: 'dʷʌnʰ',
    OaK: 'tsʰʷʌnʰ',
    CaC: 'bʌnʰ',
    gaK: 'ɴʷʌnʰ',
    IaK: 'lʷʌnʰ',
    AaC: 'pʌnʰ',
    iaK: 'χʷʌnʰ',
    NaK: 'tsʷʌnʰ',
    jbC: 'ʁʌnʰ',
    dbC: 'qʌnʰ',
    gbC: 'ɴʌnʰ',
    hbC: 'ʔʌnʰ',
    jcC: 'ʁɑnʰ',
    FcC: 'tʰɑnʰ',
    hcC: 'ʔɑnʰ',
    EcC: 'tɑnʰ',
    GcC: 'dɑnʰ',
    dcC: 'qɑnʰ',
    gcC: 'ɴɑnʰ',
    ecC: 'qʰɑnʰ',
    icC: 'χɑnʰ',
    IcC: 'lɑnʰ',
    HcC: 'nɑnʰ',
    OcC: 'tsʰɑnʰ',
    QcC: 'sɑnʰ',
    NcC: 'tsɑnʰ',
    PcC: 'dzɑnʰ',
    jcK: 'ʁʷɑnʰ',
    NcK: 'tsʷɑnʰ',
    hcK: 'ʔʷɑnʰ',
    dcK: 'qʷɑnʰ',
    OcK: 'tsʰʷɑnʰ',
    gcK: 'ɴʷɑnʰ',
    GcK: 'dʷɑnʰ',
    IcK: 'lʷɑnʰ',
    EcK: 'tʷɑnʰ',
    FcK: 'tʰʷɑnʰ',
    icK: 'χʷɑnʰ',
    QcK: 'sʷɑnʰ',
    DcC: 'mɑnʰ',
    AcC: 'pɑnʰ',
    BcC: 'pʰɑnʰ',
    CcC: 'bɑnʰ',
    HcK: 'nʷɑnʰ',
    PcK: 'dzʷɑnʰ',
    ecK: 'qʰʷɑnʰ',
    ddC: 'qa˞nʰ',
    gdC: 'ɴa˞nʰ',
    hdC: 'ʔa˞nʰ',
    VdC: 'ʂɻa˞nʰ',
    jdC: 'ʁa˞nʰ',
    DdC: 'ma˞nʰ',
    hdK: 'ʔʷa˞nʰ',
    jdK: 'ʁʷa˞nʰ',
    ddK: 'qʷa˞nʰ',
    VdK: 'ʂɻʷa˞nʰ',
    TdK: 'tʂʰɻʷa˞nʰ',
    gdK: 'ɴʷa˞nʰ',
    UdC: 'dʐɻa˞nʰ',
    BdC: 'pʰa˞nʰ',
    MdK: 'ɳʷa˞nʰ',
    TdC: 'tʂʰɻa˞nʰ',
    KdC: 'ʈʰa˞nʰ',
    deC: 'qɜ˞nʰ',
    jeC: 'ʁɜ˞nʰ',
    CeC: 'bɜ˞nʰ',
    BeC: 'pʰɜ˞nʰ',
    jeK: 'ʁʷɜ˞nʰ',
    DeC: 'mɜ˞nʰ',
    LeC: 'ɖɜ˞nʰ',
    AeC: 'pɜ˞nʰ',
    deK: 'qʷɜ˞nʰ',
    QgC: 'sɛnʰ',
    OgC: 'tsʰɛnʰ',
    igK: 'χʷɛnʰ',
    jgK: 'ʁʷɛnʰ',
    dgK: 'qʷɛnʰ',
    GgC: 'dɛnʰ',
    FgC: 'tʰɛnʰ',
    IgC: 'lɛnʰ',
    dgC: 'qɛnʰ',
    HgC: 'nɛnʰ',
    egC: 'qʰɛnʰ',
    jgC: 'ʁɛnʰ',
    ggC: 'ɴɛnʰ',
    hgC: 'ʔɛnʰ',
    NgC: 'tsɛnʰ',
    DgC: 'mɛnʰ',
    BgC: 'pʰɛnʰ',
    PgC: 'dzɛnʰ',
    hgK: 'ʔʷɛnʰ',
    EgC: 'tɛnʰ',
    igC: 'χɛnʰ',
    QfC: 'siɛnʰ',
    XfC: 'tçiɛnʰ',
    ZfC: 'dʝiɛnʰ',
    gfG: 'ŋi˞ɛnʰ',
    efC: 'kʰiɛnʰ',
    dfK: 'kʷiɛnʰ',
    kfK: 'wi˞ɛnʰ',
    DfC: 'miɛnʰ',
    YfK: 'tçʰʷiɛnʰ',
    lfK: 'ɥiɛnʰ',
    JfC: 'ʈiɛnʰ',
    cfK: 'ɲʷiɛnʰ',
    NfC: 'tsiɛnʰ',
    YfC: 'tçʰiɛnʰ',
    afC: 'çiɛnʰ',
    hfC: 'ʔiɛnʰ',
    dfO: 'kʷi˞ɛnʰ',
    ffO: 'ɡʷi˞ɛnʰ',
    IfK: 'ɭʷiɛnʰ',
    KfK: 'ʈʰʷiɛnʰ',
    AfG: 'pi˞ɛnʰ',
    VfK: 'ʂɻʷɛnʰ',
    OfK: 'tsʰʷiɛnʰ',
    CfG: 'bi˞ɛnʰ',
    RfK: 'zʷiɛnʰ',
    QfK: 'sʷiɛnʰ',
    UfK: 'dʐɻʷɛnʰ',
    SfK: 'tʂɻʷɛnʰ',
    LfK: 'ɖʷiɛnʰ',
    PfC: 'dziɛnʰ',
    RfC: 'ziɛnʰ',
    BfC: 'pʰiɛnʰ',
    ZfK: 'dʝʷiɛnʰ',
    MfC: 'ɳiɛnʰ',
    JfK: 'ʈʷiɛnʰ',
    lfC: 'jiɛnʰ',
    CfC: 'biɛnʰ',
    LfC: 'ɖiɛnʰ',
    IfC: 'ɭiɛnʰ',
    efO: 'kʰʷi˞ɛnʰ',
    XfK: 'tçʷiɛnʰ',
    AgC: 'pɛnʰ',
    QhC: 'sɛuʰ',
    FhC: 'tʰɛuʰ',
    EhC: 'tɛuʰ',
    dhC: 'qɛuʰ',
    HhC: 'nɛuʰ',
    GhC: 'dɛuʰ',
    ehC: 'qʰɛuʰ',
    IhC: 'lɛuʰ',
    ghC: 'ɴɛuʰ',
    hhC: 'ʔɛuʰ',
    ihC: 'χɛuʰ',
    QiC: 'siɛuʰ',
    XiC: 'tçiɛuʰ',
    liC: 'jiɛuʰ',
    hiC: 'ʔiɛuʰ',
    LiC: 'ɖiɛuʰ',
    ZiC: 'dʝiɛuʰ',
    fiG: 'ɡɯɛuʰ',
    BiC: 'pʰiɛuʰ',
    PiC: 'dziɛuʰ',
    DiC: 'miɛuʰ',
    OiC: 'tsʰiɛuʰ',
    IiC: 'ɭiɛuʰ',
    eiG: 'kʰɯɛuʰ',
    giG: 'ŋɯɛuʰ',
    NiC: 'tsiɛuʰ',
    DiG: 'mɯɛuʰ',
    CiC: 'biɛuʰ',
    aiC: 'çiɛuʰ',
    KiC: 'ʈʰiɛuʰ',
    AiG: 'pɯɛuʰ',
    fiC: 'ɡiɛuʰ',
    ciC: 'ɲiɛuʰ',
    jjC: 'ʁa˞uʰ',
    djC: 'qa˞uʰ',
    ijC: 'χa˞uʰ',
    JjC: 'ʈa˞uʰ',
    AjC: 'pa˞uʰ',
    ejC: 'qʰa˞uʰ',
    DjC: 'ma˞uʰ',
    BjC: 'pʰa˞uʰ',
    KjC: 'ʈʰa˞uʰ',
    VjC: 'ʂɻa˞uʰ',
    LjC: 'ɖa˞uʰ',
    MjC: 'ɳa˞uʰ',
    SjC: 'tʂɻa˞uʰ',
    CjC: 'ba˞uʰ',
    TjC: 'tʂʰɻa˞uʰ',
    hjC: 'ʔa˞uʰ',
    gjC: 'ɴa˞uʰ',
    UjC: 'dʐɻa˞uʰ',
    jkC: 'ʁɑuʰ',
    GkC: 'dɑuʰ',
    EkC: 'tɑuʰ',
    dkC: 'qɑuʰ',
    gkC: 'ɴɑuʰ',
    DkC: 'mɑuʰ',
    IkC: 'lɑuʰ',
    OkC: 'tsʰɑuʰ',
    CkC: 'bɑuʰ',
    AkC: 'pɑuʰ',
    PkC: 'dzɑuʰ',
    hkC: 'ʔɑuʰ',
    QkC: 'sɑuʰ',
    ekC: 'qʰɑuʰ',
    NkC: 'tsɑuʰ',
    ikC: 'χɑuʰ',
    HkC: 'nɑuʰ',
    dlC: 'qɑʰ',
    jlC: 'ʁɑʰ',
    NlC: 'tsɑʰ',
    ElC: 'tɑʰ',
    IlC: 'lɑʰ',
    elC: 'qʰɑʰ',
    glC: 'ɴɑʰ',
    GlC: 'dɑʰ',
    HlC: 'nɑʰ',
    QlC: 'sɑʰ',
    ilC: 'χɑʰ',
    FlC: 'tʰɑʰ',
    dlK: 'qʷɑʰ',
    jlK: 'ʁʷɑʰ',
    NlK: 'tsʷɑʰ',
    elK: 'qʰʷɑʰ',
    FlK: 'tʰʷɑʰ',
    AlC: 'pɑʰ',
    OlK: 'tsʰʷɑʰ',
    DlC: 'mɑʰ',
    HlK: 'nʷɑʰ',
    BlC: 'pʰɑʰ',
    PlK: 'dzʷɑʰ',
    glK: 'ɴʷɑʰ',
    OlK: 'tsʰʷɑʰ',
    ilK: 'χʷɑʰ',
    GlK: 'dʷɑʰ',
    ClC: 'bɑʰ',
    IlK: 'lʷɑʰ',
    ElK: 'tʷɑʰ',
    OlK: 'tsʰʷɑʰ',
    QlK: 'sʷɑʰ',
    hlC: 'ʔɑʰ',
    hlK: 'ʔʷɑʰ',
    DnC: 'ma˞ʰ',
    dnC: 'qa˞ʰ',
    hnC: 'ʔa˞ʰ',
    inC: 'χa˞ʰ',
    gnC: 'ɴa˞ʰ',
    KnC: 'ʈʰa˞ʰ',
    JnC: 'ʈa˞ʰ',
    SnC: 'tʂɻa˞ʰ',
    UnC: 'dʐɻa˞ʰ',
    RoC: 'ziaʰ',
    enC: 'qʰa˞ʰ',
    jnC: 'ʁa˞ʰ',
    PoC: 'dziaʰ',
    loC: 'jiaʰ',
    YoC: 'tçʰiaʰ',
    QoC: 'siaʰ',
    XoC: 'tçiaʰ',
    NoC: 'tsiaʰ',
    aoC: 'çiaʰ',
    boC: 'ʝiaʰ',
    AnC: 'pa˞ʰ',
    BnC: 'pʰa˞ʰ',
    jnK: 'ʁʷa˞ʰ',
    inK: 'χʷa˞ʰ',
    enK: 'qʰʷa˞ʰ',
    VnK: 'ʂɻʷa˞ʰ',
    CnC: 'ba˞ʰ',
    HnC: 'na˞ʰ',
    OoC: 'tsʰiaʰ',
    VnC: 'ʂɻa˞ʰ',
    dnK: 'qʷa˞ʰ',
    LnC: 'ɖa˞ʰ',
    gnK: 'ɴʷa˞ʰ',
    hnK: 'ʔʷa˞ʰ',
    lpC: 'jɯɑɴʰ',
    IpC: 'ɭɯɑɴʰ',
    UpC: 'dʐɻɑɴʰ',
    cpC: 'ɲɯɑɴʰ',
    apC: 'çɯɑɴʰ',
    JpC: 'ʈɯɑɴʰ',
    KpC: 'ʈʰɯɑɴʰ',
    ipC: 'xɯɑɴʰ',
    LpC: 'ɖɯɑɴʰ',
    MpC: 'ɳɯɑɴʰ',
    PpC: 'dzɯɑɴʰ',
    XpC: 'tçɯɑɴʰ',
    ZpC: 'dʝɯɑɴʰ',
    SpC: 'tʂɻɑɴʰ',
    hpC: 'ʔɯɑɴʰ',
    fpC: 'ɡɯɑɴʰ',
    YpC: 'tçʰɯɑɴʰ',
    TpC: 'tʂʰɻɑɴʰ',
    NpC: 'tsɯɑɴʰ',
    gpC: 'ŋɯɑɴʰ',
    BpC: 'pʰuɑɴʰ',
    DpC: 'muɑɴʰ',
    ipK: 'xuɑɴʰ',
    dpK: 'kuɑɴʰ',
    kpK: 'wuɑɴʰ',
    ApC: 'puɑɴʰ',
    QpC: 'sɯɑɴʰ',
    dpC: 'kɯɑɴʰ',
    epC: 'kʰɯɑɴʰ',
    OpC: 'tsʰɯɑɴʰ',
    fpK: 'ɡuɑɴʰ',
    CpC: 'buɑɴʰ',
    GqC: 'dɑɴʰ',
    IqC: 'lɑɴʰ',
    jqC: 'ʁɑɴʰ',
    hqC: 'ʔɑɴʰ',
    gqC: 'ɴɑɴʰ',
    NqC: 'tsɑɴʰ',
    CqC: 'bɑɴʰ',
    PqC: 'dzɑɴʰ',
    EqC: 'tɑɴʰ',
    eqC: 'qʰɑɴʰ',
    AqC: 'pɑɴʰ',
    FqC: 'tʰɑɴʰ',
    eqK: 'qʰʷɑɴʰ',
    HqC: 'nɑɴʰ',
    QqC: 'sɑɴʰ',
    jqK: 'ʁʷɑɴʰ',
    dqK: 'qʷɑɴʰ',
    dqC: 'qɑɴʰ',
    iqK: 'χʷɑɴʰ',
    DqC: 'mɑɴʰ',
    hqK: 'ʔʷɑɴʰ',
    hsC: 'ʔi˞aŋʰ',
    dsC: 'ki˞aŋʰ',
    fsC: 'ɡi˞aŋʰ',
    esC: 'kʰi˞aŋʰ',
    drC: 'qa˞ŋʰ',
    DsC: 'mi˞aŋʰ',
    CsC: 'bi˞aŋʰ',
    DrC: 'ma˞ŋʰ',
    jrK: 'ʁʷa˞ŋʰ',
    AsC: 'pi˞aŋʰ',
    ksK: 'wi˞aŋʰ',
    jrC: 'ʁa˞ŋʰ',
    hrC: 'ʔa˞ŋʰ',
    TrC: 'tʂʰɻa˞ŋʰ',
    JrC: 'ʈa˞ŋʰ',
    KrC: 'ʈʰa˞ŋʰ',
    ArC: 'pa˞ŋʰ',
    LrC: 'ɖa˞ŋʰ',
    VrC: 'ʂɻa˞ŋʰ',
    gsC: 'ŋi˞aŋʰ',
    CrC: 'ba˞ŋʰ',
    irC: 'χa˞ŋʰ',
    hrK: 'ʔʷa˞ŋʰ',
    StC: 'tʂɻɜ˞ŋʰ',
    AtC: 'pɜ˞ŋʰ',
    CtC: 'bɜ˞ŋʰ',
    htC: 'ʔɜ˞ŋʰ',
    gtC: 'ɴɜ˞ŋʰ',
    itK: 'χʷɜ˞ŋʰ',
    duC: 'kiaŋʰ',
    OuC: 'tsʰiaŋʰ',
    XuC: 'tçiaŋʰ',
    auC: 'çiaŋʰ',
    LuC: 'ɖiaŋʰ',
    KuC: 'ʈʰiaŋʰ',
    QuC: 'siaŋʰ',
    IuC: 'ɭiaŋʰ',
    BuC: 'pʰiaŋʰ',
    iuK: 'xʷiaŋʰ',
    AuC: 'piaŋʰ',
    CuC: 'biaŋʰ',
    PuC: 'dziaŋʰ',
    ZuC: 'dʝiaŋʰ',
    DuC: 'miaŋʰ',
    euC: 'kʰiaŋʰ',
    iuC: 'xiaŋʰ',
    NuC: 'tsiaŋʰ',
    dvC: 'qɛŋʰ',
    HvC: 'nɛŋʰ',
    QvC: 'sɛŋʰ',
    jvC: 'ʁɛŋʰ',
    GvC: 'dɛŋʰ',
    EvC: 'tɛŋʰ',
    evC: 'qʰɛŋʰ',
    FvC: 'tʰɛŋʰ',
    OvC: 'tsʰɛŋʰ',
    DvC: 'mɛŋʰ',
    hvK: 'ʔʷɛŋʰ',
    IvC: 'lɛŋʰ',
    XwC: 'tçiɯŋʰ',
    lwC: 'jiɯŋʰ',
    bwC: 'ʝiɯŋʰ',
    cwC: 'ɲiɯŋʰ',
    hwC: 'ʔɯŋʰ',
    NwC: 'tsiɯŋʰ',
    iwC: 'xɯŋʰ',
    awC: 'çiɯŋʰ',
    LwC: 'ɖiɯŋʰ',
    IwC: 'ɭiɯŋʰ',
    CwC: 'bi˞ɯŋʰ',
    YwC: 'tçʰiɯŋʰ',
    gwC: 'ŋɯŋʰ',
    ZwC: 'dʝiɯŋʰ',
    KwC: 'ʈʰiɯŋʰ',
    fwC: 'ɡɯŋʰ',
    ExC: 'tʌŋʰ',
    PxC: 'dzʌŋʰ',
    dxC: 'qʌŋʰ',
    OxC: 'tsʰʌŋʰ',
    GxC: 'dʌŋʰ',
    DxC: 'mʌŋʰ',
    AxC: 'pʌŋʰ',
    CxC: 'bʌŋʰ',
    IxC: 'lʌŋʰ',
    NxC: 'tsʌŋʰ',
    QxC: 'sʌŋʰ',
    FxC: 'tʰʌŋʰ',
    kyC: 'wuʰ',
    dyC: 'kuʰ',
    LyC: 'ɖuʰ',
    JyC: 'ʈuʰ',
    ayC: 'çuʰ',
    YyC: 'tçʰuʰ',
    RyC: 'zuʰ',
    iyC: 'xuʰ',
    XyC: 'tçuʰ',
    fyC: 'ɡuʰ',
    VyC: 'ʂɻuʰ',
    SyC: 'tʂɻuʰ',
    ByC: 'pʰuʰ',
    TyC: 'tʂʰɻuʰ',
    AyC: 'puʰ',
    KyC: 'ʈʰuʰ',
    IyC: 'ɭuʰ',
    QyC: 'suʰ',
    NyC: 'tsuʰ',
    UyC: 'dʐɻuʰ',
    PyC: 'dzuʰ',
    MyC: 'ɳuʰ',
    CyC: 'buʰ',
    lyC: 'juʰ',
    ZyC: 'dʝuʰ',
    cyC: 'ɲuʰ',
    DyC: 'muʰ',
    gyC: 'ŋuʰ',
    eyC: 'kʰuʰ',
    OyC: 'tsʰuʰ',
    jzC: 'ʁᵒuʰ',
    ezC: 'qʰᵒuʰ',
    DzC: 'mᵒuʰ',
    BzC: 'pʰᵒuʰ',
    GzC: 'dᵒuʰ',
    EzC: 'tᵒuʰ',
    HzC: 'nᵒuʰ',
    QzC: 'sᵒuʰ',
    NzC: 'tsᵒuʰ',
    FzC: 'tʰᵒuʰ',
    hzC: 'ʔᵒuʰ',
    dzC: 'qᵒuʰ',
    OzC: 'tsʰᵒuʰ',
    IzC: 'lᵒuʰ',
    izC: 'χᵒuʰ',
    CzC: 'bᵒuʰ',
    gzC: 'ɴᵒuʰ',
    PzC: 'dzᵒuʰ',
    h0C: 'ʔiuʰ',
    D0C: 'miuʰ',
    e0C: 'kʰiuʰ',
    f0C: 'ɡiuʰ',
    O1C: 'tsʰimʰ',
    N1C: 'tsimʰ',
    c1C: 'ɲimʰ',
    L1C: 'ɖimʰ',
    X1C: 'tçimʰ',
    f1G: 'ɡɯmʰ',
    d1G: 'kɯmʰ',
    M1C: 'ɳimʰ',
    h1G: 'ʔɯmʰ',
    V1C: 'ʂɻmʰ',
    K1C: 'ʈʰimʰ',
    S1C: 'tʂɻmʰ',
    T1C: 'tʂʰɻmʰ',
    g1G: 'ŋɯmʰ',
    J1C: 'ʈimʰ',
    I1C: 'ɭimʰ',
    Z1C: 'dʝimʰ',
    k1C: 'wi˞mʰ',
    a1C: 'çimʰ',
    e2C: 'qʰɔmʰ',
    d2C: 'qɔmʰ',
    j2C: 'ʁɔmʰ',
    h2C: 'ʔɔmʰ',
    F2C: 'tʰɔmʰ',
    Q2C: 'sɔmʰ',
    O2C: 'tsʰɔmʰ',
    G2C: 'dɔmʰ',
    g2C: 'ɴɔmʰ',
    E2C: 'tɔmʰ',
    H2C: 'nɔmʰ',
    I2C: 'lɔmʰ',
    i2C: 'χɔmʰ',
    N2C: 'tsɔmʰ',
    e3C: 'qʰɑmʰ',
    I3C: 'lɑmʰ',
    F3C: 'tʰɑmʰ',
    d3C: 'qɑmʰ',
    i3C: 'χɑmʰ',
    j3C: 'ʁɑmʰ',
    G3C: 'dɑmʰ',
    P3C: 'dzɑmʰ',
    E3C: 'tɑmʰ',
    Q3C: 'sɑmʰ',
    l4C: 'jiɛmʰ',
    Z4C: 'dʝiɛmʰ',
    c4C: 'ɲiɛmʰ',
    h4C: 'ʔiɛmʰ',
    A4G: 'pi˞ɛmʰ',
    g4G: 'ŋi˞ɛmʰ',
    a4C: 'çiɛmʰ',
    N4C: 'tsiɛmʰ',
    O4C: 'tsʰiɛmʰ',
    I4C: 'ɭiɛmʰ',
    K4C: 'ʈʰiɛmʰ',
    Y4C: 'tçʰiɛmʰ',
    h4G: 'ʔi˞ɛmʰ',
    P4C: 'dziɛmʰ',
    X4C: 'tçiɛmʰ',
    F5C: 'tʰɛmʰ',
    H5C: 'nɛmʰ',
    E5C: 'tɛmʰ',
    Q5C: 'sɛmʰ',
    G5C: 'dɛmʰ',
    d5C: 'qɛmʰ',
    h5C: 'ʔɛmʰ',
    N5C: 'tsɛmʰ',
    P5C: 'dzɛmʰ',
    d5C: 'qɛmʰ',
    e5C: 'qʰɛmʰ',
    I5C: 'lɛmʰ',
    g8C: 'ŋɯʌmʰ',
    i8C: 'xɯʌmʰ',
    e8C: 'kʰɯʌmʰ',
    D9C: 'muʌmʰ',
    j6C: 'ʁɜ˞mʰ',
    h6C: 'ʔɜ˞mʰ',
    S6C: 'tʂɻɜ˞mʰ',
    J6C: 'ʈɜ˞mʰ',
    e6C: 'qʰɜ˞mʰ',
    L6C: 'ɖɜ˞mʰ',
    U6C: 'dʐɻɜ˞mʰ',
    d6C: 'qɜ˞mʰ',
    M6C: 'ɳɜ˞mʰ',
    g6C: 'ɴɜ˞mʰ',
    d7C: 'qa˞mʰ',
    T7C: 'tʂʰɻa˞mʰ',
    S7C: 'tʂɻa˞mʰ',
    V7C: 'ʂɻa˞mʰ',
    i7C: 'χa˞mʰ',
    C7C: 'ba˞mʰ',
    j7C: 'ʁa˞mʰ',
    U7C: 'dʐɻa˞mʰ',
    h7C: 'ʔa˞mʰ',
    C9C: 'buʌmʰ',
    B9C: 'pʰuʌmʰ',
    d8C: 'kɯʌmʰ',
    e8C: 'kʰɯʌmʰ',
    h8C: 'ʔɯʌmʰ',
    hAD: 'ʔᵒuq',
    GAD: 'dᵒuq',
    dAD: 'qᵒuq',
    jAD: 'ʁᵒuq',
    eAD: 'qʰᵒuq',
    FAD: 'tʰᵒuq',
    EAD: 'tᵒuq',
    QAD: 'sᵒuq',
    IAD: 'lᵒuq',
    iAD: 'χᵒuq',
    PAD: 'dzᵒuq',
    OAD: 'tsʰᵒuq',
    NAD: 'tsᵒuq',
    CAD: 'bᵒuq',
    BAD: 'pʰᵒuq',
    AAD: 'pᵒuq',
    DAD: 'mᵒuq',
    ABD: 'puq',
    CBD: 'buq',
    VBD: 'ʂɻuq',
    IBD: 'ɭuq',
    LBD: 'ɖuq',
    dBD: 'kuq',
    eBD: 'kʰuq',
    ZBD: 'dʝuq',
    YBD: 'tçʰuq',
    lBD: 'ɥuq',
    fBD: 'ɡuq',
    OBD: 'tsʰuq',
    cBD: 'ɲuq',
    XBD: 'tçuq',
    aBD: 'çuq',
    iBD: 'xuq',
    JBD: 'ʈuq',
    NBD: 'tsuq',
    TBD: 'tʂʰɻuq',
    MBD: 'ɳuq',
    SBD: 'tʂɻuq',
    BBD: 'pʰuq',
    hBD: 'ʔuq',
    QBD: 'suq',
    DBD: 'muq',
    kBD: 'wuq',
    KBD: 'ʈʰuq',
    gBD: 'ŋuq',
    PBD: 'dzuq',
    hCD: 'ʔɔq',
    GCD: 'dɔq',
    ECD: 'tɔq',
    eCD: 'qʰɔq',
    jCD: 'ʁɔq',
    CCD: 'bɔq',
    QCD: 'sɔq',
    dCD: 'qɔq',
    DCD: 'mɔq',
    iCD: 'χɔq',
    HCD: 'nɔq',
    NCD: 'tsɔq',
    ACD: 'pɔq',
    ICD: 'lɔq',
    gCD: 'ɴɔq',
    XDD: 'tçuoq',
    gDD: 'ŋuoq',
    iDD: 'xuoq',
    dDD: 'kuoq',
    fDD: 'ɡuoq',
    ZDD: 'dʝuoq',
    YDD: 'tçʰuoq',
    cDD: 'ɲuoq',
    aDD: 'çuoq',
    lDD: 'ɥuoq',
    LDD: 'ɖuoq',
    IDD: 'ɭuoq',
    eDD: 'kʰuoq',
    JDD: 'ʈuoq',
    NDD: 'tsuoq',
    bDD: 'ʝuoq',
    CDD: 'buoq',
    ODD: 'tsʰuoq',
    RDD: 'zuoq',
    QDD: 'suoq',
    ADD: 'puoq',
    KDD: 'ʈʰuoq',
    dED: 'qɑ˞qʷ',
    gED: 'ɴɑ˞qʷ',
    UED: 'dʐɻɑ˞qʷ',
    SED: 'tʂɻɑ˞qʷ',
    VED: 'ʂɻɑ˞qʷ',
    JED: 'ʈɑ˞qʷ',
    AED: 'pɑ˞qʷ',
    DED: 'mɑ˞qʷ',
    CED: 'bɑ˞qʷ',
    BED: 'pʰɑ˞qʷ',
    eED: 'qʰɑ˞qʷ',
    LED: 'ɖɑ˞qʷ',
    hED: 'ʔɑ˞qʷ',
    MED: 'ɳɑ˞qʷ',
    KED: 'ʈʰɑ˞qʷ',
    IED: 'ɭɑ˞qʷ',
    jED: 'ʁɑ˞qʷ',
    iED: 'χɑ˞qʷ',
    TED: 'tʂʰɻɑ˞qʷ',
    XVD: 'tçit',
    cVD: 'ɲit',
    bVD: 'ʝit',
    LVD: 'ɖit',
    QVD: 'sit',
    hVD: 'ʔit',
    OVD: 'tsʰit',
    BVD: 'pʰit',
    dVD: 'kit',
    MVD: 'ɳit',
    lVD: 'jit',
    eVD: 'kʰit',
    iVD: 'xit',
    KVD: 'ʈʰit',
    IVD: 'ɭit',
    JVD: 'ʈit',
    PVD: 'dzit',
    TWD: 'tʂʰɻt',
    aVD: 'çit',
    NVD: 'tsit',
    DVD: 'mit',
    AVD: 'pit',
    fVH: 'ɡi˞t',
    CVD: 'bit',
    kVL: 'wi˞t',
    VVL: 'ʂɻʷt',
    YVD: 'tçʰit',
    UWD: 'dʐɻt',
    DVH: 'mi˞t',
    CVH: 'bi˞t',
    hVH: 'ʔi˞t',
    gVH: 'ŋi˞t',
    AVH: 'pi˞t',
    JVL: 'ʈʷit',
    iVH: 'xi˞t',
    dVH: 'ki˞t',
    JVD: 'ʈit',
    iVL: 'xʷit',
    bVL: 'ʝʷit',
    dVL: 'kʷit',
    PVL: 'dzʷit',
    lVL: 'ɥit',
    NVL: 'tsʷit',
    QVL: 'sʷit',
    IVL: 'ɭʷit',
    KVL: 'ʈʰʷit',
    JVL: 'ʈʷit',
    LVL: 'ɖʷit',
    YVL: 'tçʰʷit',
    OVL: 'tsʰʷit',
    SVL: 'tʂɻʷt',
    iVP: 'xʷi˞t',
    SWD: 'tʂɻt',
    VWD: 'ʂɻt',
    UWD: 'dʐɻt',
    DXD: 'mut',
    AXD: 'put',
    hXL: 'ʔut',
    dXL: 'kut',
    eXL: 'kʰut',
    fXL: 'ɡut',
    CXD: 'but',
    iXL: 'xut',
    kXL: 'wut',
    BXD: 'pʰut',
    gXL: 'ŋut',
    iYD: 'xɯt',
    dYD: 'kɯt',
    gYD: 'ŋɯt',
    fYD: 'ɡɯt',
    eYD: 'kʰɯt',
    gZL: 'ŋuʌt',
    CZD: 'buʌt',
    kZL: 'wuʌt',
    dZL: 'kuʌt',
    hZL: 'ʔuʌt',
    fZL: 'ɡuʌt',
    eZL: 'kʰuʌt',
    AZD: 'puʌt',
    DZD: 'muʌt',
    iZL: 'xuʌt',
    hZD: 'ʔɯʌt',
    iZD: 'xɯʌt',
    dZD: 'kɯʌt',
    fZD: 'ɡɯʌt',
    BZD: 'pʰuʌt',
    gZD: 'ŋɯʌt',
    DaD: 'mʌt',
    FaL: 'tʰʷʌt',
    daL: 'qʷʌt',
    CaD: 'bʌt',
    EaL: 'tʷʌt',
    FaL: 'tʰʷʌt',
    GaL: 'dʷʌt',
    haL: 'ʔʷʌt',
    iaL: 'χʷʌt',
    gaL: 'ɴʷʌt',
    BaD: 'pʰʌt',
    IaL: 'lʷʌt',
    eaL: 'qʰʷʌt',
    HaL: 'nʷʌt',
    QaL: 'sʷʌt',
    OaL: 'tsʰʷʌt',
    PaL: 'dzʷʌt',
    jbD: 'ʁʌt',
    jaL: 'ʁʷʌt',
    NaL: 'tsʷʌt',
    jcD: 'ʁɑt',
    icD: 'χɑt',
    EcD: 'tɑt',
    FcD: 'tʰɑt',
    hcD: 'ʔɑt',
    IcD: 'lɑt',
    ecD: 'qʰɑt',
    GcD: 'dɑt',
    PcD: 'dzɑt',
    gcD: 'ɴɑt',
    dcD: 'qɑt',
    QcD: 'sɑt',
    OcD: 'tsʰɑt',
    HcD: 'nɑt',
    DcD: 'mɑt',
    DcD: 'mɑt',
    gcL: 'ɴʷɑt',
    PcL: 'dzʷɑt',
    AcD: 'pɑt',
    NcD: 'tsɑt',
    dcL: 'qʷɑt',
    ecL: 'qʰʷɑt',
    jcL: 'ʁʷɑt',
    GcL: 'dʷɑt',
    icL: 'χʷɑt',
    hcL: 'ʔʷɑt',
    NcL: 'tsʷɑt',
    BcD: 'pʰɑt',
    FcL: 'tʰʷɑt',
    IcL: 'lʷɑt',
    EcL: 'tʷɑt',
    OcL: 'tsʰʷɑt',
    CcD: 'bɑt',
    jeD: 'ʁa˞t',
    SeD: 'tʂɻa˞t',
    CeD: 'ba˞t',
    eeD: 'qʰa˞t',
    jeL: 'ʁʷa˞t',
    AeD: 'pa˞t',
    JeL: 'ʈʷa˞t',
    heL: 'ʔʷa˞t',
    MeL: 'ɳʷa˞t',
    TeD: 'tʂʰɻa˞t',
    deL: 'qʷa˞t',
    deD: 'qa˞t',
    heD: 'ʔa˞t',
    VeD: 'ʂɻa˞t',
    DeD: 'ma˞t',
    ieL: 'χʷa˞t',
    MeD: 'ɳa˞t',
    geL: 'ɴʷa˞t',
    eeL: 'qʰʷa˞t',
    BeD: 'pʰa˞t',
    SeL: 'tʂɻʷa˞t',
    jdD: 'ʁɜ˞t',
    hdD: 'ʔɜ˞t',
    gdD: 'ɴɜ˞t',
    TdD: 'tʂʰɻɜ˞t',
    edD: 'qʰɜ˞t',
    idD: 'χɜ˞t',
    KdD: 'ʈʰɜ˞t',
    ddL: 'qʷɜ˞t',
    jdL: 'ʁʷɜ˞t',
    KdL: 'ʈʰʷɜ˞t',
    JdL: 'ʈʷɜ˞t',
    VdL: 'ʂɻʷɜ˞t',
    gdL: 'ɴʷɜ˞t',
    TdL: 'tʂʰɻʷɜ˞t',
    DdD: 'mɜ˞t',
    MdL: 'ɳʷɜ˞t',
    AdD: 'pɜ˞t',
    ddD: 'qɜ˞t',
    UdD: 'dʐɻɜ˞t',
    JdD: 'ʈɜ˞t',
    cdD: 'ɲɜ˞t',
    QgD: 'sɛt',
    OgD: 'tsʰɛt',
    dgD: 'qɛt',
    NgD: 'tsɛt',
    igL: 'χʷɛt',
    egL: 'qʰʷɛt',
    dgL: 'qʷɛt',
    jgL: 'ʁʷɛt',
    hgL: 'ʔʷɛt',
    GgD: 'dɛt',
    FgD: 'tʰɛt',
    jgD: 'ʁɛt',
    HgD: 'nɛt',
    PgD: 'dzɛt',
    ggD: 'ɴɛt',
    DgD: 'mɛt',
    AgD: 'pɛt',
    hgD: 'ʔɛt',
    egD: 'qʰɛt',
    igD: 'χɛt',
    BgD: 'pʰɛt',
    CgD: 'bɛt',
    EgD: 'tɛt',
    IgD: 'lɛt',
    QfD: 'siɛt',
    IfD: 'ɭiɛt',
    JfD: 'ʈiɛt',
    ffH: 'ɡi˞ɛt',
    cfD: 'ɲiɛt',
    XfD: 'tçiɛt',
    bfD: 'ʝiɛt',
    ZfD: 'dʝiɛt',
    gfH: 'ŋi˞ɛt',
    DfD: 'miɛt',
    efH: 'kʰi˞ɛt',
    AfD: 'piɛt',
    PfL: 'dzʷiɛt',
    NfL: 'tsʷiɛt',
    QfL: 'sʷiɛt',
    KfL: 'ʈʰʷiɛt',
    lfL: 'ɥiɛt',
    efL: 'kʰʷiɛt',
    hfP: 'ʔʷi˞ɛt',
    cfL: 'ɲʷiɛt',
    afL: 'çʷiɛt',
    XfL: 'tçʷiɛt',
    YfL: 'tçʰʷiɛt',
    JfL: 'ʈʷiɛt',
    IfL: 'ɭʷiɛt',
    BfD: 'pʰiɛt',
    CfH: 'bi˞ɛt',
    LfD: 'ɖiɛt',
    AfH: 'pi˞ɛt',
    VfL: 'ʂɻʷɛt',
    dfH: 'ki˞ɛt',
    afD: 'çiɛt',
    MfL: 'ɳʷiɛt',
    ifP: 'xʷi˞ɛt',
    hfL: 'ʔʷiɛt',
    dfP: 'kʷi˞ɛt',
    SfL: 'tʂɻʷɛt',
    OfL: 'tsʰʷiɛt',
    NfD: 'tsiɛt',
    VfD: 'ʂɻɛt',
    hfH: 'ʔi˞ɛt',
    KfD: 'ʈʰiɛt',
    ifH: 'xi˞ɛt',
    YfL: 'tçʰʷiɛt',
    TfD: 'tʂʰɻɛt',
    RfL: 'zʷiɛt',
    YfD: 'tçʰiɛt',
    lfD: 'jiɛt',
    UfD: 'dʐɻɛt',
    lpD: 'jɯɑq⁽ʷ⁾',
    IpD: 'ɭɯɑq⁽ʷ⁾',
    dpD: 'kɯɑq⁽ʷ⁾',
    XpD: 'tçɯɑq⁽ʷ⁾',
    apD: 'çɯɑq⁽ʷ⁾',
    cpD: 'ɲɯɑq⁽ʷ⁾',
    YpD: 'tçʰɯɑq⁽ʷ⁾',
    hpD: 'ʔɯɑq⁽ʷ⁾',
    epD: 'kʰɯɑq⁽ʷ⁾',
    gpD: 'ŋɯɑq⁽ʷ⁾',
    ZpD: 'dʝɯɑq⁽ʷ⁾',
    KpD: 'ʈʰɯɑq⁽ʷ⁾',
    QpD: 'sɯɑq⁽ʷ⁾',
    SpD: 'tʂɻɑq⁽ʷ⁾',
    NpD: 'tsɯɑq⁽ʷ⁾',
    PpD: 'dzɯɑq⁽ʷ⁾',
    OpD: 'tsʰɯɑq⁽ʷ⁾',
    fpD: 'ɡɯɑq⁽ʷ⁾',
    hpL: 'ʔuɑq⁽ʷ⁾',
    CpD: 'buɑq⁽ʷ⁾',
    ipL: 'xuɑq⁽ʷ⁾',
    kpL: 'wuɑq⁽ʷ⁾',
    dpL: 'kuɑq⁽ʷ⁾',
    JpD: 'ʈɯɑq⁽ʷ⁾',
    LpD: 'ɖɯɑq⁽ʷ⁾',
    epL: 'kʰuɑq⁽ʷ⁾',
    fpL: 'ɡuɑq⁽ʷ⁾',
    MpD: 'ɳɯɑq⁽ʷ⁾',
    BpD: 'pʰuɑq⁽ʷ⁾',
    ipD: 'xɯɑq⁽ʷ⁾',
    GqD: 'dɑq⁽ʷ⁾',
    DqD: 'mɑq⁽ʷ⁾',
    IqD: 'lɑq⁽ʷ⁾',
    FqD: 'tʰɑq⁽ʷ⁾',
    NqD: 'tsɑq⁽ʷ⁾',
    OqD: 'tsʰɑq⁽ʷ⁾',
    dqD: 'qɑq⁽ʷ⁾',
    eqD: 'qʰɑq⁽ʷ⁾',
    gqD: 'ɴɑq⁽ʷ⁾',
    BqD: 'pʰɑq⁽ʷ⁾',
    hqD: 'ʔɑq⁽ʷ⁾',
    CqD: 'bɑq⁽ʷ⁾',
    iqD: 'χɑq⁽ʷ⁾',
    QqD: 'sɑq⁽ʷ⁾',
    jqD: 'ʁɑq⁽ʷ⁾',
    PqD: 'dzɑq⁽ʷ⁾',
    AqD: 'pɑq⁽ʷ⁾',
    HqD: 'nɑq⁽ʷ⁾',
    iqL: 'χʷɑq⁽ʷ⁾',
    dqL: 'qʷɑq⁽ʷ⁾',
    hqL: 'ʔʷɑq⁽ʷ⁾',
    jqL: 'ʁʷɑq⁽ʷ⁾',
    eqL: 'qʰʷɑq⁽ʷ⁾',
    gqL: 'ɴʷɑq⁽ʷ⁾',
    IqL: 'lʷɑq⁽ʷ⁾',
    NqL: 'tsʷɑq⁽ʷ⁾',
    DrD: 'ma˞k',
    JrD: 'ʈa˞k',
    CrD: 'ba˞k',
    ArD: 'pa˞k',
    fsD: 'ɡi˞ak',
    dsD: 'ki˞ak',
    VsD: 'ʂɻak',
    TsD: 'tʂʰɻak',
    jrL: 'ʁʷa˞k',
    SrD: 'tʂɻa˞k',
    UrD: 'dʐɻa˞k',
    esD: 'kʰi˞ak',
    grD: 'ɴa˞k',
    gsD: 'ŋi˞ak',
    erD: 'qʰa˞k',
    hrD: 'ʔa˞k',
    isD: 'xi˞ak',
    KrD: 'ʈʰa˞k',
    hrL: 'ʔʷa˞k',
    BrD: 'pʰa˞k',
    irD: 'χa˞k',
    jrD: 'ʁa˞k',
    drD: 'qa˞k',
    irL: 'χʷa˞k',
    LrD: 'ɖa˞k',
    drL: 'qʷa˞k',
    hsL: 'ʔʷi˞ak',
    MrD: 'ɳa˞k',
    erL: 'qʰʷa˞k',
    CsD: 'bi˞ak',
    DtD: 'mɜ˞k',
    jtL: 'ʁʷɜ˞k',
    dtL: 'qʷɜ˞k',
    AtD: 'pɜ˞k',
    CtD: 'bɜ˞k',
    UtD: 'dʐɻɜ˞k',
    StD: 'tʂɻɜ˞k',
    UtL: 'dʐɻʷɜ˞k',
    TtD: 'tʂʰɻɜ˞k',
    etD: 'qʰɜ˞k',
    itL: 'χʷɜ˞k',
    jtD: 'ʁɜ˞k',
    dtD: 'qɜ˞k',
    JtD: 'ʈɜ˞k',
    htD: 'ʔɜ˞k',
    VtD: 'ʂɻɜ˞k',
    BtD: 'pʰɜ˞k',
    gtD: 'ɴɜ˞k',
    ItD: 'ɭɜ˞k',
    StL: 'tʂɻʷɜ˞k',
    VtL: 'ʂɻʷɜ˞k',
    ftL: 'ɢʷɜ˞k',
    MtD: 'ɳɜ˞k',
    QuD: 'siak',
    NuD: 'tsiak',
    huD: 'ʔiak',
    luD: 'jiak',
    auD: 'çiak',
    YuD: 'tçʰiak',
    ZuD: 'dʝiak',
    XuD: 'tçiak',
    LuD: 'ɖiak',
    OuD: 'tsʰiak',
    RuD: 'ziak',
    PuD: 'dziak',
    CuD: 'biak',
    luL: 'ɥiak',
    iuL: 'xʷiak',
    AuD: 'piak',
    BuD: 'pʰiak',
    buD: 'ʝiak',
    AsD: 'pi˞ak',
    JuD: 'ʈiak',
    KuD: 'ʈʰiak',
    XuL: 'tçʷiak',
    OuL: 'tsʰʷiak',
    QvD: 'sɛk',
    dvD: 'qɛk',
    BvD: 'pʰɛk',
    IvD: 'lɛk',
    EvD: 'tɛk',
    jvD: 'ʁɛk',
    gvD: 'ɴɛk',
    GvD: 'dɛk',
    FvD: 'tʰɛk',
    NvD: 'tsɛk',
    evD: 'qʰɛk',
    HvD: 'nɛk',
    PvD: 'dzɛk',
    DvD: 'mɛk',
    CvD: 'bɛk',
    AvD: 'pɛk',
    evL: 'qʰʷɛk',
    dvL: 'qʷɛk',
    OvD: 'tsʰɛk',
    ivD: 'χɛk',
    ivL: 'χʷɛk',
    KvD: 'ʈʰɛk',
    XwD: 'tçiɯk',
    LwD: 'ɖiɯk',
    IwD: 'ɭiɯk',
    KwD: 'ʈʰiɯk',
    JwD: 'ʈiɯk',
    bwD: 'ʝiɯk',
    QwD: 'siɯk',
    ZwD: 'dʝiɯk',
    awD: 'çiɯk',
    iwD: 'xɯk',
    UwD: 'dʐɻɯk',
    fwD: 'ɡɯk',
    MwD: 'ɳiɯk',
    TwD: 'tʂʰɻɯk',
    hwD: 'ʔɯk',
    VwD: 'ʂɻɯk',
    ewD: 'kʰɯk',
    dwD: 'kɯk',
    lwD: 'jiɯk',
    NwD: 'tsiɯk',
    AwD: 'pi˞ɯk',
    kwL: 'wi˞ɯk',
    iwL: 'xʷi˞ɯk',
    BwD: 'pʰi˞ɯk',
    SwD: 'tʂɻɯk',
    CwD: 'bi˞ɯk',
    gwD: 'ŋɯk',
    PwD: 'dziɯk',
    JwD: 'ʈiɯk',
    DwD: 'mi˞ɯk',
    YwD: 'tçʰiɯk',
    ExD: 'tʌk',
    NxD: 'tsʌk',
    IxD: 'lʌk',
    FxD: 'tʰʌk',
    exD: 'qʰʌk',
    GxD: 'dʌk',
    ixD: 'χʌk',
    DxD: 'mʌk',
    PxD: 'dzʌk',
    QxD: 'sʌk',
    AxD: 'pʌk',
    CxD: 'bʌk',
    jxL: 'ʁʷʌk',
    dxL: 'qʷʌk',
    hxD: 'ʔʌk',
    jxD: 'ʁʌk',
    HxD: 'nʌk',
    dxD: 'qʌk',
    OxD: 'tsʰʌk',
    BxD: 'pʰʌk',
    ixL: 'χʷʌk',
    O1D: 'tsʰip',
    Z1D: 'dʝip',
    X1D: 'tçip',
    R1D: 'zip',
    P1D: 'dzip',
    c1D: 'ɲip',
    h1D: 'ʔip',
    a1D: 'çip',
    N1D: 'tsip',
    f1H: 'ɡɯp',
    L1D: 'ɖip',
    J1D: 'ʈip',
    I1D: 'ɭip',
    d1H: 'kɯp',
    g1H: 'ŋɯp',
    e1H: 'kʰɯp',
    Q1D: 'sip',
    V1D: 'ʂɻp',
    i1H: 'xɯp',
    S1D: 'tʂɻp',
    h1H: 'ʔɯp',
    K1D: 'ʈʰip',
    k1D: 'wi˞p',
    l1D: 'jip',
    M1D: 'ɳip',
    U1D: 'dʐɻp',
    C1H: 'bi˞p',
    A1H: 'pi˞p',
    T1D: 'tʂʰɻp',
    Y1D: 'tçʰip',
    j2D: 'ʁɔp',
    d2D: 'qɔp',
    E2D: 'tɔp',
    Q2D: 'sɔp',
    G2D: 'dɔp',
    F2D: 'tʰɔp',
    P2D: 'dzɔp',
    N2D: 'tsɔp',
    I2D: 'lɔp',
    H2D: 'nɔp',
    e2D: 'qʰɔp',
    h2D: 'ʔɔp',
    i2D: 'χɔp',
    g2D: 'ɴɔp',
    O2D: 'tsʰɔp',
    Z2D: 'dʝɔp',
    h2D: 'ʔɔp',
    j3D: 'ʁɑp',
    I3D: 'lɑp',
    E3D: 'tɑp',
    F3D: 'tʰɑp',
    i3D: 'χɑp',
    H3D: 'nɑp',
    G3D: 'dɑp',
    Q3D: 'sɑp',
    g3D: 'ɴɑp',
    P3D: 'dzɑp',
    d3D: 'qɑp',
    e3D: 'qʰɑp',
    h3D: 'ʔɑp',
    O3D: 'tsʰɑp',
    d3D: 'qɑp',
    X3D: 'tçɑp',
    l4D: 'jiɛp',
    N4D: 'tsiɛp',
    a4D: 'çiɛp',
    Z4D: 'dʝiɛp',
    I4D: 'ɭiɛp',
    P4D: 'dziɛp',
    L4D: 'ɖiɛp',
    h4H: 'ʔi˞ɛp',
    M4D: 'ɳiɛp',
    Y4D: 'tçʰiɛp',
    c4D: 'ɲiɛp',
    X4D: 'tçiɛp',
    O4D: 'tsʰiɛp',
    K4D: 'ʈʰiɛp',
    f4H: 'ɡi˞ɛp',
    J4D: 'ʈiɛp',
    k4D: 'wi˞ɛp',
    e4D: 'kʰiɛp',
    V4D: 'ʂɻɛp',
    d4H: 'ki˞ɛp',
    h4D: 'ʔiɛp',
    F5D: 'tʰɛp',
    j5D: 'ʁɛp',
    d5D: 'qɛp',
    e5D: 'qʰɛp',
    G5D: 'dɛp',
    H5D: 'nɛp',
    Q5D: 'sɛp',
    I5D: 'lɛp',
    E5D: 'tɛp',
    P5D: 'dzɛp',
    N5D: 'tsɛp',
    i5D: 'χɛp',
    Q5D: 'sɛp',
    j6D: 'ʁɜ˞p',
    e6D: 'qʰɜ˞p',
    U6D: 'dʐɻɜ˞p',
    d6D: 'qɜ˞p',
    S6D: 'tʂɻɜ˞p',
    T6D: 'tʂʰɻɜ˞p',
    M6D: 'ɳɜ˞p',
    i6D: 'χɜ˞p',
    V6D: 'ʂɻɜ˞p',
    J6D: 'ʈɜ˞p',
    h6D: 'ʔɜ˞p',
    g6D: 'ɴɜ˞p',
    K6D: 'ʈʰɜ˞p',
    j7D: 'ʁa˞p',
    L7D: 'ɖa˞p',
    h7D: 'ʔa˞p',
    d7D: 'qa˞p',
    V7D: 'ʂɻa˞p',
    i7D: 'χa˞p',
    g8D: 'ŋɯʌp',
    i8D: 'xɯʌp',
    e8D: 'kʰɯʌp',
    d8D: 'kɯʌp',
    h8D: 'ʔɯʌp',
    l8D: 'jɯʌp',
    f8D: 'ɡɯʌp',
    C9D: 'buʌp',
    A9D: 'puʌp',
    B9D: 'pʰuʌp',
    e9L: 'kʰuʌp',
    M9L: 'ɳuʌp',
    K9L: 'ʈʰuʌp',
  }[音韻地位.編碼] || '?';
}

/**
 * 推導《聲音唱和圖》擬音
 * 暫無擬音說明文章，一些擬音重點在代碼中以注釋寫出
 *
 * 説明
 * 以下內容為生成推導《聲音唱和圖》擬音的函數體
 * 函數接受音韻地位，返回對應的推導《聲音唱和圖》擬音
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的推導《聲音唱和圖》擬音
 */
export function chiangxhua(音韻地位, 字頭) {
  // 代碼中的音韻表達式使用了全角半角空格用於對齊，需要進行移除
  const is = (x) => 音韻地位.屬於(x.replace(/ +/g, ' ').replace(/　/g,'').trim());

  // 如果只需要顯示音位或音值中的一個，請將另一個改為 false
  const 顯示音位 = true;
  const 顯示音值 = true;
  const 聲調用調類表示 = false; // 打開則用 ¹²³⁴ 表示（分別為平上去入），關閉則用國際音標的附加符號表示
  const 省略唇音後的w介音 = true; // 僅用於音值。打開則“八”/pʕʷɐ/ -> [pa]，關閉則 /pʕʷɐ/ -> [pwa]

  function 聲母規則() {
    if (is('幫滂母 東鍾微虞廢文元陽尤凡韻 三等')) return 'f';
    if (is('並　母 東鍾微虞廢文元陽尤凡韻 三等')) return 'v';
    if (is('明　母 　鍾微虞廢文元陽　凡韻 上聲')) return 'ˀʋ';
    if (is('明　母 　鍾微虞廢文元陽　凡韻 　　')) return 'ʋ';
    if (is('幫　母 　　')) return 'p';
    if (is('滂　母 　　')) return 'pʰ';
    if (is('並　母 仄聲')) return 'b';
    if (is('並　母 平聲')) return 'bʰ';
    if (is('明　母 上聲')) return 'ˀm';
    if (is('明　母 　　')) return 'm';
    if (is('端　母 　　')) return 't';
    if (is('透　母 　　')) return 'tʰ';
    if (is('定　母 仄聲')) return 'd';
    if (is('定　母 平聲')) return 'dʰ';
    if (is('泥孃母 上聲')) return 'ˀn';
    if (is('泥孃母 　　')) return 'n';
    if (is('知　母 　　')) return 'tɻ';
    if (is('徹　母 　　')) return 'tɻʰ';
    if (is('澄　母 仄聲')) return 'dɻ';
    if (is('澄　母 平聲')) return 'dɻʰ';
    if (is('精　母 　　')) return 'ts';
    if (is('清　母 　　')) return 'tsʰ';
    if (is('從　母 仄聲')) return 'dz';
    if (is('從　母 平聲')) return 'dzʰ';
    if (is('心　母 　　')) return 's';
    if (is('邪　母 　　')) return 'z';
    if (is('莊章母 　　')) return 'tʂ';
    if (is('初昌母 　　')) return 'tʂʰ';
    if (is('崇　母 仄聲') && '撰饌棧助寨砦乍狀驟'.includes(字頭)) return 'dʐ'; // 特例：乍
    if (is('船　母 平聲') && '晨船荼乘'.includes(字頭)) return 'dʐʰ'; // 推斷的特例
    if (is('崇常母 平聲')) return 'dʐʰ';
    if (is('生書母 　　')) return 'ʂ';
    if (is('崇常母 仄聲')) return 'ʐ';
    if (is('俟船母 　　')) return 'ʐ';
    if (is('見　母 　　')) return 'k';
    if (is('溪　母 　　')) return 'kʰ';
    if (is('羣　母 仄聲')) return 'ɡ';
    if (is('羣　母 平聲')) return 'ɡʰ';
    if (is('疑　母 上聲')) return 'ˀŋ';
    if (is('疑　母 　　')) return 'ŋ';
    if (is('曉　母 　　')) return 'x';
    if (is('云　母 　　') && '雄熊'.includes(字頭)) return 'ɣ'; // 特例：雄
    if (is('匣　母 　　') && '肴餚殽爻倄'.includes(字頭)) return ''; // 特例：爻
    if (is('匣　母 　　')) return 'ɣ';
    if (is('影　母 　　')) return 'ˀ';
    if (is('云以母 上聲')) return 'ˀ'; // 推斷
    if (is('云以母 　　')) return '';
    if (is('來　母 上聲')) return 'ˀl';
    if (is('來　母 　　')) return 'l';
    if (is('日　母 上聲')) return 'ˀɻ';
    if (is('日　母 　　')) return 'ɻ';
    throw new Error('無聲母規則');
  }

  function 介音規則() {
    // IV  等：j
    // III 等：ɉ
    // II  等：ʕ
    // I   等：無
    if (is('端精組 或 孃來母')) {
      if (is('泥母 咍韻 開口 上聲')) return ''; // 特例：乃。可能是因為 II 等被“嬭”小韻（中古孃母）佔據了。來母蟹攝二等僅“唻”一字，推測不影響咍泰韻 II 等化，故不考慮來母
      if (is('泥母 咍泰韻 開口')) return ''; // 據“乃”類推
      if (is('二等 或 咍泰寒覃談韻 開口')) return 'ʕ';
      if (is('一等')) return '';
      if (is('精組 止攝 開口')) return ''; // 中古後期發生 sjᵻ & sɉᵻ > sᵻ，變 I 等
      if (is('三四等')) return 'ɉ'; // 一律歸 III 等
    };
    if (is('知組 二等 或 莊組')) return 'ʕ';
    if (is('知組 三等 或 章組 或 日母')) return 'ɉ';
    // “茝、佁”之類字不作考慮
    // 其餘是鈍音和以母
    if (is('幫組 三等')) { // 處理輕唇化的韻
      if (is('微廢韻')) return 'j'; // [fi] 被列入 IV 等
      if (is('元凡韻')) return 'ʕ';
      if (is('東鍾虞文陽尤韻')) return ''; // 據現代方言推斷陽韻輕唇化後是 I 等
    }
    if (is('四等 或 重紐A類 或 以母 或 清幽韻')) return 'j'; // IV 等
    if (is('三等')) return 'ɉ'; // 其餘的切韻三等是 III 等
    if (is('二等')) return 'ʕ';
    if (is('一等')) return '';
    throw new Error('無介音規則');
  }

  function 唇化規則() {
    if (is('止臻攝 幫組 重紐B類')) return 'ʷ'; // 幫組 III 等推測已經發生了 ᵻ > ᵻj 的裂化（和見組合口 ᵻ > ᵻj 平行）。ᵻj III、IV 等舒聲只有合口
    if (is('幫組 流通深咸攝')) return ''; // 幫組一等除這四攝外，都算合口
    if (is('幫組 二等 梗蟹攝')) return ''; // 幫組二等這兩攝非常不像合口，不算合口
    if (is('幫組 一二等 或 合口 或 鍾虞模韻')) return 'ʷ'; // 其他一二等推測都算合口
    if (is('幫組 鍾虞文陽韻')) return 'ʷ'; // 輕唇變一等的
    if (is('幫組 元韻')) return 'ʷ'; // 輕唇變二等的（凡韻已經被排除），也是合口。因為“八”列在合口
    // 注意，咸深攝舒聲入聲都視為開口，不依原圖定義

    if (is('江韻 知莊組 或 江韻 孃來母')) return 'ʷ'; // 據《蒙古字韻》推斷，且“雙、霜”需要靠開合對立
    return '';
  }

  function 韻基規則() {
    // 遇攝
    if (is('遇攝 　　　　')) return 'ʊ ';
    // 止攝
    if (is('止攝 幫組 重紐B類')) return 'ᵻj'; // 幫組 III 等推測已經發生了 ᵻ > ᵻj 的裂化（和見組合口 ᵻ > ᵻj 平行）
    if (is('止攝 莊組　　')) return 'ᵻ '; // 衰、帥
    if (is('止攝 合口　　')) return 'ᵻj';
    if (is('止攝 　　　　')) return 'ᵻ '; // [i] 的音位也可分析为 jᵻj，但邵雍的分析是 jᵻ，排在 ᵻ 韻基一行
    // 蟹攝
    if (is('蟹攝 一二等　')) return 'ɐj'; // 二等今讀 [a] 的字不作考慮
    if (is('蟹攝 合口　　')) return 'ᵻj'; // 三四等合口。推斷
    if (is('蟹攝 　　　　')) return 'ᵻ '; // 三四等開口。“妻”排在 ᵻ 韻基一行，說明 [jej] 之類的音已經併入 [i]
    // 果假攝
    if (is('果假攝 　　　')) return 'ɐ ';
    // 流攝
    if (is('流攝 　　　　')) return 'ᵻw';
    // 效攝
    if (is('效攝 　　　　')) return 'ɐw';
    // 臻攝
    if (is('臻攝 幫組 重紐B類 入聲')) return 'ᵻj'; // 推測和止攝一樣 ᵻ > ᵻj（“筆”）
    if (is('元韻 　　　　')) return is('舒聲') ? 'ɐn' : 'ɐ ';
    if (is('臻攝 　　　　')) return is('舒聲') ? 'ᵻn' : 'ᵻ '; // ʊ 無入聲，推斷文韻幫組入聲韻基只能是 ᵻ。就像南昌話“骨”經歷 kut > kul > kuɨʔ > kuʔ 的演變。入聲（“紇”）韻基也推測為 ᵻ
    // 山攝
    if (is('山攝 　　　　')) return is('舒聲') ? 'ɐn' : 'ɐ ';
    // 曾梗攝
    if (is('曾梗攝 一二等')) return is('舒聲') ? 'ᵻŋ' : 'ᵻj'; // 入聲韻基為推測，因為“窄、虱”今不同韻所以韻基不能是 ᵻ
    if (is('曾梗攝 　　　')) return is('舒聲') ? 'ᵻŋ' : 'ᵻ '; // 三四等。開口入聲只能是 [i]，合口入聲也推測為 ᵻ。《蒙古字韻》中合口入聲是 ᵻj 韻基（ÿue 或 ue），但現代方言未見，不採用
    // 通攝
    if (is('通攝 　　　　')) return is('舒聲') ? 'ʊŋ' : 'ᵻw'; // ʊ 無入聲，推斷東一入聲韻基只能是 ᵻw。又推斷冬韻併入東一
    // 宕江韻
    if (is('宕江攝 　　　')) return is('舒聲') ? 'ɐŋ' : 'ɐw';
    // 深攝
    if (is('深攝 　　　　')) return is('舒聲') ? 'ᵻm' : 'ᵻʋ'; // 南昌話“骨”kul 也可以視為和“滾、棍”kun 只有聲調差別（因為 n、l 是同一音位）。這裡 ʋ 就是聲母中的微母，它有鼻音變體絲毫不奇怪
    // 咸攝
    if (is('咸攝 　　　　')) return is('舒聲') ? 'ɐm' : 'ɐʋ';
    throw new Error('無韻基規則');
  }

  function 聲調規則() {
    if (is('平聲 　　')) return '̀';
    if (is('上聲 全濁')) return '̌'; // 推斷全濁上已經歸陽去
    if (is('上聲 　　')) return '́';
    if (is('去聲 　　')) return '̌';
    if (is('入聲 　　')) return '̆';
    throw new Error('無聲調規則');
  }

  function 調類規則() {
    if (is('平聲 　　')) return '¹';
    if (is('上聲 全濁')) return '³'; // 推斷全濁上已經歸陽去
    if (is('上聲 　　')) return '²';
    if (is('去聲 　　')) return '³';
    if (is('入聲 　　')) return '⁴';
    throw new Error('無聲調規則');
  }

  let 聲母 = 聲母規則();
  let 介音 = 介音規則();
  let 唇化 = 唇化規則();
  let 韻基 = 韻基規則().trim(); // 移除韻母末尾可能的空格
  let 韻核 = 韻基[0];
  let 韻尾 = 韻基.substring(1);
  let 聲調 = 聲調用調類表示 ? 調類規則() : 聲調規則();

  let 音位 = '';
  let 音值 = '';

  if (聲調用調類表示) {
    音位 = 聲母 + 介音 + 唇化 + 韻核 + 韻尾 + 聲調;
  } else {
    音位 = 聲母 + 介音 + 唇化 + 韻核 + 聲調 + 韻尾;
  }

  if (顯示音位 && !顯示音值) return 音位;

  // 接下來對音位應用音系規則，生成音值

  // /ʊ/ 的實現與等無關，單獨處理
  if (韻核 === 'ʊ' && 唇化) {
    韻核 = 'u';
  }

  // /ᵻ/ 和 /ɐ/ 的實現
  if (介音 === 'j') { // IV 等
    if (韻核 === 'ᵻ') {
      if (韻尾 === 'ʋ' || 韻尾 === 'm') {
        韻核 = 'ɨ';
      } else if (唇化 && 'jw'.includes(韻尾)) { // 合口，且韻尾是 j、w 或無韻尾
        韻核 = 'ʉ';
      } else {
        韻核 = 'i';
      }
    } else if (韻核 === 'ɐ' && 韻尾 !== 'ŋ') { // jɐŋ 仍是 jɐŋ
      韻核 = 'ɛ';
    }
  } else if (介音 === 'ɉ') { // III 等
    if (韻核 === 'ᵻ') {
      韻核 = 唇化 ? 'ʉ' : 'ɨ';
    } // III 等 /ɐ/ 就實現為 [ɐ]
  } else if (介音 === 'ʕ') { // II 等
    if (韻核 === 'ᵻ') {
      if (韻尾 === 'j' || 韻尾 === 'ŋ') {
        韻核 = 'ɛ'; // 構擬音值介於 [ɛ]、[ɜ] 之間，寫 [ɛ]
      } else if (韻尾 === '') {
        if (唇化) {
          韻核 = 'ɛ';
          韻尾 = 'ɻ';
        } else {
          韻核 = 'ɻ̍';
        }
      } else {
        韻核 = 唇化 ? 'u' : 'ə'; // 同 I 等
      }
    } else if (韻核 === 'ɐ') {
      韻核 = 'a';
    }
  } else { // I 等
    if (韻核 === 'ᵻ') {
      韻核 = 唇化 ? 'u' : 'ə';
      if (韻尾 === '') {
        if (is('精組') && !唇化) {
          韻核 = 'ɹ̩';
        } else {
          韻尾 = 'ɹ'; // 卒、麧
        }
      }
    } else if (韻核 === 'ɐ') {
      if (唇化) {
        if (韻尾 === 'j') {
          韻核 = 'ɵ'; // wɐj 更加高化
        } else if (韻尾 !== 'ŋ') { // wɐŋ 仍是 wɐŋ，其他的合口韻核是 ɔ
          韻核 = 'ɔ';
        }
      } else if (韻尾 === '' && !唇化) {
        韻核 = 'ʌ'; // 歌韻更加高化
      }
    }
  }

  if (介音[0] === 'ɉ') {
    介音 = 'j̈';
  }

  // 移除 ʕ，因為 II 等已由單獨的 ɛ、a 或捲舌聲母標記
  if (介音 === 'ʕ') {
    介音 = '';
  }

  // 唇化坍縮到介音上
  if (唇化) {
    if (介音[0] === 'j') {
      介音 = 介音.replace('j', 'ɥ'); // III、IV 等
    } else {
      介音 = is('幫組') && 省略唇音後的w介音 ? '' : 'w'; // I、II 等直接是 w 介音，但唇音后省略
    }
  }

  if (聲調用調類表示) {
    音值 = 聲母 + 介音 + 韻核 + 韻尾 + 聲調;
  } else {
    音值 = 聲母 + 介音 + 韻核 + 聲調 + 韻尾;
  }

  // 移除元音同部位近音
  音值 = 音值.replace('ji', 'i');
  音值 = 音值.replace('j̈ɨ', 'ɨ');
  音值 = 音值.replace('ɥ̈ʉ', 'ʉ');
  音值 = 音值.replace('wu', 'u');

  if (!顯示音位 && 顯示音值) return 音值;

  return 音位 + ' \n[' + 音值 + ']'; // 如果兩個開關都是 false，也同時顯示音位、音值
}

/**
 * 推導《分韻撮要》擬音
 * https://ayaka.shn.hk/fanwan/
 *
 * 説明
 * 以下內容為生成推導《分韻撮要》擬音的函數體
 * 函數接受音韻地位，返回對應的推導《分韻撮要》擬音
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的推導《分韻撮要》擬音
 */
export function fanwan(音韻地位, 字頭) {
  const is = (x) => 音韻地位.屬於(x);

  function 聲母規則() {
    if (is('幫母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      return 'b';
    }
    if (is('滂母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      return 'p';
    }
    if (is('並母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      return is('平聲') ? 'p' : 'b';
    }
    if (is('明母')) return 'm';

    if (is('端母')) return 'd';
    if (is('透母')) return 't';
    if (is('定母')) return is('平聲') ? 't' : 'd';
    if (is('泥母')) return 'n';
    if (is('來母')) return 'l';

    if (is('知母')) return 'zh';
    if (is('徹母')) return 'ch';
    if (is('澄母')) return is('平聲') ? 'ch' : 'zh';
    if (is('孃母')) return 'n';

    if (is('精母')) return 'z';
    if (is('清母')) return 'c';
    if (is('從母')) return is('平聲') ? 'c' : 'z';
    if (is('心母')) return 's';
    if (is('邪母')) return is('平聲') ? 'c' : 'z'; // 塞擦音多於擦音

    if (is('莊母 二等')) return 'zh';
    if (is('初母 二等')) return 'ch';
    if (is('崇母 二等')) return is('平聲') ? 'ch' : 'zh';
    if (is('生母 二等')) return 'sh';
    if (is('俟母 二等')) return is('平聲') ? 'ch' : 'zh';

    // 莊組三等平翹音規律不明確
    if (is('莊母')) return 'z';
    if (is('初母')) return 'c';
    if (is('崇母')) return is('平聲') ? 'c' : 'z';
    if (is('生母')) return 's';
    if (is('俟母')) return is('平聲') ? 'c' : 'z';

    if (is('章母')) return 'zh';
    if (is('昌母')) return 'ch';
    if (is('常母')) return 'sh'; // 擦音多於塞擦音
    if (is('書母')) return 'sh';
    if (is('船母')) return 'sh';
    if (is('日母')) return 'nj';

    if (is('見母')) return 'g';
    if (is('溪母')) return 'h'; // 多數擦化
    if (is('羣母')) return is('平聲') ? 'k' : 'g';
    if (is('疑母')) return 'ng'; // ng 拼細音時為 nj，詳後

    if (is('曉母')) return 'h';
    if (is('匣母')) {
      if (is('合口 或 模韻')) return 'j'; // 非 yu 前為 w，詳後
      return 'h';
    }
    if (is('影云以母')) {
      if (is('三四等')) return 'j'; // 非 yu 前為 w，詳後
      return '';
    }

    throw new Error('無聲母規則');
  }

  function 韻母規則() {
    // 通攝
    if (is('東冬鍾韻')) return 'ung';

    // 江攝
    if (is('江韻 幫組')) return 'ong';
    if (is('江韻 舌齒音')) return 'oeng';
    if (is('江韻 牙喉音')) return 'ong';

    // 止攝
    if (is('支脂之微韻 幫組')) return 'i';
    if (is('支脂之微韻 開口')) return 'i'; // i 在 z/c/s 前為 ii，詳後
    if (is('支脂之微韻 合口 舌齒音')) return 'eoi';
    if (is('支脂之微韻 合口 牙喉音')) return 'ai';

    // 遇攝
    if (is('魚虞韻 幫組')) return 'u';
    if (is('魚虞韻 莊組')) return 'o';
    if (is('魚虞韻')) return 'yu';
    if (is('模韻 疑母')) return '';
    if (is('模韻')) return 'u';

    // 蟹攝
    if (is('齊韻')) return 'ai';
    if (is('祭韻 幫組')) return 'ai';
    if (is('祭韻 開口')) return 'ai';
    if (is('祭韻 合口 舌齒音')) return 'eoi';
    if (is('祭韻 合口 以母')) return 'eoi';
    if (is('祭韻 合口 牙喉音')) return 'ai';
    if (is('泰韻 幫組')) return 'ui';
    if (is('泰韻 開口 端組')) return 'aai';
    if (is('泰韻 開口 來母')) return 'aai';
    if (is('泰韻 開口 精組')) return 'oi';
    if (is('泰韻 開口 牙喉音')) return 'oi';
    if (is('泰韻 合口 舌齒音')) return 'eoi';
    if (is('泰韻 合口 疑母')) return 'oi';
    if (is('泰韻 合口 牙喉音')) return 'ui';
    if (is('佳皆夬韻')) return 'aai';
    if (is('灰韻 疑母')) return 'oi';
    if (is('灰韻')) return 'ui';
    if (is('咍韻')) return 'oi';
    if (is('廢韻')) return 'ai';

    // 臻攝
    if (is('眞韻 幫組')) return 'an';
    if (is('眞韻 開口')) return 'an';
    if (is('眞韻 合口 舌齒音')) return 'eon';
    if (is('眞韻 合口 牙喉音')) return 'an';
    if (is('臻文欣韻')) return 'an';
    if (is('元韻 幫組')) return 'aan';
    if (is('元韻 開口')) return 'in';
    if (is('元韻 合口')) return 'yun';
    if (is('魂韻 幫組')) return 'un';
    if (is('魂韻 端組')) return 'eon';
    if (is('魂韻 來母')) return 'eon';
    if (is('魂韻 精組')) return 'yun';
    if (is('魂韻 牙喉音')) return 'an';
    if (is('痕韻')) return 'an';

    // 山攝
    if (is('寒韻 幫組')) return 'un';
    if (is('寒韻 開口 舌齒音')) return 'aan';
    if (is('寒韻 開口 牙喉音')) return 'on';
    if (is('寒韻 合口 舌齒音')) return 'yun';
    if (is('寒韻 合口 牙喉音')) return 'un';
    if (is('刪山韻')) return 'aan';
    if (is('仙先韻 幫組')) return 'in';
    if (is('仙先韻 開口')) return 'in';
    if (is('仙先韻 合口')) return 'yun';

    // 效攝
    if (is('蕭宵韻')) return 'iu';
    if (is('肴韻')) return 'aau';
    if (is('豪韻')) return 'u';

    // 果攝
    if (is('歌韻 一等')) return 'o';
    if (is('歌韻 三等')) return 'e';

    // 假攝
    if (is('麻韻 二等')) return 'aa';
    if (is('麻韻 三等')) return 'e';

    // 宕攝
    if (is('陽韻 幫組')) return 'ong';
    if (is('陽韻 開口 莊組')) return 'ong';
    if (is('陽韻 開口')) return 'oeng';
    if (is('陽韻 合口')) return 'ong';
    if (is('唐韻')) return 'ong';

    // 梗攝
    if (is('庚韻 二等')) return 'ang';
    if (is('庚韻 三等 莊組')) return 'ang';
    if (is('庚韻 三等')) return 'ing';
    if (is('耕韻')) return 'ang';
    if (is('清青韻')) return 'ing';

    // 曾攝
    if (is('蒸韻')) return 'ing';
    if (is('登韻')) return 'ang';

    // 流攝
    if (is('尤侯幽韻')) return 'au';

    // 深攝
    if (is('侵韻')) return 'am'; // m 韻尾在聲母為脣音時為 n，詳後，下同

    // 咸攝
    if (is('覃談韻 幫組')) return 'aam';
    if (is('覃談韻 舌齒音')) return 'aam';
    if (is('覃談韻 牙喉音')) return 'om';
    if (is('鹽添嚴韻')) return 'im';
    if (is('咸銜凡韻')) return 'aam';

    throw new Error('無韻母規則');
  }

  function 聲調規則() {
    if (is('全清 或 次清')) {
      if (is('平聲')) return '1'; // 陰平
      if (is('上聲')) return '2'; // 陰上
      if (is('去聲')) return '3'; // 陰去
      if (is('入聲')) return '1'; // 陰入
    } else {
      if (is('平聲')) return '4'; // 陽平
      if (is('全濁 上聲')) return '6'; // 陽去，全濁上變去
      if (is('上聲')) return '5'; // 陽上
      if (is('去聲')) return '6'; // 陽去
      if (is('入聲')) return '6'; // 陽入
    }
    throw new Error('無聲調規則');
  }

  let 聲母 = 聲母規則();
  let 韻母 = 韻母規則();
  let 聲調 = 聲調規則();

  // i 在 z/c/s 前為 ii
  if (['z', 'c', 's'].includes(聲母) && 韻母 === 'i') 韻母 = 'ii';

  // ng 拼細音時為 nj
  const is細音 = ['eo', 'i', 'oe', 'u', 'yu'].some((x) => 韻母.startsWith(x));
  if (聲母 === 'ng' && is細音) 聲母 = 'nj';

  if (is('合口 或 模韻') && !['eo', 'oe', 'yu'].some((x) => 韻母.startsWith(x))) { // 合口字
    if (聲母 === 'g') 聲母 = 'gw';
    else if (聲母 === 'k') 聲母 = 'kw';
    else if (聲母 === 'h' && !韻母.startsWith('i')) 聲母 = 'f';
    else if (聲母 === 'j') 聲母 = 'w';
    else if (聲母 === '') 聲母 = 'w';
  }

  // m 韻尾在聲母為脣音時為 n
  if (is('幫組') && 韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'n';

  if (is('入聲')) {
    if (韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'p';
    else if (韻母.endsWith('n')) 韻母 = 韻母.slice(0, -1) + 't';
    else if (韻母.endsWith('ng')) 韻母 = 韻母.slice(0, -2) + 'k';
  }

  return 聲母 + 韻母 + 聲調;
}

/**
 * 推導普通話
 * https://github.com/BYVoid/ytenx/blob/master/ytenx/sync/kyonh/Dauh.txt
 *
 * 説明
 * 以下內容為生成推導普通話的函數體
 * 函數接受音韻地位，返回對應的推導普通話音
 * 部分音資料暫缺，以問號表示
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的推導普通話
 */
export function putonghua(音韻地位, 字頭) {
  return {
    EAA: 'dong1',
    GAA: 'tong2',
    JBA: 'zhong1',
    LBA: 'chong2',
    XBA: 'zhong1',
    KBA: 'chong1',
    UBA: 'chong2',
    QBA: 'song1',
    cBA: 'rong2',
    dBA: 'jiong1',
    lBA: 'yong2',
    kBA: 'yong2',
    DBA: 'meng2',
    eBA: 'qiong1',
    fBA: 'qiong2',
    CBA: 'feng2',
    ABA: 'feng1',
    BBA: 'feng1',
    YBA: 'chong1',
    IBA: 'long2',
    eAA: 'kong1',
    dAA: 'gong1',
    DAA: 'meng2',
    IAA: 'long2',
    jAA: 'hong2',
    PAA: 'cong2',
    hAA: 'weng1',
    OAA: 'cong1',
    FAA: 'tong1',
    NAA: 'zong1',
    CAA: 'peng2',
    iAA: 'hong1',
    gAA: 'weng2',
    QAA: 'song1',
    ECA: 'dong1',
    GCA: 'tong2',
    PCA: 'cong2',
    HCA: 'nong2',
    dCA: 'gong1',
    jCA: 'hong2',
    ICA: 'long2',
    NCA: 'zong1',
    QCA: 'song1',
    FCA: 'tong1',
    XDA: 'zhong1',
    IDA: 'long2',
    aDA: 'shong1',
    RDA: 'song2',
    YDA: 'chong1',
    lDA: 'yong2',
    ADA: 'feng1',
    iDA: 'xiong1',
    gDA: 'yong2',
    hDA: 'yong1',
    MDA: 'nong2',
    LDA: 'chong2',
    PDA: 'cong2',
    KDA: 'chong1',
    CDA: 'feng2',
    BDA: 'feng1',
    NDA: 'zong1',
    cDA: 'rong2',
    fDA: 'qiong2',
    ZDA: 'chong2',
    dDA: 'jiong1',
    QDA: 'song1',
    ODA: 'cong1',
    eDA: 'qiong1',
    dEA: 'jiang1',
    DEA: 'mang2',
    MEA: 'nang2',
    TEA: 'chuang1',
    AEA: 'bang1',
    jEA: 'xiang2',
    BEA: 'pang1',
    IEA: 'lang2',
    VEA: 'shuang1',
    CEA: 'pang2',
    iEA: 'xiang1',
    hEA: 'yang1',
    eEA: 'qiang1',
    LEA: 'chuang2',
    KEA: 'chuang1',
    JEA: 'zhuang1',
    gEA: 'yang2',
    UEA: 'chuang2',
    XFA: 'zhi1',
    lFA: 'yi2',
    kFI: 'wei2',
    dFM: 'gui1',
    iFM: 'hui1',
    hFM: 'wei1',
    DFE: 'mi2',
    iFI: 'hui1',
    LFI: 'chui2',
    ZFI: 'chui2',
    IFI: 'lei2',
    YFI: 'chui1',
    BFE: 'pi1',
    AFE: 'bi1',
    RFI: 'sui2',
    eFM: 'kui1',
    eFI: 'kui1',
    fFE: 'qi2',
    fFA: 'qi2',
    iFE: 'xi1',
    eFE: 'qi1',
    gFE: 'yi2',
    CFE: 'pi2',
    ZFA: 'chi2',
    cFA: 'er2',
    IFA: 'li2',
    PFA: 'ci2',
    NFA: 'zi1',
    dFE: 'ji1',
    AFA: 'bi1',
    CFA: 'pi2',
    aFA: 'shi1',
    QFA: 'si1',
    TFA: 'chi1',
    KFA: 'chi1',
    DFA: 'mi2',
    OFA: 'ci1',
    JFA: 'zhi1',
    hFE: 'yi1',
    LFA: 'chi2',
    QFI: 'sui1',
    gFM: 'wei2',
    iFA: 'xi1',
    VFA: 'shi1',
    VFI: 'shuai1',
    cFI: 'rui2',
    dFI: 'gui1',
    NFI: 'zui1',
    TFI: 'chuai1',
    JFI: 'zhui1',
    YFA: 'chi1',
    SFA: 'zhi1',
    BFA: 'pi1',
    lFI: 'wei2',
    UFA: 'chi2',
    XGA: 'zhi1',
    lGA: 'yi2',
    VGA: 'shi1',
    CGA: 'pi2',
    NGA: 'zi1',
    dGE: 'ji1',
    YGA: 'chi1',
    KGA: 'chi1',
    OGA: 'ci1',
    PGA: 'ci2',
    MGA: 'ni2',
    LGA: 'chi2',
    QGA: 'si1',
    aGA: 'shi1',
    fGE: 'qi2',
    hGA: 'yi1',
    IGA: 'li2',
    fGI: 'kui2',
    JGI: 'zhui1',
    dGM: 'gui1',
    cGI: 'rui2',
    VGI: 'shuai1',
    lGI: 'wei2',
    IGI: 'lei2',
    QGI: 'sui1',
    fGM: 'kui2',
    DGE: 'mi2',
    AGE: 'bi1',
    XGI: 'zhui1',
    ZGI: 'chui2',
    kGI: 'wei2',
    CGE: 'pi2',
    BGE: 'pi1',
    iGI: 'hui1',
    LGI: 'chui2',
    JGA: 'zhi1',
    BGA: 'pi1',
    NGI: 'zui1',
    eGM: 'kui1',
    gGE: 'yi2',
    iGA: 'xi1',
    XHA: 'zhi1',
    lHA: 'yi2',
    ZHA: 'chi2',
    gHA: 'yi2',
    QHA: 'si1',
    THA: 'chi1',
    fHA: 'qi2',
    aHA: 'shi1',
    cHA: 'er2',
    eHA: 'qi1',
    dHA: 'ji1',
    RHA: 'si2',
    IHA: 'li2',
    SHA: 'zhi1',
    iHA: 'xi1',
    hHA: 'yi1',
    KHA: 'chi1',
    LHA: 'chi2',
    YHA: 'chi1',
    PHA: 'ci2',
    NHA: 'zi1',
    UHA: 'chi2',
    WHA: 'chi2',
    DIA: 'wei2',
    iII: 'hui1',
    kII: 'wei2',
    BIA: 'fei1',
    AIA: 'fei1',
    CIA: 'fei2',
    hII: 'wei1',
    fIA: 'qi2',
    dIA: 'ji1',
    iIA: 'xi1',
    hIA: 'yi1',
    gIA: 'yi2',
    gII: 'wei2',
    dII: 'gui1',
    eII: 'kui1',
    gJA: 'yu2',
    TJA: 'chu1',
    aJA: 'shu1',
    dJA: 'ju1',
    fJA: 'qu2',
    lJA: 'yu2',
    QJA: 'xu1',
    OJA: 'qu1',
    UJA: 'chu2',
    KJA: 'chu1',
    VJA: 'shu1',
    iJA: 'xu1',
    RJA: 'xu2',
    hJA: 'yu1',
    JJA: 'zhu1',
    IJA: 'lü2',
    XJA: 'zhu1',
    LJA: 'chu2',
    cJA: 'ru2',
    NJA: 'ju1',
    eJA: 'qu1',
    SJA: 'zhu1',
    ZJA: 'chu2',
    MJA: 'nü2',
    gKA: 'yu2',
    TKA: 'chu1',
    DKA: 'wu2',
    kKA: 'yu2',
    iKA: 'xu1',
    fKA: 'qu2',
    cKA: 'ru2',
    QKA: 'xu1',
    JKA: 'zhu1',
    KKA: 'chu1',
    ZKA: 'chu2',
    lKA: 'yu2',
    eKA: 'qu1',
    XKA: 'zhu1',
    OKA: 'qu1',
    IKA: 'lü2',
    CKA: 'fu2',
    UKA: 'chu2',
    SKA: 'zhu1',
    BKA: 'fu1',
    NKA: 'ju1',
    AKA: 'fu1',
    hKA: 'yu1',
    aKA: 'shu1',
    YKA: 'chu1',
    LKA: 'chu2',
    dKA: 'ju1',
    VKA: 'shu1',
    DLA: 'mu2',
    CLA: 'pu2',
    jLA: 'hu2',
    dLA: 'gu1',
    GLA: 'tu2',
    HLA: 'nu2',
    iLA: 'hu1',
    gLA: 'wu2',
    NLA: 'zu1',
    ILA: 'lu2',
    QLA: 'su1',
    PLA: 'cu2',
    hLA: 'wu1',
    ALA: 'bu1',
    eLA: 'ku1',
    OLA: 'cu1',
    FLA: 'tu1',
    ELA: 'du1',
    BLA: 'pu1',
    PMA: 'qi2',
    IMA: 'li2',
    OMA: 'qi1',
    EMA: 'di1',
    GMA: 'ti2',
    AMA: 'bi1',
    dMA: 'ji1',
    jMA: 'xi2',
    hMA: 'yi1',
    gMA: 'yi2',
    iMA: 'xi1',
    QMA: 'xi1',
    FMA: 'ti1',
    CMA: 'pi2',
    BMA: 'pi1',
    NMA: 'ji1',
    DMA: 'mi2',
    HMA: 'ni2',
    eMA: 'qi1',
    dMI: 'gui1',
    eMI: 'kui1',
    jMI: 'hui2',
    hMI: 'wei1',
    iMI: 'hui1',
    dPA: 'jia1',
    jPA: 'xia2',
    CPA: 'pai2',
    dPI: 'gua1',
    hPI: 'wa1',
    ePI: 'kua1',
    UPA: 'chai2',
    TPA: 'chai1',
    iPI: 'hua1',
    MPA: 'nai2',
    gPA: 'ya2',
    hPA: 'ya1',
    VPA: 'shai1',
    iPA: 'xia1',
    KPA: 'chai1',
    DPA: 'mai2',
    jPI: 'hua2',
    dQA: 'jie1',
    hQA: 'ye1',
    jQA: 'xie2',
    CQA: 'pai2',
    dQI: 'guai1',
    jQI: 'huai2',
    eQI: 'kuai1',
    UQA: 'chai2',
    TQA: 'chai1',
    iQI: 'huai1',
    LQI: 'chuai2',
    DQA: 'mai2',
    SQA: 'zhai1',
    hQI: 'wai1',
    JQA: 'zhai1',
    IQA: 'lai2',
    eQA: 'qie1',
    MQA: 'nai2',
    VQA: 'shai1',
    gQA: 'ye2',
    KQA: 'chai1',
    iQA: 'xie1',
    IQI: 'lai2',
    UQI: 'chuai2',
    iSI: 'hui1',
    eSI: 'kui1',
    hSI: 'wei1',
    jSI: 'hui2',
    DSA: 'mei2',
    dSI: 'gui1',
    ISI: 'lei2',
    GSI: 'tui2',
    OSI: 'cui1',
    ESI: 'dui1',
    QSI: 'sui1',
    PSI: 'cui2',
    CSA: 'pei2',
    ASA: 'bei1',
    BSA: 'pei1',
    gSI: 'wei2',
    FSI: 'tui1',
    HSI: 'nei2',
    NSI: 'zui1',
    iTA: 'hai1',
    eTA: 'kai1',
    hTA: 'ai1',
    GTA: 'tai2',
    dTA: 'gai1',
    PTA: 'cai2',
    ITA: 'lai2',
    NTA: 'zai1',
    OTA: 'cai1',
    CTA: 'pei2',
    FTA: 'tai1',
    jTA: 'hai2',
    QTA: 'sai1',
    gTA: 'ai2',
    HTA: 'nai2',
    ETA: 'dai1',
    BTA: 'pei1',
    XVA: 'zhen1',
    KVA: 'chen1',
    hVA: 'yin1',
    QVA: 'xin1',
    ZVA: 'chen2',
    cVA: 'ren2',
    bVA: 'chen2',
    OVA: 'qin1',
    kVA: 'yin2',
    aVA: 'shen1',
    AVA: 'bin1',
    IVA: 'lin2',
    fVE: 'qin2',
    JVA: 'zhen1',
    LVA: 'chen2',
    NVA: 'jin1',
    YVA: 'chen1',
    PVA: 'qin2',
    lVA: 'yin2',
    MVA: 'nin2',
    BVA: 'pin1',
    CVA: 'pin2',
    gVE: 'yin2',
    dVE: 'jin1',
    dVM: 'jun1',
    kVI: 'yun2',
    eVM: 'qun1',
    DVE: 'min2',
    CVE: 'pin2',
    hVE: 'yin1',
    hVM: 'yun1',
    AVE: 'bin1',
    DVA: 'min2',
    XVI: 'zhun1',
    KVI: 'chun1',
    LVI: 'chun2',
    QVI: 'xun1',
    ZVI: 'chun2',
    cVI: 'run2',
    bVI: 'chun2',
    IVI: 'lun2',
    JVI: 'zhun1',
    OVI: 'qun1',
    NVI: 'jun1',
    YVI: 'chun1',
    PVI: 'qun2',
    lVI: 'yun2',
    RVI: 'xun2',
    dVI: 'jun1',
    fVA: 'qin2',
    BVE: 'pin1',
    SWA: 'zhen1',
    VWA: 'shen1',
    UWA: 'chen2',
    DXA: 'wen2',
    kXI: 'yun2',
    hXI: 'yun1',
    CXA: 'fen2',
    AXA: 'fen1',
    fXI: 'qun2',
    iXI: 'xun1',
    dXI: 'jun1',
    BXA: 'fen1',
    iYA: 'xin1',
    hYA: 'yin1',
    fYA: 'qin2',
    dYA: 'jin1',
    gYA: 'yin2',
    gZI: 'yuan2',
    kZI: 'yuan2',
    CZA: 'fan2',
    BZA: 'fan1',
    iZI: 'xuan1',
    hZI: 'yuan1',
    gZA: 'yan2',
    eZA: 'qian1',
    iZA: 'xian1',
    dZA: 'jian1',
    hZA: 'yan1',
    AZA: 'fan1',
    fZA: 'qian2',
    DZA: 'wan2',
    jaI: 'hun2',
    daI: 'gun1',
    haI: 'wen1',
    DaA: 'men2',
    QaI: 'sun1',
    NaI: 'zun1',
    PaI: 'cun2',
    EaI: 'dun1',
    FaI: 'tun1',
    GaI: 'tun2',
    OaI: 'cun1',
    gaI: 'wen2',
    CaA: 'pen2',
    AaA: 'ben1',
    IaI: 'lun2',
    eaI: 'kun1',
    iaI: 'hun1',
    BaA: 'pen1',
    HaI: 'nun2',
    jbA: 'hen2',
    dbA: 'gen1',
    hbA: 'en1',
    FbA: 'tun1',
    gbA: 'en2',
    jcA: 'han2',
    gcA: 'an2',
    EcA: 'dan1',
    hcA: 'an1',
    HcA: 'nan2',
    OcA: 'can1',
    FcA: 'tan1',
    QcA: 'san1',
    GcA: 'tan2',
    PcA: 'can2',
    dcA: 'gan1',
    IcA: 'lan2',
    ecA: 'kan1',
    icA: 'han1',
    HcI: 'nuan2',
    jcI: 'huan2',
    gcI: 'wan2',
    EcI: 'duan1',
    hcI: 'wan1',
    FcI: 'tuan1',
    QcI: 'suan1',
    GcI: 'tuan2',
    PcI: 'cuan2',
    dcI: 'guan1',
    IcI: 'luan2',
    icI: 'huan1',
    ecI: 'kuan1',
    NcI: 'zuan1',
    CcA: 'pan2',
    DcA: 'man2',
    BcA: 'pan1',
    AcA: 'ban1',
    VdA: 'shan1',
    ddI: 'guan1',
    hdI: 'wan1',
    VdI: 'shuan1',
    jdI: 'huan2',
    AdA: 'ban1',
    DdA: 'man2',
    gdA: 'yan2',
    ddA: 'jian1',
    BdA: 'pan1',
    MdI: 'nuan2',
    gdI: 'wan2',
    edA: 'qian1',
    SdI: 'zhuan1',
    VeA: 'shan1',
    deI: 'guan1',
    deA: 'jian1',
    jeA: 'xian2',
    eeA: 'qian1',
    UeA: 'chan2',
    ieA: 'xian1',
    geA: 'yan2',
    heA: 'yan1',
    AeA: 'ban1',
    IeA: 'lan2',
    MeA: 'nan2',
    IeI: 'luan2',
    JeA: 'zhan1',
    LeA: 'chan2',
    jeI: 'huan2',
    QgA: 'xian1',
    PgA: 'qian2',
    OgA: 'qian1',
    NgA: 'jian1',
    FgA: 'tian1',
    dgA: 'jian1',
    jgA: 'xian2',
    hgA: 'yan1',
    IgA: 'lian2',
    GgA: 'tian2',
    HgA: 'nian2',
    EgA: 'dian1',
    egA: 'qian1',
    ggA: 'yan2',
    DgA: 'mian2',
    CgA: 'pian2',
    hgI: 'yuan1',
    dgI: 'juan1',
    igI: 'xuan1',
    AgA: 'bian1',
    jgI: 'xuan2',
    igA: 'xian1',
    QfA: 'xian1',
    PfA: 'qian2',
    OfA: 'qian1',
    NfA: 'jian1',
    cfA: 'ran2',
    lfA: 'yan2',
    XfA: 'zhan1',
    dfA: 'jian1',
    JfA: 'zhan1',
    UfA: 'chan2',
    afA: 'shan1',
    KfA: 'chan1',
    ZfA: 'chan2',
    LfA: 'chan2',
    ifA: 'xian1',
    IfA: 'lian2',
    BfA: 'pian1',
    CfA: 'pian2',
    DfA: 'mian2',
    PfI: 'quan2',
    QfI: 'xuan1',
    NfI: 'juan1',
    ifI: 'xuan1',
    cfI: 'ruan2',
    YfI: 'chuan1',
    lfI: 'yuan2',
    RfI: 'xuan2',
    hfI: 'yuan1',
    bfI: 'chuan2',
    AfA: 'bian1',
    RfA: 'xian2',
    OfI: 'quan1',
    XfI: 'zhuan1',
    ZfI: 'chuan2',
    kfI: 'yuan2',
    SfI: 'zhuan1',
    VfI: 'shuan1',
    KfI: 'chuan1',
    ffE: 'qian2',
    efE: 'qian1',
    ffM: 'quan2',
    LfI: 'chuan2',
    IfI: 'luan2',
    efM: 'quan1',
    hfE: 'yan1',
    kfA: 'yan2',
    hfM: 'yuan1',
    YfA: 'chan2',
    JfI: 'zhuan1',
    dfM: 'juan1',
    QhA: 'xiao1',
    FhA: 'tiao1',
    EhA: 'diao1',
    GhA: 'tiao2',
    dhA: 'jiao1',
    IhA: 'liao2',
    ghA: 'yao2',
    ihA: 'xiao1',
    hhA: 'yao1',
    ehA: 'qiao1',
    QiA: 'xiao1',
    KiA: 'chao1',
    JiA: 'zhao1',
    LiA: 'chao2',
    iiE: 'xiao1',
    PiA: 'qiao2',
    diE: 'jiao1',
    NiA: 'jiao1',
    ciA: 'rao2',
    aiA: 'shao1',
    liA: 'yao2',
    ZiA: 'chao2',
    XiA: 'zhao1',
    AiA: 'biao1',
    AiE: 'biao1',
    CiA: 'piao2',
    DiA: 'miao2',
    DiE: 'miao2',
    hiA: 'yao1',
    kiA: 'yao2',
    fiE: 'qiao2',
    OiA: 'qiao1',
    hiE: 'yao1',
    eiA: 'qiao1',
    YiA: 'chao1',
    BiA: 'piao1',
    fiA: 'qiao2',
    IiA: 'liao2',
    eiE: 'qiao1',
    jjA: 'xiao2',
    djA: 'jiao1',
    UjA: 'chao2',
    MjA: 'nao2',
    VjA: 'shao1',
    DjA: 'mao2',
    ijA: 'xiao1',
    AjA: 'bao1',
    BjA: 'pao1',
    ejA: 'qiao1',
    gjA: 'yao2',
    SjA: 'zhao1',
    JjA: 'zhao1',
    TjA: 'chao1',
    CjA: 'pao2',
    hjA: 'yao1',
    KjA: 'chao1',
    IjA: 'lao2',
    LjA: 'chao2',
    jkA: 'hao2',
    dkA: 'gao1',
    IkA: 'lao2',
    ikA: 'hao1',
    DkA: 'mao2',
    FkA: 'tao1',
    EkA: 'dao1',
    QkA: 'sao1',
    CkA: 'pao2',
    AkA: 'bao1',
    GkA: 'tao2',
    NkA: 'zao1',
    gkA: 'ao2',
    PkA: 'cao2',
    hkA: 'ao1',
    HkA: 'nao2',
    ekA: 'kao1',
    OkA: 'cao1',
    BkA: 'pao1',
    dlA: 'ge1',
    OlA: 'cuo1',
    ElA: 'duo1',
    QlA: 'suo1',
    GlA: 'tuo2',
    PlA: 'cuo2',
    glA: 'e2',
    FlA: 'tuo1',
    IlA: 'luo2',
    HlA: 'nuo2',
    jlA: 'he2',
    ilA: 'he1',
    elA: 'ke1',
    hlA: 'e1',
    dlI: 'guo1',
    OlI: 'cuo1',
    ElI: 'duo1',
    QlI: 'suo1',
    ClA: 'po2',
    GlI: 'tuo2',
    DlA: 'mo2',
    PlI: 'cuo2',
    glI: 'wo2',
    FlI: 'tuo1',
    IlI: 'luo2',
    HlI: 'nuo2',
    AlA: 'bo1',
    BlA: 'po1',
    jlI: 'huo2',
    elI: 'kuo1',
    hlI: 'wo1',
    imI: 'xue1',
    hmI: 'yue1',
    emI: 'que1',
    fmA: 'qie2',
    emA: 'qie1',
    dmA: 'jie1',
    OmI: 'que1',
    NmI: 'jue1',
    fmI: 'que2',
    ImI: 'lüe2',
    DnA: 'ma2',
    YoA: 'che1',
    aoA: 'she1',
    loA: 'ye2',
    XoA: 'zhe1',
    NoA: 'jie1',
    boA: 'che2',
    jnI: 'hua2',
    dnI: 'gua1',
    inI: 'hua1',
    enI: 'kua1',
    MnA: 'na2',
    dnA: 'jia1',
    jnA: 'xia2',
    BnA: 'pa1',
    hnA: 'ya1',
    AnA: 'ba1',
    TnA: 'cha1',
    VnA: 'sha1',
    gnA: 'ya2',
    SnA: 'zha1',
    LnA: 'cha2',
    RoA: 'xie2',
    ZoA: 'che2',
    hnI: 'wa1',
    SnI: 'zhua1',
    JnI: 'zhua1',
    CnA: 'pa2',
    UnA: 'cha2',
    KnA: 'cha1',
    JnA: 'zha1',
    inA: 'xia1',
    enA: 'qia1',
    PoA: 'qie2',
    coA: 're2',
    QoA: 'xie1',
    JoA: 'zhe1',
    gnI: 'wa2',
    lpA: 'yang2',
    RpA: 'xiang2',
    IpA: 'liang2',
    ipA: 'xiang1',
    apA: 'shang1',
    CpA: 'fang2',
    XpA: 'zhang1',
    YpA: 'chang1',
    epA: 'qiang1',
    dpA: 'jiang1',
    LpA: 'chang2',
    JpA: 'zhang1',
    cpA: 'rang2',
    ApA: 'fang1',
    QpA: 'xiang1',
    NpA: 'jiang1',
    TpA: 'chuang1',
    DpA: 'wang2',
    MpA: 'niang2',
    UpA: 'chuang2',
    SpA: 'zhuang1',
    ZpA: 'chang2',
    VpA: 'shuang1',
    PpA: 'qiang2',
    OpA: 'qiang1',
    epI: 'kuang1',
    kpI: 'wang2',
    hpA: 'yang1',
    fpA: 'qiang2',
    KpA: 'chang1',
    BpA: 'fang1',
    fpI: 'kuang2',
    GqA: 'tang2',
    IqA: 'lang2',
    EqA: 'dang1',
    OqA: 'cang1',
    dqA: 'gang1',
    QqA: 'sang1',
    eqA: 'kang1',
    iqI: 'huang1',
    jqI: 'huang2',
    dqI: 'guang1',
    FqA: 'tang1',
    BqA: 'pang1',
    hqI: 'wang1',
    hqA: 'ang1',
    iqA: 'hang1',
    jqA: 'hang2',
    DqA: 'mang2',
    NqA: 'zang1',
    HqA: 'nang2',
    CqA: 'pang2',
    gqA: 'ang2',
    PqA: 'cang2',
    eqI: 'kuang1',
    AqA: 'bang1',
    drA: 'geng1',
    erA: 'keng1',
    DrA: 'meng2',
    jrI: 'hong2',
    ArA: 'beng1',
    irI: 'hong1',
    drI: 'gong1',
    CrA: 'peng2',
    irA: 'heng1',
    KrA: 'cheng1',
    TrA: 'cheng1',
    UrA: 'cheng2',
    hsA: 'ying1',
    BrA: 'ping1',
    CsA: 'ping2',
    dsA: 'jing1',
    DsA: 'ming2',
    LrA: 'cheng2',
    JrA: 'zheng1',
    ksI: 'yong2',
    AsA: 'bing1',
    isI: 'xiong1',
    esA: 'qing1',
    VrA: 'sheng1',
    fsA: 'qing2',
    gsA: 'ying2',
    jrA: 'heng2',
    MrA: 'neng2',
    dtA: 'geng1',
    etA: 'keng1',
    DtA: 'meng2',
    AtA: 'beng1',
    jtI: 'hong2',
    jtA: 'heng2',
    JtA: 'zheng1',
    htA: 'eng1',
    UtA: 'cheng2',
    TtA: 'cheng1',
    MtA: 'neng2',
    BtA: 'peng1',
    itI: 'hong1',
    LtA: 'cheng2',
    htI: 'weng1',
    CtA: 'peng2',
    gtA: 'eng2',
    StA: 'zheng1',
    OuA: 'qing1',
    PuA: 'qing2',
    NuA: 'jing1',
    luA: 'ying2',
    luI: 'yong2',
    huA: 'ying1',
    JuA: 'zheng1',
    KuA: 'cheng1',
    ZuA: 'cheng2',
    LuA: 'cheng2',
    auA: 'sheng1',
    XuA: 'zheng1',
    euA: 'qing1',
    DuA: 'ming2',
    IuA: 'ling2',
    AuA: 'bing1',
    euI: 'qiong1',
    RuA: 'xing2',
    huI: 'yong1',
    fuI: 'qiong2',
    QuI: 'xing1',
    fuA: 'qing2',
    iuI: 'xiong1',
    OvA: 'qing1',
    dvA: 'jing1',
    jvA: 'xing2',
    GvA: 'ting2',
    EvA: 'ding1',
    ivA: 'xing1',
    QvA: 'xing1',
    BvA: 'ping1',
    IvA: 'ling2',
    HvA: 'ning2',
    FvA: 'ting1',
    DvA: 'ming2',
    CvA: 'ping2',
    jvI: 'xiong2',
    dvI: 'jiong1',
    XwA: 'zheng1',
    ZwA: 'cheng2',
    LwA: 'cheng2',
    IwA: 'ling2',
    hwA: 'ying1',
    CwA: 'ping2',
    AwA: 'bing1',
    lwA: 'ying2',
    bwA: 'cheng2',
    awA: 'sheng1',
    cwA: 'reng2',
    dwA: 'jing1',
    JwA: 'zheng1',
    PwA: 'qing2',
    gwA: 'ying2',
    iwA: 'xing1',
    YwA: 'cheng1',
    fwA: 'qing2',
    VwA: 'sheng1',
    ewA: 'qing1',
    KwA: 'cheng1',
    UwA: 'cheng2',
    BwA: 'ping1',
    ExA: 'deng1',
    IxA: 'leng2',
    QxA: 'seng1',
    AxA: 'beng1',
    NxA: 'zeng1',
    DxA: 'meng2',
    PxA: 'ceng2',
    CxA: 'peng2',
    jxI: 'hong2',
    dxI: 'gong1',
    ixI: 'hong1',
    HxA: 'neng2',
    GxA: 'teng2',
    jxA: 'heng2',
    dxA: 'geng1',
    FxA: 'teng1',
    BxA: 'peng1',
    kyA: 'you2',
    hyA: 'you1',
    IyA: 'liu2',
    OyA: 'qiu1',
    lyA: 'you2',
    gyA: 'you2',
    NyA: 'jiu1',
    PyA: 'qiu2',
    QyA: 'xiu1',
    KyA: 'chou1',
    eyA: 'qiu1',
    YyA: 'chou1',
    XyA: 'zhou1',
    ZyA: 'chou2',
    cyA: 'rou2',
    ayA: 'shou1',
    ByA: 'fou1',
    dyA: 'jiu1',
    AyA: 'fou1',
    VyA: 'shou1',
    TyA: 'chou1',
    SyA: 'zhou1',
    UyA: 'chou2',
    iyA: 'xiu1',
    RyA: 'xiu2',
    LyA: 'chou2',
    JyA: 'zhou1',
    fyA: 'qiu2',
    CyA: 'fou2',
    DyA: 'miu2',
    jzA: 'hou2',
    hzA: 'ou1',
    HzA: 'nou2',
    IzA: 'lou2',
    QzA: 'sou1',
    ezA: 'kou1',
    izA: 'hou1',
    NzA: 'zou1',
    FzA: 'tou1',
    GzA: 'tou2',
    gzA: 'ou2',
    dzA: 'gou1',
    EzA: 'dou1',
    PzA: 'cou2',
    CzA: 'pou2',
    OzA: 'cou1',
    DzA: 'mou2',
    h0A: 'you1',
    f0A: 'qiu2',
    A0A: 'biao1',
    I0A: 'liu2',
    d0A: 'jiu1',
    C0A: 'piao2',
    N0A: 'jiu1',
    V0A: 'shou1',
    g0A: 'you2',
    i0A: 'xiu1',
    D0A: 'miao2',
    O1A: 'qin1',
    R1A: 'xin2',
    I1A: 'lin2',
    K1A: 'chen1',
    X1A: 'zhen1',
    L1A: 'chen2',
    J1A: 'zhen1',
    Z1A: 'chen2',
    c1A: 'ren2',
    a1A: 'shen1',
    l1A: 'yin2',
    Q1A: 'xin1',
    h1A: 'yin1',
    N1A: 'jin1',
    P1A: 'qin2',
    M1A: 'nin2',
    f1E: 'qin2',
    e1E: 'qin1',
    g1E: 'yin2',
    i1E: 'xin1',
    d1E: 'jin1',
    h1E: 'yin1',
    V1A: 'shen1',
    U1A: 'chen2',
    S1A: 'zhen1',
    T1A: 'chen1',
    Y1A: 'chen1',
    G2A: 'tan2',
    O2A: 'can1',
    H2A: 'nan2',
    h2A: 'an1',
    j2A: 'han2',
    I2A: 'lan2',
    P2A: 'can2',
    N2A: 'zan1',
    F2A: 'tan1',
    E2A: 'dan1',
    e2A: 'kan1',
    i2A: 'han1',
    Q2A: 'san1',
    d2A: 'gan1',
    g2A: 'an2',
    G3A: 'tan2',
    d3A: 'gan1',
    E3A: 'dan1',
    Q3A: 'san1',
    I3A: 'lan2',
    e3A: 'kan1',
    F3A: 'tan1',
    P3A: 'can2',
    j3A: 'han2',
    D3A: 'man2',
    i3A: 'han1',
    l4A: 'yan2',
    I4A: 'lian2',
    A4E: 'bian1',
    Q4A: 'xian1',
    O4A: 'qian1',
    X4A: 'zhan1',
    Z4A: 'chan2',
    a4A: 'shan1',
    Y4A: 'chan1',
    c4A: 'ran2',
    M4A: 'nian2',
    k4A: 'yan2',
    J4A: 'zhan1',
    K4A: 'chan1',
    h4E: 'yan1',
    e4E: 'qian1',
    g4E: 'yan2',
    N4A: 'jian1',
    P4A: 'qian2',
    f4E: 'qian2',
    h4A: 'yan1',
    R4A: 'xian2',
    V4A: 'shan1',
    L4A: 'chan2',
    f4A: 'qian2',
    F5A: 'tian1',
    E5A: 'dian1',
    G5A: 'tian2',
    I5A: 'lian2',
    e5A: 'qian1',
    d5A: 'jian1',
    j5A: 'xian2',
    H5A: 'nian2',
    i5A: 'xian1',
    j6A: 'xian2',
    d6A: 'jian1',
    V6A: 'shan1',
    h6A: 'yan1',
    g6A: 'yan2',
    i6A: 'xian1',
    J6A: 'zhan1',
    M6A: 'nan2',
    U6A: 'chan2',
    e6A: 'qian1',
    j7A: 'xian2',
    U7A: 'chan2',
    g7A: 'yan2',
    T7A: 'chan1',
    V7A: 'shan1',
    d7A: 'jian1',
    C7A: 'pan2',
    e7A: 'qian1',
    g8A: 'yan2',
    i8A: 'xian1',
    h8A: 'yan1',
    e8A: 'qian1',
    C9A: 'fan2',
    B9A: 'fan1',
    EAB: 'dong3',
    DAB: 'meng3',
    eAB: 'kong3',
    QAB: 'song3',
    FAB: 'tong3',
    NAB: 'zong3',
    jAB: 'hong4',
    hAB: 'weng3',
    HAB: 'nong3',
    AAB: 'beng3',
    IAB: 'long3',
    iAB: 'hong3',
    GAB: 'dong4',
    CAB: 'beng4',
    XDB: 'zhong3',
    KDB: 'chong3',
    IDB: 'long3',
    hDB: 'yong3',
    cDB: 'rong3',
    LDB: 'zhong4',
    JDB: 'zhong3',
    CDB: 'feng4',
    BDB: 'feng3',
    lDB: 'yong3',
    eDB: 'qiong3',
    ZDB: 'shung4',
    dDB: 'jiong3',
    QDB: 'song3',
    iDB: 'xiong3',
    ADB: 'feng3',
    ECB: 'dong3',
    DCB: 'meng3',
    fDB: 'jiong4',
    YDB: 'chong3',
    NDB: 'zong3',
    dEB: 'jiang3',
    CEB: 'bang4',
    hEB: 'yang3',
    DEB: 'mang3',
    jEB: 'xiang4',
    AEB: 'bang3',
    iEB: 'xiang3',
    XFB: 'zhi3',
    ZFB: 'shi4',
    DFF: 'mi3',
    AFF: 'bi3',
    CFF: 'pi4',
    iFN: 'hui3',
    hFN: 'wei3',
    eFN: 'kui3',
    dFN: 'gui3',
    QFJ: 'sui3',
    IFJ: 'lei3',
    fFF: 'ji4',
    hFF: 'yi3',
    dFF: 'ji3',
    eFF: 'qi3',
    gFF: 'yi3',
    kFJ: 'wei3',
    NFJ: 'zui3',
    cFJ: 'rui3',
    OFB: 'ci3',
    LFB: 'zhi4',
    QFB: 'si3',
    lFB: 'yi3',
    IFB: 'li3',
    VFB: 'shi3',
    AFB: 'bi3',
    cFB: 'er3',
    DFB: 'mi3',
    CFB: 'pi4',
    YFB: 'chi3',
    aFB: 'shi3',
    NFB: 'zi3',
    XFJ: 'zhui3',
    TFJ: 'chuai3',
    RFJ: 'sui4',
    bFB: 'shi4',
    SFB: 'zhi3',
    BFF: 'pi3',
    BFB: 'pi3',
    lFJ: 'wei3',
    PFJ: 'zui4',
    eFJ: 'kui3',
    MFB: 'ni3',
    gFN: 'wei3',
    fFN: 'gui4',
    KFB: 'chi3',
    JFB: 'zhi3',
    iFF: 'xi3',
    eFB: 'kui3',
    ZFJ: 'shui4',
    dFB: 'ji3',
    XGB: 'zhi3',
    ZGB: 'shi4',
    DGF: 'mi3',
    AGF: 'bi3',
    RGB: 'si4',
    dGF: 'ji3',
    NGB: 'zi3',
    AGB: 'bi3',
    dGN: 'gui3',
    kGJ: 'wei3',
    aGB: 'shi3',
    LGB: 'zhi4',
    QGB: 'si3',
    CGB: 'bi4',
    IGB: 'li3',
    aGJ: 'shui3',
    IGJ: 'lei3',
    fGJ: 'gui4',
    OGJ: 'cui3',
    MGB: 'ni3',
    dGJ: 'gui3',
    CGF: 'bi4',
    PGJ: 'zui4',
    BGF: 'pi3',
    cGJ: 'rui3',
    lGJ: 'wei3',
    hGF: 'yi3',
    NGJ: 'zui3',
    JGB: 'zhi3',
    KGB: 'chi3',
    iGJ: 'hui3',
    fGN: 'gui3',
    fGF: 'ji4',
    eGN: 'kui3',
    XHB: 'zhi3',
    ZHB: 'shi4',
    JHB: 'zhi3',
    iHB: 'xi3',
    dHB: 'ji3',
    lHB: 'yi3',
    RHB: 'si4',
    VHB: 'shi3',
    cHB: 'er3',
    IHB: 'li3',
    QHB: 'si3',
    aHB: 'shi3',
    LHB: 'zhi4',
    eHB: 'qi3',
    UHB: 'shi4',
    WHB: 'shi4',
    NHB: 'zi3',
    kHB: 'yi3',
    gHB: 'yi3',
    YHB: 'chi3',
    KHB: 'chi3',
    THB: 'chi3',
    SHB: 'zhi3',
    hHB: 'yi3',
    MHB: 'ni3',
    DIB: 'wei3',
    hIB: 'yi3',
    eIB: 'qi3',
    dIB: 'ji3',
    BIB: 'fei3',
    AIB: 'fei3',
    kIJ: 'wei3',
    dIJ: 'gui3',
    iIJ: 'hui3',
    gIB: 'yi3',
    iIB: 'xi3',
    hIJ: 'wei3',
    CIB: 'fei4',
    gJB: 'yu3',
    IJB: 'lü3',
    LJB: 'zhu4',
    lJB: 'yu3',
    XJB: 'zhu3',
    cJB: 'ru3',
    aJB: 'shu3',
    YJB: 'chu3',
    JJB: 'zhu3',
    QJB: 'xu3',
    KJB: 'chu3',
    MJB: 'nü3',
    iJB: 'xu3',
    fJB: 'ju4',
    VJB: 'shu3',
    TJB: 'chu3',
    SJB: 'zhu3',
    UJB: 'shu4',
    PJB: 'ju4',
    hJB: 'yu3',
    dJB: 'ju3',
    RJB: 'xu4',
    eJB: 'qu3',
    bJB: 'shu4',
    ZJB: 'shu4',
    OJB: 'qu3',
    NJB: 'ju3',
    gKB: 'yu3',
    kKB: 'yu3',
    PKB: 'ju4',
    AKB: 'fu3',
    DKB: 'wu3',
    CKB: 'fu4',
    BKB: 'fu3',
    LKB: 'zhu4',
    iKB: 'xu3',
    ZKB: 'shu4',
    lKB: 'yu3',
    XKB: 'zhu3',
    hKB: 'yu3',
    eKB: 'qu3',
    JKB: 'zhu3',
    cKB: 'ru3',
    fKB: 'ju4',
    VKB: 'shu3',
    dKB: 'ju3',
    OKB: 'qu3',
    IKB: 'lü3',
    QKB: 'xu3',
    UKB: 'shu4',
    DLB: 'mu3',
    FLB: 'tu3',
    GLB: 'du4',
    ILB: 'lu3',
    OLB: 'cu3',
    ELB: 'du3',
    dLB: 'gu3',
    gLB: 'wu3',
    CLB: 'bu4',
    PLB: 'zu4',
    NLB: 'zu3',
    iLB: 'hu3',
    hLB: 'wu3',
    eLB: 'ku3',
    HLB: 'nu3',
    jLB: 'hu4',
    BLB: 'pu3',
    ALB: 'bu3',
    PMB: 'ji4',
    IMB: 'li3',
    FMB: 'ti3',
    BMB: 'pi3',
    NMB: 'ji3',
    EMB: 'di3',
    GMB: 'di4',
    HMB: 'ni3',
    QMB: 'xi3',
    OMB: 'qi3',
    eMB: 'qi3',
    jMB: 'xi3',
    DMB: 'mi3',
    CMB: 'bi4',
    hMB: 'yi3',
    gMB: 'yi3',
    AMB: 'bi3',
    jPB: 'xia4',
    DPB: 'mai3',
    ePB: 'qia3',
    LPB: 'zhai4',
    MPB: 'nai3',
    CPB: 'bai3',
    hPB: 'ya3',
    APB: 'bai3',
    dPB: 'jia3',
    VPB: 'shai3',
    dPJ: 'gua3',
    LPJ: 'zhuai4',
    jPJ: 'hua4',
    iPJ: 'hua3',
    jQB: 'xie4',
    eQB: 'qie3',
    gQB: 'ye3',
    hQB: 'ye3',
    iSJ: 'hui3',
    hSJ: 'wei3',
    ISJ: 'lei3',
    GSJ: 'dui4',
    PSJ: 'zui4',
    DSB: 'mei3',
    FSJ: 'tui3',
    jSJ: 'hui4',
    eSJ: 'kui3',
    ESJ: 'dui3',
    HSJ: 'nei3',
    gSJ: 'wei3',
    OSJ: 'cui3',
    CSB: 'bei4',
    NSJ: 'zui3',
    iTB: 'hai3',
    eTB: 'kai3',
    NTB: 'zai3',
    GTB: 'dai4',
    HTB: 'nai3',
    dTB: 'gai3',
    jTB: 'hai4',
    BTB: 'pei3',
    OTB: 'cai3',
    ETB: 'dai3',
    DTB: 'mei3',
    PTB: 'zai4',
    hTB: 'ai3',
    FTB: 'tai3',
    ITB: 'lai3',
    CTB: 'bei4',
    XVB: 'zhen3',
    KVB: 'chen3',
    ZVB: 'shen4',
    cVB: 'ren3',
    aVB: 'shen3',
    IVB: 'lin3',
    LVB: 'zhen4',
    dVB: 'jin3',
    PVB: 'jin4',
    NVB: 'jin3',
    CVB: 'bin4',
    gVF: 'yin3',
    OVB: 'qin3',
    fVN: 'jun4',
    lVB: 'yin3',
    DVF: 'min3',
    DVB: 'min3',
    kVJ: 'yun3',
    XVJ: 'zhun3',
    lVJ: 'yun3',
    QVJ: 'xun3',
    cVJ: 'run3',
    YVJ: 'chun3',
    bVJ: 'shun4',
    KVJ: 'chun3',
    IVJ: 'lun3',
    eVJ: 'qun3',
    eVB: 'qin3',
    aVJ: 'shun3',
    iVB: 'xin3',
    UVB: 'shen4',
    JVB: 'zhen3',
    DXB: 'wen3',
    AXB: 'fen3',
    CXB: 'fen4',
    BXB: 'fen3',
    hXJ: 'yun3',
    gXJ: 'yun3',
    kXJ: 'yun3',
    eXJ: 'qun3',
    hYB: 'yin3',
    dYB: 'jin3',
    SWB: 'zhen3',
    eYB: 'qin3',
    fYB: 'jin4',
    TWB: 'chen3',
    gYB: 'yin3',
    iYB: 'xin3',
    gZJ: 'yuan3',
    kZJ: 'yuan3',
    hZB: 'yan3',
    dZB: 'jian3',
    fZB: 'jian4',
    eZB: 'qian3',
    gZB: 'yan3',
    iZB: 'xian3',
    DZB: 'wan3',
    AZB: 'fan3',
    fZJ: 'juan4',
    hZJ: 'yuan3',
    eZJ: 'quan3',
    iZJ: 'xuan3',
    CZB: 'fan4',
    jaJ: 'hun4',
    BaB: 'pen3',
    OaJ: 'cun3',
    AaB: 'ben3',
    QaJ: 'sun3',
    NaJ: 'zun3',
    haJ: 'wen3',
    GaJ: 'dun4',
    PaJ: 'zun4',
    daJ: 'gun3',
    FaJ: 'tun3',
    eaJ: 'kun3',
    IaJ: 'lun3',
    CaB: 'ben4',
    DaB: 'men3',
    HaJ: 'nun3',
    iaJ: 'hun3',
    jbB: 'hen4',
    ebB: 'ken3',
    dbB: 'gen3',
    jcB: 'han4',
    EcB: 'dan3',
    QcB: 'san3',
    GcB: 'dan4',
    PcB: 'zan4',
    dcB: 'gan3',
    IcB: 'lan3',
    ecB: 'kan3',
    icB: 'han3',
    NcB: 'zan3',
    jcJ: 'huan4',
    EcJ: 'duan3',
    hcJ: 'wan3',
    FcJ: 'tuan3',
    QcJ: 'suan3',
    dcJ: 'guan3',
    IcJ: 'luan3',
    ecJ: 'kuan3',
    HcJ: 'nuan3',
    NcJ: 'zuan3',
    CcB: 'ban4',
    DcB: 'man3',
    AcB: 'ban3',
    GcJ: 'duan4',
    HcB: 'nan3',
    BcB: 'pan3',
    VdB: 'shan3',
    hdJ: 'wan3',
    AdB: 'ban3',
    SdB: 'zhan3',
    MdB: 'nan3',
    jdB: 'xian4',
    jdJ: 'huan4',
    CdB: 'ban4',
    DdB: 'man3',
    UdB: 'shan4',
    gdB: 'yan3',
    TdB: 'chan3',
    BdB: 'pan3',
    UdJ: 'shuan4',
    VeB: 'shan3',
    jeB: 'xian4',
    deB: 'jian3',
    TeB: 'chan3',
    UeB: 'shan4',
    geB: 'yan3',
    SeB: 'zhan3',
    TeJ: 'chuan3',
    eeB: 'qian3',
    QgB: 'xian3',
    FgB: 'tian3',
    EgB: 'dian3',
    hgB: 'yan3',
    GgB: 'dian4',
    dgB: 'jian3',
    jgB: 'xian4',
    igB: 'xian3',
    DgB: 'mian3',
    HgB: 'nian3',
    AgB: 'bian3',
    jgJ: 'xuan4',
    dgJ: 'juan3',
    CgB: 'bian4',
    egJ: 'quan3',
    egB: 'qian3',
    ggB: 'yan3',
    QfB: 'xian3',
    lfB: 'yan3',
    PfB: 'juan4',
    JfB: 'zhan3',
    XfB: 'zhan3',
    MfB: 'nian3',
    OfB: 'qian3',
    YfB: 'chan3',
    efB: 'qian3',
    dfF: 'jian3',
    ZfB: 'shan4',
    NfB: 'jian3',
    cfB: 'ran3',
    RfB: 'xian4',
    IfB: 'lian3',
    gfF: 'yan3',
    ffF: 'jian4',
    CfF: 'bian4',
    DfB: 'mian3',
    AfB: 'bian3',
    NfJ: 'juan3',
    PfJ: 'juan4',
    AfF: 'bian3',
    lfJ: 'yuan3',
    IfJ: 'luan3',
    JfJ: 'zhuan3',
    dfN: 'juan3',
    ffN: 'juan4',
    cfJ: 'ruan3',
    YfJ: 'chuan3',
    ZfJ: 'shuan4',
    LfJ: 'zhuan4',
    XfJ: 'zhuan3',
    QfJ: 'xuan3',
    UfJ: 'shuan4',
    ffJ: 'juan4',
    ifJ: 'xuan3',
    CfB: 'bian4',
    DfF: 'mian3',
    KfB: 'chan3',
    BfF: 'pian3',
    afB: 'shan3',
    hfF: 'yan3',
    LfB: 'zhan4',
    UfB: 'shuan4',
    QhB: 'xiao3',
    dhB: 'jiao3',
    EhB: 'diao3',
    IhB: 'liao3',
    FhB: 'tiao3',
    ihB: 'xiao3',
    hhB: 'yao3',
    HhB: 'niao3',
    jhB: 'xiao4',
    GhB: 'diao4',
    ehB: 'qiao3',
    NhB: 'jiao3',
    QiB: 'xiao3',
    LiB: 'zhao4',
    XiB: 'zhao3',
    hiF: 'yao3',
    KiB: 'chao3',
    aiB: 'shao3',
    ciB: 'rao3',
    CiB: 'biao4',
    YiB: 'chao3',
    BiB: 'piao3',
    DiB: 'miao3',
    ZiB: 'shao4',
    diF: 'jiao3',
    AiF: 'biao3',
    AiB: 'biao3',
    CiF: 'biao4',
    liB: 'yao3',
    OiB: 'qiao3',
    NiB: 'jiao3',
    fiF: 'jiao4',
    IiB: 'liao3',
    BiF: 'piao3',
    hiB: 'yao3',
    ejB: 'qiao3',
    jjB: 'xiao4',
    AjB: 'bao3',
    MjB: 'nao3',
    DjB: 'mao3',
    djB: 'jiao3',
    SjB: 'zhao3',
    hjB: 'yao3',
    CjB: 'bao4',
    gjB: 'yao3',
    UjB: 'shao4',
    TjB: 'chao3',
    JjB: 'zhao3',
    VjB: 'shao3',
    jkB: 'hao4',
    CkB: 'bao4',
    IkB: 'lao3',
    FkB: 'tao3',
    GkB: 'dao4',
    HkB: 'nao3',
    QkB: 'sao3',
    EkB: 'dao3',
    OkB: 'cao3',
    NkB: 'zao3',
    PkB: 'zao4',
    dkB: 'gao3',
    ikB: 'hao3',
    DkB: 'mao3',
    AkB: 'bao3',
    hkB: 'ao3',
    ekB: 'kao3',
    gkB: 'ao3',
    dlB: 'ge3',
    OlB: 'cuo3',
    ElB: 'duo3',
    QlB: 'suo3',
    GlB: 'duo4',
    glB: 'e3',
    FlB: 'tuo3',
    IlB: 'luo3',
    HlB: 'nuo3',
    jlB: 'he4',
    ilB: 'he3',
    elB: 'ke3',
    hlB: 'e3',
    NlB: 'zuo3',
    dlJ: 'guo3',
    ElJ: 'duo3',
    QlJ: 'suo3',
    GlJ: 'duo4',
    FlJ: 'tuo3',
    DlB: 'mo3',
    PlJ: 'zuo4',
    glJ: 'wo3',
    IlJ: 'luo3',
    hlJ: 'wo3',
    HlJ: 'nuo3',
    AlB: 'bo3',
    BlB: 'po3',
    jlJ: 'huo4',
    ilJ: 'huo3',
    elJ: 'kuo3',
    ClB: 'bo4',
    OlJ: 'cuo3',
    NlJ: 'zuo3',
    DnB: 'ma3',
    XoB: 'zhe3',
    loB: 'ye3',
    gnB: 'ya3',
    dnB: 'jia3',
    VnB: 'sha3',
    hnB: 'ya3',
    RoB: 'xie4',
    jnB: 'xia4',
    QoB: 'xie3',
    NmJ: 'jue3',
    DoB: 'mie3',
    OoB: 'qie3',
    inB: 'xia3',
    ZoB: 'she4',
    enB: 'qia3',
    CnB: 'ba4',
    aoB: 'she3',
    NoB: 'jie3',
    AnB: 'ba3',
    jnJ: 'hua4',
    dnJ: 'gua3',
    gnJ: 'wa3',
    coB: 're3',
    SnB: 'zha3',
    JnB: 'zha3',
    UnB: 'sha4',
    YoB: 'che3',
    MnB: 'na3',
    enJ: 'kua3',
    TnJ: 'chua3',
    KnJ: 'chua3',
    VnJ: 'shua3',
    KnB: 'cha3',
    InB: 'la3',
    lpB: 'yang3',
    RpB: 'xiang4',
    NpB: 'jiang3',
    IpB: 'liang3',
    hpB: 'yang3',
    fpB: 'jiang4',
    gpB: 'yang3',
    TpB: 'chuang3',
    QpB: 'xiang3',
    XpB: 'zhang3',
    VpB: 'shuang3',
    ipB: 'xiang3',
    YpB: 'chang3',
    dpB: 'jiang3',
    LpB: 'zhang4',
    KpB: 'chang3',
    dpJ: 'guang3',
    cpB: 'rang3',
    apB: 'shang3',
    BpB: 'fang3',
    DpB: 'wang3',
    ApB: 'fang3',
    hpJ: 'wang3',
    kpJ: 'wang3',
    ipJ: 'huang3',
    OpB: 'qiang3',
    JpB: 'zhang3',
    ZpB: 'shang4',
    fpJ: 'guang4',
    CpB: 'fang4',
    GqB: 'dang4',
    QqB: 'sang3',
    dqJ: 'guang3',
    AqB: 'bang3',
    NqB: 'zang3',
    HqB: 'nang3',
    jqB: 'hang4',
    FqB: 'tang3',
    DqB: 'mang3',
    EqB: 'dang3',
    IqB: 'lang3',
    hqB: 'ang3',
    eqB: 'kang3',
    hqJ: 'wang3',
    jqJ: 'huang4',
    BqB: 'pang3',
    iqJ: 'huang3',
    dqB: 'gang3',
    gqB: 'ang3',
    PqB: 'zang4',
    OqB: 'cang3',
    iqB: 'hang3',
    eqJ: 'kuang3',
    drB: 'geng3',
    AsB: 'bing3',
    dsB: 'jing3',
    hsB: 'ying3',
    VsB: 'sheng3',
    ksJ: 'yong3',
    isJ: 'xiong3',
    DsB: 'ming3',
    dsJ: 'jiong3',
    jrB: 'heng4',
    DrB: 'meng3',
    drJ: 'gong3',
    ArB: 'beng3',
    JrB: 'zheng3',
    LrB: 'zheng4',
    hrJ: 'weng3',
    IrB: 'leng3',
    jrJ: 'hong4',
    CrB: 'beng4',
    erJ: 'kong3',
    MrB: 'neng3',
    dtB: 'geng3',
    DtB: 'meng3',
    jtB: 'heng4',
    CtB: 'beng4',
    BtB: 'peng3',
    PuB: 'jing4',
    XuB: 'zheng3',
    KuB: 'cheng3',
    luB: 'ying3',
    fuB: 'jing4',
    luJ: 'yong3',
    IuB: 'ling3',
    duB: 'jing3',
    AuB: 'bing3',
    euJ: 'qiong3',
    NuB: 'jing3',
    huB: 'ying3',
    OuB: 'qing3',
    QuB: 'xing3',
    DuB: 'ming3',
    LuB: 'zheng3',
    jvJ: 'xiong4',
    dvJ: 'jiong3',
    DvB: 'ming3',
    EvB: 'ding3',
    GvB: 'ding4',
    FvB: 'ting3',
    PvB: 'jing4',
    hvJ: 'yong3',
    evB: 'qing3',
    HvB: 'ning3',
    jvB: 'xing4',
    QvB: 'xing3',
    AvB: 'bing3',
    evJ: 'qiong3',
    dvB: 'jing3',
    BvB: 'ping3',
    hvB: 'ying3',
    IvB: 'ling3',
    CvB: 'bing4',
    ivJ: 'xiong3',
    gvB: 'ying3',
    XwB: 'zheng3',
    KwB: 'cheng3',
    fwB: 'jing4',
    VwB: 'sheng3',
    ExB: 'deng3',
    BxB: 'peng3',
    exB: 'keng3',
    HxB: 'neng3',
    kyB: 'you3',
    IyB: 'liu3',
    MyB: 'niu3',
    KyB: 'chou3',
    JyB: 'zhou3',
    iyB: 'xiu3',
    dyB: 'jiu3',
    ayB: 'shou3',
    YyB: 'chou3',
    PyB: 'jiu4',
    CyB: 'fou4',
    AyB: 'fou3',
    eyB: 'qiu3',
    cyB: 'rou3',
    fyB: 'jiu4',
    LyB: 'zhou4',
    lyB: 'you3',
    ZyB: 'shou4',
    QyB: 'xiu3',
    hyB: 'you3',
    NyB: 'jiu3',
    VyB: 'shou3',
    XyB: 'zhou3',
    ByB: 'fou3',
    SyB: 'zhou3',
    UyB: 'shou4',
    TyB: 'chou3',
    jzB: 'hou4',
    DzB: 'mou3',
    CzB: 'bou4',
    EzB: 'dou3',
    FzB: 'tou3',
    dzB: 'gou3',
    gzB: 'ou3',
    AzB: 'bou3',
    HzB: 'nou3',
    QzB: 'sou3',
    izB: 'hou3',
    BzB: 'pou3',
    hzB: 'ou3',
    IzB: 'lou3',
    NzB: 'zou3',
    ezB: 'kou3',
    GzB: 'dou4',
    OzB: 'cou3',
    h0B: 'you3',
    d0B: 'jiu3',
    f0B: 'jiu4',
    O1B: 'qin3',
    L1B: 'zhen4',
    e1B: 'qin3',
    I1B: 'lin3',
    Q1B: 'xin3',
    K1B: 'chen3',
    N1B: 'jin3',
    c1B: 'ren3',
    X1B: 'zhen3',
    a1B: 'shen3',
    b1B: 'shen4',
    Z1B: 'shen4',
    Y1B: 'chen3',
    T1B: 'chen3',
    M1B: 'nin3',
    f1F: 'jin4',
    d1F: 'jin3',
    P1B: 'jin4',
    g1F: 'yin3',
    V1B: 'shen3',
    A1F: 'bin3',
    h1F: 'yin3',
    B1F: 'pin3',
    e1F: 'qin3',
    i1F: 'xin3',
    J1B: 'zhen3',
    l1B: 'yin3',
    U1B: 'shen4',
    d2B: 'gan3',
    G2B: 'dan4',
    h2B: 'an3',
    H2B: 'nan3',
    F2B: 'tan3',
    P2B: 'zan4',
    O2B: 'can3',
    g2B: 'an3',
    N2B: 'zan3',
    Q2B: 'san3',
    e2B: 'kan3',
    j2B: 'han4',
    I2B: 'lan3',
    E2B: 'tan3',
    i2B: 'han3',
    d3B: 'gan3',
    I3B: 'lan3',
    F3B: 'tan3',
    E3B: 'dan3',
    G3B: 'dan4',
    O3B: 'can3',
    D3B: 'man3',
    N3B: 'zan3',
    P3B: 'zan4',
    i3B: 'han3',
    h3B: 'an3',
    e3B: 'kan3',
    l4B: 'yan3',
    I4B: 'lian3',
    i4F: 'xian3',
    A4F: 'bian3',
    X4B: 'zhan3',
    e4F: 'qian3',
    g4F: 'yan3',
    f4F: 'jian4',
    d4F: 'jian3',
    h4B: 'yan3',
    c4B: 'ran3',
    a4B: 'shan3',
    K4B: 'chan3',
    h4F: 'yan3',
    P4B: 'jian4',
    O4B: 'qian3',
    N4B: 'jian3',
    e4B: 'qian3',
    Z4B: 'shan4',
    F5B: 'tian3',
    H5B: 'nian3',
    E5B: 'dian3',
    G5B: 'dian4',
    e5B: 'qian3',
    I5B: 'lian3',
    j5B: 'xian4',
    O5B: 'qian3',
    d5B: 'jian3',
    D5B: 'mian3',
    g8B: 'yan3',
    e8B: 'qian3',
    h8B: 'yan3',
    j6B: 'xian4',
    L6B: 'zhan4',
    e6B: 'qian3',
    d6B: 'jian3',
    U6B: 'shan4',
    I6B: 'lan3',
    S6B: 'zhan3',
    T6B: 'chan3',
    i6B: 'xian3',
    V6B: 'shan3',
    h6B: 'yan3',
    M6B: 'nian3',
    K6B: 'chan3',
    j7B: 'xian4',
    e7B: 'qian3',
    V7B: 'shan3',
    h7B: 'yan3',
    T7B: 'chan3',
    i7B: 'xian3',
    U7B: 'shan4',
    C9B: 'fan4',
    D9B: 'man3',
    A9B: 'fan3',
    e9J: 'quan3',
    B9B: 'fan3',
    K9J: 'chuan3',
    QAC: 'song4',
    CBC: 'feng4',
    dAC: 'gong4',
    IAC: 'long4',
    EAC: 'dong4',
    eAC: 'kong4',
    NAC: 'zong4',
    hAC: 'weng4',
    OAC: 'cong4',
    GAC: 'dong4',
    FAC: 'tong4',
    LBC: 'zhong4',
    ABC: 'feng4',
    eBC: 'qiong4',
    DBC: 'meng4',
    iBC: 'xiong4',
    DAC: 'meng4',
    OBC: 'cong4',
    BBC: 'feng4',
    PAC: 'zong4',
    JBC: 'zhong4',
    jAC: 'hong4',
    HAC: 'nong4',
    XBC: 'zhong4',
    YBC: 'chong4',
    UBC: 'shong4',
    iAC: 'hong4',
    QCC: 'song4',
    NCC: 'zong4',
    FCC: 'tong4',
    DCC: 'meng4',
    jCC: 'hong4',
    lDC: 'yong4',
    RDC: 'song4',
    CDC: 'feng4',
    fDC: 'jiong4',
    ADC: 'feng4',
    dDC: 'jiong4',
    hDC: 'yong4',
    JDC: 'zhong4',
    NDC: 'zong4',
    KDC: 'chong4',
    cDC: 'rong4',
    XDC: 'zhong4',
    LDC: 'zhong4',
    IDC: 'long4',
    eDC: 'qiong4',
    PDC: 'zong4',
    MDC: 'nong4',
    dEC: 'jiang4',
    jEC: 'xiang4',
    JEC: 'zhuang4',
    LEC: 'zhuang4',
    KEC: 'chuang4',
    UEC: 'shuang4',
    BEC: 'pang4',
    TEC: 'chuang4',
    VEC: 'shuang4',
    XFC: 'zhi4',
    CFC: 'bi4',
    XFK: 'zhui4',
    IFC: 'li4',
    ZFC: 'shi4',
    NFC: 'zi4',
    QFC: 'si4',
    kFK: 'wei4',
    dFO: 'gui4',
    BFG: 'pi4',
    AFG: 'bi4',
    CFG: 'bi4',
    IFK: 'lei4',
    dFG: 'ji4',
    AFC: 'bi4',
    fFG: 'ji4',
    OFC: 'ci4',
    lFC: 'yi4',
    gFG: 'yi4',
    BFC: 'pi4',
    PFC: 'zi4',
    JFC: 'zhi4',
    hFG: 'yi4',
    LFK: 'zhui4',
    YFK: 'chui4',
    iFG: 'xi4',
    eFC: 'qi4',
    hFC: 'yi4',
    aFC: 'shi4',
    VFC: 'shi4',
    eFK: 'kui4',
    hFO: 'wei4',
    gFO: 'wei4',
    iFO: 'hui4',
    hFK: 'wei4',
    ZFK: 'shui4',
    dFC: 'ji4',
    cFK: 'rui4',
    YFC: 'chi4',
    QFK: 'sui4',
    JFK: 'zhui4',
    lFK: 'wei4',
    MFK: 'nei4',
    dFK: 'gui4',
    SFC: 'zhi4',
    iFK: 'hui4',
    eFG: 'qi4',
    XGC: 'zhi4',
    kGK: 'wei4',
    DGG: 'mi4',
    RGK: 'sui4',
    NGK: 'zui4',
    QGK: 'sui4',
    IGK: 'lei4',
    AGG: 'bi4',
    fGO: 'gui4',
    BGG: 'pi4',
    CGG: 'bi4',
    dGO: 'gui4',
    VGK: 'shuai4',
    eGO: 'kui4',
    iGO: 'hui4',
    ZGC: 'shi4',
    IGC: 'li4',
    MGC: 'ni4',
    BGC: 'pi4',
    gGG: 'yi4',
    JGC: 'zhi4',
    eGC: 'qi4',
    LGC: 'zhi4',
    DGC: 'mi4',
    KGC: 'chi4',
    dGG: 'ji4',
    fGG: 'ji4',
    fGK: 'gui4',
    OGK: 'cui4',
    cGC: 'er4',
    NGC: 'zi4',
    OGC: 'ci4',
    hGG: 'yi4',
    QGC: 'si4',
    eGG: 'qi4',
    dGK: 'gui4',
    CGC: 'bi4',
    iGK: 'hui4',
    AGC: 'bi4',
    PGK: 'cui4',
    iGG: 'xi4',
    lGC: 'yi4',
    bGC: 'shi4',
    PGC: 'zi4',
    LGK: 'zhui4',
    YGK: 'chui4',
    lGK: 'wei4',
    JGK: 'zhui4',
    YGC: 'chi4',
    aGC: 'shi4',
    TGK: 'chuai4',
    aGK: 'shui4',
    XHC: 'zhi4',
    LHC: 'zhi4',
    RHC: 'si4',
    OHC: 'ci4',
    QHC: 'si4',
    aHC: 'shi4',
    SHC: 'zhi4',
    IHC: 'li4',
    PHC: 'zi4',
    KHC: 'chi4',
    cHC: 'er4',
    VHC: 'shi4',
    THC: 'chi4',
    lHC: 'yi4',
    JHC: 'zhi4',
    ZHC: 'shi4',
    UHC: 'shi4',
    fHC: 'ji4',
    YHC: 'chi4',
    hHC: 'yi4',
    dHC: 'ji4',
    iHC: 'xi4',
    gHC: 'yi4',
    eHC: 'qi4',
    DIC: 'wei4',
    dIK: 'gui4',
    kIK: 'wei4',
    gIK: 'wei4',
    AIC: 'fei4',
    BIC: 'fei4',
    eIK: 'kui4',
    hIK: 'wei4',
    iIK: 'hui4',
    CIC: 'fei4',
    dIC: 'ji4',
    gIC: 'yi4',
    eIC: 'qi4',
    iIC: 'xi4',
    fIC: 'ji4',
    hIC: 'yi4',
    gJC: 'yu4',
    IJC: 'lü4',
    dJC: 'ju4',
    OJC: 'qu4',
    eJC: 'qu4',
    ZJC: 'shu4',
    aJC: 'shu4',
    JJC: 'zhu4',
    XJC: 'zhu4',
    VJC: 'shu4',
    hJC: 'yu4',
    LJC: 'zhu4',
    fJC: 'ju4',
    QJC: 'xu4',
    UJC: 'zhu4',
    NJC: 'ju4',
    SJC: 'zhu4',
    cJC: 'ru4',
    lJC: 'yu4',
    iJC: 'xu4',
    MJC: 'nü4',
    TJC: 'chu4',
    YJC: 'chu4',
    KJC: 'chu4',
    RJC: 'xu4',
    gKC: 'yu4',
    hKC: 'yu4',
    ZKC: 'shu4',
    LKC: 'zhu4',
    CKC: 'fu4',
    XKC: 'zhu4',
    dKC: 'ju4',
    iKC: 'xu4',
    aKC: 'shu4',
    lKC: 'yu4',
    cKC: 'ru4',
    BKC: 'fu4',
    DKC: 'wu4',
    NKC: 'ju4',
    fKC: 'ju4',
    kKC: 'yu4',
    PKC: 'ju4',
    VKC: 'shu4',
    AKC: 'fu4',
    OKC: 'qu4',
    JKC: 'zhu4',
    eKC: 'qu4',
    TKC: 'chu4',
    KKC: 'chu4',
    QKC: 'xu4',
    IKC: 'lü4',
    DLC: 'mu4',
    GLC: 'du4',
    ILC: 'lu4',
    ELC: 'du4',
    FLC: 'tu4',
    dLC: 'gu4',
    gLC: 'wu4',
    jLC: 'hu4',
    QLC: 'su4',
    PLC: 'zu4',
    HLC: 'nu4',
    ALC: 'bu4',
    hLC: 'wu4',
    BLC: 'pu4',
    OLC: 'cu4',
    eLC: 'ku4',
    CLC: 'bu4',
    iLC: 'hu4',
    NLC: 'zu4',
    NMC: 'ji4',
    EMC: 'di4',
    PMC: 'ji4',
    FMC: 'ti4',
    GMC: 'di4',
    OMC: 'qi4',
    QMC: 'xi4',
    gMC: 'yi4',
    dMC: 'ji4',
    jMC: 'xi4',
    eMC: 'qi4',
    hMC: 'yi4',
    DMC: 'mi4',
    AMC: 'bi4',
    jMK: 'hui4',
    dMK: 'gui4',
    iMK: 'hui4',
    BMC: 'pi4',
    CMC: 'bi4',
    IMC: 'li4',
    HMC: 'ni4',
    iMC: 'xi4',
    NNC: 'ji4',
    QNK: 'sui4',
    kNK: 'wei4',
    cNK: 'rui4',
    XNK: 'zhui4',
    VNK: 'shuai4',
    ONK: 'cui4',
    lNK: 'wei4',
    JNK: 'zhui4',
    aNK: 'shui4',
    CNC: 'bi4',
    TNK: 'chuai4',
    RNK: 'sui4',
    ANC: 'bi4',
    dNO: 'gui4',
    DNC: 'mi4',
    VNC: 'shi4',
    YNC: 'chi4',
    XNC: 'zhi4',
    ZNC: 'shi4',
    lNC: 'yi4',
    hNG: 'yi4',
    gNC: 'yi4',
    LNC: 'zhi4',
    INC: 'li4',
    eNG: 'qi4',
    aNC: 'shi4',
    dNG: 'ji4',
    JNC: 'zhi4',
    fNG: 'ji4',
    ZNK: 'shui4',
    KNC: 'chi4',
    LNK: 'zhui4',
    gNG: 'yi4',
    NNK: 'zui4',
    BNC: 'pi4',
    eUK: 'kui4',
    iUK: 'hui4',
    FOC: 'tai4',
    dOC: 'gai4',
    gOC: 'ai4',
    hOC: 'ai4',
    HOC: 'nai4',
    GOC: 'dai4',
    jOC: 'hai4',
    EOC: 'dai4',
    AOC: 'bei4',
    BOC: 'pei4',
    jOK: 'hui4',
    GOK: 'dui4',
    dOK: 'kui4',
    NOK: 'zui4',
    iOK: 'hui4',
    eOK: 'kui4',
    IOK: 'lei4',
    gOK: 'wei4',
    EOK: 'dui4',
    hOK: 'wei4',
    POK: 'zui4',
    QOK: 'sui4',
    OOK: 'cui4',
    COC: 'bei4',
    eOC: 'kai4',
    OOC: 'cai4',
    IOC: 'lai4',
    iOC: 'hai4',
    FOK: 'tai4',
    DOC: 'mei4',
    dPK: 'gua4',
    dPC: 'jia4',
    jPC: 'xia4',
    DPC: 'mai4',
    jPK: 'hua4',
    TPC: 'chai4',
    gPC: 'ya4',
    iPC: 'xia4',
    APC: 'bai4',
    CPC: 'bai4',
    UPC: 'zhai4',
    BPC: 'pai4',
    SPC: 'zhai4',
    ePC: 'qia4',
    VPC: 'shai4',
    iPK: 'hua4',
    JPK: 'zhuai4',
    dQK: 'kuai4',
    hQC: 'ye4',
    SQC: 'zhai4',
    dQC: 'jie4',
    MQC: 'nai4',
    iQC: 'xie4',
    jQC: 'xie4',
    gQC: 'ye4',
    eQK: 'kuai4',
    AQC: 'bai4',
    BQC: 'pai4',
    jQK: 'huai4',
    gQK: 'wai4',
    CQC: 'bai4',
    DQC: 'mai4',
    iQK: 'huai4',
    VQC: 'shai4',
    eQC: 'qie4',
    KQK: 'chuai4',
    dRK: 'gua4',
    eRK: 'kuai4',
    DRC: 'mai4',
    jRK: 'huai4',
    CRC: 'bai4',
    hRK: 'wai4',
    TRK: 'chuai4',
    dRC: 'jie4',
    KRC: 'chai4',
    hRC: 'ye4',
    VRC: 'shai4',
    LRC: 'zhai4',
    iRC: 'xie4',
    iRK: 'huai4',
    ARC: 'bai4',
    ORK: 'cuai4',
    URC: 'zhai4',
    jRC: 'xie4',
    GSK: 'dui4',
    CSC: 'bei4',
    DSC: 'mei4',
    BSC: 'pei4',
    iSK: 'hui4',
    ESK: 'dui4',
    OSK: 'cui4',
    NSK: 'zui4',
    hSK: 'wei4',
    FSK: 'tui4',
    dSK: 'gui4',
    jSK: 'hui4',
    eSK: 'kui4',
    QSK: 'sui4',
    HSK: 'nei4',
    ISK: 'lei4',
    ASC: 'bei4',
    gSK: 'wei4',
    GTC: 'dai4',
    NTC: 'zai4',
    DTC: 'mei4',
    QTC: 'sai4',
    FTC: 'tai4',
    dTC: 'gai4',
    eTC: 'kai4',
    gTC: 'ai4',
    hTC: 'ai4',
    jTC: 'hai4',
    HTC: 'nai4',
    ETC: 'dai4',
    ITC: 'lai4',
    OTC: 'cai4',
    PTC: 'zai4',
    iTC: 'hai4',
    AUC: 'fei4',
    BUC: 'fei4',
    hUK: 'wei4',
    CUC: 'fei4',
    fUK: 'gui4',
    gUC: 'yi4',
    XVC: 'zhen4',
    QVC: 'xin4',
    cVC: 'ren4',
    lVC: 'yin4',
    IVC: 'lin4',
    AVC: 'bin4',
    LVC: 'zhen4',
    ZVC: 'shen4',
    aVC: 'shen4',
    eVG: 'qin4',
    RVC: 'xin4',
    gVG: 'yin4',
    NVC: 'jin4',
    iVG: 'xin4',
    JVC: 'zhen4',
    fVG: 'jin4',
    TWC: 'chen4',
    hVC: 'yin4',
    KVC: 'chen4',
    BVC: 'pin4',
    OVC: 'qin4',
    eVC: 'qin4',
    dVK: 'jun4',
    XVK: 'zhun4',
    QVK: 'xun4',
    RVK: 'xun4',
    NVK: 'jun4',
    aVK: 'shun4',
    cVK: 'run4',
    bVK: 'shun4',
    DXC: 'wen4',
    kXK: 'yun4',
    iXK: 'xun4',
    BXC: 'fen4',
    AXC: 'fen4',
    hXK: 'yun4',
    dXK: 'jun4',
    fXK: 'jun4',
    CXC: 'fen4',
    iYC: 'xin4',
    dYC: 'jin4',
    fYC: 'jin4',
    hYC: 'yin4',
    gYC: 'yin4',
    gZK: 'yuan4',
    hZK: 'yuan4',
    AZC: 'fan4',
    eZK: 'quan4',
    DZC: 'wan4',
    CZC: 'fan4',
    BZC: 'fan4',
    dZC: 'jian4',
    hZC: 'yan4',
    iZC: 'xian4',
    iZK: 'xuan4',
    fZC: 'jian4',
    kZK: 'yuan4',
    gZC: 'yan3',
    fZK: 'juan4',
    dZK: 'juan4',
    jaK: 'hun4',
    EaK: 'dun4',
    QaK: 'sun4',
    eaK: 'kun4',
    HaK: 'nun4',
    haK: 'wen4',
    DaC: 'men4',
    PaK: 'zun4',
    daK: 'gun4',
    BaC: 'pen4',
    GaK: 'dun4',
    OaK: 'cun4',
    CaC: 'ben4',
    gaK: 'wen4',
    IaK: 'lun4',
    AaC: 'ben4',
    iaK: 'hun4',
    NaK: 'zun4',
    jbC: 'hen4',
    dbC: 'gen4',
    gbC: 'en4',
    hbC: 'en4',
    jcC: 'han4',
    FcC: 'tan4',
    hcC: 'an4',
    EcC: 'dan4',
    GcC: 'dan4',
    dcC: 'gan4',
    gcC: 'an4',
    ecC: 'kan4',
    icC: 'han4',
    IcC: 'lan4',
    HcC: 'nan4',
    OcC: 'can4',
    QcC: 'san4',
    NcC: 'zan4',
    PcC: 'san4',
    jcK: 'huan4',
    NcK: 'zuan4',
    hcK: 'wan4',
    dcK: 'guan4',
    OcK: 'cuan4',
    gcK: 'wan4',
    GcK: 'duan4',
    IcK: 'luan4',
    EcK: 'duan4',
    FcK: 'tuan4',
    icK: 'huan4',
    QcK: 'suan4',
    DcC: 'man4',
    AcC: 'ban4',
    BcC: 'pan4',
    CcC: 'ban4',
    HcK: 'nuan4',
    PcK: 'zuan4',
    ecK: 'kuan4',
    ddC: 'jian4',
    gdC: 'yan4',
    hdC: 'yan4',
    VdC: 'shan4',
    jdC: 'xian4',
    DdC: 'man4',
    hdK: 'wan4',
    jdK: 'huan4',
    ddK: 'guan4',
    VdK: 'shuan4',
    TdK: 'chuan4',
    gdK: 'wan4',
    UdC: 'zhan4',
    BdC: 'pan4',
    MdK: 'nuan4',
    TdC: 'chan4',
    KdC: 'chan4',
    deC: 'jian4',
    jeC: 'xian4',
    CeC: 'ban4',
    BeC: 'pan4',
    jeK: 'huan4',
    DeC: 'man4',
    LeC: 'zhan4',
    AeC: 'ban4',
    deK: 'guan4',
    QgC: 'xian4',
    OgC: 'qian4',
    igK: 'xuan4',
    jgK: 'xuan4',
    dgK: 'juan4',
    GgC: 'dian4',
    FgC: 'tian4',
    IgC: 'lian4',
    dgC: 'jian4',
    HgC: 'nian4',
    egC: 'qian4',
    jgC: 'xian4',
    ggC: 'yan4',
    hgC: 'yan4',
    NgC: 'jian4',
    DgC: 'mian4',
    BgC: 'pian4',
    PgC: 'jian4',
    hgK: 'yuan4',
    EgC: 'dian4',
    igC: 'xian4',
    QfC: 'xian4',
    XfC: 'zhan4',
    ZfC: 'shan4',
    gfG: 'yan4',
    efC: 'qian4',
    dfK: 'juan4',
    kfK: 'yuan4',
    DfC: 'mian4',
    YfK: 'chuan4',
    lfK: 'yuan4',
    JfC: 'zhan4',
    cfK: 'ruan4',
    NfC: 'jian4',
    YfC: 'chan4',
    afC: 'shan4',
    hfC: 'yan4',
    dfO: 'juan4',
    ffO: 'juan4',
    IfK: 'luan4',
    KfK: 'chuan4',
    AfG: 'bian4',
    VfK: 'shuan4',
    OfK: 'quan4',
    CfG: 'bian4',
    RfK: 'xuan4',
    QfK: 'xuan4',
    UfK: 'shuan4',
    SfK: 'zhuan4',
    LfK: 'zhuan4',
    PfC: 'jian4',
    RfC: 'xian4',
    BfC: 'pian4',
    ZfK: 'shuan4',
    MfC: 'nian4',
    JfK: 'zhuan4',
    lfC: 'yan4',
    CfC: 'bian4',
    LfC: 'zhan4',
    IfC: 'lian4',
    efO: 'quan4',
    XfK: 'zhuan4',
    AgC: 'bian4',
    QhC: 'xiao4',
    FhC: 'tiao4',
    EhC: 'diao4',
    dhC: 'jiao4',
    HhC: 'niao4',
    GhC: 'diao4',
    ehC: 'qiao4',
    IhC: 'liao4',
    ghC: 'yao4',
    hhC: 'yao4',
    ihC: 'xiao4',
    QiC: 'xiao4',
    XiC: 'zhao3',
    liC: 'yao4',
    hiC: 'yao4',
    LiC: 'zhao4',
    ZiC: 'shao4',
    fiG: 'jiao4',
    BiC: 'piao4',
    PiC: 'jiao4',
    DiC: 'miao4',
    OiC: 'qiao4',
    IiC: 'liao4',
    eiG: 'qiao4',
    giG: 'yao4',
    NiC: 'jiao4',
    DiG: 'miao4',
    CiC: 'biao4',
    aiC: 'shao4',
    KiC: 'chao4',
    AiG: 'biao4',
    fiC: 'jiao4',
    ciC: 'rao4',
    jjC: 'xiao4',
    djC: 'jiao4',
    ijC: 'xiao4',
    JjC: 'zhao4',
    AjC: 'bao4',
    ejC: 'qiao4',
    DjC: 'mao4',
    BjC: 'pao4',
    KjC: 'chao4',
    VjC: 'shao4',
    LjC: 'zhao4',
    MjC: 'nao4',
    SjC: 'zhao4',
    CjC: 'bao4',
    TjC: 'chao4',
    hjC: 'yao4',
    gjC: 'yao4',
    UjC: 'zhao4',
    jkC: 'hao4',
    GkC: 'dao4',
    EkC: 'dao4',
    dkC: 'gao4',
    gkC: 'ao4',
    DkC: 'mao4',
    IkC: 'lao4',
    OkC: 'cao4',
    CkC: 'bao4',
    AkC: 'bao4',
    PkC: 'zao4',
    hkC: 'ao4',
    QkC: 'sao4',
    ekC: 'kao4',
    NkC: 'zao4',
    ikC: 'hao4',
    HkC: 'nao4',
    dlC: 'ge4',
    jlC: 'he4',
    NlC: 'zuo4',
    ElC: 'duo4',
    IlC: 'luo4',
    elC: 'ke4',
    glC: 'e4',
    GlC: 'duo4',
    HlC: 'nuo4',
    QlC: 'suo4',
    ilC: 'he4',
    FlC: 'tuo4',
    dlK: 'guo4',
    jlK: 'huo4',
    NlK: 'zuo4',
    elK: 'kuo4',
    FlK: 'tuo4',
    AlC: 'bo4',
    OlK: 'cuo4',
    DlC: 'mo4',
    HlK: 'nuo4',
    BlC: 'po4',
    PlK: 'zuo4',
    glK: 'wo4',
    ilK: 'huo4',
    GlK: 'duo4',
    ClC: 'bo4',
    IlK: 'luo4',
    ElK: 'duo4',
    QlK: 'suo4',
    hlC: 'e4',
    hlK: 'wo4',
    DnC: 'ma4',
    dnC: 'jia4',
    hnC: 'ya4',
    inC: 'xia4',
    gnC: 'ya4',
    KnC: 'cha4',
    JnC: 'zha4',
    SnC: 'zha4',
    UnC: 'zha4',
    RoC: 'xie4',
    enC: 'qia4',
    jnC: 'xia4',
    PoC: 'jie4',
    loC: 'ye4',
    YoC: 'che4',
    QoC: 'xie4',
    XoC: 'zhe4',
    NoC: 'jie4',
    aoC: 'she4',
    boC: 'she4',
    AnC: 'ba4',
    BnC: 'pa4',
    jnK: 'hua4',
    inK: 'hua4',
    enK: 'kua4',
    VnK: 'shua4',
    CnC: 'ba4',
    HnC: 'na4',
    OoC: 'qie4',
    VnC: 'sha4',
    dnK: 'gua4',
    LnC: 'zha4',
    gnK: 'wa4',
    hnK: 'wa4',
    lpC: 'yang4',
    IpC: 'liang4',
    UpC: 'zhuang4',
    cpC: 'rang4',
    apC: 'shang4',
    JpC: 'zhang4',
    KpC: 'chang4',
    ipC: 'xiang4',
    LpC: 'zhang4',
    MpC: 'niang4',
    PpC: 'jiang4',
    XpC: 'zhang4',
    ZpC: 'shang4',
    SpC: 'zhuang4',
    hpC: 'yang4',
    fpC: 'jiang4',
    YpC: 'chang4',
    TpC: 'chuang4',
    NpC: 'jiang4',
    gpC: 'yang4',
    BpC: 'fang4',
    DpC: 'wang4',
    ipK: 'huang4',
    dpK: 'guang4',
    kpK: 'wang4',
    ApC: 'fang4',
    QpC: 'xiang4',
    dpC: 'jiang4',
    epC: 'qiang4',
    OpC: 'qiang4',
    fpK: 'guang4',
    CpC: 'fang4',
    GqC: 'dang4',
    IqC: 'lang4',
    jqC: 'hang4',
    hqC: 'ang4',
    gqC: 'ang4',
    NqC: 'zang4',
    CqC: 'bang4',
    PqC: 'zang4',
    EqC: 'dang4',
    eqC: 'kang4',
    AqC: 'bang4',
    FqC: 'tang4',
    eqK: 'kuang4',
    HqC: 'nang4',
    QqC: 'sang4',
    jqK: 'huang4',
    dqK: 'guang4',
    dqC: 'kang4',
    iqK: 'huang4',
    DqC: 'mang4',
    hqK: 'ang4',
    hsC: 'ying4',
    dsC: 'jing4',
    fsC: 'jing4',
    esC: 'qing4',
    drC: 'geng4',
    DsC: 'ming4',
    CsC: 'bing4',
    DrC: 'meng4',
    jrK: 'hong4',
    AsC: 'bing4',
    ksK: 'yong4',
    jrC: 'heng4',
    hrC: 'eng4',
    TrC: 'cheng4',
    JrC: 'zheng4',
    KrC: 'cheng4',
    ArC: 'beng4',
    LrC: 'zheng4',
    VrC: 'sheng4',
    gsC: 'ying4',
    CrC: 'beng4',
    irC: 'heng4',
    hrK: 'weng4',
    StC: 'zheng4',
    AtC: 'beng4',
    CtC: 'beng4',
    htC: 'eng4',
    gtC: 'eng4',
    itK: 'hung4',
    duC: 'jing4',
    OuC: 'qing4',
    XuC: 'zheng4',
    auC: 'sheng4',
    LuC: 'zheng4',
    KuC: 'cheng4',
    QuC: 'xing4',
    IuC: 'ling4',
    BuC: 'ping4',
    iuK: 'xiong4',
    AuC: 'bing4',
    CuC: 'bing4',
    PuC: 'jing4',
    ZuC: 'sheng4',
    DuC: 'ming4',
    euC: 'qing4',
    iuC: 'xing4',
    NuC: 'jing4',
    dvC: 'jing4',
    HvC: 'ning4',
    QvC: 'xing4',
    jvC: 'xing4',
    GvC: 'ding4',
    EvC: 'ding4',
    evC: 'qing4',
    FvC: 'ting4',
    OvC: 'qing4',
    DvC: 'ming4',
    hvK: 'yong4',
    IvC: 'ling4',
    XwC: 'zheng4',
    lwC: 'ying4',
    bwC: 'sheng4',
    cwC: 'reng4',
    hwC: 'ying4',
    NwC: 'jing4',
    iwC: 'xing4',
    awC: 'sheng4',
    LwC: 'zheng4',
    IwC: 'xing4',
    CwC: 'bing4',
    YwC: 'cheng4',
    gwC: 'ying4',
    ZwC: 'sheng4',
    KwC: 'cheng4',
    fwC: 'jing4',
    ExC: 'deng4',
    PxC: 'zeng4',
    dxC: 'geng4',
    OxC: 'ceng4',
    GxC: 'deng4',
    DxC: 'meng4',
    AxC: 'beng4',
    CxC: 'beng4',
    IxC: 'leng4',
    NxC: 'zeng4',
    QxC: 'seng4',
    FxC: 'teng4',
    kyC: 'you4',
    dyC: 'jiu4',
    LyC: 'zhou4',
    JyC: 'zhou4',
    ayC: 'shou4',
    YyC: 'chou4',
    RyC: 'xiu4',
    iyC: 'xiu4',
    XyC: 'zhou4',
    fyC: 'jiu4',
    VyC: 'shou4',
    SyC: 'zhou4',
    ByC: 'fou4',
    TyC: 'chou4',
    AyC: 'fou4',
    KyC: 'chou4',
    IyC: 'liu4',
    QyC: 'xiu4',
    NyC: 'jiu4',
    UyC: 'zhou4',
    PyC: 'jiu4',
    MyC: 'niu4',
    CyC: 'fou4',
    lyC: 'you4',
    ZyC: 'shou4',
    cyC: 'rou4',
    DyC: 'miu4',
    gyC: 'you4',
    eyC: 'qiu4',
    OyC: 'qiu4',
    jzC: 'hou4',
    ezC: 'kou4',
    DzC: 'mou4',
    BzC: 'pou4',
    GzC: 'dou4',
    EzC: 'dou4',
    HzC: 'nou4',
    QzC: 'sou4',
    NzC: 'zou4',
    FzC: 'tou4',
    hzC: 'ou4',
    dzC: 'gou4',
    OzC: 'cou4',
    IzC: 'lou4',
    izC: 'hou4',
    CzC: 'bou4',
    gzC: 'ou4',
    PzC: 'zou4',
    h0C: 'you4',
    D0C: 'miao4',
    e0C: 'qiu4',
    f0C: 'jiu4',
    O1C: 'qin4',
    N1C: 'jin4',
    c1C: 'ren4',
    L1C: 'zhen4',
    X1C: 'zhen4',
    f1G: 'jin4',
    d1G: 'jin4',
    M1C: 'nin4',
    h1G: 'yin4',
    V1C: 'shen4',
    K1C: 'chen4',
    S1C: 'zhen4',
    T1C: 'chen4',
    g1G: 'yin4',
    J1C: 'zhen4',
    I1C: 'lin4',
    Z1C: 'shen4',
    k1C: 'yin4',
    a1C: 'shen4',
    e2C: 'kan4',
    d2C: 'gan4',
    j2C: 'han4',
    h2C: 'an4',
    F2C: 'tan4',
    Q2C: 'san4',
    O2C: 'can4',
    G2C: 'dan4',
    g2C: 'an4',
    E2C: 'dan4',
    H2C: 'nan4',
    I2C: 'lan4',
    i2C: 'han4',
    N2C: 'zan4',
    e3C: 'kan4',
    I3C: 'lan4',
    F3C: 'tan4',
    d3C: 'gan4',
    i3C: 'han4',
    j3C: 'han4',
    G3C: 'dan4',
    P3C: 'zan4',
    E3C: 'dan4',
    Q3C: 'san4',
    l4C: 'yan4',
    Z4C: 'shan4',
    c4C: 'ran4',
    h4C: 'yan4',
    A4G: 'bian4',
    g4G: 'yan4',
    a4C: 'shan4',
    N4C: 'jian4',
    O4C: 'qian4',
    I4C: 'lian4',
    K4C: 'chan4',
    Y4C: 'chan4',
    h4G: 'yan4',
    P4C: 'jian4',
    X4C: 'zhan4',
    F5C: 'tian4',
    H5C: 'nian4',
    E5C: 'dian4',
    Q5C: 'xian4',
    G5C: 'dian4',
    d5C: 'jian4',
    h5C: 'yan4',
    N5C: 'jian4',
    P5C: 'jian4',
    e5C: 'qian4',
    I5C: 'lian4',
    g8C: 'yan4',
    i8C: 'xian4',
    e8C: 'qian4',
    D9C: 'wan4',
    j6C: 'xian4',
    h6C: 'yan4',
    S6C: 'zhan4',
    J6C: 'zhan4',
    e6C: 'qian4',
    L6C: 'zhan4',
    U6C: 'zhan4',
    d6C: 'jian4',
    M6C: 'nan4',
    g6C: 'yan4',
    d7C: 'jian4',
    T7C: 'chan4',
    S7C: 'zhan4',
    V7C: 'shan4',
    i7C: 'xian4',
    C7C: 'ban4',
    j7C: 'xian4',
    U7C: 'zhan4',
    h7C: 'zheng3',
    C9C: 'fan4',
    B9C: 'fan4',
    d8C: 'jian4',
    h8C: 'yan4',
    hAD: 'wu',
    GAD: 'du2',
    dAD: 'gu',
    jAD: 'hu2',
    eAD: 'ku',
    FAD: 'tu',
    EAD: 'du',
    QAD: 'su',
    IAD: 'lu4',
    iAD: 'hu',
    PAD: 'zu2',
    OAD: 'cu',
    NAD: 'zu',
    CAD: 'bu2',
    BAD: 'pu',
    AAD: 'bu',
    DAD: 'mu4',
    ABD: 'fu',
    CBD: 'fu2',
    VBD: 'shu',
    IBD: 'lü4',
    LBD: 'zhu2',
    dBD: 'ju',
    eBD: 'qu',
    ZBD: 'shu2',
    YBD: 'chu',
    lBD: 'yu4',
    fBD: 'ju2',
    OBD: 'cu',
    cBD: 'ru4',
    XBD: 'zhu',
    aBD: 'shu',
    iBD: 'xu',
    JBD: 'zhu',
    NBD: 'zu',
    TBD: 'chu',
    MBD: 'nü4',
    SBD: 'zhu',
    BBD: 'fu',
    hBD: 'yu',
    QBD: 'su',
    DBD: 'mu4',
    kBD: 'yu4',
    KBD: 'chu',
    gBD: 'yu4',
    PBD: 'zu2',
    hCD: 'wu',
    GCD: 'du2',
    ECD: 'du',
    eCD: 'ku',
    jCD: 'hu2',
    CCD: 'bu2',
    QCD: 'su',
    dCD: 'ku',
    DCD: 'mu4',
    iCD: 'hu',
    HCD: 'nu4',
    NCD: 'zu',
    ACD: 'bu',
    ICD: 'lu4',
    gCD: 'wu4',
    XDD: 'zhu',
    gDD: 'yu4',
    iDD: 'xu',
    dDD: 'ju',
    fDD: 'ju2',
    ZDD: 'shu2',
    YDD: 'chu',
    cDD: 'ru4',
    aDD: 'shu',
    lDD: 'yu4',
    LDD: 'zhu2',
    IDD: 'lü4',
    eDD: 'qu',
    JDD: 'zhu',
    NDD: 'zu',
    bDD: 'shu2',
    CDD: 'fu2',
    ODD: 'cu',
    RDD: 'su2',
    QDD: 'su',
    ADD: 'fu',
    KDD: 'chu',
    dED: 'jue',
    gED: 'yue4',
    UED: 'zhuo2',
    SED: 'zhuo',
    VED: 'shuo',
    JED: 'zhuo',
    AED: 'bo',
    DED: 'mo4',
    CED: 'bo2',
    BED: 'po',
    eED: 'que',
    LED: 'zhuo2',
    hED: 'yue',
    MED: 'nuo4',
    KED: 'chuo',
    IED: 'luo4',
    jED: 'xue2',
    iED: 'xue',
    TED: 'chuo',
    XVD: 'zhi',
    cVD: 'ri4',
    bVD: 'shi2',
    LVD: 'zhi2',
    QVD: 'xi',
    hVD: 'yi',
    OVD: 'qi',
    BVD: 'pi',
    dVD: 'ji',
    MVD: 'ni4',
    lVD: 'yi4',
    eVD: 'qi',
    iVD: 'xi',
    KVD: 'chi',
    IVD: 'li4',
    JVD: 'zhi',
    PVD: 'ji2',
    TWD: 'che',
    aVD: 'shi',
    NVD: 'ji',
    DVD: 'mi4',
    AVD: 'bi',
    fVH: 'ji2',
    CVD: 'bi2',
    kVL: 'yu4',
    VVL: 'shuai',
    YVD: 'chi',
    UWD: 'zhe2',
    DVH: 'mi4',
    CVH: 'bi2',
    hVH: 'yi',
    gVH: 'yi4',
    AVH: 'bi',
    JVL: 'zhu',
    iVH: 'xi',
    dVH: 'ji',
    iVL: 'xu',
    bVL: 'shu2',
    dVL: 'ju',
    PVL: 'ju2',
    lVL: 'yu4',
    NVL: 'ju',
    QVL: 'xu',
    IVL: 'lü4',
    KVL: 'chu',
    LVL: 'zhu2',
    YVL: 'chu',
    OVL: 'qu',
    SVL: 'zhuai',
    iVP: 'xu',
    SWD: 'zhe',
    VWD: 'she',
    DXD: 'wu4',
    AXD: 'fu',
    hXL: 'yu',
    dXL: 'ju',
    eXL: 'qu',
    fXL: 'ju2',
    CXD: 'fu2',
    iXL: 'xu',
    kXL: 'yu4',
    BXD: 'fu',
    gXL: 'yu4',
    iYD: 'xi',
    dYD: 'ji',
    gYD: 'yi4',
    fYD: 'ji2',
    eYD: 'qi',
    gZL: 'yue4',
    CZD: 'fa2',
    kZL: 'yue4',
    dZL: 'jue',
    hZL: 'yue',
    fZL: 'jue2',
    eZL: 'que',
    AZD: 'fa',
    DZD: 'wa4',
    iZL: 'xue',
    hZD: 'ye',
    iZD: 'xie',
    dZD: 'jie',
    fZD: 'jie2',
    BZD: 'fa',
    gZD: 'ye4',
    DaD: 'mo4',
    FaL: 'tu',
    daL: 'gu',
    CaD: 'bo2',
    EaL: 'du',
    GaL: 'du2',
    haL: 'wu',
    iaL: 'hu',
    gaL: 'wu4',
    BaD: 'pu',
    IaL: 'lu4',
    eaL: 'ku',
    HaL: 'nu4',
    QaL: 'su',
    OaL: 'cu',
    PaL: 'zu2',
    jbD: 'he2',
    jaL: 'hu2',
    NaL: 'zu',
    jcD: 'he2',
    icD: 'he',
    EcD: 'da',
    FcD: 'ta',
    hcD: 'e',
    IcD: 'la4',
    ecD: 'ke',
    GcD: 'da2',
    PcD: 'za2',
    gcD: 'e4',
    dcD: 'ge',
    QcD: 'sa',
    OcD: 'ca',
    HcD: 'na4',
    DcD: 'mo4',
    gcL: 'wo4',
    PcL: 'zuo2',
    AcD: 'bo',
    NcD: 'zuo',
    dcL: 'guo',
    ecL: 'kuo',
    jcL: 'huo2',
    GcL: 'duo2',
    icL: 'huo',
    hcL: 'wo',
    NcL: 'zuo',
    BcD: 'po',
    FcL: 'tuo',
    IcL: 'luo4',
    EcL: 'duo',
    OcL: 'cuo',
    CcD: 'bo2',
    jeD: 'xia2',
    SeD: 'zha',
    CeD: 'ba2',
    eeD: 'qia',
    jeL: 'hua2',
    AeD: 'ba',
    JeL: 'zhua',
    heL: 'wa',
    MeL: 'na4',
    TeD: 'chua',
    deL: 'gua',
    deD: 'jia',
    heD: 'ya',
    VeD: 'sha',
    DeD: 'ma',
    ieL: 'xia',
    MeD: 'na4',
    geL: 'wa4',
    eeL: 'kua',
    BeD: 'pa',
    SeL: 'zhua',
    jdD: 'xia2',
    hdD: 'ya',
    gdD: 'ya4',
    TdD: 'cha',
    edD: 'qia',
    idD: 'xia',
    KdD: 'cha',
    ddL: 'gua',
    jdL: 'hua2',
    KdL: 'chua',
    JdL: 'zhua',
    VdL: 'shua',
    gdL: 'wa4',
    TdL: 'chua',
    DdD: 'ma4',
    MdL: 'na4',
    AdD: 'ba',
    ddD: 'jia',
    UdD: 'zha2',
    JdD: 'zha',
    QgD: 'xie',
    OgD: 'qie',
    dgD: 'jie',
    NgD: 'jie',
    igL: 'xue',
    egL: 'que',
    dgL: 'jue',
    jgL: 'xue2',
    hgL: 'yue',
    GgD: 'die2',
    FgD: 'tie',
    jgD: 'xie2',
    HgD: 'nie4',
    PgD: 'jie2',
    ggD: 'ye4',
    DgD: 'mie4',
    AgD: 'bie',
    hgD: 'ye',
    egD: 'qie',
    igD: 'xie',
    BgD: 'pie',
    CgD: 'bie2',
    EgD: 'die',
    IgD: 'lie4',
    QfD: 'xie',
    IfD: 'lie4',
    JfD: 'zhe',
    ffH: 'jie2',
    cfD: 're4',
    XfD: 'zhe',
    bfD: 'she2',
    ZfD: 'she2',
    gfH: 'ye4',
    DfD: 'mie4',
    efH: 'qie',
    AfD: 'bie',
    PfL: 'jue2',
    NfL: 'jue',
    QfL: 'xue',
    KfL: 'chuo',
    lfL: 'yue4',
    efL: 'que',
    hfP: 'yue',
    cfL: 'ruo4',
    afL: 'shuo',
    XfL: 'zhuo',
    YfL: 'chuo',
    JfL: 'zhuo',
    IfL: 'lüe4',
    BfD: 'pie',
    CfH: 'bie2',
    LfD: 'zhe2',
    AfH: 'bie',
    VfL: 'shuo',
    dfH: 'jue',
    afD: 'she',
    MfL: 'nüe4',
    ifP: 'xue',
    hfL: 'yue',
    dfP: 'jue',
    SfL: 'zhuo',
    OfL: 'que',
    NfD: 'jue',
    VfD: 'she',
    hfH: 'ye',
    KfD: 'che',
    ifH: 'xue',
    TfD: 'che',
    RfL: 'xue2',
    YfD: 'che',
    lfD: 'ye4',
    UfD: 'zhe2',
    lpD: 'yue4',
    IpD: 'lüe4',
    dpD: 'jue',
    XpD: 'zhuo',
    apD: 'shuo',
    cpD: 'ruo4',
    YpD: 'chuo',
    hpD: 'yue',
    epD: 'que',
    gpD: 'yue4',
    ZpD: 'shuo2',
    KpD: 'chuo',
    QpD: 'xue',
    SpD: 'zhuo',
    NpD: 'jue',
    PpD: 'jue2',
    OpD: 'que',
    fpD: 'jue2',
    hpL: 'yue',
    CpD: 'fo2',
    ipL: 'xue',
    kpL: 'yue4',
    dpL: 'jue',
    JpD: 'zhuo',
    LpD: 'zhuo2',
    epL: 'que',
    fpL: 'jue2',
    MpD: 'nüe4',
    BpD: 'fo',
    ipD: 'xue',
    GqD: 'duo2',
    DqD: 'mo4',
    IqD: 'luo4',
    FqD: 'tuo',
    NqD: 'zuo',
    OqD: 'cuo',
    dqD: 'ge',
    eqD: 'ke',
    gqD: 'e4',
    BqD: 'po',
    hqD: 'e',
    CqD: 'bo2',
    iqD: 'he',
    QqD: 'suo',
    jqD: 'he2',
    PqD: 'zuo2',
    AqD: 'bo',
    HqD: 'nuo4',
    iqL: 'huo',
    dqL: 'guo',
    hqL: 'wo',
    jqL: 'huo2',
    eqL: 'kuo',
    gqL: 'wo4',
    IqL: 'luo4',
    NqL: 'zuo',
    DrD: 'mo4',
    JrD: 'zhe',
    CrD: 'bo2',
    ArD: 'bo',
    fsD: 'ji2',
    dsD: 'ji',
    VsD: 'shi',
    TsD: 'chi',
    jrL: 'huo2',
    SrD: 'zha',
    UrD: 'zhe2',
    esD: 'qi',
    grD: 'e4',
    gsD: 'yi4',
    erD: 'ke',
    hrD: 'e',
    isD: 'xi',
    KrD: 'che',
    hrL: 'wo',
    BrD: 'po',
    irD: 'he',
    jrD: 'he2',
    drD: 'ge',
    irL: 'huo',
    LrD: 'zhe2',
    drL: 'ge',
    hsL: 'yu',
    MrD: 'nuo4',
    erL: 'kuo',
    CsD: 'bi2',
    DtD: 'mo4',
    jtL: 'huo2',
    dtL: 'guo',
    AtD: 'bo',
    CtD: 'bo2',
    UtD: 'zhe2',
    StD: 'zhe',
    UtL: 'zhuo2',
    TtD: 'che',
    etD: 'ke',
    itL: 'huo',
    jtD: 'he2',
    dtD: 'ge',
    JtD: 'zhe',
    htD: 'e',
    VtD: 'she',
    BtD: 'po',
    gtD: 'e4',
    ItD: 'le4',
    StL: 'zhuo',
    VtL: 'shuo',
    MtD: 'ne4',
    QuD: 'xi',
    NuD: 'ji',
    huD: 'yi',
    luD: 'yi4',
    auD: 'shi',
    YuD: 'chi',
    ZuD: 'shi2',
    XuD: 'zhi',
    LuD: 'zhi2',
    OuD: 'qi',
    RuD: 'xi2',
    PuD: 'ji2',
    CuD: 'bi2',
    luL: 'yu4',
    iuL: 'xu',
    AuD: 'bi',
    BuD: 'pi',
    buD: 'shi2',
    AsD: 'bi',
    JuD: 'zhi',
    KuD: 'chi',
    XuL: 'zhu',
    OuL: 'qu',
    QvD: 'xi',
    dvD: 'ji',
    BvD: 'pi',
    IvD: 'li4',
    EvD: 'di',
    jvD: 'xi2',
    gvD: 'yi4',
    GvD: 'di2',
    FvD: 'ti',
    NvD: 'ji',
    evD: 'qi',
    HvD: 'ni4',
    PvD: 'ji2',
    DvD: 'mi4',
    CvD: 'bi2',
    AvD: 'bi',
    evL: 'qu',
    dvL: 'ju',
    OvD: 'qi',
    ivD: 'xi',
    ivL: 'xu',
    XwD: 'zhi',
    LwD: 'zhi2',
    IwD: 'li4',
    KwD: 'chi',
    JwD: 'zhi',
    bwD: 'shi2',
    QwD: 'xi',
    ZwD: 'shi2',
    awD: 'shi',
    iwD: 'xi',
    UwD: 'zhe2',
    fwD: 'ji2',
    MwD: 'ni4',
    TwD: 'che',
    hwD: 'yi',
    VwD: 'she',
    ewD: 'qi',
    dwD: 'ji',
    lwD: 'yi4',
    NwD: 'ji',
    AwD: 'bi',
    kwL: 'yu4',
    iwL: 'xu',
    BwD: 'pi',
    SwD: 'zhe',
    CwD: 'bi2',
    gwD: 'yi4',
    PwD: 'ji2',
    DwD: 'mi4',
    YwD: 'chi',
    ExD: 'de',
    NxD: 'ze',
    IxD: 'le4',
    exD: 'ke',
    GxD: 'de2',
    ixD: 'he',
    DxD: 'mo4',
    PxD: 'ze2',
    QxD: 'se',
    AxD: 'bo',
    CxD: 'bo2',
    jxL: 'huo2',
    dxL: 'guo',
    hxD: 'e',
    jxD: 'he2',
    HxD: 'ne4',
    dxD: 'ge',
    OxD: 'ce',
    BxD: 'po',
    ixL: 'huo',
    O1D: 'qi',
    Z1D: 'shi2',
    X1D: 'zhi',
    R1D: 'xi2',
    P1D: 'ji2',
    c1D: 'ri4',
    h1D: 'yi',
    a1D: 'shi',
    N1D: 'ji',
    f1H: 'ji2',
    L1D: 'zhi2',
    J1D: 'zhi',
    I1D: 'li4',
    d1H: 'ji',
    g1H: 'yi4',
    e1H: 'qi',
    Q1D: 'xi',
    V1D: 'she',
    i1H: 'xi',
    h1H: 'yi',
    K1D: 'chi',
    k1D: 'yi4',
    l1D: 'yi4',
    M1D: 'ni4',
    U1D: 'zhe2',
    C1H: 'bi2',
    A1H: 'bi',
    T1D: 'che',
    Y1D: 'chi',
    j2D: 'he2',
    d2D: 'ge',
    E2D: 'da',
    Q2D: 'sa',
    G2D: 'da2',
    F2D: 'ta',
    P2D: 'za2',
    N2D: 'za',
    H2D: 'na4',
    e2D: 'ke',
    h2D: 'e',
    i2D: 'he',
    g2D: 'e4',
    O2D: 'ca',
    j3D: 'he2',
    I3D: 'la4',
    E3D: 'da',
    F3D: 'ta',
    i3D: 'he',
    H3D: 'na4',
    G3D: 'da2',
    Q3D: 'sa',
    g3D: 'e4',
    P3D: 'za2',
    d3D: 'ge',
    e3D: 'ke',
    h3D: 'e',
    O3D: 'ca',
    l4D: 'ye4',
    N4D: 'jie',
    a4D: 'she',
    Z4D: 'she2',
    I4D: 'lie4',
    P4D: 'jie2',
    L4D: 'zhe2',
    h4H: 'ye',
    M4D: 'nie4',
    Y4D: 'che',
    c4D: 're4',
    X4D: 'zhe',
    O4D: 'qie',
    K4D: 'che',
    f4H: 'jie2',
    J4D: 'zhe',
    k4D: 'ye4',
    e4D: 'qie',
    V4D: 'she',
    d4H: 'jie',
    h4D: 'ye',
    F5D: 'tie',
    j5D: 'xie2',
    d5D: 'jie',
    e5D: 'qie',
    G5D: 'die2',
    H5D: 'nie4',
    Q5D: 'xie',
    I5D: 'lie4',
    E5D: 'die',
    P5D: 'jie2',
    N5D: 'jie',
    i5D: 'xie',
    j6D: 'xia2',
    e6D: 'qia',
    U6D: 'zha2',
    d6D: 'jia',
    S6D: 'zha',
    T6D: 'cha',
    M6D: 'na4',
    i6D: 'xia',
    V6D: 'sha',
    J6D: 'zha',
    h6D: 'ya',
    g6D: 'ya4',
    K6D: 'cha',
    j7D: 'xia2',
    L7D: 'zha',
    h7D: 'ya',
    d7D: 'jia',
    V7D: 'sha',
    i7D: 'xia',
    g8D: 'ye4',
    i8D: 'xie',
    e8D: 'qie',
    d8D: 'jie',
    h8D: 'ye',
    f8D: 'jie2',
    C9D: 'fa2',
    A9D: 'fa',
    B9D: 'fa',
  }[音韻地位.編碼] || '?';
}

/**
 * 推導廣州音
 * https://ayaka.shn.hk/teoi/
 *
 * 説明
 * 以下內容為生成推導廣州音的函數體
 * 函數接受音韻地位，返回對應的推導廣州音
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的推導廣州音
 */
export function gwongzau(音韻地位, 字頭) {
  const is = (x) => 音韻地位.屬於(x);

  function 聲母規則() {
    if (is('幫母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      return 'b';
    }
    if (is('滂母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      return 'p';
    }
    if (is('並母')) {
      if (is('東韻 三等 或 鍾微虞廢文元陽尤凡韻')) return 'f';
      if (is('平聲')) return 'p';
      return 'b';
    }
    if (is('明母')) return 'm';

    if (is('端母')) return 'd';
    if (is('透母')) return 't';
    if (is('定母')) return is('平聲') ? 't' : 'd';
    if (is('泥母')) return 'n';
    if (is('來母')) return 'l';

    if (is('知母')) return 'z';
    if (is('徹母')) return 'c';
    if (is('澄母')) return is('平聲') ? 'c' : 'z';
    if (is('孃母')) return 'n';

    if (is('精母')) return 'z';
    if (is('清母')) return 'c';
    if (is('從母')) return is('平聲') ? 'c' : 'z';
    if (is('心母')) return 's';
    if (is('邪母')) return is('平聲') ? 'c' : 'z'; // 塞擦音多於擦音

    if (is('莊母')) return 'z';
    if (is('初母')) return 'c';
    if (is('崇母')) return is('平聲') ? 'c' : 'z';
    if (is('生母')) return 's';
    if (is('俟母')) return is('平聲') ? 'c' : 'z';

    if (is('章母')) return 'z';
    if (is('昌母')) return 'c';
    if (is('常母')) return 's'; // 擦音多於塞擦音
    if (is('書母')) return 's';
    if (is('船母')) return 's';
    if (is('日母')) return 'j';

    if (is('見母')) return 'g';
    if (is('溪母')) return 'h'; // 多數擦化
    if (is('羣母')) return is('平聲') ? 'k' : 'g';
    if (is('疑母')) return 'ng'; // ng 拼細音時為 j，詳後

    if (is('曉母')) return 'h';
    if (is('匣母')) {
      if (is('合口 或 模韻')) return 'j'; // 非 yu 前為 w，詳後
      return 'h';
    }
    if (is('影云以母')) {
      if (is('三四等')) return 'j'; // 非 yu 前為 w，詳後
      return '';
    }

    throw new Error('無聲母規則');
  }

  function 韻母規則() {
    // 通攝
    if (is('東冬鍾韻')) return 'ung';

    // 江攝
    if (is('江韻 幫組')) return 'ong';
    if (is('江韻 舌齒音')) return 'oeng';
    if (is('江韻 牙喉音')) return 'ong';

    // 止攝
    if (is('支脂之微韻 幫組')) return 'ei';
    if (is('支脂之微韻 開口 舌齒音 端組')) return 'ei';
    if (is('支脂之微韻 開口 舌齒音 來母')) return 'ei';
    if (is('支脂之微韻 開口 舌齒音 孃母')) return 'ei';
    if (is('支脂之微韻 開口 舌齒音')) return 'i';
    if (is('支脂之微韻 開口 牙喉音 疑母')) return 'i';
    if (is('支脂之微韻 開口 牙喉音 影母')) return 'i';
    if (is('支脂之微韻 開口 牙喉音 云母')) return 'i';
    if (is('支脂之微韻 開口 牙喉音 以母')) return 'i';
    if (is('支脂之微韻 開口 牙喉音')) return 'ei';
    if (is('支脂之微韻 合口 舌齒音')) return 'eoi';
    if (is('支脂之微韻 合口 牙喉音')) return 'ai';

    // 遇攝
    if (is('魚虞韻 幫組 幫滂並母')) return 'u';
    if (is('魚虞韻 幫組 明母')) return 'ou';
    if (is('魚虞韻 舌齒音 端組')) return 'eoi';
    if (is('魚虞韻 舌齒音 來母')) return 'eoi';
    if (is('魚虞韻 舌齒音 孃母')) return 'eoi';
    if (is('魚虞韻 舌齒音 精組')) return 'eoi';
    if (is('魚虞韻 舌齒音 莊組')) return 'o';
    if (is('魚虞韻 舌齒音')) return 'yu';
    if (is('魚虞韻 牙喉音 見溪羣母')) return 'eoi';
    if (is('魚虞韻 牙喉音 曉匣母')) return 'eoi';
    if (is('魚虞韻 牙喉音')) return 'yu';
    if (is('模韻 脣音')) return 'ou';
    if (is('模韻 舌齒音')) return 'ou';
    if (is('模韻 牙喉音 疑母')) return '';
    if (is('模韻 牙喉音')) return 'u';

    // 蟹攝
    if (is('齊韻')) return 'ai';
    if (is('祭韻 幫組')) return 'ai';
    if (is('祭韻 開口')) return 'ai';
    if (is('祭韻 合口 舌齒音')) return 'eoi';
    if (is('祭韻 合口 以母')) return 'eoi';
    if (is('祭韻 合口 牙喉音')) return 'ai';
    if (is('泰韻 幫組')) return 'ui';
    if (is('泰韻 開口 舌齒音 精組')) return 'oi';
    if (is('泰韻 開口 舌齒音')) return 'aai';
    if (is('泰韻 開口 牙喉音')) return 'oi';
    if (is('泰韻 合口 舌齒音')) return 'eoi';
    if (is('泰韻 合口 牙喉音 疑母')) return 'oi';
    if (is('泰韻 合口 牙喉音')) return 'ui';
    if (is('佳皆夬韻 幫組')) return 'aai';
    if (is('佳皆夬韻 開口')) return 'aai';
    if (is('佳皆夬韻 合口 舌齒音')) return 'eoi';
    if (is('佳皆夬韻 合口')) return 'aai';
    if (is('灰韻 舌齒音')) return 'eoi';
    if (is('灰韻 疑母')) return 'oi';
    if (is('灰韻')) return 'ui';
    if (is('咍韻')) return 'oi';
    if (is('廢韻')) return 'ai';

    // 臻攝
    if (is('眞韻 幫組')) return 'an';
    if (is('眞韻 開口')) return 'an';
    if (is('眞韻 合口 舌齒音')) return 'eon';
    if (is('眞韻 合口 牙喉音')) return 'an';
    if (is('臻文欣韻')) return 'an';
    if (is('元韻 幫組')) return 'aan';
    if (is('元韻 開口')) return 'in';
    if (is('元韻 合口')) return 'yun';
    if (is('魂韻 幫組')) return 'un';
    if (is('魂韻 端組')) return 'eon';
    if (is('魂韻 來母')) return 'eon';
    if (is('魂韻 精組')) return 'yun';
    if (is('魂韻 牙喉音')) return 'an';
    if (is('痕韻')) return 'an';

    // 山攝
    if (is('寒韻 幫組')) return 'un';
    if (is('寒韻 開口 舌齒音')) return 'aan';
    if (is('寒韻 開口 牙喉音')) return 'on';
    if (is('寒韻 合口 舌齒音')) return 'yun';
    if (is('寒韻 合口 牙喉音')) return 'un';
    if (is('刪山韻')) return 'aan';
    if (is('仙先韻 幫組')) return 'in';
    if (is('仙先韻 開口')) return 'in';
    if (is('仙先韻 合口')) return 'yun';

    // 效攝
    if (is('蕭宵韻')) return 'iu';
    if (is('肴韻')) return 'aau';
    if (is('豪韻')) return 'ou';

    // 果攝
    if (is('歌韻 一等')) return 'o';
    if (is('歌韻 三等 脣音')) return 'e';
    if (is('歌韻 三等 開口')) return 'e';
    if (is('歌韻 三等 合口')) return 'oe';

    // 假攝
    if (is('麻韻 二等')) return 'aa';
    if (is('麻韻 三等')) return 'e';

    // 宕攝
    if (is('陽韻 幫組')) return 'ong';
    if (is('陽韻 開口 莊組')) return 'ong';
    if (is('陽韻 開口')) return 'oeng';
    if (is('陽韻 合口')) return 'ong';
    if (is('唐韻')) return 'ong';

    // 梗攝
    if (is('庚韻 二等')) return 'ang';
    if (is('庚韻 三等 莊組')) return 'ang';
    if (is('庚韻 三等')) return 'ing';
    if (is('耕韻')) return 'ang';
    if (is('清青韻')) return 'ing';

    // 曾攝
    if (is('蒸韻')) return 'ing';
    if (is('登韻')) return 'ang';

    // 流攝
    if (is('尤侯幽韻')) return 'au';

    // 深攝
    if (is('侵韻')) return 'am'; // m 韻尾在聲母為脣音時為 n，詳後，下同

    // 咸攝
    if (is('覃談韻 幫組')) return 'aam';
    if (is('覃談韻 舌齒音')) return 'aam';
    if (is('覃談韻 牙喉音')) return 'om'; // -om 併入 -am，詳後
    if (is('鹽添嚴韻')) return 'im';
    if (is('咸銜凡韻')) return 'aam';

    throw new Error('無韻母規則');
  }

  function 聲調規則() {
    if (is('全清 或 次清')) {
      if (is('平聲')) return '1'; // 陰平
      if (is('上聲')) return '2'; // 陰上
      if (is('去聲')) return '3'; // 陰去
      if (is('入聲')) return '1'; // 陰入。長元音為 3，詳後
    } else {
      if (is('平聲')) return '4'; // 陽平
      if (is('全濁 上聲')) return '6'; // 陽去，全濁上變去
      if (is('上聲')) return '5'; // 陽上
      if (is('去聲')) return '6'; // 陽去
      if (is('入聲')) return '6'; // 陽入
    }
    throw new Error('無聲調規則');
  }

  function is長元音(韻母) {
    if (['aam', 'aan', 'im', 'in', 'om', 'on', 'ong', 'oeng', 'un', 'yun'].includes(韻母)) return true;
    if (['am', 'an', 'ang', 'eon', 'ing', 'ung'].includes(韻母)) return false;
    throw new Error('無法判斷元音長短：' + 韻母);
  }

  let 聲母 = 聲母規則();
  let 韻母 = 韻母規則();
  let 聲調 = 聲調規則();

  // ng 拼細音時為 j
  const is細音 = ['eo', 'i', 'oe', 'u', 'yu'].some((x) => 韻母.startsWith(x));
  if (聲母 === 'ng' && is細音) 聲母 = 'j';

  // 陰入分化
  if (is('入聲') && 聲調 === '1' && is長元音(韻母)) 聲調 = '3';

  if (is('合口 或 模韻') && !['eo', 'oe', 'yu'].some((x) => 韻母.startsWith(x))) { // 合口字
    if (聲母 === 'g' && !韻母.startsWith('u')) 聲母 = 'gw';
    else if (聲母 === 'k' && !韻母.startsWith('u')) 聲母 = 'kw';
    else if (聲母 === 'h' && !韻母.startsWith('i')) 聲母 = 'f';
    else if (聲母 === 'j') 聲母 = 'w';
    else if (聲母 === '') 聲母 = 'w';
  }

  // -om 併入 -am
  if (韻母 === 'om') 韻母 = 'am';

  // m 韻尾在聲母為脣音時為 n
  if (is('幫組') && 韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'n';

  if (is('入聲')) {
    if (韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'p';
    else if (韻母.endsWith('n')) 韻母 = 韻母.slice(0, -1) + 't';
    else if (韻母.endsWith('ng')) 韻母 = 韻母.slice(0, -2) + 'k';
  }

  return 聲母 + 韻母 + 聲調;
}

/**
 * 綾香思考音系
 * https://ayaka.shn.hk/v8/
 *
 * 説明
 * 以下內容為生成綾香思考音系的函數體
 * 函數接受音韻地位，返回對應的綾香思考音系拼音
 *
 * @param {Qieyun.音韻地位} 音韻地位 切韻音系音韻地位
 * @param {string} 字頭 字頭
 * @return {string} 音韻地位對應的綾香思考音系
 */
export function ayaka_v8(音韻地位, 字頭) {
  // 1. 開關

  const 開關 = {
    // '平假名': 地 ち
    // '片假名': 地 チ
    // '日本式羅馬字': 地 ti
    // '平文式羅馬字': 地 chi
    書寫系統: '日本式羅馬字',

    // true: 迥 ク𛅥ィ
    // false: 迥 クヱィ
    // 僅當開啓假名時生效
    ヰヱヲ小假名: true,

    // null: 宙 チウ tiu；南 ダム dam；愁 スウ suu
    // '日語': 宙 チュウ tyuu；南 ダン dan；愁 スウ suu
    // '綾香': 宙 チュウ tyuu；南 ダム dam；愁 ス su
    音變: '綾香',

    // null: 無聲調
    // '漢音八聲': 八聲家的聲調
    // '綾香': 綾香的聲調
    聲調: '綾香',
  };

  // 2. 輔助函數

  const 假名表 = {
    a:  'ア', i:  'イ', u:  'ウ', e:  'エ', o:  'オ',
    ka: 'カ', ki: 'キ', ku: 'ク', ke: 'ケ', ko: 'コ',
    ga: 'ガ', gi: 'ギ', gu: 'グ', ge: 'ゲ', go: 'ゴ',
    sa: 'サ', si: 'シ', su: 'ス', se: 'セ', so: 'ソ',
    za: 'ザ', zi: 'ジ', zu: 'ズ', ze: 'ゼ', zo: 'ゾ',
    ta: 'タ', ti: 'チ', tu: 'ツ', te: 'テ', to: 'ト',
    da: 'ダ', di: 'ヂ', du: 'ヅ', de: 'デ', do: 'ド',
    na: 'ナ', ni: 'ニ', nu: 'ヌ', ne: 'ネ', no: 'ノ',
    pa: 'ハ', pi: 'ヒ', pu: 'フ', pe: 'ヘ', po: 'ホ',
    ba: 'バ', bi: 'ビ', bu: 'ブ', be: 'ベ', bo: 'ボ',
    ma: 'マ', mi: 'ミ', mu: 'ム', me: 'メ', mo: 'モ',
    ya: 'ヤ',           yu: 'ユ',           yo: 'ヨ',
    ra: 'ラ', ri: 'リ', ru: 'ル', re: 'レ', ro: 'ロ',
    wa: 'ワ', wi: 'ヰ',           we: 'ヱ', wo: 'ヲ',
  };

  const 拗音表 = {
    wya: 'ヰャ', wyo: 'ヰョ',
    ya: 'ャ', yu: 'ュ', yo: 'ョ',
    wa: 'ヮ', wi: '𛅤', we: '𛅥', wo: '𛅦',
  };

  const 韻尾表 = {
    '': '', i: 'イ', u: 'ウ',
    m: 'ム', n: 'ン', ng: 'ゥ', // ng: 'ィ',
    p: 'フ', t: 'ツ', k: 'ク', // k: 'キ',
  };

  function roma2kata(s) {
    const r = /^([kgsztdnpbmyrw]?w??[yw]?)([aiueo])([ptkmngiu]*)$/g; // 將音節分為韻頭、主要元音及韻尾
    const match = r.exec(s);
    if (match == null) {
      throw new Error('無法轉換為假名：' + s);
    }
    const { 1: 韻頭, 2: 主要元音, 3: 韻尾 } = match;
    let 假名韻尾 = 韻尾表[韻尾];
    if (主要元音 === 'e') {
      if (韻尾 === 'k') 假名韻尾 = 'キ';
      if (韻尾 === 'ng') 假名韻尾 = 'ィ';
    }
    if (韻頭.length <= 1) {
      return 假名表[韻頭 + 主要元音] + 假名韻尾;
    }
    const 填充元音 = 韻頭[1] === 'w' ? 'u' : 'i'; // 韻頭[1] 只能為 w 或 y
    return 假名表[韻頭[0] + 填充元音] + 拗音表[韻頭.slice(1) + 主要元音] + 假名韻尾;
  }

  function kata2hira(s) {
    const diff = 'ぁ'.charCodeAt(0) - 'ァ'.charCodeAt(0);
    return [...s].map((c) => {
      switch (c) {
        case '𛅤': return '𛅐';
        case '𛅥': return '𛅑';
        case '𛅦': return '𛅒';
        default: return String.fromCharCode(c.charCodeAt(0) + diff);
      }
    }).join('');
  }

  function small2large(s) {
    return [...s].map((c) => {
      switch (c) {
        case '𛅤': return 'ヰ';
        case '𛅥': return 'ヱ';
        case '𛅦': return 'ヲ';
        default: return c;
      }
    }).join('');
  }

  // 3. 推導規則

  const is = (x) => 音韻地位.屬於(x);

  function 聲母規則() {
    // 幫組
    if (is('幫滂並母')) return 'p';
    if (is('明母')) return is('梗攝') && !is('庚耕青韻 入聲') ? 'm' : 'b';

    // 端組、來母
    if (is('端透定母')) return 't';
    if (is('泥母')) return is('梗攝') ? 'n' : 'd';
    if (is('來母')) return 'r';

    // 知組
    if (is('知徹澄母')) return 't';
    if (is('孃母')) return is('梗攝') ? 'n' : 'd';

    // 精組
    if (is('精清從心邪母')) return 's';

    // 莊組
    if (is('莊初崇生俟母')) return 's';

    // 章組、日母
    if (is('章昌船書常母')) return 's';
    if (is('日母')) return 'z';

    // 見組
    if (is('見溪羣母')) return 'k';
    if (is('疑母')) return 'g';

    // 影組、以母
    if (is('影母')) return '';
    if (is('曉匣母')) return 'k';
    if (is('云以母')) return '';

    throw new Error('無聲母規則');
  }

  function 韻母規則() {
    // 通攝
    if (is('東韻 一等')) return 'ong';
    if (is('東韻 三等 幫滂並母')) return is('舒聲') ? 'ong' : 'uk';
    if (is('東韻 三等 明母')) return 'ong';
    if (is('東韻 三等 精莊章組')) return 'yung';
    if (is('東韻 三等 舌齒音')) return is('舒聲') ? 'yung' : 'ik';
    if (is('東韻 三等 影母')) return is('舒聲') ? 'yung' : 'wik';
    if (is('東韻 三等 牙喉音')) return is('舒聲') ? 'yung' : 'ik';
    if (is('冬韻')) return 'ong';
    if (is('鍾韻 脣音')) return 'ong';
    if (is('鍾韻')) return 'yong';

    // 江攝
    if (is('江韻')) return 'ang';

    // 止攝
    if (is('支脂之微韻 脣音')) return 'i';
    if (is('支脂之微韻 開口')) return 'i';
    if (is('支脂之微韻 合口 舌齒音')) return 'ui';
    if (is('支脂之微韻 合口 牙喉音')) return 'wi';

    // 遇攝
    if (is('魚韻 莊組')) return 'o';
    if (is('魚韻 舌齒音')) return 'yo';
    if (is('魚韻 牙喉音')) return 'yo';
    if (is('虞韻 脣音')) return 'u';
    if (is('虞韻 來母')) return 'u';
    if (is('虞韻 知組')) return 'yuu';
    if (is('虞韻 莊組')) return 'u';
    if (is('虞韻 舌齒音')) return 'yu';
    if (is('虞韻 以母')) return 'yu';
    if (is('虞韻 牙喉音')) return 'u';
    if (is('模韻 幫組')) return 'o';
    if (is('模韻 舌齒音')) return 'o';
    if (is('模韻 影母')) return 'wo';
    if (is('模韻 牙喉音')) return 'o';

    // 蟹攝
    if (is('祭齊韻 脣音')) return 'ei';
    if (is('祭齊韻 開口')) return 'ei';
    if (is('祭齊韻 合口')) return 'wei';
    if (is('泰佳皆夬韻 脣音')) return 'ai';
    if (is('泰佳皆夬韻 開口')) return 'ai';
    if (is('泰佳皆夬韻 合口')) return 'wai';
    if (is('灰韻')) return 'wai';
    if (is('咍韻')) return 'ai';
    if (is('廢韻 脣音')) return 'ei';
    if (is('廢韻 開口')) return 'ei';
    if (is('廢韻 合口')) return 'wai';

    // 臻攝
    if (is('眞韻 脣音')) return 'in';
    if (is('眞韻 開口')) return 'in';
    if (is('眞韻 合口 來母')) return 'in';
    if (is('眞韻 合口 莊組')) return is('舒聲') ? 'on' : 'it';
    if (is('眞韻 合口 舌齒音')) return is('舒聲') ? 'yun' : 'ot';
    if (is('眞韻 合口 云母')) return 'win';
    if (is('眞韻 合口 牙喉音')) return 'in';
    if (is('臻韻')) return 'in';
    if (is('文韻')) return 'un';
    if (is('欣韻')) return 'in';
    if (is('元韻 脣音')) return 'an';
    if (is('元韻 開口')) return 'en';
    if (is('元韻 合口')) return 'wen';
    if (is('魂痕韻')) return 'on';

    // 山攝
    if (is('寒刪山韻 脣音')) return 'an';
    if (is('寒刪山韻 開口')) return 'an';
    if (is('寒刪山韻 合口')) return 'wan';
    if (is('仙先韻 脣音')) return 'en';
    if (is('仙先韻 開口')) return 'en';
    if (is('仙先韻 合口')) return 'wen';

    // 效攝
    if (is('蕭宵韻')) return 'eu';
    if (is('肴韻')) return 'au';
    if (is('豪韻 脣音')) return 'ou';
    if (is('豪韻 舌齒音')) return 'au';
    if (is('豪韻 牙喉音')) return 'au';

    // 果攝
    if (is('歌韻 一等 脣音')) return 'a';
    if (is('歌韻 一等 開口')) return 'a';
    if (is('歌韻 一等 合口')) return 'wa';
    if (is('歌韻 三等 開口')) return 'ya';
    if (is('歌韻 三等 合口')) return 'wa';

    // 假攝
    if (is('麻韻 二等 脣音')) return 'a';
    if (is('麻韻 二等 開口')) return 'a';
    if (is('麻韻 二等 合口')) return 'wa';
    if (is('麻韻 三等 脣音')) return 'ya';
    if (is('麻韻 三等 開口')) return 'ya';

    // 宕攝
    if (is('陽韻 脣音')) return 'ang';
    if (is('陽韻 開口 莊組')) return 'ang';
    if (is('陽韻 開口 舌齒音')) return 'yang';
    if (is('陽韻 開口 牙喉音')) return 'yang';
    if (is('陽韻 合口 舌齒音')) return 'ang';
    if (is('陽韻 合口 影云母')) return 'wang';
    if (is('陽韻 合口 牙喉音')) return 'wyang';
    if (is('唐韻 脣音')) return 'ang';
    if (is('唐韻 開口')) return 'ang';
    if (is('唐韻 合口')) return 'wang';

    // 梗攝
    if (is('庚韻 二等 脣音')) return 'ang';
    if (is('庚韻 二等 開口')) return 'ang';
    if (is('庚韻 二等 合口')) return 'wang';
    if (is('庚韻 三等 脣音')) return 'eng';
    if (is('庚韻 三等 開口')) return 'eng';
    if (is('庚韻 三等 合口')) return 'weng';
    if (is('耕韻')) return 'wang';
    if (is('清青韻 脣音')) return 'eng';
    if (is('清青韻 開口')) return 'eng';
    if (is('清青韻 合口')) return 'weng';

    // 曾攝
    if (is('蒸韻 脣音')) return 'yong';
    if (is('蒸韻 開口 莊組')) return 'ong';
    if (is('蒸韻 開口 舌齒音')) return 'yong';
    if (is('蒸韻 開口 牙喉音')) return 'yong';
    if (is('蒸韻 合口 影云母')) return 'yong';
    if (is('蒸韻 合口 牙喉音')) return 'wyong';
    if (is('登韻')) return 'ong';

    // 流攝
    if (is('尤韻 幫滂並母')) return 'uu';
    if (is('尤韻 明母')) return 'ou';
    if (is('尤韻 莊組')) return 'uu';
    if (is('尤韻 舌齒音')) return 'iu';
    if (is('尤韻 牙喉音')) return 'iu';
    if (is('侯韻')) return 'ou';
    if (is('幽韻')) return 'iu';

    // 深攝
    if (is('侵韻')) return 'im';

    // 咸攝
    if (is('覃談韻')) return 'am';
    if (is('鹽添嚴韻')) return 'em';
    if (is('咸銜凡韻')) return 'am';

    throw new Error('無韻母規則');
  }

  function 聲調規則() {
    if (開關.聲調 === '漢音八聲') {
      if (!is('全濁')) {
        if (is('平聲')) return '꜀';
        if (is('上聲')) return '꜂';
        if (is('去聲')) return '꜄';
        if (is('入聲')) return '꜆';
      } else {
        if (is('平聲')) return '꜁';
        if (is('上聲')) return '꜃';
        if (is('去聲')) return '꜅';
        if (is('入聲')) return '꜇';
      }
      throw new Error('無聲調規則');
    } else if (開關.聲調 === '綾香') {
      if (is('全清 或 次清')) {
        if (is('平聲')) return 'ˉ';
        if (is('上聲')) return 'ˊ';
        if (is('去聲')) return 'ˋ';
        if (is('入聲')) return is('通梗曾深攝 或 眞臻文欣痕韻 或 魂韻 牙喉音') ? 'ˉ' : 'ˇ';
      } else {
        if (is('平聲')) return 'ˇ';
        if (is('上聲')) return is('全濁') ? 'ˇ' : 'ˊ';
        if (is('去聲')) return 'ˇ';
        if (is('入聲')) return 'ˇ';
      }
      throw new Error('無聲調規則');
    }
    return '';
  }

  let 聲母 = 聲母規則();
  let 韻母 = 韻母規則();
  let 聲調 = 聲調規則();

  if (is('入聲')) {
    if (韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'p';
    else if (韻母.endsWith('n')) 韻母 = 韻母.slice(0, -1) + 't';
    else if (韻母.endsWith('ng')) 韻母 = 韻母.slice(0, -2) + 'k';
  }

  if (韻母.startsWith('w') && (!is('牙喉音') || is('重紐A類 或 以母'))) 韻母 = 韻母.slice(1);

  // 4. 音變規則

  if (開關.音變 === '日語') {
    if (韻母.startsWith('w')) 韻母 = 韻母.slice(1); // 園 wen -> en

    if (韻母.endsWith('p')) 韻母 = 韻母.slice(0, -1) + 'u'; // 鄴 gep -> geu
    else if (韻母.endsWith('m')) 韻母 = 韻母.slice(0, -1) + 'n'; // 南 dam -> dan
    else if (韻母.endsWith('eng')) 韻母 = 韻母.slice(0, -2) + 'i'; // 生 seng -> sei
    else if (韻母.endsWith('ng')) 韻母 = 韻母.slice(0, -2) + 'u'; // 相 syang -> syau
  }

  if (['日語', '綾香'].includes(開關.音變)) {
    if (韻母.endsWith('au')) 韻母 = 韻母.slice(0, -2) + 'ou'; // 高 kau -> kou
    else if (韻母.endsWith('iu')) 韻母 = 韻母.slice(0, -2) + 'yuu'; // 宙 tiu -> tyuu
    else if (韻母.endsWith('eu')) 韻母 = 韻母.slice(0, -2) + 'you'; // 遙 eu -> you
    else if (韻母.endsWith('ang')) 韻母 = 韻母.slice(0, -3) + 'ong'; // 相 syang -> syong
  }

  if (開關.音變 === '綾香') {
    if (韻母 === 'uu') 韻母 = 韻母.slice(0, -1); // 愁 suu -> su
    else if (韻母.endsWith('yung')) 韻母 = 韻母.slice(0, -1); // 窮 kyung -> kyun
  }

  let 聲韻;

  if (['平假名', '片假名'].includes(開關.書寫系統)) {
    聲韻 = roma2kata(聲母 + 韻母);
    if (!開關.ヰヱヲ小假名) 聲韻 = small2large(聲韻);
    if (開關.書寫系統 === '平假名') 聲韻 = kata2hira(聲韻);
  } else {
    if (['日語', '綾香'].includes(開關.音變)) {
      if (聲母 === 'p') 聲母 = 'h'; // 甫 pu -> hu
    }

    if (開關.音變 === '日語') {
      if (韻母.endsWith('t')) 韻母 = 韻母 + 'u'; // 遏 at -> atu
      else if (韻母.endsWith('ek')) 韻母 = 韻母 + 'i'; // 席 sek -> seki
      else if (韻母.endsWith('k')) 韻母 = 韻母 + 'u'; // 澤 tak -> taku
    }

    if (開關.音變 === '綾香') {
      if (聲母 === 'r') 聲母 = 'l'; // 籟 rai -> lai
      if (韻母.endsWith('t')) 韻母 = 韻母.slice(0, -1) + 's'; // 遏 at -> as
      else if (韻母.endsWith('p')) 韻母 = 韻母.slice(0, -1) + 'f'; // 鄴 gep -> gef

      if (韻母.endsWith('eng')) 韻母 = 韻母.slice(0, -3) + 'ein'; // 傾 keng -> kein
      else if (韻母.endsWith('ong')) 韻母 = 韻母.slice(0, -3) + 'oun'; // 同 tong -> toun
    }

    if (開關.書寫系統 === '平文式羅馬字') {
      if (['日語', '綾香'].includes(開關.音變)) {
        if (聲母 === 's' && 韻母.startsWith('i')) 聲母 = 'sh'; // 四 si -> shi
        else if (聲母 === 'z' && 韻母.startsWith('i')) 聲母 = 'j'; // 人 zin -> jin
        else if (聲母 === 't' && 韻母.startsWith('i')) 聲母 = 'ch'; // 地 ti -> chi
        else if (聲母 === 't' && 韻母.startsWith('u')) 聲母 = 'ts'; // 追 tui -> tsui
        else if (聲母 === 'h' && 韻母.startsWith('u')) 聲母 = 'f'; // 甫 hu -> fu
        else if (聲母 === 's' && 韻母.startsWith('y')) { 聲母 = 'sh'; 韻母 = 韻母.slice(1); } // 小 syou -> shou
        else if (聲母 === 'z' && 韻母.startsWith('y')) { 聲母 = 'j'; 韻母 = 韻母.slice(1); } // 繞 zyou -> jou
        else if (聲母 === 't' && 韻母.startsWith('y')) { 聲母 = 'ch'; 韻母 = 韻母.slice(1); } // 兆 tyou -> chou
      }

      if (開關.音變 === '日語') {
        if (聲母 === 'd' && 韻母.startsWith('i')) 聲母 = 'j'; // 膩 di -> ji
        else if (聲母 === 'd' && 韻母.startsWith('y')) { 聲母 = 'j'; 韻母 = 韻母.slice(1); } // 紐 dyuu -> juu

        if (韻母.endsWith('tu')) 韻母 = 韻母.slice(0, -1) + 'su'; // 遏 atu -> atsu
      }
    }

    聲韻 = 聲母 + 韻母;
  }

  return [...'꜀꜁꜂꜃'].includes(聲調) ? 聲調 + 聲韻 : 聲韻 + 聲調;
}

